/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <gtk/gtk.h>
#include <libgen.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_dbio.h"
#include "xplab_maintenance_fe.h"
#include "xplab_image.h"
#include "xplab_preferences_fe.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/
#define XPLAB_SECONDS_PER_DAY (24*60*60)

/* local prototypes ********************************************************/
static void write_db_def(FILE *file, db_def db);
static void write_tps_lst(FILE *file, tps_lst tpl);
static void write_tps_def(FILE *file, tps_def tps);
static void write_media_att(FILE *file, media_att media);
static void write_cls_lst(FILE *file, cls_lst cl);
static void write_cls_def(FILE *file, cls_def cls);
static void write_cls_att(FILE *file, cls_att att);
static void write_nb_lst(FILE *file, nb_lst nbl);
static void write_nb_def(FILE *file, nb_def nbd);
static void write_pic_def(FILE *file, pic_def pat);
static void write_area_def(FILE *file, area_def ar);
static void write_nb_chn_lst(FILE *file, nb_chn_lst ncl);
static void write_nb_chn_def(FILE *file, nb_chn_def ncd);
static int write_bmp(FILE *file, PICT *bmp, area_def ar);
static int write_pack(FILE *file, ulong *patt, area_def ar);
static int write_bitlevel(FILE *file, PICT *bpat, area_def ar);
static int read_db_def(FILE *file, db_def *db, char *db_version);
static int read_db_version(FILE *file, char *db_version);
static int read_tps_lst(FILE *file, tps_lst *tpl, char *db_version);
static int read_tps_def(FILE *file, tps_def *tps, char *db_version);
static int read_media_att(FILE *file, media_att *media, char *db_version);
static int read_cls_lst(FILE *file, cls_lst *cll);
static int read_cls_def(FILE *file, cls_def *cls);
static int read_cls_att(FILE *file, cls_att *att);
static int read_nb_lst(FILE *file, nb_lst *nbl);
static int read_nb_def(FILE *file, nb_def *nbd);
static int read_pic_def(FILE *file, pic_def *pat);
static int read_area_def(FILE *file, area_def *ar);
static int read_nb_chn_lst(FILE *file, nb_chn_lst *ncl);
static int read_nb_chn_def(FILE *file, nb_chn_def *ncd);
static int read_bmp(FILE *file, area_def ar, PICT **bmp);
static int read_pack(FILE *file, area_def ar, ulong **pack);
static int read_bitlevel(FILE *file, area_def ar, PICT **bpat);
static PICT *unpack_bmp(ulong *pack, area_def ar);
static ulong *pack_bmp(area_def ar, PICT *bmp);
static void xplab_write_recognition_file(FILE *file, char *db_name, 
					 char *tps_ident, 
					 char *image_filename, 
					 int mborder, int nborder, 
					 int mweight, int thold, 
					 rb_lst rblst, cand_lst hit_lst, 
					 char *options);
static void write_rb_lst(FILE *file, rb_lst rblst);
static void write_rb_def(FILE *file, rb_def rb);
static void write_ival_lst(FILE *file, ival_lst ivl);
static void writeline_sector_def(FILE *file, sector_def sc);
static void writeline_ival_def(FILE *file, ival_def iv);
static void write_cand_lst(FILE *file, cand_lst *kl);
static void write_cand_def(FILE *file, cand_def kr);
static void writeline_cv_def_i(FILE *file, cv_def_i kdi);
static void write_int_lst(FILE *file, int_lst il);
static int read_rb_lst(FILE *file, rb_lst *rbl);
static int read_rb_def(FILE *file, rb_def *rbd);
static int read_ival_lst(FILE *file, ival_lst *ivl);
static int read_sector_def(FILE *file, sector_def *sec);
static int read_ival_def(FILE *file, ival_def *iv);
static int read_cand_lst(FILE *file, cand_lst *kl);
static int read_cand_def(FILE *file, cand_def *kr);
static int read_cv_def_i(FILE *file, cv_def_i *kdi);
static int read_int_lst(FILE *file, int_lst *il);
static int test_string_length(FILE *file, int len);
static int copy_file(char *src, char *dst, int verbose_if_error);
static char *read_line(FILE *file);
static int file_modified(char *fname, int past_seconds);

/* Functions ***************************************************************/

/* Writes the DB to a file.

Revision: 23.07.2004 E.Sack

Input : - DB
Output: - DB and backup dumped
        - flag (0: DB dumped, 1: DB not dumped)
Postcondition: - structure must be consistent to read_db_file.

Implementation: Database files are version-controlled.
*/
int
write_db_file(db_def db)
{
  int  err, clos;
  FILE *file;

  err = check_name(db.db_fname);
  assert(!err);

  back_up_file(db.db_fname);

  file = xplab_open_file(db.db_fname, "w");
  if (!file) {
    return 1;
  }

  string_to_log(main_log, "dump DB '", db.db_fname, "' ...");

  write_db_def(file, db);

  clos = xplab_close_file(file, db.db_fname);

  if (!clos) {
    string_to_log(main_log, " ok\n", NULL, NULL);
  }

  return 0;
}



/* Makes a copy of a disc file if there is any.

Revision: 23.07.2004 E.Sack

Input : - pointer to file name string
Output: - file copied
*/
void
back_up_file(char *fname)
{
  int  err;
  char *buf = NULL;

  err = check_name(fname);
  if (err) {
    return;
  }

  cat_string(fname, FILE_BCK_DEFAULT, "", &buf);
  err = file_modified(buf, XPLAB_SECONDS_PER_DAY);

  if (!err) {               /* default backup once per day if there is any */
    copy_file(fname, buf, 0);
  }

  FREEMEM(buf);
}



/* Checks wether a file is updated for a past period of time.

Revision: 02.07.2004 E.Sack

Input : - pointer to file name
        - period of seconds
Output: - flag (0: no modification, 1: modification in the past period)
*/
static int
file_modified(char *fname, int past_seconds)
{
  int    d, err;
  double sec;
  time_t now = time((time_t *)NULL);
  struct stat buf;

  err = check_name(fname);
  assert(!err);

  d = stat(fname, &buf);

  if (d) {                       /* assuming an update is ok,              */
    return 0;                    /* further to check by an access function */
  }

  sec = difftime(now, buf.st_mtime);           /* difference is in seconds */

  d = sec < past_seconds ? 1 : 0;

  return d;
}



/* Dumps a DB to a file.

Revision: 09.05.2003 E.Sack

Input : - pointer to stream
        - DB
Output: - DB dumped
*/
static void 
write_db_def(FILE *file, db_def db) 
{
  int d;

  assert(file);
  assert(strlen(DB_VERSION_CURRENT));

  /* always write current DB version */
  d = fprintf(file, "%s\n", DB_VERSION_CURRENT);
  assert(d > 0);

  write_tps_lst(file, db.tpl);
}



/* Dumps a document list to a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - document list
Output: - document list dumped
*/
static void
write_tps_lst(FILE *file, tps_lst tpl)
{
  int d, j;

  assert(file);

  d = fprintf(file, "%d\n", tpl.cnt);         /* number of test pattern sets */
  assert(d > 0);

  if (!tpl.cnt) {
    assert(!tpl.tps);
    return;
  }
  else {
    assert(tpl.cnt>0);
    assert(tpl.tps);
  }

  for (j = 0; j < tpl.cnt; j++) {
    write_tps_def(file, tpl.tps[j]);
  }
}



/* Dumps a document definition to a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - document definition
Output: - document definition dumped
*/
static void
write_tps_def(FILE *file, tps_def tps)
{
  int d;

  assert(file);

  d = check_name(tps.ident);
  assert(!d);

  d = fprintf(file, "%s ", tps.ident);

  write_media_att(file, tps.media);
  write_cls_lst(file, tps.cl);                         /* store class list */
}



/* Dumps document features to a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - document features
Output: - document features dumped
Postcondition: - 'opos' not written
*/
static void
write_media_att(FILE *file, media_att media)
{
  int d;

  assert(file);
  assert(XPLAB_IS_METHOD(media.in_line.method));
  assert(media.in_line.linespacing.period.l_phase>=0);
  assert(media.in_line.linespacing.period.h_phase>=0);
  assert(media.dpi<=DPI_MAX);

  d = fprintf(file, "%d %d %d %d ", 
	      media.in_line.method,
	      media.in_line.linespacing.period.l_phase,
	      media.in_line.linespacing.period.h_phase,
	      media.dpi);
  assert(d > 0);
}



/* Dumps a class list to a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - class list
Output: - stream updated
*/
static void 
write_cls_lst(FILE *file, cls_lst cll) 
{
  int d, i;

  assert(file);

  d = fprintf(file, "%d\n", cll.cnt);
  assert(d > 0);

  if (!cll.cnt) {
    assert(!cll.cls);
    return;
  }
  else {
    assert(cll.cnt>0);
    assert(cll.cls);
  }

  for (i = 0; i < cll.cnt; i++) {
    write_cls_def(file, cll.cls[i]);
  }
}



/* Wrapper for xplab_read_db_file,

Revision: 03.10.2004 E.Sack

Input : - pointer to DB filename
        - flag (0: don't renew old DB version, 1: renew old DB version)
        - verbose flag
        - pointer to flag
Output: - DB updated
        - flag updated (0: no error, 1: error)
Postcondition: - DB is empty if file is not readable
*/
db_def
xplab_read_db(char *db_name, int renew, int verbose, int *serious_error) 
{
  int    d;
  char   *buf = NULL, *version_ctrl = NULL;
  db_def db;                                                   /* database */
  FILE   *db_file;

  *serious_error = 0;

  init_db_def(&db);
   
  d = check_name(db_name);

  if (d) {
    if (verbose) {
      xplab_text_insert(main_log, "WARNING: DB file name undefined\n");
    }
    return db;
  }

  d = access(db_name, F_OK);
  if (d) {
    if (verbose) {
      string_to_log(main_log, "WARNING: DB '", db_name, "' not available\n");
    }
    return db;
  }

  db_file = xplab_open_file(db_name, "r");
  if (!db_file) {
    return db;
  }

  db = xplab_read_db_file(db_name, db_file, &version_ctrl, verbose);

  d = xplab_close_file(db_file, db_name);

  if (d) {
    free_db_def(&db);
    FREEMEM(version_ctrl);
    return db;
  }

  d = check_name(db.db_fname);

  if (d) {
    if (verbose) {
      string_to_log(main_log, "ERROR: '", db_name, "' not a XPLAB database\n");
    }
    free_db_def(&db);
    FREEMEM(version_ctrl);
    *serious_error = 1;
    return db;                                              /* DB is empty */
  }

  d = check_name(version_ctrl);
  assert(!d);

  d = strcmp(version_ctrl, DB_VERSION_CURRENT);

  if (d && renew) {                                /* old version detected */
    string_to_log(main_log, "WARNING: DB '", db_name, "' has old version '");
    string_to_log(main_log, "", version_ctrl, "' - renewed with '" 
		  DB_VERSION_CURRENT "\n");

    cat_string(db.db_fname, OLD_VERSION_BCK, "", &buf);

    d = access(buf, F_OK);
    if (!d) {
      string_to_log(main_log, "WARNING: file '", buf, 
		    "' exists - no files overwritten\n");
    }
    else {
      d = copy_file(db.db_fname, buf, 1);                 /* backup old DB */

      if (d) {                                    /* old DB could be saved */
	write_db_file(db);                                 /* write new DB */
      }
    }
  }

  FREEMEM(version_ctrl);
  FREEMEM(buf);

  return db;                                                    /* DB read */
}



/* Reads version controlled database. XPLAB tries to be downward compatibel
   and reads old versions, which are automatically converted and rewritten
   in the current format. Old database files are moved. 

Revision: 29.07.2005 E.Sack

Input : - pointer to DB filename
        - pointer to DB file stream
        - pointer to DB version string pointer
        - verbose flag
Output: - DB
        - version string updated
Postcondition: - DB is empty if file is not readable
               - version string contains DB version read
               - DB file name without path
Implementation: Database is version-controlled.
*/
db_def
xplab_read_db_file(const char *db_name, FILE *db_file, char **version_ctrl, 
		   int verbose) 
{
  int    d;
  char   *basec = NULL, *bname;
  db_def db;                                                   /* database */

  init_db_def(&db);

  assert(db_name);
  assert(db_file);
  assert(version_ctrl);
  FREEMEM(*version_ctrl);

  if (verbose) {
    string_to_log(main_log, "read DB '", db_name, "' ...");
  }

  /*************************************************************************/
  /***** for every DB version change there must be an entry in the loop ****/
  /*************************************************************************/
  while (1) {                                        /* detect DB versions */
    add_string(DB_VERSION_CURRENT, version_ctrl);
    d = read_db_def(db_file, &db, *version_ctrl);
    if (!d) {                                    /* older version detected */
      break;
    }

    FREEMEM(*version_ctrl);
    add_string(PROGVERSION_0_0_4, version_ctrl);
    d = read_db_def(db_file, &db, *version_ctrl);
    if (!d) {                                    /* older version detected */
      break;
    }

    FREEMEM(*version_ctrl);
    add_string(PROGVERSION_0_0_3, version_ctrl);
    d = read_db_def(db_file, &db, *version_ctrl);
    if (!d) {                                    /* older version detected */
      break;
    }

    FREEMEM(*version_ctrl);
    add_string(PROGVERSION_0_0_2, version_ctrl);
    d = read_db_def(db_file, &db, *version_ctrl);
    if (!d) {                                    /* older version detected */
      break;
    }

    if (verbose) {
      string_to_log(main_log, "\nERROR: DB '", db_name, 
		    "' has unknown format\n");
    }

    FREEMEM(*version_ctrl);
    free_db_def(&db);
    return db;
    break;                                            /* not a readable DB */
  }

  add_string(db_name, &basec);                   /* enroll current DB name */
  bname = basename(basec);                         /* see "man 3 basename" */
  add_string(bname, &db.db_fname);
  FREEMEM(basec);

  if (verbose) {
    xplab_text_insert(main_log, " ok\n");
  }

  return db;                                                /* DB readable */
}



/* Copy file

Revision: 29.07.2005 E.Sack

Input : - pointer to source filename
        - pointer to destination filename
        - flag (0: verbose not if error, 1: verbose if error)
Output: - destination file written to disc
        - Flag (0: error, 1: no error)
*/
static int
copy_file(char *src, char *dst, int verbose_if_error) 
{
  int  d, f1, f2, n, N=0;
  char buf[1024];

  assert(src);
  assert(dst);
  assert(src!=dst);
  assert(strlen(src));
  assert(strlen(dst));

  d = strcmp(src,dst);
  assert(d);                     /* assert different strings */

  if ((f1 = open(src, O_RDONLY, 0)) == -1) {
    if (verbose_if_error) {
      string_to_log(main_log, "\nDBIO: can't open '", src, "'\n\n");
    }
    return 0;
  }

  if ((f2 = creat(dst, 0666)) == -1) {
    if (verbose_if_error) {
      string_to_log(main_log, "\nDBIO: can't create '", dst, "' with mode 0666\n\n");
    }

    d = close(f1);
    if (d && verbose_if_error) {
      string_to_log(main_log, "\nDBIO: error closing '", src, "'\n\n");
    }

    return 0;
  }

  while ((n = read(f1, buf, 1024)) > 0) {
    N+=n;
    if (write(f2, buf, n) != n) {
      string_to_log(main_log, "\nDBIO: write error on file  '", dst, "'\n\n");

      d = close(f1);
      if (d && verbose_if_error) {
	string_to_log(main_log, "\nDBIO: error closing '", src, "'\n\n");
      }

      d = close(f2);
      if (d && verbose_if_error) {
	string_to_log(main_log, "\nDBIO: error closing '", dst, "'\n\n");
      }

      return 0;
    }
  }

  d = close(f1);
  if (d && verbose_if_error) {
    string_to_log(main_log, "\nDBIO: error closing '", src, "'\n\n");
  }

  d = close(f2);
  if (d && verbose_if_error) {
    string_to_log(main_log, "\nDBIO: error closing '", dst, "'\n\n");
  }

  string_to_log(main_log, "DBIO: copied '", src, "' ");
  string_to_log(main_log, "to '", dst, "'\n");

  return 1;
}



/* Reads DB with integrated version control, which assures downward
   compatibility.

Revision: 09.05.2003 E.Sack

Input : - pointer to stream
        - pointer to DB
        - pointer to DB version
Output: - DB updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_db_def(FILE *file, db_def *db, char *db_version) 
{
  int d, version_error;

  assert(file);
  assert(db);
  assert(db_version);

  init_db_def(db);

  rewind(file);                                           /* read whole DB */
  clearerr(file);                                 /* clear previous errors */

  version_error = read_db_version(file, db_version);

  if (version_error) {
    free_db_def(db);
    return version_error;
  }

  version_error = read_tps_lst(file, &db->tpl, db_version);

  if (version_error) {
    free_db_def(db);
    return version_error;
  }

  d = ferror(file);                                         /* file error? */

  if (d) {
    free_db_def(db);
    clearerr(file);
    return 1;                                     /* unexpected file error */
  }

  d = fgetc(file);                                            /* file end? */

  if (d != EOF) {
    free_db_def(db);
    return 1;                                    /* unexpected input value */
  }

  /* update current version                                                */
  add_string(DB_VERSION_CURRENT, &db->db_version);

  return 0;
}



/* Reads and checks the DB version from a file.

Usage: Whenever a file contains not the current DB version, it is to
       update and this function delivers a trigger for that. For update
       it is assumed to call this function repeatedly starting with the
       latest DB version up to the oldest one.

Revision: 09.05.2003 E.Sack

Input : - pointer to stream
        - pointer to version string
Output: - Flag (0: input succeeded, >0: input not succeeded) */
static int
read_db_version(FILE *file, char *db_version)
{
  int    d, d1, major, minor, sub;
  char   version_read[MAXCHAR], *ptr;
  fpos_t file_offset;

  assert(file);
  assert(db_version);

  d = test_string_length(file, MAXCHAR);             /* test first DB line */

  if (d) {
    return 1;                                   /* unexpected input format */
  }

  d = fgetpos(file, &file_offset);                  /* current file offset */

  if (d) {
    return 1;                                           /* file read error */
  }

  ptr = fgets(version_read, sizeof(version_read), file);/* read DB version */

  if (!ptr) {
    return 1;                                   /* unexpected input format */
  }

  /* not all old formats are version controlled */
  d = strncmp(version_read, DB_VERSION_CURRENT, strlen(DB_VERSION_CURRENT));
  d1 = strcmp(db_version, DB_VERSION_CURRENT);  

  if (d && !d1) {                                  /* triggering DB update */
    return 1;                                        /* not current format */
  }

  d = sscanf(version_read, "%d.%d.%d\n",              /* DB format control */
	     &major,
	     &minor,
	     &sub);

  if (d != 3) {
    d = strcmp(db_version,                   /* representative for old     */
	       PROGVERSION_0_0_2);           /* DB formats without version */

    if (d) {
      return 1;                                   /* not current DB format */
    }
    else {                                               /* old DB formats */
      d = fsetpos(file, &file_offset);               /* reset file pointer */

      if (d) {
	return 1;                                        /* file set error */
      }
    }
  }

  return 0;
}



/* Reads document list from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to document list
        - pointer to DB version
Output: - document list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_tps_lst(FILE *file, tps_lst *tpl, char *db_version)
{
  int d, tmp, m, version_error = 0;

  assert(file);
  assert(tpl);
  assert(db_version);

  init_tps_lst(tpl);

  d = fscanf(file, "%d\n", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0 || tmp > TPS_MAX) {                         /* be reasonable */
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  tpl->cnt = tmp;                                        /* list not empty */

  CALLOC(tps_def, tpl->cnt, tpl->tps);

  for (m = 0; m < tpl->cnt; m++) {                       /* read documents */
    version_error = read_tps_def(file, &tpl->tps[m], db_version);

    if (version_error) {
      break;
    }
  }

  return version_error;
}



/* Read document list.

Revision: 25.07.2002 E.Sack

Input : - pointer to file
        - pointer to document list
        - pointer to DB version
Output: - document list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_tps_def(FILE *file, tps_def *tps, char *db_version)
{
  int  d, version_error = 0;
  char temp_id[MAXCHAR];

  assert(file);
  assert(tps);
  assert(db_version);

  init_tps_def(tps);

  d = test_string_length(file, sizeof(temp_id));

  if (d) {
    return 1;                                   /* unexpected input format */
  }

  d = fscanf(file, "%s", temp_id);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (!strlen(temp_id)) {
    return 1;                                    /* unexpected input value */
  }

  add_string(temp_id, &tps->ident);

  version_error = read_media_att(file, &tps->media, db_version);

  if (version_error) {
    return version_error;
  }

  version_error = read_cls_lst(file, &tps->cl);

  if (version_error) {
    return version_error;
  }

  gen_back_refs(tps->cl);              /* generate backward references NNB */

  return 0;
}



/* Reads document attributes.

Revision: 28.03.2003 E.Sack

Input : - pointer to file
        - pointer to document attributes
        - pointer to DB version
Output: - document attributes updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_media_att(FILE *file, media_att *media, char *db_version)
{
  int d2, d3, d, tmp1, tmp2, tmp3;

  init_media_att(media);

  d3 = strcmp(db_version, PROGVERSION_0_0_3);           /* version control */
  d2 = strcmp(db_version, PROGVERSION_0_0_2);

  if (d3 && d2) {
    d = fscanf(file, "%d", &tmp1);

    if (d != 1) {
      return 1;                                 /* unexpected input format */
    }

    if (!XPLAB_IS_METHOD((INLINE_METHOD)tmp1)) {
      return 1;                                  /* unexpected input value */
    }

    media->in_line.method = (INLINE_METHOD)tmp1;               /* input ok */

    d = fscanf(file, "%d%d", &tmp1, &tmp2);

    if (d != 2) {
      return 1;                                 /* unexpected input format */
    }

    if (tmp1 < 0 || tmp2 < 0) {
      return 1;                                  /* unexpected input value */
    }

    media->in_line.linespacing.period.l_phase = tmp1;          /* input ok */
    media->in_line.linespacing.period.h_phase = tmp2;
    /* 'opos' not read because it is a temporary variable                  */
  }

  if (!d3) {
    d = fscanf(file, "%d", &tmp1);

    if (d != 1) {
      return 1;                                 /* unexpected input format */
    }

    if (!XPLAB_IS_METHOD((INLINE_METHOD)tmp1)) {
      return 1;                                  /* unexpected input value */
    }

    media->in_line.method = (INLINE_METHOD)tmp1;               /* input ok */

    d = fscanf(file, "%d%d%d", &tmp1, &tmp2, &tmp3);

    if (d != 3) {
      return 1;                                 /* unexpected input format */
    }

    if (tmp1 < 0 || tmp2 < 0) {
      return 1;                                  /* unexpected input value */
    }

    media->in_line.linespacing.period.l_phase = tmp1;          /* input ok */
    media->in_line.linespacing.period.h_phase = tmp2;
    media->in_line.linespacing.opos = tmp3;
  }

  if (!d2) {                                  /* current version using old */
    media->in_line.method = OVERLAPPING;      /* defaults.                 */
  }

  d = fscanf(file, "%d", &tmp1);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp1 > DPI_MAX) {
    return 1;                                    /* unexpected input value */
  }

  media->dpi = tmp1;                                           /* input ok */

  return 0;
}



/* Generate backward references.
    forward reference: NNB belongs to several CNB(i), i = 1,2,3...
   backward reference: CNB consists of several NNB(j), j = 1,2,3...

Revision: 25.07.2002 E.Sack

Input : - NB list
Output: - NB list updated
*/
void 
gen_back_refs(cls_lst clst)
{
  int chn_idx, h, cnb_c, nnb_c, cnb_n, nnb_n;

  for (cnb_c = 0; cnb_c < clst.cnt; cnb_c++) {
    for (cnb_n = 0; cnb_n < clst.cls[cnb_c].nbl.cnt; cnb_n++) {
      for (nnb_c = 0; nnb_c < clst.cnt; nnb_c++) {
	for (nnb_n = 0; nnb_n < clst.cls[nnb_c].nbl.cnt; nnb_n++) {
          for (h = 0; h < clst.cls[nnb_c].nbl.nb[nnb_n].cnb.cnt; h++) {
            if (clst.cls[nnb_c].nbl.nb[nnb_n].cnb.cd[h].cdx == cnb_c &&
		clst.cls[nnb_c].nbl.nb[nnb_n].cnb.cd[h].ndx == cnb_n) {
	      chn_idx = clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cnt;
	      if (chn_idx) {
		REALLOC(nb_chn_def, chn_idx + 1,
			clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cd);
	      }
	      else {
		CALLOC(nb_chn_def, chn_idx + 1, 
		       clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cd);
	      }

	      clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_idx].cdx = nnb_c;
	      clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_idx].ndx = nnb_n;
	      clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_idx].x = 
		clst.cls[nnb_c].nbl.nb[nnb_n].cnb.cd[h].x;
	      clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_idx].y = 
		clst.cls[nnb_c].nbl.nb[nnb_n].cnb.cd[h].y;
	      clst.cls[cnb_c].nbl.nb[cnb_n].nnb.cnt++;
	    }
	  }
	}
      }
    }
  }
}



/* Dumps a class to a file.

Revision: 27.07.2002 E.Sack

Input : - pointer to stream
        - class
Output: - class dumped
Postcondition: - class kernel not written
*/
static void 
write_cls_def(FILE *file, cls_def cls) 
{
  assert(file);

  write_cls_att(file, cls.att);
  write_nb_lst(file, cls.nbl);
}



/* Dumps class attributes to a file.

Revision: 13.02.2004 E.Sack

Input : - pointer to stream
        - class attributes
Output: - class attributes dumped
*/
static void 
write_cls_att(FILE *file, cls_att att) 
{
  int d, len;

  assert(file);

  assert(att.class);                           /* empty string not allowed */
  len = strlen(att.class);  
  assert(len>0 && len<MAX_LEN_CLASS_NAME);

  assert(att.sign);                            /* empty string not allowed */
  len = strlen(att.sign);  
  assert(len>0 && len<MAX_LEN_CLASS_TRANS);

  assert(att.crem);                            /* empty string not allowed */
  len = strlen(att.crem);  
  assert(len>0 && len<MAX_LEN_CLASS_COMMENT);

  assert(att.alignment == UNDERLINE ||
	 att.alignment == BASELINE ||
	 att.alignment == MIDDLELINE ||
	 att.alignment == UPPERLINE);

  assert(att.printable == 0 || att.printable == 1);
  assert(att.fixed == 0 || att.fixed == 1);

  d = fprintf(file, "%s %s %s ", att.class, att.sign, att.crem);
  assert(d > 0);

  d = fprintf(file, "%d  %d  %d\n", att.alignment, att.printable, att.fixed);
  assert(d > 0);
}



/* Dumps NB (pattern) list to a file.

Revision: 20.09.2002 E.Sack

Input : - pointer to stream
        - NB list
Output: - NB list dumped
*/
static void 
write_nb_lst(FILE *file, nb_lst nbl) 
{
  int m, d;

  assert(file);

  d = fprintf(file, "%d\n", nbl.cnt);
  assert(d > 0);

  if (!nbl.cnt) {
    assert(!nbl.nb);
    return;
  }
  else {
    assert(nbl.cnt>0);
    assert(nbl.nb);
  }

  for (m = 0; m < nbl.cnt; m++) {
    write_nb_def(file, nbl.nb[m]);
  }
}



/* Dumps NB to a file.

Revision: 20.09.2002 E.Sack

Input : - pointer to stream
        - NB
Output: - NB dumped
Postcondition: - NB backward references NNB are not written
               - NB auxiliary datum is not written
*/
static void 
write_nb_def(FILE *file, nb_def nbd) 
{
  assert(file);

  write_pic_def(file, nbd.pat);
  write_nb_chn_lst(file, nbd.cnb);
}



/* Dumps NB chain list to a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - NB chain list
Output: - NB chain list dumped
*/
static void 
write_nb_chn_lst(FILE *file, nb_chn_lst ncl) 
{
  int m, d;

  assert(file);

  d = fprintf(file, "%d\n", ncl.cnt);
  assert(d > 0);

  if (!ncl.cnt) {
    assert(!ncl.cd);
    return;
  }
  else {
    assert(ncl.cnt>0);
    assert(ncl.cd);
  }

  for (m = 0; m < ncl.cnt; m++) {
    write_nb_chn_def(file, ncl.cd[m]);
  }
}



/* Dumps NB chain to a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - NB chain
Output: - NB chain dumped
*/
static void 
write_nb_chn_def(FILE *file, nb_chn_def ncd) 
{
  int d;

  assert(file);
  assert(ncd.cdx>=0);
  assert(ncd.ndx>=0);

  d = fprintf(file, "%d %d %d %d\n", ncd.cdx, ncd.ndx, ncd.x, ncd.y);
  assert(d > 0);
}



/* Die NB-Muster werden in eine Datei geschrieben.

Input : - Dateizeiger
        - Zeiger auf Muster
        - NB-Gre
Output: - Datei auf Stand
        - number of non-zero pattern
*/
static int
write_pack(FILE *file, ulong *patt, area_def ar) 
{
  int n, k, d, l, leer;

  assert(patt);

  for (k = 0, n = 0, leer = 0; k < ar.height; k++) {     /* writing long */
    for (l = 0; l < ar.width; l+=32, n++) {
      if (patt[n]) leer++;
      d = fprintf(file, "%08lx ", patt[n]);
      assert(d > 0);
    }

    d = fprintf(file, "\n");
    assert(d > 0);
  }

  return leer;
}



/* Die b-Werte werden in eine Datei geschrieben. Das Format ist 
   wertunabhngig, da die Bitebenen einzel gespeichert werden, um Platz
   zu sparen.

Input : - Dateizeiger
        - Zeiger auf b-Werteliste
        - NB-Gre
Output: - Datei auf Stand
        - number of non-zero pattern (-1 not acquired)
*/
static int
write_bitlevel(FILE *file, PICT *bpat, area_def ar)
{
  int  lvl = 1<<MAXBLVL;                          /* e.g. 1<<7 = 2^7 = 128 */
  int  max, j, k, d;
  PICT *z = 0;

  if(!bpat) {
    d = fprintf(file, "%d\n", 0);
    assert(d > 0);
    return -1;                                         /* no b-values used */
  }

  for (j = 0, max = 0; j < ar.width*ar.height; j++) {
    assert(bpat[j] < lvl);                  /* allowed number of bitlevels */
    max = XPMAX(max, bpat[j]);
  }

  if(!max) {
    d = fprintf(file, "%d\n", 0);
    assert(d > 0);
    return -1;                                        /* no b-values used */
  }

  for (j = 0, lvl = 0; j < MAXBLVL; j++) {
    if (max & 1) {
      lvl++;                                 /* level for biggest number */
    }
    max >>= 1;
  }

  d = fprintf(file, "%d\n", lvl);
  assert(d > 0);

  assert(lvl);                          /* sollte mit bpat bereinstimmen */

  CALLOC(PICT, ar.width*ar.height, z);                /* value to bitlevel */

  for (j = 0, d = 0; j < lvl; j++) {                      /* all bitlevel */
    for (k = 0; k < ar.width*ar.height; k++) {     /* extracting bitlevel */
      z[k] = bpat[k] & 1;
      bpat[k] >>= 1;
    }

    d += write_bmp(file, z, ar);
  }

  FREEMEM(z);

  return d;
}



/* Die Bitmap wird in eine Datei geschrieben.

Input : - Dateizeiger
        - Zeiger auf Bitmap
        - Bitmap-Gre
Output: - Datei auf Stand
        - number of non-zero pattern
*/
static int
write_bmp(FILE *file, PICT *bmp, area_def ar)
{
  ulong *u;
  int   d;

  assert(bmp);

  u = pack_bmp(ar, bmp);                               /* packing to long */
  d = write_pack(file, u, ar);

  FREEMEM(u);

  return d;
}



/* Die Bitmap wird in eine Datei geschrieben.

Input : - Dateizeiger
        - Zeiger auf Bitmap
        - Bitmap-Gre
Output: - Datei auf Stand
*/
static void 
write_pic_def(FILE *file, pic_def pat)
{
  int d;

  write_area_def(file, pat.ar);

  d = write_bmp(file, pat.mem, pat.ar);                     /* NB pattern */
  assert(d);                                                /* no pattern */

  d = write_bitlevel(file, pat.bvl, pat.ar);
  assert(d);                                                /* no pattern */
}



/* Reads class list from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to class list
Output: - class list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_cls_lst(FILE *file, cls_lst *cll)
{
  int d, tmp, m, version_error = 0;

  assert(file);
  assert(cll);

  init_cls_lst(cll);

  d = fscanf(file, "%d\n", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0) {
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  cll->cnt = tmp;                                        /* list not empty */

  CALLOC(cls_def, cll->cnt, cll->cls);

  for (m = 0; m < cll->cnt; m++) {
    version_error = read_cls_def(file, &cll->cls[m]);        /* read class */

    if (version_error) {
      break;
    }
  }

  return version_error;
}



/* Reads class from a file.

Revision: 19.07.2002 E.Sack

Input : - pointer to file
        - pointer to class
Output: - class updated
        - Flag (0: DB has current version, >0: DB has unknown version)
*/
static int
read_cls_def(FILE *file, cls_def *cls)
{
  int version_error = 0;

  assert(file);
  assert(cls);

  init_cls_def(cls);

  version_error = read_cls_att(file, &cls->att);

  if (version_error) {
    return version_error;
  }

  version_error = read_nb_lst(file, &cls->nbl);

  return version_error;
}



/* Reads class attributes from a file.

Revision: 19.07.2002 E.Sack

Input : - pointer to file
        - pointer to class attributes
Output: - class attributes updated
        - Flag (0: DB has current version, >0: DB has unknown version)
*/
static int
read_cls_att(FILE *file, cls_att *att)
{
  int  d, tmp1, tmp2, tmp3;
  char crem_tmp[MAX_LEN_CLASS_COMMENT], sign_tmp[MAX_LEN_CLASS_TRANS];
  char class_tmp[MAX_LEN_CLASS_NAME];

  assert(file);
  assert(att);

  init_cls_att(att);

  d = test_string_length(file, MAX_LEN_CLASS_NAME);

  if (d) {
    return 1;                                   /* unexpected input format */
  }

  d = fscanf(file, "%s", class_tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  add_string(class_tmp, &att->class);

  d = test_string_length(file, MAX_LEN_CLASS_TRANS);

  if (d) {
    return 1;                                   /* unexpected input format */
  }

  d = fscanf(file, "%s", sign_tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  add_string(sign_tmp, &att->sign);

  d = test_string_length(file, MAX_LEN_CLASS_COMMENT);

  if (d) {
    return 1;                                   /* unexpected input format */
  }

  d = fscanf(file, "%s", crem_tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  add_string(crem_tmp, &att->crem);

  d = fscanf(file, "%d%d%d\n", 
	     &tmp1,                                           /* alignment */
	     &tmp2, 
	     &tmp3);

  if (d != 3) {
    return 1;                                   /* unexpected input format */
  }

  if ((ALIGMNT)tmp1 != UNDERLINE &&
      (ALIGMNT)tmp1 != BASELINE &&
      (ALIGMNT)tmp1 != MIDDLELINE &&
      (ALIGMNT)tmp1 != UPPERLINE) {
    return 1;                                    /* unexpected input value */
  }

  att->alignment = (ALIGMNT)tmp1;

  if (tmp2 != 0 && tmp2 != 1) {
    return 1;                                    /* unexpected input value */
  }

  att->printable = tmp2;

  if (tmp3 != 0 && tmp3 != 1) {
    return 1;                                    /* unexpected input value */
  }

  att->fixed = tmp3;

  return 0;
}



/* Reads NB list from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to NB list
Output: - NB list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_nb_lst(FILE *file, nb_lst *nbl)
{
  int d, tmp, m, version_error = 0;

  assert(file);
  assert(nbl);

  init_nb_lst(nbl);

  d = fscanf(file, "%d\n", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0) {
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  nbl->cnt = tmp;                                        /* list not empty */

  CALLOC(nb_def, nbl->cnt, nbl->nb);

  for (m = 0; m < nbl->cnt; m++) {
    version_error = read_nb_def(file, &nbl->nb[m]);

    if (version_error) {
      break;
    }
  }

  return version_error;
}



/* Reads NB from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to NB
Output: - NB updated
        - Flag (0: input succeeded, >0: input not succeeded)
Postcondition: - backward references NNB are not read
               - auxiliary datum is not read
*/
static int
read_nb_def(FILE *file, nb_def *nbd)
{
  int version_error = 0;

  assert(file);
  assert(nbd);

  init_nb_def(nbd);

  version_error = read_pic_def(file, &nbd->pat);

  if (version_error) {
    return version_error;
  }

  version_error = read_nb_chn_lst(file, &nbd->cnb);

  return version_error;
}



/* Reads NB chain list from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to NB chain list
Output: - NB chain list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_nb_chn_lst(FILE *file, nb_chn_lst *ncl)
{
  int d, tmp, m, version_error = 0;

  assert(file);
  assert(ncl);

  init_nb_chn_lst(ncl);

  d = fscanf(file, "%d\n", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0) {
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  ncl->cnt = tmp;                                        /* list not empty */

  CALLOC(nb_chn_def, ncl->cnt, ncl->cd);

  for (m = 0; m < ncl->cnt; m++) {
    version_error = read_nb_chn_def(file, &ncl->cd[m]);

    if (version_error) {
      break;
    }
  }

  return version_error;
}



/* Reads NB chain from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to NB chain
Output: - NB chain updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_nb_chn_def(FILE *file, nb_chn_def *ncd)
{
  int d, tmp1, tmp2, tmp3, tmp4;

  assert(file);
  assert(ncd);

  init_nb_chn_def(ncd);

  d = fscanf(file, "%d %d %d %d\n", &tmp1, &tmp2, &tmp3, &tmp4);

  if (d != 4) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp1 < 0 || tmp2 < 0) {
    return 1;                                    /* unexpected input value */
  }

  ncd->cdx = tmp1;
  ncd->ndx = tmp2;
  ncd->x = tmp3;
  ncd->y = tmp4;

  return 0;
}



/* Reads packed pattern image from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - image size
        - pointer to packed pattern store
Output: - packed pattern store updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_pack(FILE *file, area_def ar, ulong **pack)
{
  int k, d, n, l;

  assert(file);
  assert(ar.width>0);
  assert(ar.height>0);
  assert(pack);

  CALLOC(ulong, XPMOD32(ar.width)*ar.height, *pack);

  for (k = n = 0; k < ar.height; k++) {
    for (l = 0; l < ar.width; l+=32, n++) {
      d = fscanf(file, "%lx", &(*pack)[n]);

      if (d != 1) {
	return 1;                               /* unexpected input format */
      }
    }

    fscanf(file, "\n");
  }

  return 0;
}



/* Reads b-values from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - image size
        - pointer to pattern store
Output: - pattern store updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_bitlevel(FILE *file, area_def ar, PICT **bpat)
{
  int  lvl, k, d, j, len, version_error = 0;
  PICT *mem = 0;

  assert(file);
  assert(ar.width>0);
  assert(ar.height>0);
  assert(bpat);

  d = fscanf(file, "%d\n", &lvl);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (!lvl) {
    return 0;                                                   /* no data */
  }

  len = ar.width*ar.height;
  CALLOC(PICT, len, *bpat);

  for (j = 0; j < lvl; j++) {                              /* all bitlevel */
    version_error = read_bmp(file, ar, &mem);

    if (version_error) {
      FREEMEM(mem);
      return version_error;
    }

    for (k = 0; k < len; k++) {                     /* extracting bitlevel */
      (*bpat)[k] <<= 1;
      (*bpat)[k] += mem[k];                       /* contains only one bit */
    }

    FREEMEM(mem);
  }

  return 0;
}



/* Reads packed pattern image from a file and converts it to bitmapped.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - image size
        - pointer to bitmap store
Output: - bitmap store updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_bmp(FILE *file, area_def ar, PICT **bmp)
{
  int   version_error;
  ulong *u = 0;

  assert(file);
  assert(ar.width>0);
  assert(ar.height>0);
  assert(bmp);

  version_error = read_pack(file, ar, &u);

  if (!version_error) {
    *bmp = unpack_bmp(u, ar);
  }

  FREEMEM(u);

  return version_error;
}



/* Reads pattern image from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to pattern store
Output: - pattern store updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int 
read_pic_def(FILE *file, pic_def *pat)
{
  int version_error = 0;

  init_pic_def(pat);

  version_error = read_area_def(file, &pat->ar);

  if (version_error) {
    return version_error;
  }

  version_error = read_bmp(file, pat->ar, &pat->mem);

  if (version_error) {
    return version_error;
  }

  version_error = read_bitlevel(file, pat->ar, &pat->bvl);

  return version_error;
}



/* Unpacks integer array to bitmap image.

Revision: 12.09.2004 E.Sack

Input : - pointer to integer array
        - memory size
Output: - pointer to bitmap
*/
static PICT*
unpack_bmp(ulong *pack, area_def ar) 
{
  int  i, j, k, l, mod, n;
  PICT *s = 0;

  assert(pack);
  assert(ar.width > 0); assert(ar.height > 0);

  mod = XPMOD32(ar.width);

  CALLOC(PICT, ar.width*ar.height, s);

  for (i = 0, n = 0; i < ar.height; i++) {                      /* rows    */
    for (j = 0, l = 0; j < mod; j++, n++) {                     /* columns */
      for (k = 31; k >= 0 && l < ar.width; k--, l++) {          /* words   */
	s[l + i*ar.width] = (pack[n] & 1<<k)>>k;
      }
    }
  }

  return s;
}



/* Packs bitmap image to integer array.

Revision: 12.09.2004 E.Sack

Input : - memory size
        - pointer to bitmap
Output: - pointer to integer array
*/
static ulong*
pack_bmp(area_def ar, PICT *bmp)
{
  int   i, j, k, l, n, z;
  int   mod = XPMOD32(ar.width);             /* Rest auch noch abspeichern */
  ulong *pack = 0;
   
  assert(bmp);

  for (i = 0; i < ar.width*ar.height; i++) {  
    assert(bmp[i] < 2);                                  /* test of bitmap */
  }

  CALLOC(ulong, mod*ar.height, pack);

  for (i = 0, n = 0; i < ar.height; i++) {                      /* rows    */
    for (j = 0, l = 0; j < mod; j++, n++) {                     /* columns */
      for (k = 31; k >= 0 && l < ar.width; k--, l++) {          /* words   */
        z = bmp[31-k+(j<<5)+i*ar.width];                      /* short cut */
	assert(!z || z == 1);                            /* only BMP {0,1} */
        pack[n] |= z << k;
      }
    }
  }
   
  return pack;
}



/* All recognition data are written to a file.

Revision: 08.11.2003 E.Sack

Input : - pointer to DB name
        - pointer to TPS ID
        - pointer to name of image file
        - mask border size
        - maximal allowed lag
        - RB list
        - candidate hitlist
Output: - data file written

Implementation: As recognition files are temporary, their format is not 
                version-controlled.
*/
void
xplab_write_recognition(char *db_name, char *tps_ident, char *image_filename, 
			int mborder, int nborder, int mweight, int thold, 
			rb_lst rblst, cand_lst hit_lst, char *options)
{
  int  d;
  char *replay_filename = NULL;
  FILE *file;

  d = check_name(image_filename);
  assert(!d);

  replay_filename = default_replay_filename(image_filename);

  d = check_name(replay_filename);
  assert(!d);

  file = xplab_open_file(replay_filename, "w");
  if (!file) {                          /* no backup-copy for replay files */
    FREEMEM(replay_filename);
    return;
  }

  string_to_log(main_log, "write replay file '", replay_filename, " ...");

  xplab_write_recognition_file(file, db_name, tps_ident, image_filename, 
			       mborder, nborder, mweight, thold, rblst, 
			       hit_lst, options);

  d = xplab_close_file(file, replay_filename);

  if (!d) {
    xplab_text_insert(main_log, " ok");
  }

  xplab_text_insert(main_log, "\n");

  FREEMEM(replay_filename);
}



/* Writes recognition data to a flie stream.

Revision: 02.07.2004 E.Sack

Input : - pointer to file stream
        - pointer to DB name
        - pointer to TPS ID
        - pointer to image filename
        - mask border size
        - maximal allowed lag
        - mask weight
        - threshold
        - RB list
        - candidate hitlist
        - pointer to user options string
Output: - file stream updated
*/
static void
xplab_write_recognition_file(FILE *file, char *db_name, char *tps_ident, 
			     char *image_filename, int mborder, int nborder, 
			     int mweight, int thold, rb_lst rblst, 
			     cand_lst hit_lst, char *options)
{
  int  d;

  d = fprintf(file, XPLAB_REPLAY_SIGNATURE "\n");
  assert(d > 0);

  d = check_name(options);
  assert(!d);
  d = fprintf(file, "%s\n", options); 
  assert(d > 0);

  d = fprintf(file, "%s\n%s\n%s\n", image_filename, db_name, tps_ident); 
  assert(d > 0);

  assert(mborder>=0);
  assert(nborder>=0);
  assert(mweight>=0);

  d = fprintf(file, "%d %d %d %d\n", mborder, nborder, mweight, thold); 
  assert(d > 0);

  write_rb_lst(file, rblst);
  write_cand_lst(file, &hit_lst);
}



/* Write RB list in a file.

Input : - pointer to file stream
        - RB list
Output: - file stream updated
*/
static void
write_rb_lst(FILE *file, rb_lst rblst)
{
  int j, d;

  assert(rblst.cnt);

  write_area_def(file, rblst.rar);

  d = fprintf(file, "%d\n", rblst.cnt);               /* list may be empty */
  assert(d > 0);

  for (j = 0; j < rblst.cnt; j++) {
    write_rb_def(file, rblst.rb[j]);
  }
}



/* Write area in a file.

Input : - pointer to file
        - area
Output: - file updated
*/
static void
write_area_def(FILE *file, area_def ar)
{
  int d;

  assert(ar.width>0);
  assert(ar.height>0);

  d = fprintf(file, "%d %d\n", ar.width, ar.height);
  assert(d > 0);
}



/* Write RB in a file.

Input : - pointer to file
        - RB
Output: - file updated
*/
static void
write_rb_def(FILE *file, rb_def rb)
{
  int d;

  assert(rb.rld <= CHAIN);

  d = fprintf(file, "%d ", rb.rld);
  assert(d > 0);

  writeline_sector_def(file, rb.sc);
  write_ival_lst(file, rb.gridy);
}



/* Dumps a interval list to a file.

Revision: 05.01.2003 E.Sack

Input : - pointer to stream
        - interval list
Output: - stream updated
*/
static void 
write_ival_lst(FILE *file, ival_lst ivl) 
{
  int d, i;

  assert(file);

  d = fprintf(file, "%d ", ivl.cnt);
  assert(d > 0);

  if (!ivl.ival) {                                    /* consistency check */
    assert(!ivl.cnt);
  }
  else {
    assert(ivl.cnt>0);
  }

  for (i = 0; i < ivl.cnt; i++) {                     /* list may be empty */
    writeline_ival_def(file, ivl.ival[i]);
  }

  d = fprintf(file, "\n");
  assert(d > 0);
}



/* Write sector in a file.

Input : - pointer to file
        - sector
Output: - file updated
*/
static void
writeline_sector_def(FILE *file, sector_def sc)
{
  writeline_ival_def(file, sc.column);
  writeline_ival_def(file, sc.line);
}



/* Write interval in a file.

Input : - pointer to file
        - interval
Output: - file updated
*/
static void
writeline_ival_def(FILE *file, ival_def iv)
{
  int d;

  d = fprintf(file, "%d %d ", iv.start, iv.end);
  assert(d > 0);
}



/* Write candidate list in a file.

Revision: 19.08.2002 E.Sack

Input : - pointer to file
        - pointer to candidate list
Output: - file updated
Postcondition: - if pointer to candidate list is NULL, an empty list is
                 written
*/
static void
write_cand_lst(FILE *file, cand_lst *kl)
{
  int d, j;

  if (kl) {                                             /* pointer defined */
    d = fprintf(file, "%d\n", kl->cnt);               /* list may be empty */
    assert(d > 0);

    for (j = 0; j < kl->cnt; j++) {
      write_cand_def(file, kl->c[j]);
    }
  }
  else {                                             /* specify empty list */
    d = fprintf(file, "%d\n", 0);
    assert(d > 0);
  }
}



/* Write candidate in a file.

Revision: 23.02.2003 E.Sack

Input : - pointer to file
        - candidate
Output: - file updated
*/
static void
write_cand_def(FILE *file, cand_def kr)
{
  int d;

  writeline_cv_def_i(file, kr.m);

  assert(kr.rbn>=0);
  assert(kr.n>=0 || 
	 kr.n==SUBSTRB_FLAG || 
	 kr.n==SUBSTNB_FLAG || 
	 kr.n==SUBSTUC_FLAG);
  assert(kr.c>=0 || kr.c==SUBCLS_FLAG);

  d = fprintf(file, "%d %d %d\n", kr.rbn, kr.n, kr.c);
  assert(d > 0);

  writeline_ival_def(file, kr.line);

  writeline_sector_def(file, kr.roi);
  writeline_sector_def(file, kr.ext);
  writeline_sector_def(file, kr.knl);
  writeline_sector_def(file, kr.hdx);

  d = fprintf(file, "\n");
  assert(d > 0);

  write_int_lst(file, kr.crb);
  write_cand_lst(file, kr.alt);
}



/* Write an integer list in a file.

Revision: 19.08.2002 E.Sack

Input : - pointer to file
        - integer list
Output: - file updated
*/
static void
write_int_lst(FILE *file, int_lst il)
{
  int d, j;

  assert(il.cnt>=0);

  d = fprintf(file, "%d ", il.cnt);
  assert(d > 0);

  for (j = 0; j < il.cnt; j++) {                      /* list may be empty */
    d = fprintf(file, "%d ", il.nr[j]);
    assert(d > 0);
  }

  d = fprintf(file, "\n");
  assert(d > 0);
}



/* Write correlation features in a file.

Input : - pointer to file
        - correlation features
Output: - file updated
*/
static void
writeline_cv_def_i(FILE *file, cv_def_i kdi)
{
  int d;

  assert(kdi.cv>=0.0);

  d = fprintf(file, "%f %d %d ", kdi.cv, kdi.x, kdi.y);
  assert(d > 0);
}



/* Wrapper for xplab_read_recognition.

Revision: 07.11.2003 E.Sack

Input : - pointer to replay filename
        - pointer to RB list
        - pointer to DB name
        - pointer to TPS ID
        - pointer to name of input image filename
        - pointer to mask border size
        - pointer to maximal allowed lag
        - pointer to candidate hitlist
        - pointer to user options string
        - verbose flag
Output: - RB list updated
        - DB name updated
        - TPS ID updated
        - name of input image filename updated
        - mask border size updated
        - maximal allowed lag updated
        - candidate hitlist updated
        - user options string updated
        - Flag (0: no error, 1: error)

Implementation: As recognition files are temporary, their format is not 
                version-controlled.
*/
int
xplab_read_recognition(char *rfile, rb_lst *rblst, char **db_name, 
		       char **tps_ident, char **filename, int *mborder, 
		       int *nborder, int *mweight, int *thold, 
		       cand_lst *hit_lst, char **options, int verbose)
{
  int  error;
  FILE *file;

  assert(rfile);

  file = xplab_open_file(rfile, "r");
  if (file == NULL) {
    return 1;
  }

  error = xplab_read_recognition_file(rfile, file, rblst, db_name, tps_ident, 
				      filename, mborder, nborder, mweight, 
				      thold, hit_lst, options, verbose);
  xplab_close_file(file, rfile);

  return error;
}



/* All recognition data are read in.

Revision: 07.11.2003 E.Sack

Input : - pointer to replay filename
        - pointer to replay file stream
        - pointer to RB list
        - pointer to DB name
        - pointer to TPS ID
        - pointer to name of input image filename
        - pointer to mask border size
        - pointer to maximal allowed lag
        - pointer to candidate hitlist
        - pointer to user options string
        - verbose flag
Output: - RB list updated
        - DB name updated
        - TPS ID updated
        - name of input image filename updated
        - mask border size updated
        - maximal allowed lag updated
        - candidate hitlist updated
        - user options string updated
        - Flag (0: no error, 1: error)

Implementation: As recognition files are temporary, there format is not 
                version-controlled.
*/
int
xplab_read_recognition_file(const char *rfile, FILE *file, rb_lst *rblst, 
			    char **db_name, char **tps_ident, 
			    char **filename, int *mborder, int *nborder, 
			    int *mweight, int *thold, cand_lst *hit_lst, 
			    char **options, int verbose)
{
  int  d, error = 0, read_error = 0, tmp1, tmp2, tmp3, tmp4;
  char *ptr, temp_line[MAXCHAR<<1], *buf = 0;

  assert(rfile);
  assert(file);
  assert(rblst);
  assert(db_name);
  assert(tps_ident);
  assert(filename);
  assert(mborder);
  assert(nborder);
  assert(mweight);
  assert(thold);
  assert(hit_lst);
  assert(options);

  if (verbose) {
    string_to_log(main_log, "read '", rfile, "' ...");
  }

  while(1) {
    buf = read_line(file);                /* read replay file magic number */

    if (buf == NULL) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    d = strncmp(buf, XPLAB_REPLAY_MAGIC, strlen(XPLAB_REPLAY_MAGIC));
    FREEMEM(buf);

    if (d) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    buf = read_line(file); 

    if (buf == NULL) {
      error = 1;                                 /* unexpected input value */
      break;
    }
    else {
      *options = buf;
    }

    ptr = fgets(temp_line, sizeof(temp_line), file);
    if (!ptr) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    ptr = strchr(temp_line, '\n');                        /* strip newline */
    if (ptr) {
      *ptr = 0;
    }

    read_error = check_name(temp_line);
    if (read_error) {
      string_to_log(main_log, "\nDBIO: no filename specified\n\n", "", "");
      error = 1;                                 /* unexpected input value */
      break;
    }

    add_string(temp_line, filename);

    ptr = fgets(temp_line, sizeof(temp_line), file);
    if (!ptr) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    ptr = strchr(temp_line, '\n');                        /* strip newline */
    if (ptr) {
      *ptr = 0;
    }

    read_error = check_name(temp_line);
    if (read_error) {
      string_to_log(main_log, "\nDBIO: no DB name specified\n\n", "", "");
      error = 1;                                 /* unexpected input value */
      break;
    }

    add_string(temp_line, db_name);

    ptr = fgets(temp_line, sizeof(temp_line), file);
    if (!ptr) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    ptr = strchr(temp_line, '\n');                        /* strip newline */
    if (ptr) {
      *ptr = 0;
    }

    read_error = check_name(temp_line);
    if (read_error) {
      string_to_log(main_log, "\nDBIO: no TPS ID specified\n\n", "", "");
      error = 1;                                 /* unexpected input value */
      break;
    }

    add_string(temp_line, tps_ident);

    ptr = fgets(temp_line, sizeof(temp_line), file);
    if (!ptr) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    d = sscanf(temp_line, "%d%d%d%d ", &tmp1, &tmp2, &tmp3, &tmp4); 

    if (d != 4) {
      error = 1;                                /* unexpected input format */
      break;
    }

    if (tmp1 < 0 || tmp2 < 0 || tmp3 < 0) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    if (tmp4 < THOLD_MIN || tmp4 > THOLD_MAX) {
      error = 1;                                 /* unexpected input value */
      break;
    }

    *mborder = tmp1;
    *nborder = tmp2;
    *mweight = tmp3;
    *thold = tmp4;

    read_error = read_rb_lst(file, rblst);

    if (read_error) {
      error = 1;
      break;
    }

    read_error = read_cand_lst(file, hit_lst);

    if (read_error) {
      error = 1;
      break;
    }

    if (verbose) {
      string_to_log(main_log, "", "", " ok");
    }

    break;
  }

  if (verbose) {
    string_to_log(main_log, "", "", "\n");
  }

  return error;
}



/* Reads RB list from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to RB list
Output: - RB list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_rb_lst(FILE *file, rb_lst *rbl)
{
  int j, d, tmp, read_error = 0;

  assert(file);
  assert(rbl);

  init_rb_lst(rbl);

  read_error = read_area_def(file, &rbl->rar);

  if (read_error) {
    return read_error;
  }

  d = fscanf(file, "%d ", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0 || tmp > RB_MAX) {                          /* be reasonable */
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  rbl->cnt = tmp;                                        /* list not empty */

  CALLOC(rb_def, rbl->cnt, rbl->rb);

  for (j = 0; j < rbl->cnt; j++) {
    read_error = read_rb_def(file, &rbl->rb[j]);

    if (read_error) {
      break;
    }
  }

  return read_error;
}



/* Reads area from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to area
Output: - area updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_area_def(FILE *file, area_def *ar)
{
  int d, tmp1, tmp2;

  d = fscanf(file, "%d%d\n", &tmp1, &tmp2);

  if (d != 2) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp1 <= 0 || tmp2 <= 0) {
    return 1;                                    /* unexpected input value */
  }

  ar->width = tmp1;
  ar->height = tmp2;

  return 0;
}



/* Reads a RB from a file.

Revision: 01.10.2002 E.Sack

Input : - pointer to stream
        - pointer to RB
Output: - RB updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_rb_def(FILE *file, rb_def *rbd)
{
  int d, tmp, read_error = 0;

  assert(file);
  assert(rbd);

  init_rb_def(rbd);

  d = fscanf(file, "%d ", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0 || (RB_STATUS)tmp > CHAIN) {
    return 1;                                    /* unexpected input value */
  }

  rbd->rld = (RB_STATUS)tmp;

  read_error = read_sector_def(file, &rbd->sc);

  if (read_error) {
    return read_error;
  }

  read_error = read_ival_lst(file, &rbd->gridy);

  return read_error;
}



/* Reads interval list from a file.

Revision: 05.01.2003 E.Sack

Input : - pointer to stream
        - pointer to interval list
Output: - interval list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_ival_lst(FILE *file, ival_lst *ivl)
{
  int j, d, tmp, read_error = 0;

  assert(file);
  assert(ivl);

  init_ival_lst(ivl);

  d = fscanf(file, "%d ", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0) {
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  ivl->cnt = tmp;                                        /* list not empty */

  CALLOC(ival_def, ivl->cnt, ivl->ival);

  for (j = 0; j < ivl->cnt; j++) {
    read_error = read_ival_def(file, &ivl->ival[j]);

    if (read_error) {
      return read_error;
    }
  }

  return read_error;
}



/* Reads a sector from a file.

Revision: 01.10.2002 E.Sack

Input : - pointer to stream
        - pointer to sector
Output: - sector updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_sector_def(FILE *file, sector_def *sec)
{
  int read_error = 0;

  assert(file);
  assert(sec);

  init_sector_def(sec);

  read_error = read_ival_def(file, &sec->column);

  if (read_error) {
    return read_error;
  }

  read_error = read_ival_def(file, &sec->line);

  return read_error;
}



/* Reads an interval from a file.

Revision: 01.10.2002 E.Sack

Input : - pointer to stream
        - pointer to interval
Output: - interval updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_ival_def(FILE *file, ival_def *iv)
{
  int d, tmp1, tmp2;

  init_ival_def(iv);

  d = fscanf(file, "%d %d ", &tmp1, &tmp2);

  if (d != 2) {
    return 1;                                   /* unexpected input format */
  }

  iv->start = tmp1;
  iv->end = tmp2;

  return 0;
}



/* Reads candidate list from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to candidate list
Output: - candidate list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_cand_lst(FILE *file, cand_lst *kl)
{
  int d, j, tmp, read_error = 0;

  assert(file);
  assert(kl);

  init_cand_lst(kl);

  d = fscanf(file, "%d ", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0) {
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  kl->cnt = tmp;

  CALLOC(cand_def, kl->cnt, kl->c);

  for (j = 0; j < kl->cnt; j++) {
    read_error = read_cand_def(file, &kl->c[j]);

    if (read_error) {
      return read_error;
    }
  }

  return 0;
}



/* Reads a candidate from a file.

Revision: 23.02.2003 E.Sack

Input : - pointer to stream
        - pointer to candidate
Output: - candidate updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_cand_def(FILE *file, cand_def *kr)
{
  int d, tmp1, tmp2, tmp3, read_error = 0;

  assert(file);
  assert(kr);

  init_cand_def(kr);

  read_error = read_cv_def_i(file, &kr->m);

  if (read_error) {
    return read_error;
  }

  d = fscanf(file, "%d%d%d ", &tmp1, &tmp2, &tmp3);

  if (d != 3) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp1 < 0) {
    return 1;                                    /* unexpected input value */
  }

  if(!(tmp2>=0 || 
       tmp2==SUBSTRB_FLAG || 
       tmp2==SUBSTNB_FLAG || 
       tmp2==SUBSTUC_FLAG)) {
    return 1;                                    /* unexpected input value */
  }

  if(!(tmp3>=0 || 
       tmp3==SUBCLS_FLAG)) {
    return 1;                                    /* unexpected input value */
  }

  kr->rbn = tmp1;
  kr->n = tmp2;
  kr->c = tmp3;

  read_error = read_ival_def(file, &kr->line);

  if (read_error) {
    return read_error;
  }

  read_error = read_sector_def(file, &kr->roi);

  if (read_error) {
    return read_error;
  }

  read_error = read_sector_def(file, &kr->ext);

  if (read_error) {
    return read_error;
  }

  read_error = read_sector_def(file, &kr->knl);

  if (read_error) {
    return read_error;
  }

  read_error = read_sector_def(file, &kr->hdx);

  if (read_error) {
    return read_error;
  }

  read_error = read_int_lst(file, &kr->crb);

  if (read_error) {
    return read_error;
  }

  CALLOC(cand_lst, 1, kr->alt);
  read_error = read_cand_lst(file, kr->alt);

  if (!kr->alt->cnt) {
    FREEMEM(kr->alt);
  }

  if (read_error) {
    return read_error;
  }

  return read_error;
}



/* Reads an integer list from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to integer list
Output: - integer list updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_int_lst(FILE *file, int_lst *il)
{
  int d, j, tmp;

  assert(file);
  assert(il);

  init_int_lst(il);

  d = fscanf(file, "%d ", &tmp);

  if (d != 1) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp < 0) {
    return 1;                                    /* unexpected input value */
  }

  if (tmp == 0) {
    return 0;                                                /* list empty */
  }

  il->cnt = tmp;                                         /* list not empty */

  CALLOC(int, il->cnt, il->nr);

  for (j = 0; j < il->cnt; j++) {
    d = fscanf(file, "%d ", &il->nr[j]);

    if (d != 1) {
      return 1;                                 /* unexpected input format */
    }
  }

  return 0;
}



/* Reads correlation features from a file.

Revision: 29.09.2002 E.Sack

Input : - pointer to stream
        - pointer to correlation features
Output: - correlation features updated
        - Flag (0: input succeeded, >0: input not succeeded)
*/
static int
read_cv_def_i(FILE *file, cv_def_i *kdi)
{
  int    d, tmp2, tmp3;
  double tmp1;

  assert(file);
  assert(kdi);

  d = fscanf(file, "%lf%d%d ", &tmp1, &tmp2, &tmp3);

  if (d != 3) {
    return 1;                                   /* unexpected input format */
  }

  if (tmp1 < 0.0) {
    return 1;                                    /* unexpected input value */
  }

  kdi->cv = tmp1;
  kdi->x = tmp2;
  kdi->y = tmp3;

  return 0;
}



/* Reads a string from the current file offset and reposition the offset
   after completion. The purpose is to avoid a possible buffer overflow.

Input : - pointer to file
        - length
Output: - flag (0: input acceptable, 1: input not acceptable)
*/
static int
test_string_length(FILE *file, int len)
{
  int    d, err = 0, test_len = 2*len;
  char   *buf1 = 0, *buf2 = 0, *ptr = 0;
  fpos_t pos;

  assert(file);
  assert(len>1);                                  /* more than trailing \0 */

  d = fgetpos(file, &pos);                          /* current file offset */

  if (d) {
    return 1;
  }

  CALLOC(char, test_len, buf1);

  ptr = fgets(buf1, test_len, file);

  if (!ptr) {
    err++;
  }
  else {
    assert(ptr == buf1);

    CALLOC(char, test_len, buf2);

    d = sscanf(buf1, "%s", buf2);

    if (d != 1) {
      err++;
    }
    else {
      d = strlen(buf2);

      if (d > len-1) {
	err++;
      }
    }

    FREEMEM(buf2);
  }

  FREEMEM(buf1);

  d = fsetpos(file, &pos);                            /* reset file offset */

  if (d) {
    err++;
  }

  return err;
}



/* Reads a line from a file.

Revision: 07.11.2003 E.Sack

Input : - pointer to stream
Output: - pointer to string
*/
static char*
read_line(FILE *file)
{
  int  character;
  char *line = NULL;

  while (1) {
    character = fgetc(file); 

    if (character == EOF) {
      break;
    }

    if (character == '\n') {
      break;
    }

    add_character(character, &line);
  }

  return line;
}



/* Closes a file stream.

Revision: 07.11.2003 E.Sack

Input : - pointer to file stream
        - pointer to file name
Output: - flag (0: file closed without error, 1: file closing error)
*/
int
xplab_close_file(FILE *file, const char *filename)
{
  int  d = 0, err;

  if (file) {
    err = check_name(filename);

    d = fclose(file);

    if (d) {
      if (err) {
	string_to_log(main_log, "\nERROR: cannot close file\n", "", "");
      }
      else {
	string_to_log(main_log, "\nERROR: cannot close file '", filename, "'\n");
      }
    }
  }

  return d;
}



/* Opens a file stream.

Revision: 09.01.2005 E.Sack

Input : - pointer to file name
Output: - pointer to file stream (0: file not opened)
*/
FILE *
xplab_open_file(const char *filename, char *oflag)
{
  int  d;
  FILE *file = NULL;

  d = check_name(filename);

  if (d) {
    return file;
  }

  d = check_name(oflag);

  if (d) {
    return file;
  }

  if (strstr(oflag,"r")) {
    d = access(filename, F_OK);                        /* read access file */

    if (d) {
      string_to_log(main_log, "WARNING: file '", filename, 
		    "' not found\n");
      return file;
    }
  }
  else if (strstr(oflag,"w")) {
    d = access(".", W_OK);                       /* write access directory */

    if (d) {
      string_to_log(main_log, 
		    "WARNING: no write permission in current directory\n","","");
      return file;
    }

    d = access(filename, F_OK);

    if (!d) {                                               /* file exists */
      d = access(filename, W_OK);                     /* write access file */

      if (d) {
	string_to_log(main_log, 
		      "WARNING: no write permission to file '", filename,"'\n");
	return file;
      }
    }
  }                                           /* something else, no checks */

  file = fopen(filename, oflag);
  if (!file) {
    string_to_log(main_log, "WARNING: cannot open '", filename, 
		  (strstr(oflag,"w")==NULL) ? "' for read\n":"' for write\n");
    return file;
  }

  return file;
}



