/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_image.h"
#include "xplab_algorithm.h"
#include "xplab_report.h"
#include "xplab_maintenance_fe.h"
#include "xplab_dchannel.h"
#include "xplab_tprint.h"
#include "xplab_preferences_fe.h"
#include "xplab_dbio.h"

/* local prototypes ********************************************************/
static char *get_magic_pnm(PNM_FORMAT form);
static void write_pnm_header(FILE *bild_output, PNM_FORMAT form, area_def clp);
static int check_pic_setup(char *name, pic_def src, area_def clp, int x_offset, 
			   int y_offset, PNM_FORMAT form);
static PNM_FORMAT test_format_col_def(pic_def src, int rawbit);
static PNM_FORMAT test_format_grey(PICT *mem, area_def mem_size, area_def clp, 
				   int x_offset, int y_offset, int rawbit);
static void threshold_info(hst_def hst, int threshold, int histo_thold);
static int count_float(double max, double add, double *value, int carry);
static pic_def read_P1_data(FILE *file, area_def size);
static pic_def read_P2_data(FILE *file, area_def size);
static pic_def read_P3_data(FILE *file, area_def size);
static pic_def read_PXPLAB2_data(FILE *file, area_def size);
static pic_def read_P4_data(FILE *file, area_def size);
static pic_def read_P5_data(FILE *file, area_def size);
static pic_def read_PXPLAB5_data(FILE *file, area_def size);
static pic_def read_P6_data(FILE *file, area_def size);
static void write_P1_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			  int y_offset);
static void write_P2_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			  int y_offset);
static void write_P3_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			  int y_offset);
static void write_PXPLAB2_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			       int y_offset);
static void write_P4_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			  int y_offset);
static void write_P5_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			  int y_offset);
static void write_P6_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			  int y_offset);
static void write_PXPLAB5_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
			       int y_offset);
static int max_of_color_sum(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
			    int dx, int dy);
static int max_of_grey_sum(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
			   int dx, int dy);
static void norm_color_pic(pic_def dst, area_def cut, int sx, int sy, int dx, 
			   int dy, double norm);
static void norm_grey_pic(pic_def dst, area_def cut, int sx, int sy, int dx, int dy, 
			  double norm);
static int trueline(char s[], int max, FILE *fp);
static pic_def session_to_image(pic_def ses);
static int session_to_bitmap(pic_def sessimg);
static void squeeze_hst_def(hst_def *hin, int max_len);
static gboolean squeeze_double_lst(double_lst *dlst, int max_len);
/* Functions **************************************************************/

/* Converts grey value image to color image.

Input : - image
Output: - color image 
Postcondition: - no output if input isn't grey value.
*/
pic_def
grey_to_color(pic_def pin)
{
  int        x, xx, y, yy, del_ibmp = 0;
  pic_def    pout, ibmp;
  PNM_FORMAT input_type;
  PICT       grey;

  init_pic_def(&pout);
  ibmp = pin;

  input_type = automatic_format(pin, pin.ar, 0, 0, xctrl.cfg.rawbit);

  if (input_type != P5 && input_type != P2) {
    if (input_type != P1 && input_type != P4) {       /* input is not grey */
      return pout;                                  /* input is not bitmap */
    }
    else {
      ibmp = bitmap_to_grey(pin);                   /* input is bitmap b/w */
      del_ibmp = 1;
    }
  }

  pout.ar = ibmp.ar;
  CALLOC(col_def, AREA_SIZE(pout.ar), pout.col);

  for (y = 0; y < ibmp.ar.height; y++) {
    yy = y*ibmp.ar.width;

    for (x = 0; x < ibmp.ar.width; x++) {
      xx = x + yy;
      grey = ibmp.mem[xx];

      pout.col[xx].r = grey;
      pout.col[xx].g = grey;
      pout.col[xx].b = grey;
    }
  }

  if (del_ibmp == 1) {
    free_pic_def(&ibmp);
  }

  return pout;
}



/* Converts binary image to grey value image.

Input : - image
Output: - grey image 
Postcondition: - no output if input isn't binary.
*/
pic_def
bitmap_to_grey(pic_def pin)
{
  int        x, xx, y, yy;
  pic_def    pout;
  PNM_FORMAT input_type;

  init_pic_def(&pout);

  input_type = automatic_format(pin, pin.ar, 0, 0, xctrl.cfg.rawbit);

  if (input_type != P4 && input_type != P1) {
    return pout;
  }

  pout.ar = pin.ar;
  CALLOC(PICT, AREA_SIZE(pout.ar), pout.mem);

  for (y = 0; y < pin.ar.height; y++) {
    yy = y*pin.ar.width;

    for (x = 0; x < pin.ar.width; x++) {
      xx = x + yy;

      pout.mem[xx] = (pin.mem[xx] == FOREGROUND) ? 0 : PICT_MAX;
    }
  }

  return pout;
}



/* Adds the clipping of two color images. The destination image
   is cut to pixel maximum.

Revision: 28.12.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
void 
add_color(pic_def src, pic_def dst, area_def cut, int sx, int sy, int dx, 
	  int dy)
{
  int x, y, ydy, ysy, xd, xs, sum;

  assert(src.col);
  assert(dst.col);
  assert(cut.width >= 0); 
  assert(cut.height >= 0); 
  assert(dx >= 0); 
  assert(dy >= 0); 
  assert(sx >= 0); 
  assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;
      xs = x+sx + ysy;

      sum = (int)dst.col[xd].r + (int)src.col[xs].r;
      dst.col[xd].r = (PICT)XPMIN((int)PICT_MAX, sum);

      sum = (int)dst.col[xd].g + (int)src.col[xs].g;
      dst.col[xd].g = (PICT)XPMIN((int)PICT_MAX, sum);

      sum = (int)dst.col[xd].b + (int)src.col[xs].b;
      dst.col[xd].b = (PICT)XPMIN((int)PICT_MAX, sum);
    }
  }
}



/* Adds the clipping of a binary to color image. Only foreground
   pixel of binary image are patched.

Revision: 30.12.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
void 
add_bitmap_to_color(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
		    int dx, int dy, col_def foreground_color)
{
  int  x, y, ydy, ysy, xd;
  PICT bin;

  assert(src.mem);
  assert(dst.col);
  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;
      bin = src.mem[x+sx + ysy];

      dst.col[xd] = (bin == FOREGROUND) ? foreground_color : dst.col[xd];
    }
  }
}



/* Adds the clipping of two color images. The destination image
   is normalized.

Revision: 28.12.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
void 
add_color_norm(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
	       int dx, int dy)
{
  int     x, y, ydy, ysy, xd, xs, sum, max;
  double  norm;
  pic_def ptemp;

  init_pic_def(&ptemp);

  assert(src.col);
  assert(dst.col);
  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  max = max_of_color_sum(src, dst, cut, sx, sy, dx, dy);
  max = max?max:1;

  if (max < PICT_MAX) {
    norm = 1.0;                                            /* don't spread */
  }
  else {
    norm = (double)PICT_MAX/(double)max;                    /* but squeeze */
    norm = XPMAX(norm, 1.0);                                /* don't erase */
  }

  copy_pic_def(&src, &ptemp);
  norm_color_pic(ptemp, ptemp.ar, 0, 0, 0, 0, norm);   /* source unchanged */
  norm_color_pic(dst, dst.ar, 0, 0, 0, 0, norm);

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;
      xs = x+sx + ysy;

      sum = (int)dst.col[xd].r + (int)src.col[xs].r;
      dst.col[xd].r = (PICT)XPMIN((int)PICT_MAX, sum);

      sum = (int)dst.col[xd].g + (int)src.col[xs].g;
      dst.col[xd].g = (PICT)XPMIN((int)PICT_MAX, sum);

      sum = (int)dst.col[xd].b + (int)src.col[xs].b;
      dst.col[xd].b = (PICT)XPMIN((int)PICT_MAX, sum);
    }
  }

  free_pic_def(&ptemp);
}



/* Normalizes the clipping of a color image. 

Revision: 28.12.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
static void
norm_color_pic(pic_def dst, area_def cut, int sx, int sy, int dx, int dy, 
	       double norm)
{
  int x, y, ydy, xd;

  assert(dst.col);
  assert(norm>=1.0);
  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;

      dst.col[xd].r = (PICT)((double)dst.col[xd].r * norm);
      dst.col[xd].g = (PICT)((double)dst.col[xd].g * norm);
      dst.col[xd].b = (PICT)((double)dst.col[xd].b * norm);
    }
  }
}



/* Normalizes the clipping of a grey image. 

Revision: 10.01.2003 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
static void
norm_grey_pic(pic_def dst, area_def cut, int sx, int sy, int dx, int dy, 
	      double norm)
{
  int x, y, ydy, xd;

  assert(dst.mem);
  assert(norm>=1.0);
  assert(cut.width >= 0); 
  assert(cut.height >= 0); 
  assert(dx >= 0); 
  assert(dy >= 0); 
  assert(sx >= 0); 
  assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;

      dst.mem[xd] = (PICT)((double)dst.mem[xd] * norm);
    }
  }
}



/* Calculates in a clipping the maximum of a color sum of two images. 

Revision: 28.12.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
static int
max_of_color_sum(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
		 int dx, int dy)
{
  int x, y, ydy, ysy, xd, xs, max = 0;

  assert(src.col);
  assert(dst.col);
  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;
      xs = x+sx + ysy;

      max = XPMAX(max, (int)dst.col[xd].r + (int)src.col[xs].r);
      max = XPMAX(max, (int)dst.col[xd].g + (int)src.col[xs].g);
      max = XPMAX(max, (int)dst.col[xd].b + (int)src.col[xs].b);
    }
  }

  return max;
}



/* Calculates in a clipping the maximum of a grey sum of two images. 

Revision: 10.02.2003 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
static int
max_of_grey_sum(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
	        int dx, int dy)
{
  int x, y, ydy, ysy, xd, xs, max = 0;

  assert(src.mem);
  assert(cut.width >= 0); 
  assert(cut.height >= 0); 
  assert(dx >= 0); 
  assert(dy >= 0); 
  assert(sx >= 0); 
  assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;
      xs = x+sx + ysy;

      max = XPMAX(max, (int)dst.mem[xd] + (int)src.mem[xs]);
    }
  }

  return max;
}



/* Adds the clipping of two grey value images. The destination image
   is cut to pixel maximum.

Revision: 28.12.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
void 
add_grey(pic_def src, pic_def dst, area_def cut, int sx, int sy, int dx, 
	 int dy)
{
  int x, y, ydy, ysy, xx, sum;

  assert(src.mem);
  assert(dst.mem);
  assert(cut.width >= 0); 
  assert(cut.height >= 0); 
  assert(dx >= 0); 
  assert(dy >= 0); 
  assert(sx >= 0); 
  assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xx = x+dx + ydy;
      sum = (int)dst.mem[xx] + (int)src.mem[x+sx + ysy];
      dst.mem[xx] = (PICT)XPMIN(sum, (int)PICT_MAX);
    }
  }
}



/* Adds the clipping of two color images. The destination image
   is normalized.

Revision: 28.12.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
void 
add_grey_norm(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
	      int dx, int dy)
{
  int     x, y, ydy, ysy, xd, xs, sum, max;
  double  norm;
  pic_def ptemp;

  init_pic_def(&ptemp);

  assert(src.mem);
  assert(dst.mem);
  assert(cut.width >= 0); 
  assert(cut.height >= 0); 
  assert(dx >= 0); 
  assert(dy >= 0); 
  assert(sx >= 0); 
  assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  max = max_of_grey_sum(src, dst, cut, sx, sy, dx, dy);
  max = max?max:1;

  if (max < PICT_MAX) {
    norm = 1.0;                                            /* don't spread */
  }
  else {
    norm = (double)PICT_MAX/(double)max;                    /* but squeeze */
    norm = XPMAX(norm, 1.0);                                /* don't erase */
  }

  copy_pic_def(&src, &ptemp);
  norm_grey_pic(ptemp, ptemp.ar, 0, 0, 0, 0, norm);    /* source unchanged */
  norm_grey_pic(dst, dst.ar, 0, 0, 0, 0, norm);

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xd = x+dx + ydy;
      xs = x+sx + ysy;

      sum = (int)dst.mem[xd] + (int)src.mem[xs];
      dst.mem[xd] = (PICT)XPMIN((int)PICT_MAX, sum);
    }
  }

  free_pic_def(&ptemp);
}



/* Adds the clipping of a binary to grey value image. Only foreground
   pixel of binary image are patched.

Revision: 30.12.2002 E.Sack

Input : - binary source image
        - grey destination image
        - clipping
        - offset of clipping regarding source and destination
Output: - destination image updated
*/
void 
add_bitmap_to_grey(pic_def src, pic_def dst, area_def cut, int sx, int sy, 
		int dx, int dy)
{
  int  x, y, ydy, ysy, xx;
  PICT bin;

  assert(src.mem);
  assert(dst.mem);
  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xx = x+dx + ydy;
      bin = src.mem[x+sx + ysy];

      dst.mem[xx] = (bin == FOREGROUND) ? 0 : dst.mem[xx];
    }
  }
}



/* The areas of two bitmaped images are added.

Revision: 24.02.2002 E.Sack

Input : - source bitmap
        - destination bitmap
        - clipping
        - offsets
Output: - destination bitmap updated
*/
void 
add_bmps(pic_def src, pic_def dst, area_def cut, int sx, int sy, int dx, 
	 int dy) 
{
  int  x, y, ydy, ysy, xx;
  PICT bin;

  assert(src.mem);
  assert(dst.mem);
  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cut.width  <= dst.ar.width); 
  assert(dy+cut.height <= dst.ar.height); 
  assert(sx+cut.width  <= src.ar.width); 
  assert(sy+cut.height <= src.ar.height); 

  for (y = 0; y < cut.height; y++) {
    ydy = (y+dy)*dst.ar.width;
    ysy = (y+sy)*src.ar.width;

    for (x = 0; x < cut.width; x++) {
      xx = x+dx + ydy;
      bin = src.mem[x+sx + ysy];

      dst.mem[xx] = (bin == FOREGROUND) ? bin : dst.mem[xx];
    }
  }
}



/* Sums all pixels 'OR' in an image clipping.

Revision: 23.08.2003 E.Sack

Input : - clipping
        - image
Output: - 'OR' of all pixels
*/
PICT
xget_pixel_or(sector_def f, pic_def pic)
{
  int  z, zz, s, width = pic.ar.width;
  PICT pixel_or = 0;

  assert(f.column.start <= f.column.end);
  assert(f.line.start <= f.line.end);
  assert(f.column.start >= 0);
  assert(f.line.start >= 0);
  assert(f.column.end < pic.ar.width);
  assert(f.line.end < pic.ar.height);

  for (z = f.line.start; z <= f.line.end; z++) {
    zz = z*width;

    for (s = f.column.start; s <= f.column.end; s++) {
      pixel_or |= pic.mem[s+zz];
    }
  }

  return pixel_or;
}



/* Exchanges pixel values.

Revision: 16.08.2003 E.Sack

Input : - clipping
        - search mark
        - substitution mark
        - image
        - flag (0: exchange inside clipping, 1: exchange outside clipping)
Output: - image updated
        - flag (0: no mark found, 1: mark found)
*/
int
set_pixel(sector_def f, PICT from, PICT to, pic_def pic, int outside)
{
  int z, zz, s, mark_found = 0, width = pic.ar.width;

  assert(f.column.start <= f.column.end);
  assert(f.line.start <= f.line.end);
  assert(f.column.start >= 0);
  assert(f.line.start >= 0);
  assert(f.column.end < pic.ar.width);
  assert(f.line.end < pic.ar.height);

  switch (outside) {
  case 0:                                      /* exchange inside clipping */
    for (z = f.line.start; z <= f.line.end; z++) {
      zz = z*width;

      for (s = f.column.start; s <= f.column.end; s++) {
	if (pic.mem[s+zz] == from) {                           /* set mark */
	  pic.mem[s+zz] = to;
	  mark_found = 1;
	}
      }
    }

    break;

  case 1:                                     /* exchange outside clipping */
    for (z = 0; z < pic.ar.height; z++) {
      zz = z*width;

      if (z >= f.line.start && z <= f.line.end) {
	for (s = 0; s < f.column.start; s++) {
	  if (pic.mem[s+zz] == from) {                    /* set mark left */
	    pic.mem[s+zz] = to;
	    mark_found++;
	  }
	}

	for (s = f.column.end+1; s < pic.ar.width; s++) {
	  if (pic.mem[s+zz] == from) {                   /* set mark right */
	    pic.mem[s+zz] = to;
	    mark_found++;
	  }
	}

	continue;
      }

      for (s = 0; s < pic.ar.width; s++) {
	if (pic.mem[s+zz] == from) {              /* set mark top & bottom */
	  pic.mem[s+zz] = to;
	  mark_found++;
	}
      }
    }

    break;

  default: assert(0); break;
  }

  return mark_found;
}



/* Changes color.

Revision: 30.07.2002 E.Sack

Input : - sector
        - source color
        - destination color
        - image
Output: - sector updated
        - flag (0: no source color found, 1: source color found)
*/
int 
change_color(sector_def f, col_def from, col_def to, pic_def pic)
{
  int z, s, mark_found, width = pic.ar.width, idx;

  assert(f.column.start <= f.column.end);
  assert(f.line.start <= f.line.end);
  assert(f.column.start >= 0);
  assert(f.line.start >= 0);
  assert(f.column.end < pic.ar.width);
  assert(f.line.end < pic.ar.height);

  for (z = f.line.start, mark_found = 0; z <= f.line.end; z++) {
    for (s = f.column.start; s <= f.column.end; s++) {
      idx = s+z*width;

      if (pic.col[idx].r == from.r &&
          pic.col[idx].g == from.g &&
	  pic.col[idx].b == from.b) {
	pic.col[idx] = to;
	mark_found = 1;
      }
    }
  }

  return mark_found;
}



/* Wrapper for set_pixel(), exchanges certain pixel values.

Revision: 23.08.2002 E.Sack

Input : - image
        - certain pixel value
        - exchange value
Output: - image updated
        - flag (0: no certain pixel values found, 1: pixel values found)
*/
int 
set_all_pixel(pic_def pic, PICT from, PICT to)
{
  int        d;
  sector_def sec;

  assert(pic.ar.width > 0);
  assert(pic.ar.height > 0);

  AREA_TO_SECTOR(pic.ar, sec);

  d = set_pixel(sec, from, to, pic, 0);

  return d;
}



/* Resets pixels.

Revision: 26.08.2003 E.Sack

Input : - clipping
        - from-marking
        - to-marking
        - image
Output: - clipping updated
        - flag (0: no mark found, 1: mark found)
*/
int
unset_pixel(sector_def f, PICT from, PICT to, pic_def pic)
{
  int z, zz, s, mark_found, width = pic.ar.width;

  assert(f.column.start <= f.column.end);
  assert(f.line.start <= f.line.end);
  assert(f.column.start >= 0);
  assert(f.line.start >= 0);
  assert(f.column.end < pic.ar.width);
  assert(f.line.end < pic.ar.height);

  for (z = f.line.start, mark_found = 0; z <= f.line.end; z++) {
    zz = z*width;

    for (s = f.column.start; s <= f.column.end; s++) {
      if (pic.mem[s+zz] != from) {                        /* unset all but */
	pic.mem[s+zz] = to;
	mark_found = 1;
      }
    }
  }

  return mark_found;
}



/* Copies an image without its empty border.

Revision: 03.03.2002 E.Sack

Input : - image source
        - pointer to destination image
Output: - destination image updated
        - Flag (0: nothing copied, 1: image copied)
Precondition: - source and destination may be identic
Postcondition: nothing changed if flag==0
*/
int
xcopy_and_fit_pic(pic_def src, pic_def *dst)
{
  int i, ii, j, y_min, y_max, x_min, x_max;

  for (i = 0, y_min = -1; i < src.ar.height; i++) {
    ii = i*src.ar.width;

    for (j = 0; j < src.ar.width; j++) {
      if (src.mem[j + ii] == FOREGROUND) {
        y_min = i;                                         /* upper border */
	i = src.ar.height;                               /* stop condition */
 	break;
      }
    }
  }

  if (y_min < 0) {
    return 0;                                               /* image empty */
  }

  for (i = y_max = src.ar.height-1; i >= 0; i--) {
    ii = i*src.ar.width;

    for (j = 0; j < src.ar.width; j++) {
      if (src.mem[j + ii] == FOREGROUND) {
        y_max = i;                                         /* lower border */
	i = -1;                                          /* stop condition */
	break;
      }
    }
  }

  for (j = x_min = 0; j < src.ar.width; j++) {
    for (i = y_min; i <= y_max; i++) {
      if (src.mem[j + i*src.ar.width] == FOREGROUND) {
	x_min = j;                                          /* left border */
	j = src.ar.width;                                /* stop condition */
	break;
      }
    }
  }

  for (j = x_max = src.ar.width-1; j >= 0; j--) {
    for (i = y_min; i <= y_max; i++) {
      if (src.mem[j + i*src.ar.width] == FOREGROUND) {
        x_max = j;                                         /* right border */
	j = -1;                                          /* stop condition */
	break;
      }
    }
  }

  {
    int     bvl_true = src.bvl ? 1 : 0;
    pic_def z;

    init_pic_def(&z);

    z.ar.width  = x_max - x_min + 1;                           /* new size */
    z.ar.height = y_max - y_min + 1;

    z.org.x = src.org.x + x_min;                   /* adapt absolut origin */
    z.org.y = src.org.y + y_min;

    CALLOC(PICT, AREA_SIZE(z.ar), z.mem);

    if (bvl_true) {
      CALLOC(PICT, AREA_SIZE(z.ar), z.bvl);    /* source has second memory */
    }

    copy_bmps(src, z, z.ar, x_min, y_min, 0, 0);       /* temporary buffer */

    FREEMEM(dst->mem);/* may be image source to make on-place op. possible */
    FREEMEM(dst->bvl);
    dst->ar = z.ar;
    dst->org = z.org;
    CALLOC(PICT, AREA_SIZE(dst->ar), dst->mem);

    if (bvl_true) {
      CALLOC(PICT, AREA_SIZE(dst->ar), dst->bvl);
    }

    copy_bmps(z, *dst, dst->ar, 0, 0, 0, 0);

    FREEMEM(z.mem);
    FREEMEM(z.bvl);
  }

  return 1;
}



/* Copies image data clipping into a float matrix.

Input : - source image
        - pointer to float array
        - clipping
        - offset
Output: - array updated
*/
void 
copy_bmp_to_matrix(pic_def src, double **dst, area_def cut, 
		   int sx, int sy, int dx, int dy)
{
  int i, j, pval, sw = src.ar.width, y;

  assert(cut.width >= 0); assert(cut.height >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(sx+cut.width  <= sw); 
  assert(sy+cut.height <= src.ar.height); 
  assert(src.mem);
  assert(dst);

  for (i = 0; i < cut.height; i++) {
    y = (i+sy)*sw;

    for (j = 0; j < cut.width; j++) {
      pval = src.mem[j+sx + y];
      assert(pval == FOREGROUND || pval == BACKGROUND); /* test for bitmap */
      dst[i+dy][j+dx] = pval == FOREGROUND ? 1.0 : 0.0;
    }
  }
}



/* Copies bitmap memory.

Revision: 05.08.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - clipping offset
Output: - destination image updated
*/
void 
copy_bmps(pic_def src, pic_def dst, area_def cut, int sx, int sy, int dx, 
	  int dy)
{
  int i, id, is, j, dd, ss, dw = dst.ar.width, sw = src.ar.width;
  int ch = cut.height, cw = cut.width;

  assert(src.mem);
  assert(dst.mem);
  assert(cw >= 0); assert(ch >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cw <= dw); 
  assert(dy+ch <= dst.ar.height); 
  assert(sx+cw <= sw); 
  assert(sy+ch <= src.ar.height); 

  dd = dy*dw + dx;
  ss = sy*sw + sx;

  for (i = 0; i < ch; i++) {
    id = i*dw + dd;
    is = i*sw + ss;

    for (j = 0; j < cw; j++) {
      dst.mem[j + id] = src.mem[j + is];
    }
  }

  if (!src.bvl || !dst.bvl) {
    return;                                         /* no b-values wanted */
  }

  for (i = 0; i < ch; i++) {
    id = i*dw + dd;
    is = i*sw + ss;

    for (j = 0; j < cw; j++) {
      dst.bvl[j + id] = src.bvl[j + is];
    }
  }
}

   

/* Copies color memory.

Revision: 05.01.2003 E.Sack

Input : - source image
        - destination image
        - clipping
        - clipping offset
Output: - destination image updated
*/
void 
copy_color(pic_def src, pic_def dst, area_def cut, int sx, int sy, int dx, 
	   int dy)
{
  int i, id, is, j, dd, ss, dw = dst.ar.width, sw = src.ar.width;
  int ch = cut.height, cw = cut.width;

  assert(src.col);
  assert(dst.col);
  assert(cw >= 0); assert(ch >= 0); 
  assert(dx >= 0); assert(dy >= 0); 
  assert(sx >= 0); assert(sy >= 0); 
  assert(dx+cw <= dw); 
  assert(dy+ch <= dst.ar.height); 
  assert(sx+cw <= sw); 
  assert(sy+ch <= src.ar.height); 

  dd = dy*dw + dx;
  ss = sy*sw + sx;

  for (i = 0; i < ch; i++) {
    id = i*dw + dd;
    is = i*sw + ss;

    for (j = 0; j < cw; j++) {
      dst.col[j + id] = src.col[j + is];
    }
  }
}

   

/* Copys grey to color images.

Revision: 04.08.2002 E.Sack

Input : - source image
        - destination image
        - clipping
        - foreground color        
        - background color
        - clipping offsets for images
        - rawbit
        - verbose flag
Output: - destination image updated
*/
void
copy_grey_to_color(pic_def src, pic_def dst, area_def clp, col_def color, 
		   col_def backgnd, int src_x_offset, int src_y_offset, 
		   int dst_x_offset, int dst_y_offset, int rawbit)
{
  int        y, ys, yd, xd, x, src_start, dst_start;
  int        src_width = src.ar.width, dst_width = dst.ar.width;
  PICT       pixel;
  PNM_FORMAT format;

  assert(src.mem);
  assert(dst.col);
  assert(clp.width >= 0); 
  assert(clp.height >= 0); 
  assert(dst_x_offset >= 0); 
  assert(dst_x_offset < dst_width);
  assert(dst_y_offset >= 0); 
  assert(dst_y_offset < dst.ar.height); 
  assert(src_x_offset >= 0); 
  assert(src_x_offset < src_width);
  assert(src_y_offset >= 0);
  assert(src_y_offset < src.ar.height); 
  assert(dst_x_offset+clp.width  <= dst_width); 
  assert(dst_y_offset+clp.height <= dst.ar.height); 
  assert(src_x_offset+clp.width  <= src_width); 
  assert(src_y_offset+clp.height <= src.ar.height); 

  format = test_format_grey(src.mem, src.ar, clp, src_x_offset, 
			    src_y_offset, rawbit);

  src_start = src_x_offset + src_y_offset*src_width;
  dst_start = dst_x_offset + dst_y_offset*dst_width;

  for (y = 0; y < clp.height; y++) {
    ys = y*src_width + src_start;
    yd = y*dst_width + dst_start;

    for (x = 0; x < clp.width; x++) {
      xd = x+yd;
      pixel = src.mem[x+ys];

      if (format == P4 || format == P1) {               /* bitmap to color */
        dst.col[xd] = (pixel == FOREGROUND) ? color : backgnd;
	continue;
      }

      if (format == P5 || format == P2) {                 /* grey to color */
        dst.col[xd].r = pixel;
        dst.col[xd].g = pixel;
        dst.col[xd].b = pixel;
	continue;
      }

      assert(0);                                     /* unknown PNM format */
    }
  }
}

   

/* Converts a gray/bmp-image to a colored image.

Revision: 06.01.2003 E.Sack

Input : - pointer to image
        - flag (TRUE: don't free BMP memory, else free)
Output: - image updated
Postcondition: - image bmp-buffer freed.
*/
void
convert_to_color(pic_def *ptmp, gboolean copy_bmp)
{
  pic_def ptmpcol;

  init_pic_def(&ptmpcol);

  if (ptmp->mem) {                                             /* gray image */
    ptmpcol = grey_to_color(*ptmp);
    assert(ptmpcol.col);

    if (copy_bmp == FALSE) {
      FREEMEM(ptmp->mem);                                 /* free bmp-buffer */
    }

    ptmp->col = ptmpcol.col;                            /* pointer hand-over */
    ptmpcol.col = NULL;                               /* delete pointer copy */
  }

  free_pic_def(&ptmpcol);
}



/* Sets an image color memory to a value.

Revision: 11.12.2004 E.Sack

Input : - size
        - pointer to memory
        - value
Output: - memory updated
*/
void
set_color(int len, col_def *dst, col_def value)
{
   while(len--) {
     *dst++ = value;
   }
}
   


/* Sets an image memory to a value.

Revision: 02.01.2003 E.Sack

Input : - size
        - pointer to memory
        - value
Output: - memory updated
*/
void
set_bitmap(int len, PICT *dst, PICT value)
{
   while(len--) {
     *dst++ = value;
   }
}
   


/* Verifies a bitmap.

Revision: 02.01.2003 E.Sack

Input : - image
Output: - verification successful or program execution stop
*/
void
verify_bitmap(pic_def src, area_def clp, int x_offset, int y_offset)
{
  int  y, yy, x, err = 0;
  PICT pixel;

  assert(src.mem);
  assert(clp.width > 0); assert(clp.height > 0); 
  assert(x_offset >= 0); assert(y_offset >= 0); 
  assert(x_offset+clp.width  <= src.ar.width); 
  assert(y_offset+clp.height <= src.ar.height); 
  assert(src.ar.width > 0);
  assert(src.ar.height > 0);

  for (y = 0; y < clp.height; y++) {
    yy = (y+y_offset)*src.ar.width;

    for (x = 0; x < clp.width; x++) {
      pixel = src.mem[x+x_offset + yy];

      if (pixel != BACKGROUND && pixel != FOREGROUND) {
	string_to_log(main_log, err?"":"\n", "", "");
	int_to_log(main_log, "ERROR: pixel[", x, ",");
	int_to_log(main_log, "", y, "] = ");
	int_to_log(main_log, "", pixel, " is not a bitmap as ought to be\n");
	err++;
      }
    }
  }
}



/* Checks for format and settings of a PNM image file.

Revision: 22.12.2004 E.Sack

Input : - pointer to filename
        - image
        - clipping
        - clipping offset
        - PNM format
Output: - messages to log
        - flag (0: (extended) PNM image, 1: not a (extended) PNM image)
*/
static int
check_pic_setup(char *name, pic_def src, area_def clp, int x_offset, 
		int y_offset, PNM_FORMAT format)
{
  assert(name);
  assert(PNMWIDTH > 1);
  assert(x_offset >= 0); assert(y_offset >= 0); 
  assert(x_offset+clp.width  <= src.ar.width); 
  assert(y_offset+clp.height <= src.ar.height); 

  if (AREA_SIZE(src.ar) <= 0) {
    string_to_log(main_log, "image for file '", name, "' is empty\n");
    return 1;
  }

  if (AREA_SIZE(clp) <= 0) {
    string_to_log(main_log, "image clipping for file '", name, "' is empty\n");
    return 1;
  }

  if (AREA_SIZE(clp) > MAXWIDTH*MAXHEIGHT  ||
      clp.width > XPMAX(MAXWIDTH, MAXHEIGHT) ||
      clp.height > XPMAX(MAXWIDTH, MAXHEIGHT)) {
    string_to_log(main_log, "ERROR: image size outrun:\n", "", "");
    int_to_log(main_log, "width ", clp.width, " ");
    int_to_log(main_log, "[", MAXWIDTH, "]  ");
    int_to_log(main_log, "height ", clp.height, " ");
    int_to_log(main_log, "[", MAXHEIGHT, "]\n");
    string_to_log(main_log, "file '", name, "' not writable\n");
    return 1;
  }

  switch(format) {
  case P1:
  case P4:
    verify_bitmap(src, clp, x_offset, y_offset);
    break;

  case PXPLAB2:
  case PXPLAB5:
  case P2:
  case P5:
    assert(src.mem);
    break;

  case P3:
  case P6:
    assert(src.col);
    break;

  default:
    int_to_log(main_log, "ERROR: unknown PNM format ID: ", format, "\n");
    return 1;
    break;
  }

  return 0;
}



/* Gets a format macro to string.

Revision: 05.08.2002 E.Sack

Input : - format
Output: - pointer to string
*/
static char *
get_magic_pnm(PNM_FORMAT form)
{
  char *magic_pbm = NULL;

  switch(form) {
  case P1:                                        /* ascii                */
    add_string("P1", &magic_pbm);
    break;

  case P2:                                        /* ascii                */
    add_string("P2", &magic_pbm);
    break;

  case PXPLAB2:                                   /* ascii                */
    add_string("PXPLAB2", &magic_pbm);
    break;

  case P3:                                        /* ascii                */
    add_string("P3", &magic_pbm);
    break;

  case P4:                                        /* binary               */
    add_string("P4", &magic_pbm);
    break;

  case P5:                                        /* binary               */
    add_string("P5", &magic_pbm);
    break;

  case PXPLAB5:                                   /* binary               */
    add_string("PXPLAB5", &magic_pbm);
    break;

  case P6:                                        /* binary               */
    add_string("P6", &magic_pbm);
    break;

  default:
    assert(0);                                                   /* what? */
    break;
  }

  return magic_pbm;
}



/* Writes a PNM header to file.

Revision: 09.07.2004 E.Sack

Input : - pointer to file stream
        - extended PNM format
        - PNM size
Output: - file updated
*/
static void
write_pnm_header(FILE *bild_output, PNM_FORMAT form, area_def clp)
{
  int  d;
  char *magic_pbm;

  magic_pbm = get_magic_pnm(form);

  d = fprintf(bild_output, "%s\n", magic_pbm);      /* write general Header */
  assert(d > 0);

  if (!(form == PXPLAB2 || form == PXPLAB5)) {
    d = fprintf(bild_output, XV_DUMMY_STRING);
    assert(d > 0);
  }

  d = fprintf(bild_output, "%d %d\n", clp.width, clp.height);
  assert(d > 0);

  switch(form) {
  case P1:                                                        /* ascii  */
  case P4:                                                        /* binary */
    break;

  case PXPLAB2:                                                   /* ascii  */
  case PXPLAB5:                                                   /* binary */
  case P2:                                                        /* ascii  */
  case P5:                                                        /* binary */
  case P3:                                                        /* ascii  */
  case P6:                                                        /* binary */
    d = fprintf(bild_output, "%d\n", 255);         /* default MGV for XPLAB */
    assert(d > 0);
    break;

  default:
    assert(0);                                                     /* what? */
    break;
  }

  FREEMEM(magic_pbm);
}



/* Selects the optimal format of the image for storing.

Revision: 04.08.2002 E.Sack

Input : - pointer to image
        - clipping
        - clipping offset
        - preset format
        - verbose flag
Output: - format selected
*/
PNM_FORMAT
automatic_format(pic_def src, area_def clp, int x_offset, int y_offset, 
		 int rawbit)
{
  int        mem_cnt = 0;
  PNM_FORMAT format = PNM_FORMAT_UNDEF;

  mem_cnt += !(src.mem==0) + !(src.bvl==0) + !(src.col==0);

  switch (mem_cnt) {                      /* investigate memory allocation */
  case 0:
    string_to_log(main_log, "ERROR: no image allocated - ", "", 
		  "automatic image format failed\n");
    break;

  case 1:
    if (src.mem) {
      format = test_format_grey(src.mem, src.ar, clp, x_offset, y_offset, 
				rawbit);
      break;
    }

    if (src.bvl) {
      format = test_format_grey(src.bvl, src.ar, clp, x_offset, y_offset, 
				rawbit);
      break;
    }

    if (src.col) {
      format = test_format_col_def(src, rawbit);
    }

    break;

  case 2: case 3:
    string_to_log(main_log, "ERROR: multiple image buffers allocated - ", "", 
		  "automatic image format failed\n");
    format = PNM_FORMAT_MIXED;
    break;

    default: assert(0); break;                                 /* algo bug */
  }

  return format;
}



/* Tests the PNM format of a memory.

Revision: 02.01.2003 E.Sack

Input : - pointer to memory
        - memory size
        - clipping
        - clipping offset
        - rawbit
        - verbose flag
Output: - format
Precondition: - memory organized in 2-D
*/
static PNM_FORMAT
test_format_grey(PICT *mem, area_def mem_size, area_def clp, int x_offset, 
		 int y_offset, int rawbit)
{
  int  y, x, yy, mem_width = mem_size.width;
  int  clp_x_end = clp.width+x_offset, clp_y_end = clp.height+y_offset;
  PICT data;

  for (y = y_offset; y < clp_y_end; y++) {          /* setup bitmap levels */
    yy = y*mem_width;

    for (x = x_offset; x < clp_x_end; x++) {
      data = mem[x + yy];

      if (data != FOREGROUND && data != BACKGROUND) {
	return rawbit?P5:P2;                                       /* grey */
      }
    }
  }

  return rawbit?P4:P1;                                           /* bitmap */
}



/* Tests the PNM format of a memory. If a color image is a grey one in
   reality, it is converted to a real grey one to save memory space.

Revision: 05.08.2002 E.Sack

Input : - image
        - rawbit
Output: - format
*/
static PNM_FORMAT
test_format_col_def(pic_def src, int rawbit)
{
  assert(src.col);

  return rawbit?P6:P3;
}



/* Dumps PNM file.

Revision: 04.08.2002 E.Sack

Input : - image
        - pointer to filename
        - clipping
        - clipping offset
        - output format
        - verbose flag
Output: - image dumped to disc
*/
void 
dump_pic_def(pic_def src, char *name, area_def clp, int x_offset, 
	     int y_offset, int verbose, PNM_FORMAT forcef)
{
  int        d;
  FILE       *bild_output;
  PNM_FORMAT format;

  if (forcef == PNM_FORMAT_UNDEF) {
    format = automatic_format(src, clp, x_offset, y_offset, xctrl.cfg.rawbit);
  }
  else {
    format = forcef;
  }

  d = check_pic_setup(name, src, clp, x_offset, y_offset, format);
  if (d) {
    return;
  }

  bild_output = xplab_open_file(name, "w");
  if (!bild_output) {
    return;
  }

  write_pnm_header(bild_output, format, clp);

  switch(format) {
  case P1: write_P1_data(bild_output, src, clp, x_offset, y_offset); break;
  case P2: write_P2_data(bild_output, src, clp, x_offset, y_offset); break;
  case PXPLAB2: write_PXPLAB2_data(bild_output, src, clp, x_offset, y_offset); break;
  case P3: write_P3_data(bild_output, src, clp, x_offset, y_offset); break;
  case P4: write_P4_data(bild_output, src, clp, x_offset, y_offset); break;
  case P5: write_P5_data(bild_output, src, clp, x_offset, y_offset); break;
  case PXPLAB5: write_PXPLAB5_data(bild_output, src, clp, x_offset, y_offset); break;
  case P6: write_P6_data(bild_output, src, clp, x_offset, y_offset); break;
  default: assert(0); break;
  }

  if (verbose) {
    string_to_log(main_log, "image dumped to '", name, "' ...");
  }

  d = xplab_close_file(bild_output, name);

  if (!d && verbose) {
    char *magic_nbr;

    magic_nbr = get_magic_pnm(format);

    string_to_log(main_log, " ok  (", magic_nbr, "  ");
    int_to_log(main_log, "", clp.width, " x ");
    int_to_log(main_log, "", clp.height, ")\n");

    FREEMEM(magic_nbr);
  }
}



/* Writes a XPLAB training session file.

Revision: 07.07.2003 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - file stream updated
Usage : - This function wraps that one for writing the according PNM 
          file format. The main reason for doing so, is, to get a 
          different format key avoiding a misinterpretation. I found
          this as the simplest maintainable solution.
*/
static void
write_PXPLAB5_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
		   int y_offset)
{
  write_P5_data(file, pic, clp, x_offset, y_offset);
}



/* Writes a PPM raw color file to disc (P6, color values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - file stream updated
Precondition: - clipping must be placed correctly.
*/
static void
write_P6_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
	      int y_offset)
{
  int d, y, x;
  int yy, cy, clp_width = clp.width, pic_width = pic.ar.width;
  int clp_y_end = clp.height+y_offset;

  assert(pic.col);
  assert(file);
  assert(x_offset>=0);
  assert(y_offset>=0);
  assert(pic.ar.width>0);
  assert(pic.ar.height>0);

  for (y = y_offset; y < clp_y_end; y++) {
    yy = y*pic_width + x_offset;
    cy = yy+clp_width;

    for (x = yy; x < cy; x++) {
      d = fprintf(file, P6_FORMAT, pic.col[x].r, pic.col[x].g, pic.col[x].b);
      assert(d > 0);
    }
  }
}



/* Writes a PGM raw graymap file to disc (P5, grey values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - image
*/
static void
write_P5_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
	      int y_offset)
{
  int byte, yy, cy, d, y, x, clp_width = clp.width;
  int clp_y_end = clp.height+y_offset, pic_width = pic.ar.width;

  assert(file);
  assert(x_offset>=0);
  assert(y_offset>=0);
  assert(pic.ar.width>0);
  assert(pic.ar.height>0);
  assert(clp.width+x_offset<=pic.ar.width);
  assert(clp_y_end<=pic.ar.height);
  assert(pic.mem);

  for (y = y_offset; y < clp_y_end; y++) {
    yy = y*pic_width + x_offset;
    cy = yy+clp_width;

    for (x = yy; x < cy; x++) {
      byte = pic.mem[x] & 0xff;
      d = fputc(byte, file);
      assert(d==byte);
    }
  }
}



/* Writes a PBM raw bitmap file to disc (P4, black/white values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - image
Precondition: - clipping must be placed correctly.
*/
static void
write_P4_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
	      int y_offset)
{
  int d, y, byte_no, bit_no, l, mod8, byte, xx = x_offset+7;
  int yy, clp_y_end = clp.height+y_offset, pic_width = pic.ar.width;
  int clp_width = clp.width;

  assert(file);
  assert(x_offset>=0);
  assert(y_offset>=0);
  assert(pic.ar.width>0);
  assert(pic.ar.height>0);

  mod8 = XPMOD8(clp.width);

  for (y = y_offset; y < clp_y_end; y++) {
    yy = y*pic_width + xx;

    for (byte_no = 0, l = 0; byte_no < mod8; byte_no++) {
      for (bit_no = 7, byte = 0; bit_no >= 0 && l < clp_width; bit_no--,l++) {
	byte |= pic.mem[yy-bit_no+(byte_no<<3)] << bit_no;
      }
      d = fputc(byte, file);
      assert(d==byte);
    }
  }

}



/* Writes a XPLAB training session file.

Revision: 07.07.2003 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - file stream updated
Usage : - This function wraps that one for writing the according PNM 
          file format. The main reason for doing so, is, to get a 
          different format key avoiding a misinterpretation. I found
          this as the simplest maintainable solution.
*/
static void
write_PXPLAB2_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
		   int y_offset)
{
  write_P2_data(file, pic, clp, x_offset, y_offset);
}



/* Writes a PPM ASCII color file to disc (P3, color values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - file stream updated
Precondition: - clipping must be placed correctly.
*/
static void
write_P3_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
	      int y_offset)
{
  int d, y, x, k, l, pnm_len, yy;
  int clp_x_end, clp_y_end = clp.height+y_offset, clp_width = clp.width;
  int pic_width = pic.ar.width;

  assert(pic.col);
  assert(file);
  assert(x_offset>=0);
  assert(y_offset>=0);
  assert(pic.ar.width>0);
  assert(pic.ar.height>0);

  pnm_len = PNMWIDTH/6;
  assert(pnm_len > 1);

  for (y = y_offset, k = 0; y < clp_y_end; y++) {
    yy = y*pic_width + x_offset;
    clp_x_end = yy + clp_width;

    for (x = yy, l = 0; x < clp_x_end; x++, l++) {
      d = fprintf(file, P3_FORMAT, pic.col[x].r, pic.col[x].g, pic.col[x].b);
      assert(d > 0);

      if (l >= pnm_len-2) {
	d = fputc('\n', file);
	assert(d=='\n');

	l = -1;
	k = 0;
      }
      else {
	k = 1;
      }
    }

    if (k) {
      d = fputc('\n', file);
      assert(d=='\n');
    }
  }
}



/* Writes a PBM ASCII bitmap file to disc (P1, black/white values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - image
Precondition: - clipping must be placed correctly.
*/
static void
write_P1_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
	      int y_offset)
{
  int d, y, yy, x, k, l;
  int clp_y_end = clp.height+y_offset, clp_x_end = clp.width+x_offset;
  int pic_width = pic.ar.width;

  assert(file);
  assert(x_offset>=0);
  assert(y_offset>=0);
  assert(pic.ar.width>0);
  assert(pic.ar.height>0);

  for (y = y_offset, k = 0; y < clp_y_end; y++) {
    yy = y*pic_width;

    for (x = x_offset, l = 0; x < clp_x_end; x++, l++) {
      d = fprintf(file, P1_FORMAT, pic.mem[x + yy]);
      assert(d > 0);

      if (l >= PNMWIDTH-2) {
	d = fputc('\n', file);
	assert(d=='\n');

	l = -1;
	k = 0;
      }
      else {
	k = 1;
      }
    }

    if (k) {
      d = fputc('\n', file);
      assert(d=='\n');
    }
  }
}



/* Writes a PGM ASCII graymap file to disc (P2, grey values). As file
   format is technically equal to P1 this is a wrapper.

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image
        - clipping
        - clipping offset
Output: - image
*/
static void
write_P2_data(FILE *file, pic_def pic, area_def clp, int x_offset, 
	      int y_offset)
{
  assert(file);
  assert(x_offset>=0);
  assert(y_offset>=0);
  assert(pic.ar.width>0);
  assert(pic.ar.height>0);

  write_P1_data(file, pic, clp, x_offset, y_offset);
}



/* Wrapper for PGM file from 2-D double values.

Revision: 12.07.2002 E.Sack

Input : - image
        - filename
Output: - image dumped to disc
Postcondition: - if all values are equal, only background is dumped.
*/
void
dump_pic_dbl(pic_dbl pic, char *fname)
{
  pic_def ppic;

  if (!pic.mem) {
    return;
  }

  assert(fname);
  assert(AREA_SIZE(pic.ar)>0);

  ppic = convert_pic_dbl(pic);

  dump_pic_def(ppic, fname, pic.ar, 0, 0, xctrl.cfg.verbose > BE_QUIET,
	       PNM_FORMAT_UNDEF);

  free_pic_def(&ppic);
}



/* Converts 2-D double values to integer.

Revision: 18.04.2003 E.Sack

Input : - image with doubles
Output: - image with integers
*/
pic_def
convert_pic_dbl(pic_dbl dpic)
{
  int     y, yy, x, xx, pic_width = dpic.ar.width;
  int     max_idx, min_idx, pic_height = dpic.ar.height;
  double  norm, max, min;
  pic_def ipic;

  assert(dpic.mem);
  assert(AREA_SIZE(dpic.ar)>0);

  init_pic_def(&ipic);

  max_idx = max_index_double_left(dpic.mem, AREA_SIZE(dpic.ar));
  max = dpic.mem[max_idx];
  min_idx = min_index_double_left(dpic.mem, AREA_SIZE(dpic.ar));
  min = dpic.mem[min_idx];

  assert(max>=min);

  norm = (max > min) ? (double)PICT_MAX/(max - min) : 1.0;

  if (max == min) {
    if (min > (double)PICT_MAX) {
      min -= (double)PICT_MAX;              /* grey value becomes truncated */
    }
    else {
      min = 0.0;                                     /* grey value keeps on */
    }
  }
 
  ipic.ar = dpic.ar;
  CALLOC(PICT, AREA_SIZE(ipic.ar), ipic.mem);

  for (y = 0; y < pic_height; y++) {
    yy = y*pic_width;

    for (x = 0; x < pic_width; x++) {
      xx = x + yy;

      ipic.mem[xx] = (PICT)(0.5 + (dpic.mem[xx] - min) * norm);
    }
  }

  return ipic;
}



/* Binarisation of a graymap image to a binary one.

Revision: 05.08.2002 E.Sack

Input : - image
        - threshold (exclusive)
Output: - image updated
*/
void 
level_patt(pic_def pic, int threshold) 
{
  int i, len = pic.ar.width * pic.ar.height;

  for (i = 0; i < len; i++) {
    pic.mem[i] = (pic.mem[i] > threshold) ? FOREGROUND : BACKGROUND;      
  }
}



/* Converts a PGM image to a PBM one, assuming black text of white paper.
   Then, PGM white is PBM background, PGM black is PBM foreground.
   PBM has not really a color, just distinguish between foreground and 
   background pixel, whichever their color is.

Revision: 05.08.2002 E.Sack

Input : - threshold
        - image
Output: - image updated
*/
void 
grey_to_bitmap(int threshold, pic_def pic) 
{
  int i, len = AREA_SIZE(pic.ar);

  assert(pic.mem);

  for (i = 0; i < len; i++) {
    pic.mem[i] = (pic.mem[i] > threshold) ? BACKGROUND : FOREGROUND;      
  }
}



/* Reads the header of a PNM file.

Revision: 02.11.2003 E.Sack

Input : - pointer to stream
        - pointer to size
        - pointer to maximal color value
Output: - size updated
        - maximal color value updated
*/
PNM_FORMAT 
read_pnm_hdr(FILE *file, area_def *ar, int *mcv) 
{
  int        d;
  char       buf[MAXCHAR];
  PNM_FORMAT format = PNM_FORMAT_UNDEF;

  assert(file);
  assert(ar);
  assert(mcv);

  trueline(buf, MAXCHAR, file);

  if (!strcmp(buf, "P1")) {
    format = P1;
  }
  else if (!strcmp(buf, "P2")) {
    format = P2;
  }
  else if (!strcmp(buf, "P3")) {
    format = P3;
  }
  else if (!strcmp(buf, "P4")) {
    format = P4;
  }
  else if (!strcmp(buf, "P5")) {
    format = P5;
  }
  else if (!strcmp(buf, "P6")) {
    format = P6;
  }
  else if (!strcmp(buf, "PXPLAB2")) {
    format = PXPLAB2;
  }
  else if (!strcmp(buf, "PXPLAB5")) {
    format = PXPLAB5;
  }
  else {
    format = PNM_FORMAT_UNDEF;
    return format;
  }

  trueline(buf, MAXCHAR, file);

  d = sscanf(buf, "%d %d", &ar->width, &ar->height);
  assert(d == 2); 
  assert(ar->width > 0); 
  assert(ar->height > 0); 

  if (format == P6                                    /* special treatment */
      || format == P3
      || format == P2
      || format == P5
      || format == PXPLAB2
      || format == PXPLAB5) {
    trueline(buf, MAXCHAR, file);
    d = sscanf(buf, "%d", mcv);
    assert(d == 1);
  }
  else *mcv = 1;

  return format;
}   



/* Reads a PNM image file.

Revision: 29.07.2005 E.Sack

Input : - pointer to filename
        - pointer to max. grey value
        - verbose flag
Output: - image
        - max. grey value updated
*/
img_typ
read_pnm_file(char *input, int *mgv, int verbose) 
{
  int      clos, err;
  FILE     *file;
  img_typ  timg;
  area_def size;

  init_img_typ(&timg);

  assert(mgv);

  err = check_name(input);
  assert(!err);

  file = xplab_open_file(input, "r");
  if (!file) {
    return timg;
  }

  if (verbose) {
    xplab_text_insert(main_log, "read image from '");
    xplab_text_insert(main_log, input);
    xplab_text_insert(main_log, "' ...");
  }

  timg.typ = read_pnm_hdr(file, &size, mgv);

  switch (timg.typ) {
  case P1:      timg.pimg = read_P1_data(file, size); break;
  case P2:      timg.pimg = read_P2_data(file, size); break;
  case PXPLAB2: timg.pimg = read_PXPLAB2_data(file, size); break;
  case P3:      timg.pimg = read_P3_data(file, size); break;
  case P4:      timg.pimg = read_P4_data(file, size); break;
  case P5:      timg.pimg = read_P5_data(file, size); break;
  case PXPLAB5: timg.pimg = read_PXPLAB5_data(file, size); break;
  case P6:      timg.pimg = read_P6_data(file, size); break;
  case PNM_FORMAT_UNDEF:
  default: 
    xplab_text_insert(main_log, 
		      "\nWARNING: unknown image format in file '");
    xplab_text_insert(main_log, input);
    xplab_text_insert(main_log, "'\n");
    return timg;
    break;
  }

  clos = xplab_close_file(file, input);

  if (!clos && verbose) {
    char *magic_nbr = 0, buf1[1024];

    magic_nbr = get_magic_pnm(timg.typ);

    g_snprintf(buf1, sizeof(buf1), " ok  (%s  %d x %d)\n", magic_nbr, 
	     size.width, size.height);
    xplab_text_insert(main_log, buf1);

    FREEMEM(magic_nbr);
  }

  return timg;
}



/* Reads a PPM raw color file from disc (P6, color values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
*/
static pic_def
read_P6_data(FILE *file, area_def size)
{
  int     i, len, red, green, blue;
  pic_def pic;

  init_pic_def(&pic);

  pic.ar = size;
  len = pic.ar.width*pic.ar.height;
  CALLOC(col_def, len, pic.col);

  for (i = 0; i < len; i++) {
    red   = fgetc(file);
    green = fgetc(file);
    blue  = fgetc(file);
      
    if (blue == EOF) {
      xplab_text_insert(main_log, "\nfile end reached precociously\n");
      break;
    }
      
    pic.col[i].r = (PICT)red;
    pic.col[i].g = (PICT)green;
    pic.col[i].b = (PICT)blue;
  }
      
  return pic;
}



/* Reads a PPM ASCII color file from disc (P3, color values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
*/
static pic_def
read_P3_data(FILE *file, area_def size)
{
  int     i, c, len, red, green, blue;
  pic_def pic;

  init_pic_def(&pic);

  pic.ar = size;
  len = pic.ar.width*pic.ar.height;
  CALLOC(col_def, len, pic.col);

  for (i = 0; i < len; i++) {
    c = fscanf(file, "%d%d%d", &red, &green, &blue);
    assert(c == 3);

    pic.col[i].r = (PICT)red;
    pic.col[i].g = (PICT)green;
    pic.col[i].b = (PICT)blue;
  }

  return pic;
}



/* Reads a XPLAB training session file.

Revision: 07.07.2003 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
Usage : - This function wraps that one for reading the according PNM 
          file format. The main reason for doing so, is, to get a 
          different format key avoiding a misinterpretation. I found
          this as the simplest maintainable solution.
*/
static pic_def
read_PXPLAB2_data(FILE *file, area_def size)
{
  pic_def pic;

  pic = read_P2_data(file, size);

  return pic;
}



/* Reads a XPLAB training session file.

Revision: 07.07.2003 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
Usage : - This function wraps that one for reading the according PNM 
          file format. The main reason for doing so, is, to get a 
          different format key avoiding a misinterpretation. I found
          this as the simplest maintainable solution.
*/
static pic_def
read_PXPLAB5_data(FILE *file, area_def size)
{  
  pic_def pic;

  pic = read_P5_data(file, size);

  return pic;
}



/* Reads a PGM ASCII graymap file from disc (P2, grey values). As file
   format is technically equal to P1 this is a wrapper.

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
*/
static pic_def
read_P2_data(FILE *file, area_def size)
{
  pic_def pic;

  init_pic_def(&pic);

  pic = read_P1_data(file, size);

  return pic;
}



/* Reads a PGM raw graymap file from disc (P5, grey values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
*/
static pic_def
read_P5_data(FILE *file, area_def size)
{
  int     i, c, len, mod8;
  pic_def pic;

  init_pic_def(&pic);

  pic.ar = size;
  len = pic.ar.width*pic.ar.height;
  CALLOC(PICT, len, pic.mem);

  mod8 = XPMOD8(pic.ar.width);

  for (i = 0; i < len; i++) {
    c = fgetc(file);

    assert(c != EOF);

    pic.mem[i] = (PICT)c;
  }

  return pic;
}



/* Reads a PBM ASCII bitmap file from disc (P1, black/white values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
*/
static pic_def
read_P1_data(FILE *file, area_def size)
{
  int     i, c, len;
  pic_def pic;

  init_pic_def(&pic);

  pic.ar = size;
  len = pic.ar.width*pic.ar.height;
  CALLOC(PICT, len, pic.mem);

  for (i = 0; i < len; i++) {
    if (fscanf(file, "%d", &c) != 1) {
      xplab_text_insert(main_log, "\nfile end reached precociously\n");
      break;
    }
      
    pic.mem[i] = (PICT)c;
  }

  return pic;
}



/* Reads a PBM raw bitmap file from disc (P4, black/white values).

Revision: 05.08.2002 E.Sack

Input : - pointer to file stream
        - image size
Output: - image
*/
static pic_def
read_P4_data(FILE *file, area_def size)
{
  int     i, y, c, j, k, l, mod8;
  pic_def pic;

  init_pic_def(&pic);

  pic.ar = size;
  CALLOC(PICT, pic.ar.width*pic.ar.height, pic.mem);

  mod8 = XPMOD8(pic.ar.width);

  for (i = 0; i < pic.ar.height; i++) {
    for (j = l = 0; j < mod8; j++) {
      y = i*pic.ar.width;

      c = fgetc(file);

      assert(c != EOF);

      for (k = 7; k >= 0 && l < pic.ar.width; k--, l++) {
	pic.mem[l + y] = (PICT)((c & 1<<k)>>k);
      }
    }
  }

  return pic;
}



/* Converts a PPM color image (RGB) in a PGM one.

Revision: 05.08.2002 E.Sack

Input : - pointer to image
Output: - image updated
        - flag (0: nothing converted, 1: image converted)
Postcondition: - converted image contains only a PGM.
*/
int
rgb_to_grey(pic_def *pic, int verbose) 
{
  int     i, len = AREA_SIZE(pic->ar);
  double  red = 0.0, green = 0.0, blue = 0.0;
  pic_def tmp;

  if (!pic->col) {
    return 0;                                   /* there is no color image */
  }

  init_pic_def(&tmp);
  tmp.ar = pic->ar;
  CALLOC(PICT, len, tmp.mem);

  if (verbose) {
    double_to_log("RGB scaling: ", xctrl.cfg.rgb_norm.r, " + ");
    double_to_log("", xctrl.cfg.rgb_norm.g, " + ");
    double_to_log("", xctrl.cfg.rgb_norm.b, " = ");
    double_to_log("", xctrl.cfg.rgb_norm.r + xctrl.cfg.rgb_norm.g + 
		  xctrl.cfg.rgb_norm.b, "\n");
  }

  for (i = 0; i < len; i++) {
    red   = (double)pic->col[i].r;
    green = (double)pic->col[i].g;
    blue  = (double)pic->col[i].b;

    tmp.mem[i] = (int)(xctrl.cfg.rgb_norm.r*red + xctrl.cfg.rgb_norm.g*green
		       + xctrl.cfg.rgb_norm.b*blue);
  }

  copy_pic_def(&tmp, pic);
  free_pic_def(&tmp);

  return 1;
}



/* Binarisation of multi-valued images using threshold methods. Only 
   extended PNM format accepted.

Revision: 21.12.2004 E.Sack

Input : - pointer to image
        - maximal grey value
        - pointer to filename
        - threshold type
        - pointer to session image
        - pointer to grey-scaled histogram
Output: - image updated
        - session image updated
        - grey-scaled histogram updated
Precondition: - pointer to session image may be NULL
              - pointer to grey-scaled histogram may be NULL
*/
void
pnm_to_bitmap(img_typ *timg, int mgv, char *filename, int histo_thold, 
	      img_typ *tses, hst_def *grey_hist, hst_def *th_hist)
{
  int threshold;

  assert(timg);

  if (grey_hist) {                                 /* always delete old data */
    free_hst_def(grey_hist);
  }

  if (th_hist) {                                   /* always delete old data */
    free_hst_def(th_hist);
  }

  switch (timg->typ) {
  case PXPLAB2:
  case PXPLAB5:
    {
      img_typ ttmp = xdup_img_typ(*timg);         /* duplicate session image */

      if (tses) {                                          /* only if wanted */
	free_img_typ(tses);
	*tses = xdup_img_typ(ttmp);          /* copy session image to output */
      }

      free_img_typ(timg);
      timg->pimg = session_to_image(ttmp.pimg);/* convert to original bitmapped */

      if (!AREA_SIZE(timg->pimg.ar)) {
	*timg = xdup_img_typ(ttmp);                               /* restore */
      }

      free_img_typ(&ttmp);
    }

    break;

  case P3:
  case P6:
    rgb_to_grey(&timg->pimg, 1);                                      /* RGB */

    dc_source_pic(filename, timg->pimg, DC_BIT_GV);

    /* fall through to grey map processing                                   */
   
  case P2:                                                        /* greymap */
  case P5:
    {
      hst_def gv_hist;                                /* only for statistics */

      gv_hist = histogram_8bit(timg->pimg.mem, timg->pimg.ar);         
      xreport("histogram features are:\n");
      xreport("bins: %d\n", gv_hist.hst.cnt);
      xreport("maximum: H(%d) = %5f\n", gv_hist.max_bin, gv_hist.maxval);

      if (histo_thold == THOLD_DEFAULT) {         /* method to get threshold */
	hst_def slope;                   
	slope = slope_mean(timg->pimg, mgv);
	threshold = slope.th_bin;

	if (th_hist) {                                  /* deliver histogram */
	  free_hst_def(th_hist);
	  *th_hist = slope;
	}
	else {
	  free_hst_def(&slope);
	}
      }
      else {
	threshold = histo_thold;                          /* threshold given */
      }

      gv_hist.th_bin = threshold;                          /* init threshold */

      threshold_info(gv_hist, threshold, histo_thold);      /* documentation */

      if (xctrl.cfg.histo_write) {
	char *buf = NULL;

	cat_string(filename, HSTNAME_DEFAULT, "", &buf);
	write_histo(gv_hist, 1, buf, "");
	FREEMEM(buf);
      }

      grey_to_bitmap(threshold, timg->pimg);                 /* GV -> Bitmap */

      if (grey_hist) {                                  /* deliver histogram */
	free_hst_def(grey_hist);
	*grey_hist = gv_hist;
      }
      else {
	free_hst_def(&gv_hist);
      }
    }

    break;

  case P1:                                                         /* Bitmap */
  case P4: 
    break;

  case PNM_FORMAT_UNDEF:
  default: assert(0); break;
  }

  timg->typ = ((timg->typ != P1) && (timg->typ != P4)) ? P4 : timg->typ;
}


  
/* Extracts from a session image its original bitmapped.

Revision: 09.07.2004 E.Sack

Input : - session image
Output: - bitmapped image 
*/
static pic_def
session_to_image(pic_def ses)
{
  int     pxl_found;
  pic_def bmp;

  init_pic_def(&bmp);

  bmp = xdup_pic_def(ses);                        /* duplicate input image */

  pxl_found = session_to_bitmap(bmp);      /* get original bitmapped image */

  if (!pxl_found) {
    free_pic_def(&bmp);
  }

  return bmp;
}



/* Converts a session image to its untrained bitmapped one.

Revision: 09.07.2004 E.Sack

Input : - session image
Output: - session image updated as flagged
        - number of changed pixels
Postcondition: - if there are no changed pixels, image is left unchanged.
*/
int
session_to_untrained(pic_def sessimg)
{
  int        found;
  sector_def msec;

  AREA_TO_SECTOR(sessimg.ar, msec);

  found = set_pixel_masked(msec, XPLAB_TRAINED_PIXEL, BACKGROUND, 
			   sessimg, FALSE);
  found += set_pixel_masked(msec, (PICT)~XPLAB_TRAINED_PIXEL, FOREGROUND, 
			    sessimg, FALSE);

  return found;
}



/* Converts a session image to its original bitmapped one.

Revision: 09.07.2004 E.Sack

Input : - session image
Output: - session image updated as flagged
        - number of changed pixels
Postcondition: - if there are no changed pixels, image is left unchanged.
*/
static int
session_to_bitmap(pic_def sessimg)
{
  int found;

  sector_def msec;

  AREA_TO_SECTOR(sessimg.ar, msec);

  found = set_pixel_masked(msec, (PICT)~BACKGROUND, FOREGROUND, sessimg, 
			   FALSE);

  return found;
}



/* Counts and exchanges pixel values within a clipping by a mask. A trace
   flag prevents any changes to enable counting only.

Revision: 29.07.2005 E.Sack

Input : - clipping
        - search mask
        - substitution mark
        - image
        - flag (FALSE: count and change, TRUE: count only)
Output: - image updated
        - number of found pixels
*/
int 
set_pixel_masked(sector_def f, PICT mask, PICT to, pic_def pic,
		 gboolean trace_only)
{
  int x, xx, y, yy, mark_found, width = pic.ar.width;

  assert(f.column.start <= f.column.end);
  assert(f.line.start <= f.line.end);
  assert(f.column.start >= 0);
  assert(f.line.start >= 0);
  assert(f.column.end < pic.ar.width);
  assert(f.line.end < pic.ar.height);

  for (y = f.line.start, mark_found = 0; y <= f.line.end; y++) {
    yy = y*width;

    for (x = f.column.start; x <= f.column.end; x++) {
      if (pic.mem[xx = x+yy] & mask) {                /* test masked pixel */
	if (trace_only == FALSE) {
	  pic.mem[xx] = to;                          /* change pixel value */
	}

	mark_found++;                                             /* count */
      }
    }
  }

  return mark_found;
}



/* Histogram is dumped to a file.

Revision: 10.12.2004 E.Sack

Input : - histogram
        - threshold
Output: - file
*/
static void
threshold_info(hst_def hst, int threshold, int histo_thold)
{
  int    i;
  double dark, light;

  for (i = 0, dark = 0; i <= threshold; i++) {
    dark += hst.hst.dbl[i];
  }

  for (light = 0; i < hst.hst.cnt; i++) {
    light += hst.hst.dbl[i];
  }

  xreport("histogram sums dark: %f  light: %f\n", dark, light);
  xreport("%s threshold for binarisation is %d\n", 
             histo_thold==THOLD_DEFAULT?"calculated":"given", threshold);

  if (dark < light)
    xreport("document: black text on white ground\n\n");
  else if (dark > light)
    xreport("document: white text on black ground\n\n");
  else
    xreport("Warning: cannot decide background\n\n");   
}



/* Simple color table calculation. Calculates the subdivisions S_r, S_g, S_b
   of the basic colors RGB regarding the requested object colors G including
   background.

   0 < G = g^3, g = x.y = x + y (x integer, 0 <= y < 1 fractional part of g,
                                 g geometrical mean of G).

   G <= S = S_r*S_g*S_b = (x+a)(x+b)(x+c).        I. Condition

   D = (x+a-g)^2 + (x+b-g)^2 + (x+c-g)^2 
     = (a-y)^2 + (b-y)^2 + (c-y)^2 
   D --> minimum                                  II. Condition

   NOT a == b == c                                III. Condition

   Cond. I. ensures enaugh requested colores. 
   Cond. II. ensures equal distributed colors.
   Cond. III. ensures no grey values used.

   Every basic color is divided in the calculated subdivisions. Colors used
   starts with the lowest color value, that is not black, up to the required
   number of colors. Background therefore is white. 

Revision: 16.03.2002 E.Sack

Input : - number of colors G
Output: - RGB coefficients
*/
rgb_coef
rgb_subdivision(int col_cnt)
{
  int      i, pol_cnt;
  double   dif, dif_min, g_mean, g_test, gy_frac, gx_int, s_r, s_g, s_b;
  rgb_coef abc;
  rgb_coef abc_tab[] = {{0.0, 0.0, 1.0},        /* table ensures cond. III. */
                        {0.0, 1.0, 1.0},    /* closed solution be preferred */
                        {0.0, 1.0, 2.0},
                        {0.0, 1.0, 3.0},
                        {0.0,-1.0, 2.0},
                        {1.0, 1.0, 2.0}};

  assert(col_cnt > 0);
  assert(col_cnt <= 0xFFFFFF);                  /* no more colors available */

  pol_cnt = sizeof(abc_tab)/sizeof(rgb_coef);

  g_mean = pow((double)col_cnt, 1.0/3.0);           /* third root of G is g */
  gy_frac = modf(g_mean, &gx_int);

  for (i = 0, dif_min = -1.0; i < pol_cnt; i++) {
    s_r = gx_int + abc_tab[i].r;                 /* calculate polynom terms */
    s_g = gx_int + abc_tab[i].g;
    s_b = gx_int + abc_tab[i].b;

    g_test = s_r*s_g*s_b;                        /* calculate polynom value */

    if (g_test < (double)col_cnt) {                         /* condition I. */
      continue;
    }

    dif = (s_r-g_mean)*(s_r-g_mean) +                  /* distance function */
          (s_g-g_mean)*(s_g-g_mean) +
          (s_b-g_mean)*(s_b-g_mean);

    if (dif_min < 0.0) {                                   /* condition II. */
      dif_min = dif;
      abc = abc_tab[i];
    }
    else if (dif_min > dif) {
      dif_min = dif;
      abc = abc_tab[i];
    }
  }

  assert(dif_min >= 0.0);

  abc.r += gx_int;                                          /* subdivisions */
  abc.g += gx_int;
  abc.b += gx_int;
  assert(abc.r>0);
  assert(abc.g>0);
  assert(abc.b>0);
  return abc;
}



/* Up counter for float with carry, programmable increment and maximum.

Revision: 16.03.2002 E.Sack

Input : - maximal counter value
        - increment
        - pointer to counter value
        - carry input
Output: - counter value updated
        - carry output
*/
static int
count_float(double max, double add, double *value, int carry)
{
  assert(max>0.0);

  if (!carry) {                             /* no carry input so keep value */
    return 0;
  }

  *value += add;                                         /* increment value */

  if (*value < max) {
    return 0;                                            /* no carry output */
  }

  *value = fmod(*value, max);                    /* reset value on overflow */

  return 1;                                        /* set carry on overflow */
}



/* Up counter for float colors, programmable increment and maximum.

Revision: 16.03.2002 E.Sack

Input : - maximal color value
        - RGB increments
        - pointer to RGB colors
Output: - RGB colors updated
        - integer RGB colors
Postcondition: - counter overflow not handled with a carry
*/
col_def
color_counter(double max, rgb_coef add, rgb_coef *color)
{
  int     carry;
  col_def icolor;

  carry = count_float(max, add.r, &color->r, 1);                 /* fastest */

  carry = count_float(max, add.g, &color->g, carry);

  count_float(max, add.b, &color->b, carry);                     /* slowest */

  icolor = convert_rgb_coef(*color);              /* conversion for comfort */

  return icolor;
}



/* Repeatable random color generator ensuring used colors are exclusiv.
   According to the proposal in manual page `man rand'.

Revision: 16.03.2002 E.Sack

Input : - maximal color value
Output: - random color
*/
col_def
random_color(int max, col_def *lst, int len, col_def defc)
{
  int     i, repeat;
  double  fac = (double)max/(double)RAND_MAX;
  col_def icolor;

  do {
    icolor.r = (int)(fac*(double)rand());
    icolor.g = (int)(fac*(double)rand());
    icolor.b = (int)(fac*(double)rand());

    for (i = repeat = 0; i < len; i++) {               /* search used color */
      if (compare_col_def(icolor, lst[i])) {
	repeat = 1;
	break;
      }
    }
  } while (repeat);

  for (i = 0; i < len; i++) {                        /* enroll unused color */
    if (compare_col_def(defc, lst[i])) {
      lst[i] = icolor;
      break;
    }
  }

  return icolor;
}



/* Reads one text line of a file up to line feed suppressing comments. A
   comment starts with # and ends with line feed. It may start anywhere in 
   line.

Revision: 31.12.2004 E.Sack

Input : - pointer to line buffer
        - maximal length of line buffer
        - pointer to file
Output: - line buffer updated
        - number of characters read
*/
static int 
trueline(char s[], int max, FILE *fp) 
{
  int c = EOF, cnt = 0, comment = 0, lead = 1;
   
  assert(max > 0);
  assert(fp);
  assert(s);

  do {
    while (cnt < max-1) {
      c = getc(fp);

      if (lead && isspace(c)) {                  /* suppress leading space */
	continue;                                         /* stop sampling */
      }
      else {
	lead = 0;
      }

      switch (c) {
      case '\r': continue;                             /* carriage return  */
      case '\n': break;                        /* line feed == end of line */
      case EOF: break;                                 /* end of file      */ 
      case '#':                                        /* start of comment */
	comment = 1;
	continue;
      }

      if (c == '\n' || c == EOF) {
	break;                                        /* stop reading line */
      }

      if (comment) {
	continue;                                         /* stop sampling */
      }

      s[cnt++] = c;                                            /* sampling */
    }

    if (c == EOF) {
      break;                                          /* stop reading file */
    }

    comment = 0;                                  /* prepare for next line */
    lead = 1;                                     /* prepare for next line */
  } while (!cnt);                             /* ignore comments only line */

  s[cnt] = '\0';                                     /* string termination */

  return cnt;                                             /* buffer length */
}



/* Draws a perpendicular line into an image.

Revision: 11.12.2004 E.Sack

Input : - image
        - foreground color
        - foreground color mask
        - start coordinate
        - end coordinate
Output: - image updated
Postcondition: - Orientation of coordinates (bottom up) corrected.
Bug: - Besenham algo should be used.
*/
void 
draw_line(pic_def img, col_def bgnd, col_def fgnd, col_def fmsk, c_def start, 
	  c_def end)
{
  int x, y, idx;

  assert(start.x>=0 && start.x<img.ar.width);
  assert(start.y>=0 && start.y<img.ar.height);
  assert(start.x<=end.x);
  assert(end.x>=0 && end.x<img.ar.width);
  assert(end.y>=0 && end.y<img.ar.height);
  assert(start.y<=end.y);

  for (x = start.x; x <= end.x; x++) {
    start.y = img.ar.height - 1 - start.y;                    /* turn down */
    end.y = img.ar.height - 1 - end.y;

    for (y = end.y; y <= start.y; y++) {
      idx = x + y*img.ar.width;

      img.col[idx].r = fmsk.r ? fgnd.r : bgnd.r;
      img.col[idx].g = fmsk.g ? fgnd.g : bgnd.g;
      img.col[idx].b = fmsk.b ? fgnd.b : bgnd.b;
    }
  }
}



/* Draws a histogram into an image.

Revision: 21.12.2004 E.Sack

Input : - histogram
        - minimal image width
        - image height
Output: - image
Postcondition: - as the output image is meant to be displayed on a monitor
                 screen, its size is limited.
Usage: - this function is usable for any histograms.
       - minimal image width to avoid narrow image in consequence of a
         sparse histogram.
       - image height to normalize histogram.
*/
pic_def
histo_to_image(hst_def hin, int min_width, int height)
{
#define MAX_HISTO_IMG_WIDTH 1024/* should be enough to visualise histograms */
  int     len;
  double  scal;
  c_def   start, end;
  pic_def img;
  hst_def hsqueeze;
  col_def bgnd = {PICT_MAX, PICT_MAX, PICT_MAX};        /* white background */
  col_def fgnd = {0, 0, 0};                             /* black foreground */
  col_def fmsk = {1, 1, 1};                             /* foreground mask  */
  col_def th_fgnd = fgnd, th_fmsk = {0,1,1};   /* threshold line foreground */
  col_def th_bgnd = fgnd, th_bmsk = {1,0,1};   /* threshold line background */

  init_pic_def(&img);
  init_hst_def(&hsqueeze);

  if (hin.maxval <= 0.0) {
    return img;                                         /* empty histogram */
  }

  copy_hst_def(&hin, &hsqueeze);                          /* protect input */
  init_hst_def(&hin);                                     /* protect input */

  min_width = XPMIN(min_width, MAX_HISTO_IMG_WIDTH);

  squeeze_hst_def(&hsqueeze, MAX_HISTO_IMG_WIDTH);/* GTK cannot handle every size */

  img.ar.width = XPMAX(hsqueeze.hst.cnt, min_width); /* avoid narrow image */
  img.ar.height = height;

  CALLOC(col_def, AREA_SIZE(img.ar), img.col);          /* histogram image */

  set_color(AREA_SIZE(img.ar), img.col, bgnd);  /* set image to background */

  assert(hsqueeze.maxval>0.0);
  scal = (double)(img.ar.height-1)/hsqueeze.maxval;

  len = 0;
  while (len < hsqueeze.hst.cnt) {
    start.x = len;
    start.y = 0;

    end.x = start.x;
    end.y = (int)(hsqueeze.hst.dbl[len] * scal);                  /* scale */

    if (end.y == 0) {
      if (hsqueeze.th_bin != THOLD_DEFAULT && hsqueeze.th_bin == len) {
	start.y = 0;
	end.y = img.ar.height-1;                                /* maximum */
	draw_line(img, bgnd, th_bgnd, th_bmsk, start, end);
      }
    }
    else {
      if (hsqueeze.th_bin != THOLD_DEFAULT && hsqueeze.th_bin == len) {
	draw_line(img, bgnd, th_fgnd, th_fmsk, start, end);

	start.y = end.y + 1;

	if (start.y < img.ar.height) {
	  end.y = img.ar.height-1;                              /* maximum */
	  draw_line(img, bgnd, th_bgnd, th_bmsk, start, end);
	}
      }
      else {
	draw_line(img, bgnd, fgnd, fmsk, start, end);/* orientation corrected */
      }
    }

    len++;
  }

  free_hst_def(&hsqueeze);

  return img;
}


/* Maximises the image contrast.

Revision: 18.12.2004 E.Sack

Input : - image
Output: - image updated
*/
void
max_contrast_rgb(pic_def img)
{
  int    len, size = AREA_SIZE(img.ar);
  PICT   max = 0, min;
  double scal;

  if (size <= 0) {
    return;                                                 /* image empty */
  }

  len = size;
  while (len--) {                                               /* maximum */
    max = XPMAX(max, img.col[len].r);
    max = XPMAX(max, img.col[len].g);
    max = XPMAX(max, img.col[len].b);
  }

  if (max == 0) {
    return;                                                 /* no contrast */
  }

  min = max;                                                    /* init    */
  len = size;
  while (len--) {                                               /* minimum */
    min = XPMIN(min, img.col[len].r);
    min = XPMIN(min, img.col[len].g);
    min = XPMIN(min, img.col[len].b);
  }

  scal = (double)PICT_MAX/(double)(max-min);

  len = size;
  while (len--) {
    img.col[len].r = (PICT)((double)(img.col[len].r-min) * scal);
    img.col[len].g = (PICT)((double)(img.col[len].g-min) * scal);
    img.col[len].b = (PICT)((double)(img.col[len].b-min) * scal);
  }
}



/* Squeezes a histogram and fits its features.

Revision: 18.12.2004 E.Sack

Input : - pointer to histogram
        - maximal length
Output: - histogram updated
Postcondition: - if the histogram lenght input is less than the maximal
                 length, the histogram is left unchanged.
*/
static void
squeeze_hst_def(hst_def *hin, int max_len) 
{
  int      indx;
  gboolean changed;

  assert(hin);
  assert(max_len > 0);

  changed = squeeze_double_lst(&hin->hst, max_len);   /* squeeze histogram */

  if (changed == FALSE) {
    return;                                    /* histogram not to squeeze */
  }

                                               /* reset histogram features */
  indx = max_index_double_left(hin->hst.dbl, hin->hst.cnt);

  hin->max_bin = indx;                               /* features to change */
  hin->maxval = hin->hst.dbl[indx];

  hin->th_bin    = hin->th_bin;                      /* features unchanged */
  hin->start_bin = hin->start_bin;
  hin->set_cnt   = hin->set_cnt;
}



/* Squeezes a list using the maximum of subsets.

Revision: 29.07.2005 E.Sack

Input : - pointer to input list
        - destination list length
Output: - list updated as flagged
        - flag (FALSE if input list is shorter than destination list length 
                else TRUE)
Postcondition: - if the input list lenght is less than the destination list
                 length request, the input list is left unchanged else the
                 input list is overwritten with the destination list.
*/
static gboolean
squeeze_double_lst(double_lst *dlst, int max_len) 
{
  int        i, j, istart, iend;
  double     step, max, start;
  double_lst dtmp;

  assert(dlst);
  assert(max_len > 0);

  if (!dlst->dbl) {                                   /* consistency check */
    assert(!dlst->cnt);
  }
  else {
    assert(dlst->cnt>0);
  }

  if (dlst->cnt <= max_len) {
    return FALSE;                  /* length of list is satisfying request */
  }

  init_double_lst(&dtmp);

  dtmp.cnt = max_len;
  CALLOC(double, dtmp.cnt, dtmp.dbl);                     /* squeezed list */

  step = (double)dlst->cnt/(double)max_len;

  start = 0.0;
  for (j = 0; j < max_len; j++) {                /* all destination values */
    istart = (int)(start + 0.5);
    iend = XPMIN((int)(start + step + 0.5), dlst->cnt);

    max = 0.0;
    for (i = istart; i < iend; i++) {                            /* subset */
      max = XPMAX(max, dlst->dbl[i]);
    }

    dtmp.dbl[j] = max;                               /* fill squeezed list */

    start += step;
  }

  move_double_lst(&dtmp, dlst);                      /* copy squeezed list */

  return TRUE;
}



/* Draws histograms into an image.

Revision: 15.12.2004 E.Sack

Input : - histograms
        - background color
        - foreground color
        - foreground color mask
        - image height
Output: - image
Precondition: - size of histograms limited to 8bit.
Usage: - this function is to be used currently for color histograms.
*/
pic_def
histos_to_image(hst_def histr, hst_def histg, hst_def histb, col_def bgnd, 
		col_def fgnd, int height)
{
  int     idx, red, green, blue, min, mid, max;
  double  scal, maxval;
  c_def   start, end;
  pic_def img;
  col_def fmsk;

  init_pic_def(&img);

  img.ar.width = PICT_MAX+1;
  img.ar.height = height;

  CALLOC(col_def, AREA_SIZE(img.ar), img.col);          /* histogram image */

  set_color(AREA_SIZE(img.ar), img.col, bgnd);

  assert(histr.hst.cnt <= PICT_MAX+1);
  assert(histg.hst.cnt <= PICT_MAX+1);
  assert(histb.hst.cnt <= PICT_MAX+1);

  maxval = XPMAX(XPMAX(histr.maxval, histg.maxval), histb.maxval);

  if (maxval <= 0.0) {
    return img;                                         /* empty histogram */
  }

  scal = (double)(img.ar.height-1)/maxval;

  for (idx = 0; idx < PICT_MAX+1; idx++) {
    red = green = blue = 0;

    if (idx >= histr.start_bin &&                        /* validate index */
	idx < histr.start_bin + histr.hst.cnt) {
      red = histr.hst.dbl[idx];                         /* histogram value */
    }

    if (idx >= histg.start_bin &&                        /* validate index */
	idx < histg.start_bin + histg.hst.cnt) {
      green = histg.hst.dbl[idx];                       /* histogram value */
    }

    if (idx >= histb.start_bin &&                        /* validate index */
	idx < histb.start_bin + histb.hst.cnt) {
      blue = histb.hst.dbl[idx];                        /* histogram value */
    }

    if (red + green + blue) {                         /* some values exist */
      start.x = end.x = idx;
      start.y = 0;

      min = XPMIN(XPMIN(red, green), blue);
      max = XPMAX(XPMAX(red, green), blue);
      mid = red + green + blue - min - max;

      end.y = (int)((double)min * scal);                          /* scale */

      fmsk.r = red > 0;                       /* minimal histogram value */
      fmsk.g = green > 0;
      fmsk.b = blue > 0;

      if (fmsk.r && fmsk.g && fmsk.b) {
	draw_line(img, bgnd, fgnd, fmsk, start, end);
      }
      else {
	draw_line(img, fgnd, bgnd, fmsk, start, end);
      }

      if (min < mid) {
	fmsk.r = red > min;                         /* mid histogram value */
	fmsk.g = green > min;
	fmsk.b = blue > min;

	start.y = end.y;
	end.y = (int)((double)mid * scal);                        /* scale */

	if (fmsk.r && fmsk.g && fmsk.b) {
	  draw_line(img, bgnd, fgnd, fmsk, start, end);
	}
	else {
	  draw_line(img, fgnd, bgnd, fmsk, start, end);
	}
      }                                      /* else there is no mid value */

      fmsk.r = red > mid;                       /* maximal histogram value */
      fmsk.g = green > mid;
      fmsk.b = blue > mid;

      start.y = end.y;
      end.y = (int)((double)max * scal);                          /* scale */

      if (mid == max) {
	continue;                               /* maximum already reached */
      }

      if (fmsk.r && fmsk.g && fmsk.b) {
	draw_line(img, bgnd, fgnd, fmsk, start, end);
      }
      else {
	draw_line(img, fgnd, bgnd, fmsk, start, end);
      }
    }
  }

  return img;
}
