/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

/*#define NDEBUG*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_algorithm.h"
#include "xplab_recall_fe.h"
#include "xplab_recall_be.h"
#include "xplab_output.h"
#include "xplab_report.h"
#include "xplab_dbio.h"
#include "xplab_image.h"
#include "xplab_maintenance_fe.h"
#include "xplab_dchannel.h"
#include "xplab_tprint.h"
#include "xplab_training_fe.h"
#include "xplab_preferences_fe.h"
#include "xplab_selection.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/
#define XPLAB_RECALL_TITLE_ITEM    "recall"
#define XPLAB_RECALL_WINDOW_WIDTH  511
#define XPLAB_RECALL_WINDOW_HEIGHT 171

typedef struct {
  GtkWidget *top_recall, *rlog, *log_dump, *log_clear, *rb_use, *rb_tiles;
  GtkWidget *start_dump_button, *rb_color, *rb_chains_menu, **caller;
  gchar     *db_name;
  gboolean  cactive, cstop;
  img_typ   rimg;
  tps_def   tps;
  cand_lst  hit_lst;
  prog_ctrl ctrl;
  tiles_def rcl;
  run_def   match_run, tiles_run;
} rec_def;

/* local prototypes ********************************************************/
static void init_rec_def(rec_def *re);
static void free_rec_def(rec_def *re);
static void recall_setup(prog_ctrl ctrl, int match);
static void recall_label_text(GtkWidget *label);

/* Functions assembling widgets */
static void recall_window(void);
static void recall_menu_bar(GtkWidget *vbox);
static void recall_menu_entry_file(GtkWidget *menu_bar);
static void recall_file_menu_item_chains(GtkWidget *file_menu);
static GtkWidget *recall_file_menu_item_use(GtkWidget *submenu);
static GtkWidget *recall_file_menu_item_tiles(GtkWidget *submenu);
static GtkWidget *recall_file_menu_item_color(GtkWidget *submenu);
static void recall_file_menu_item_quit(GtkWidget *file_menu);
static void recall_control(GtkWidget *vbox);

/* Handlers/Callbacks */
static void cb_recall_stop(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_recall_dump(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_recall_clear(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_check_recall_log(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_recall_phase(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_sensitivity(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_recall_check(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);

/* local data **************************************************************/

rec_def re;                                            /* recall structure */

/* Functions ***************************************************************/

static void
init_rec_def(rec_def *re)
{
  assert(re);

  re->top_recall = NULL;                                     /* zero pointer */
  re->rlog = NULL;                                           /* zero pointer */
  re->log_dump = NULL;                                       /* zero pointer */
  re->log_clear = NULL;                                      /* zero pointer */
  re->rb_use = NULL;                                         /* zero pointer */
  re->rb_tiles = NULL;                                       /* zero pointer */
  re->start_dump_button = NULL;                              /* zero pointer */
  re->rb_color = NULL;                                       /* zero pointer */
  re->rb_chains_menu = NULL;                                 /* zero pointer */
  re->caller = NULL;                                         /* zero pointer */

  re->db_name = NULL;                                        /* zero pointer */

  re->cactive = re->cstop = FALSE;

  init_img_typ(&re->rimg);
  init_tps_def(&re->tps);
  init_cand_lst(&re->hit_lst);
  init_prog_ctrl(&re->ctrl);
  init_tiles_def(&re->rcl);

  init_run_def(&re->match_run);
  init_run_def(&re->tiles_run);
}

static void
free_rec_def(rec_def *re)
{
  assert(re);
                           /* don't free GTK stuff */

  FREEMEM(re->db_name);

  free_img_typ(&re->rimg);
  free_tps_def(&re->tps);
  free_cand_lst(&re->hit_lst);
  free_tiles_def(&re->rcl);
  free_prog_ctrl(&re->ctrl);

  init_rec_def(re);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - menu
Output: - menu item
*/
void
xplab_recall_menu(GtkWidget *match_menu, GtkWidget **caller)
{
  GtkWidget *match_menu_recall;

  match_menu_recall = gtk_menu_item_new_with_label(XPLAB_RECALL_TITLE_ITEM XPLAB_MENU_EXT);
  gtk_menu_shell_append(GTK_MENU_SHELL(match_menu), match_menu_recall);

  g_signal_connect(G_OBJECT(match_menu_recall), "activate",
		   G_CALLBACK(cb_recall_phase), (gpointer)caller);
}



/* Starts recall phase.

Revision: 09.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_recall_phase(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean raised;

  assert(user_data);
  assert(menuitem);

  raised = xplab_window_raise(&re.top_recall, 1);

  if (raised == FALSE) {
    xplab_free_recall();        /* must free here because of sudden run stop */
    re.caller = (GtkWidget **)user_data;              /* last succeeded call */
    recall_window();                                           /* top window */
  }
}



/* Calculates RB chains.

Revision: 12.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_recall_start_dump(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  if (*re.tiles_run.pactive == TRUE) {                  /* already running */
    return;
  }

  start_run_def(&re.tiles_run);

  recall_setup(xctrl, 0);

  stop_run_def(&re.tiles_run);
}



/* Creates top window for matching.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - top window
*/
static void
recall_window(void)
{
  GtkWidget *vbox;

  /* set the matching top level widget */
  re.top_recall = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(re.top_recall), XPLAB_RECALL_TITLE_ITEM);
  gtk_window_set_default_size(GTK_WINDOW(re.top_recall), 
			      XPLAB_RECALL_WINDOW_WIDTH, 
			      XPLAB_RECALL_WINDOW_HEIGHT);
  gtk_widget_set_name(re.top_recall, XPLAB_RC_RECALL_LOG);     /* for Gtk RC */

  /* put into top level widget a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(re.top_recall), vbox);

  /* put into vertical box a menubar */
  recall_menu_bar(vbox);

  /* put into vertical box a scrolled window */
  re.rlog = scrolled_text(vbox);

  recall_control(vbox);

  g_signal_emit_by_name((gpointer)re.rb_chains_menu, "activate");

  gtk_widget_show_all(re.top_recall);

  g_signal_connect(G_OBJECT(re.top_recall), "delete-event",
		   G_CALLBACK(cb_recall_top_delete), NULL);

  xplab_set_window_icon(re.top_recall);                     /* window icon */
}



/* Builds a progress bar.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
recall_control(GtkWidget *vbox)
{
  GtkWidget *hbox, *vbox1;

  /* put into inbox a label */
  re.match_run.plabel = gtk_label_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox), re.match_run.plabel, FALSE, FALSE, 0);

  /* put into inbox a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox1, FALSE, FALSE, 0);

  /* put into vertical box a start dump button */
  re.start_dump_button = 
    xset_clicked_button(vbox1, GTK_SIGNAL_FUNC(cb_recall_start_dump), 
			DIALOG_BUTTON_DUMP, DESC1_RECALL_DUMP,
			DESC2_RECALL_DUMP, NULL).button;

  /* put into vertical box a start matching button */
  xset_clicked_button(vbox1, GTK_SIGNAL_FUNC(cb_recall_start_matching), 
		      SINGLE_BUTTON_MATCHING, DESC1_RECALL_START, 
		      DESC2_RECALL_START, NULL);

  /* put into horizontal box a stop button */
  xset_clicked_button(hbox, GTK_SIGNAL_FUNC(cb_recall_stop), SINGLE_BUTTON_STOP, 
		      DESC1_STOP_ACTION, DESC2_STOP_ACTION, NULL);

  /* put into horizontal box a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox1, TRUE, TRUE, 1);

  /* put into vertical box two progress bars */
  re.tiles_run.pbar = xplab_progress(vbox1);
  re.match_run.pbar = xplab_progress(vbox1);
}



/* Starts pattern matching.

Revision: 30.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_recall_start_matching(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  if (*re.match_run.pactive == TRUE) {                  /* already running */
    return;
  }

  start_run_def(&re.tiles_run);/* contains preprocessing for matching phase*/
  start_run_def(&re.match_run);

  recall_setup(xctrl, 1);

  stop_run_def(&re.tiles_run);
  stop_run_def(&re.match_run);
}



/* Recognition phase of pattern matching methods. For documentation
   image-tiles can be dumped using the table below:

     user input   |            setting flags               |
   ---------------+----------------------------------------+
   matching  dump | tiles_to_dump   use_tiles_for_matching |
   ---------------+----------------------------------------+----------------
       0      0   |      X                    X            | nothing happens
   ---------------+----------------------------------------+----------------
       0      1   |      0                    X            | note 1)
   ---------------+----------------------------------------+----------------
       0      1   |      1                    X            | note 2)
   ---------------+----------------------------------------+----------------
       1      0   |      X                    0            | note 3)
   ---------------+----------------------------------------+----------------
       1      0   |      X                    1            | note 4)
   ---------------+----------------------------------------+----------------
       1      1   |      X                    X            | note 5)
   ---------------+----------------------------------------+----------------
   1) Warning message, because nothing flagged for dump.
   2) Dump flagged sort of tiles to disk as images.
   3) Normal matching phase without using tiles, tiles dumped as flagged
      like note 2).
   4) Matching phase using tiles, tiles dumped as flagged like note 2).
   5) This state is avoided. The first occurrence of matching or dump,
      respectively, inhibits the other until its work has been done.

Revision: 12.11.2003 E.Sack

Input : - control structure
        - flag (0: only RB chains, 1: matching)
Output: - matching results
*/
static void
recall_setup(prog_ctrl ctrl, int match)
{
  int    tps_idx;
  db_def db_tmp;

  copy_prog_ctrl(&ctrl, &re.ctrl);    /* settings copy to ensure consistence */
  db_tmp = xplab_request_db(NULL, NULL, XPLAB_DB_RO, NULL);       /* DB copy */

  free_img_typ(&re.rimg);
  re.rimg = xplab_request_image(P1, NULL, NULL, 0, NULL);      /* image copy */

  tps_idx = check_recall_resources(db_tmp, re.ctrl.cfg.tps_id, 
				   re.ctrl.set.image_filename, 
			           re.rimg.pimg, re.ctrl.set.rblst);
  if (tps_idx < 0) {
    free_db_def(&db_tmp);
    return;
  }

  FREEMEM(re.db_name);
  add_string(db_tmp.db_fname, &re.db_name);                    /* copy input */

  copy_tps_def(&db_tmp.tpl.tps[tps_idx], &re.tps, FALSE);      /* copy input */
  free_db_def(&db_tmp);

  recall_label_text(re.match_run.plabel);

  free_tiles_def(&re.rcl);
  re.rcl = rb_chain_list(re.tps.cl, re.ctrl.set.rblst, TILE_DIV_DEFAULT, 
			 &re.rimg.pimg, re.ctrl.cfg.rb_chain, match, 
			 re.ctrl.cfg.mad.mborder+re.ctrl.cfg.mad.nborder, 
			 re.ctrl.set.image_filename, re.ctrl.cfg.rawbit, 
			 &re.tiles_run);

  if (!match || re.cactive == FALSE) {           /* calculate only RB chains */
    return;
  }

  reset_rb_lst(re.ctrl.set.rblst);

  free_cand_lst(&re.hit_lst);
  re.hit_lst = xtiled_pattern_matching(re.ctrl.cfg.mad, 
				       re.ctrl.cfg.single_line, 
				       re.ctrl.set.rblst, 
				       re.rcl, re.tps.cl, 
				       &re.rimg, re.tps.ident, 
				       re.ctrl.set.image_filename, 
				       re.ctrl.cfg.cdomain, 
 				       re.ctrl.cfg.rb_chain, &re.match_run,
				       re.ctrl.cfg.single_nb,
				       re.ctrl.cfg.match_method);

  if (*re.match_run.pactive == FALSE) {
    return;
  }

  if (re.ctrl.cfg.replay_write) {
    xplab_write_recognition(re.db_name, re.tps.ident, 
			    re.ctrl.set.image_filename, 
			    re.ctrl.cfg.mad.mborder, 
			    re.ctrl.cfg.mad.nborder, 
			    re.ctrl.cfg.mad.maskweight, 
			    re.ctrl.cfg.histo_thold, re.ctrl.set.rblst, 
			    re.hit_lst, re.ctrl.set.user_options);
  }

  xrecognition_out(re.ctrl.set.rblst, re.hit_lst, re.tps, 
		   re.ctrl.cfg.mad.mborder, re.ctrl.cfg.mad.nborder, 
		   re.rimg.pimg, re.ctrl.set.image_filename, 
		   re.ctrl.cfg.line_method, &re.match_run, re.rlog, 
		   re.ctrl.cfg.renew_method);

  time_to_log(main_log, XPLAB_RECALL_TITLE_ITEM " finished after ", 
	      re.match_run.start_time, "\n");
}



/* Builds label string.

Revision: 10.06.2004 E.Sack

Input : - -
Output: - label string updated
*/
static void
recall_label_text(GtkWidget *label)
{
  char *buf = NULL;

  cat_string("image: ", re.ctrl.set.image_filename, "  ", &buf);
  cat_string("DB: ", re.db_name, "", &buf);

  xplab_label_set_text(label, buf);
  FREEMEM(buf);
}



/* Initializes the recall phase.

Revision: 30.04.2004 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_init_recall(void)
{
  init_rec_def(&re);

  re.tiles_run.pactive = &re.cactive;
  re.tiles_run.pstop = &re.cstop;
  re.match_run.pactive = &re.cactive;
  re.match_run.pstop = &re.cstop;
}



/* Frees the recall phase.

Revision: 14.11.2003 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_free_recall(void)
{
  free_rec_def(&re);

  xplab_init_recall();
}



/* Stops pattern matching (recall phase) inclusive progress bars update.

Revision: 25.04.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_recall_stop(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  if (re.tiles_run.start_time != (time_t)0) {
    stop_run_def(&re.tiles_run);
  }

  if (re.match_run.start_time != (time_t)0) {
    stop_run_def(&re.match_run);
  }
}



/* Builds a menu bar.

Revision: 05.08.2005 E.Sack

Input : - container
Output: - container updated
*/
static void
recall_menu_bar(GtkWidget *vbox)
{
  GtkWidget *menu_bar;

  menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);

  recall_menu_entry_file(menu_bar);                             /* menu file */
  xplab_menu_help(menu_bar, &re.top_recall);                    /* menu help */
}



/* Builds an entry for a menu bar.

Revision: 05.08.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static void
recall_menu_entry_file(GtkWidget *menu_bar)
{
  GtkWidget *file_entry, *file_menu, *sep;

  file_entry = gtk_menu_item_new_with_label(MENU_BUTTON_FILE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  file_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), file_menu);

  xplab_preferences_menu(file_menu, &re.top_recall);     /* item preferences */
  xplab_selection_menu(file_menu, &re.top_recall);         /* item selection */
  file_menu_item_log(file_menu,                                  /* item log */
		     GTK_SIGNAL_FUNC(cb_recall_dump), &re.log_dump,
		     GTK_SIGNAL_FUNC(cb_recall_clear), &re.log_clear, NULL);
  recall_file_menu_item_chains(file_menu);                 /* item RB chains */

  sep = gtk_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), sep);

  recall_file_menu_item_quit(file_menu);                        /* item quit */

  g_signal_connect(G_OBJECT(file_entry), "activate",
		   G_CALLBACK(cb_check_recall_log), NULL);
}



/* Checks log menu entry.

Revision: 21.10.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_check_recall_log(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int           len;
  GtkTextBuffer *buffer;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_TEXT_VIEW(re.rlog) == TRUE) {
    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(re.rlog));/* fetch default buffer */
    len = gtk_text_buffer_get_char_count(buffer);

    gtk_widget_set_sensitive(re.log_clear, len ? TRUE : FALSE);
    gtk_widget_set_sensitive(re.log_dump, len ? TRUE : FALSE);
  }
}



/* Clears recall log.

Revision: 02.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_recall_clear(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  clear_gtk_text(re.rlog);
}



/* Dumps recall log.

Revision: 15.11.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_recall_dump(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  char *txt_file = NULL;

  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  cat_string(re.ctrl.set.image_filename, XPLAB_RECALL_LOGNAME_EXT, "", 
	     &txt_file);

  dump_gtk_text(re.rlog, txt_file);

  FREEMEM(txt_file);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - menu
Output: - menu updated
*/
static void
recall_file_menu_item_chains(GtkWidget *file_menu)
{
  GtkWidget *submenu;

  re.rb_chains_menu = gtk_menu_item_new_with_label("RB chains");
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), re.rb_chains_menu);

  submenu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(re.rb_chains_menu), submenu);

  re.rb_use = recall_file_menu_item_use(submenu);
  re.rb_tiles = recall_file_menu_item_tiles(submenu);
  re.rb_color = recall_file_menu_item_color(submenu);

  g_signal_connect(G_OBJECT(re.rb_chains_menu), "activate",
		   G_CALLBACK(cb_sensitivity), NULL);
}



/* Sets sensitivity.

Revision: 23.04.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_sensitivity(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean sens = FALSE;

  STSFY_GCC(user_data);

  assert(menuitem);

  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(re.rb_use)) == TRUE) {
    xctrl.cfg.rb_chain |= RB_CHAIN_USE;
  }
  else {
    xctrl.cfg.rb_chain &= ~RB_CHAIN_USE;
  }

  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(re.rb_tiles)) == TRUE) {
    sens = TRUE;
    xctrl.cfg.rb_chain |= RB_CHAIN_DUMP_TILES;
  }
  else {
    xctrl.cfg.rb_chain &= ~RB_CHAIN_DUMP_TILES;
  }

  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(re.rb_color)) == TRUE) {
    sens = TRUE;
    xctrl.cfg.rb_chain |= RB_CHAIN_DUMP_COLOR;
  }
  else {
    xctrl.cfg.rb_chain &= ~RB_CHAIN_DUMP_COLOR;
  }

  gtk_widget_set_sensitive(re.start_dump_button, sens);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - pointer to menu item
*/
static GtkWidget *
recall_file_menu_item_use(GtkWidget *submenu)
{
  GtkWidget *rb_use;

  rb_use = gtk_check_menu_item_new_with_label("use");
  gtk_menu_shell_append(GTK_MENU_SHELL(submenu), rb_use);

  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(rb_use),
				 (re.ctrl.cfg.rb_chain & RB_CHAIN_USE)? TRUE:FALSE);

  g_signal_connect(G_OBJECT(rb_use), "toggled", 
		   G_CALLBACK(cb_recall_check), NULL);

  return rb_use;
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - pointer to menu item
*/
static GtkWidget *
recall_file_menu_item_tiles(GtkWidget *submenu)
{
  GtkWidget *rb_tiles;

  rb_tiles = gtk_check_menu_item_new_with_label("dump tiles");
  gtk_menu_shell_append(GTK_MENU_SHELL(submenu), rb_tiles);

  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(rb_tiles),
				 (re.ctrl.cfg.rb_chain & RB_CHAIN_DUMP_TILES) ?
				 TRUE:FALSE);

  g_signal_connect(G_OBJECT(rb_tiles), "toggled",
		   G_CALLBACK(cb_recall_check), NULL);

  return rb_tiles;
}



/* Wrapper for sensitivity.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_recall_check(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(checkmenuitem);

  g_signal_emit_by_name((gpointer)re.rb_chains_menu, "activate");
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - pointer to menu item
*/
static GtkWidget *
recall_file_menu_item_color(GtkWidget *submenu)
{
  GtkWidget *rb_color;

  rb_color = gtk_check_menu_item_new_with_label("dump color");
  gtk_menu_shell_append(GTK_MENU_SHELL(submenu), rb_color);

  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(rb_color),
				 (re.ctrl.cfg.rb_chain & RB_CHAIN_DUMP_COLOR) ?
				 TRUE:FALSE);

  g_signal_connect(G_OBJECT(rb_color), "toggled",
		   G_CALLBACK(cb_recall_check), NULL);

  return rb_color;
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - pointer to menu item
*/
static void
recall_file_menu_item_quit(GtkWidget *file_menu)
{
  GtkWidget *file_menu_quit;

  file_menu_quit = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);

  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_quit);

  g_signal_connect(G_OBJECT(file_menu_quit), "activate",
		   G_CALLBACK(cb_recall_quit), NULL);
}



/* Quits recall phase.

Revision: 23.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_recall_quit(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  cb_recall_stop(NULL, NULL);                             /* sudden run stop */

  if (GTK_IS_WIDGET(re.top_recall) == TRUE) {
    xplab_window_raise(re.caller, xctrl.cfg.raise);
    xplab_widget_destroy(&re.top_recall);

    /* cannot free data here because of sudden run stop                      */

  }
}



/* Deletes top window.

Revision: 30.10.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
gboolean
cb_recall_top_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(event);
  assert(widget);

  cb_recall_quit(NULL, NULL);

  return FALSE;
}
