/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <stdarg.h>
#include <time.h>
#include <float.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_statistic_be.h"
#include "xplab_statistic_fe.h"
#include "xplab_main_be.h"
#include "xplab_training_be.h"
#include "xplab_preferences_fe.h"
#include "xplab_image.h"
#include "xplab_algorithm.h"
#include "xplab_report.h"

/* local definitions *******************************************************/
#define XPLAB_STATISTIC_ICON_BGND    {222, 255, 222}

/* local prototypes ********************************************************/
static pic_def create_icon(pic_def icon_img, int icon_size_x, 
			   int icon_size_y, int icon_border,
			   gboolean has_aspect, gboolean max_contrast);
static void simage_to_state(bc_img *image_data_path, gint redraw_mag);
static void menu_popup_quit(GtkWidget *menu_widget, GtkSignalFunc del_cb,
			    GtkWidget *del_obj);
static void menu_dump_image(GtkWidget *menu_widget, GtkSignalFunc dump_cb,
			    gpointer udata);
static void menu_dump_histo(GtkWidget *menu_widget, 
			    GtkSignalFunc dump_histo_cb, bc_img *image_data);
static void preview_icon(GtkWidget *icon_box, int icon_size_x, int icon_size_y, int icon_border,
			 GtkSignalFunc button_cb, GtkSignalFunc enter_cb, 
			 GtkSignalFunc leave_cb, bc_img *image_data_path, 
			 char *preview_name, gboolean has_aspect, 
			 gboolean max_contrast);
static void prefs_window(GtkWidget *preview);

static void cb_stat_magnitude(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_dump_image(GTK_CONVENTION_MENUITEM_RTS_VOID);
static gboolean cb_expose_icon(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN);

/* local data **************************************************************/

/* Functions **************************************************************/

/* Exchanges the image in an image window.

Revision: 07.11.2004 E.Sack

Input : - pointer to data path
        - pointer to mouse CB
        - pointer to expose CB
Output: - window updated
Postcondition: - the top window and the associated data path are left
                 unchanged.
*/
void
rebuild_drawing(bc_img *image_data_path, GtkSignalFunc mouse_cb, 
		GtkSignalFunc expose_cb)
{
  area_def area;

  assert(image_data_path);

  if (GTK_IS_WIDGET(image_data_path->topwin) == FALSE) {
    return;
  }

  assert(GTK_IS_WIDGET(image_data_path->scroll.h_viewport) == TRUE);

  /* destroy viewport (drawing container) and associated signal handlers     */
  xplab_widget_destroy(&image_data_path->scroll.h_viewport);

  assert(GTK_IS_WIDGET(image_data_path->scroll.v_viewport) == TRUE);

  area = image_data_path->timg.pimg.ar;                     /* original size */

  if (image_data_path->magnitude > 0.0) {
    area.width  *= image_data_path->magnitude;            /* zooming defined */
    area.height *= image_data_path->magnitude;
  }

  image_data_path->scroll.sdrawing =   /* allocate new drawing in old window */
    horizontal_viewport(image_data_path->scroll.v_viewport,
			&image_data_path->scroll.h_viewport, 
			area, mouse_cb, expose_cb);
}



/* Sets the magnitude of backing image pixels. The complete image buffer is
   simulated by allocating only its visible part.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_stat_magnitude(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  bc_img      *image_data_path;
  const gchar *name;

  if (gtk_check_menu_item_get_active(checkmenuitem) == TRUE) {
    gboolean is_act;
    GSList   *radio_grp;

    image_data_path = (bc_img *)user_data;

    if (image_data_path == NULL) {
      return;
    }

    radio_grp =                                      /* don't free this list */
      gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(checkmenuitem));
    assert(radio_grp);

    do {                                               /* scan radio buttons */
      is_act = 
	gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(radio_grp->data));

      if (is_act == TRUE) {
	break;
      }

      radio_grp = radio_grp->next;
    } while (radio_grp);

    assert(is_act == TRUE);

    name =                                         /* don't free this string */
      gtk_label_get_text(GTK_LABEL(gtk_bin_get_child(GTK_BIN(radio_grp->data))));

    if (!strcmp(name, XPLAB_MAGN_X1_NAME)) {
      image_data_path->magnitude = XPLAB_MAGN_X1_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_X2_NAME)) {
      image_data_path->magnitude = XPLAB_MAGN_X2_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_X4_NAME)) {
      image_data_path->magnitude = XPLAB_MAGN_X4_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_D2_NAME)) {
      image_data_path->magnitude = XPLAB_MAGN_D2_VALUE;
    }
    else if (!strcmp(name, XPLAB_MAGN_D4_NAME)) {
      image_data_path->magnitude = XPLAB_MAGN_D4_VALUE;
    }
    else {
      return;                                     /* magnitude not supported */
    }

    rebuild_drawing(image_data_path,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
  }
}



/* Dumps the statistics image to hard disc.

Revision: 22.12.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dump_image(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  char   *name = NULL;
  bc_img *image_data_path;

  STSFY_GCC(menuitem);

  image_data_path = (bc_img *)user_data;

  if (image_data_path == NULL) {
    return;                                                    /* no data */
  }

  switch (image_data_path->dfmt) {
  case STATS_IMAGE_COLOR: 
    add_string(STATS_ICOL_KEY, &name);
    break;
  case STATS_IMAGE_GREY:
    add_string(STATS_IGRY_KEY, &name);
    break;
  case STATS_IMAGE_BMP: 
    add_string(STATS_IBMP_KEY, &name);
    break;
  case STATS_HISTO_COLOR: 
    add_string(STATS_HCOL_KEY, &name);
    break;
  case STATS_HISTO_GREY: 
    add_string(STATS_HGRY_KEY, &name);
    break;
  case STATS_HISTO_BMP: 
    add_string(STATS_HBMP_KEY, &name);
    break;
  case STATS_HISTO_RBS:
    add_string(STATS_HRBS_KEY, &name);
    break;
  case STATS_HISTO_SLOPE: 
    add_string(STATS_HSLOPE_KEY, &name);
    break;
  default:
    assert(0);
    break;
  }

  add_string(".image", &name);

  dump_pic_def(image_data_path->timg.pimg, name, image_data_path->timg.pimg.ar,
	       0, 0, 1, PNM_FORMAT_UNDEF);

  FREEMEM(name);
}



/* Creates the pop-up menu for statistic windows.

Revision: 05.08.2005 E.Sack

Input : - pointer to image data
Output: - image data updated
Usage:  - the pop-up menu is contained in an unbound widget.
Postcondition: - pop-up menu destroyed separately.
*/
GtkWidget *
popup_root(bc_img *image_data)
{
  GtkWidget *menu_bar, *file_entry;

  menu_bar = gtk_menu_bar_new();

  file_entry = gtk_menu_item_new_with_label("actions");
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  image_data->popmenu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), image_data->popmenu);

  menu_popup_quit(image_data->popmenu, GTK_SIGNAL_FUNC(cb_delete), 
		  image_data->topwin);

  menu_dump_image(image_data->popmenu, GTK_SIGNAL_FUNC(cb_dump_image),
		  (gpointer)image_data);

  menu_dump_histo(image_data->popmenu, GTK_SIGNAL_FUNC(cb_dump_histo),
		  image_data);

  xplab_menu_magn(image_data->popmenu, GTK_SIGNAL_FUNC(cb_stat_magnitude),
		  (gpointer)image_data, XPLAB_MAGN_X1_VALUE);

  gtk_widget_show_all(menu_bar);            /* container easily destroy-able */

  return menu_bar;
}



/* Creates a menu entry.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
        - callback function
        - pointer to callback object
Output: - menu updated
Usage: - a submenu entry is created containing a quit button.
*/
static void
menu_popup_quit(GtkWidget *menu_widget, GtkSignalFunc del_cb, 
		GtkWidget *del_obj)
{
  GtkWidget *menu_item;

  menu_item = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_widget), menu_item);

  g_signal_connect_swapped(menu_item, "activate", del_cb, del_obj);
}



/* Creates a menu entry.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
        - callback function
        - pointer to user data
Output: - menu updated
*/
static void
menu_dump_image(GtkWidget *menu_widget, GtkSignalFunc dump_cb, 
		gpointer udata)
{
  GtkWidget *menu_item;

  menu_item = gtk_menu_item_new_with_label("dump image");
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_widget), menu_item);

  g_signal_connect(G_OBJECT(menu_item), "activate", 
		   G_CALLBACK(dump_cb), udata);
}


/* Creates a menu entry.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
        - callback function
        - pointer to user data
Output: - menu updated
*/
static void
menu_dump_histo(GtkWidget *menu_widget, GtkSignalFunc dump_histo_cb, 
		bc_img *image_data)
{
  GtkWidget *menu_item;

  switch (image_data->dfmt) {
  case STATS_IMAGE_COLOR: 
  case STATS_IMAGE_GREY:
  case STATS_IMAGE_BMP: 
    return;                                      /* no histogram available */
    break;

  case STATS_HISTO_COLOR: 
  case STATS_HISTO_GREY: 
  case STATS_HISTO_BMP: 
  case STATS_HISTO_RBS:
  case STATS_HISTO_SLOPE: 
    break;

  default:
    assert(0);
    break;
  }

  menu_item = gtk_menu_item_new_with_label("dump histo");
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_widget), menu_item);

  g_signal_connect(G_OBJECT(menu_item), "activate", 
		   G_CALLBACK(dump_histo_cb), (gpointer)image_data);
}



/* Contains all format requests and their actions for statistics. All
   requests depends on the original image and its dependants.

Revision: 15.01.2005 E.Sack

Input : - pointer to broadcasted format
        - pointer to broadcast widget
        - pointer to request comment
Output: - format request updated
Precondition: - the broadcast widget may have a deferred mapping, what is
                handled by a double pointer.
              - pointer to widget to be exposed may be NULL (no exposition).
Postcondition: - all images requested are transformed to colored types.
               - some internal data buffers may be updated.
Usage: - the broadcast widget receives the broadcast signal that carries 
         new infos about the requested image which then should be 
         requested again for rebuilding its stistics. The widget enables
         the broadcasting signal.
*/
gboolean
stats_request_formats(bc_img *bci, GtkWidget **bc_widget, 
		      char *request_comment)
{
  PNM_FORMAT fmt;
  gboolean   is_original = FALSE, keep_request = FALSE;

  assert(bci);
  assert(bc_widget);

  if (bci->timg.typ == PNM_FORMAT_UNDEF) {
    keep_request = TRUE;                                /* no image loaded */
  }

  stats_release_image(bci, FALSE);  /* one request for bc_widget is enaugh */

  free_img_typ(&bci->timg);    /* free local image buffer (might be empty) */

  switch (bci->dfmt) {
  case STATS_IMAGE_COLOR: 
    bci->timg = xplab_request_image(PNM_FORMAT_ORIGINAL, request_comment, 
				    &bci->key, 0, bc_widget);

    if (bci->timg.pimg.col == NULL) {
      stats_release_image(bci, keep_request);
      free_img_typ(&bci->timg);
      break;                        /* original image is not a colored one */
    }

    /* no further actions on color image required                          */
    assert(bci->timg.typ == P6 || bci->timg.typ == P3);    /* check format */
    bci->timg.typ = P6;                                      /* set format */
    bci->img_changed = TRUE;
    is_original = TRUE;                           /* original format found */
    break;

  case STATS_IMAGE_GREY: 
    bci->timg = xplab_request_image(PNM_FORMAT_ORIGINAL, request_comment, 
				    &bci->key, 0, bc_widget);

    if (bci->timg.pimg.col) {  /* convert color image to grey-scaled image */
      assert(bci->timg.typ == P6 || bci->timg.typ == P3);  /* check format */
      rgb_to_grey(&bci->timg.pimg, 0);                              /* RGB */
    }
    else if (bci->timg.pimg.mem == NULL || 
	!(bci->timg.typ == P2 || bci->timg.typ == P5)) {   /* check format */
      stats_release_image(bci, keep_request);
      free_img_typ(&bci->timg);
      break;                    /* original image is not a grey-scaled one */
    }

    /* no further actions on grey-scaled image required                    */
    convert_to_color(&bci->timg.pimg, FALSE);              /* define color */
    bci->timg.typ = P6;                                      /* set format */
    bci->img_changed = TRUE;
    is_original = TRUE;                           /* original format found */
    break;

  case STATS_IMAGE_BMP: 
    bci->timg = xplab_request_image(P1, request_comment, 
				    &bci->key, 0, bc_widget);

    if (bci->timg.pimg.mem == NULL || 
	!(bci->timg.typ == P1 || bci->timg.typ == P4)) {   /* check format */
      stats_release_image(bci, keep_request);
      free_img_typ(&bci->timg);
      break;                            /* original image is not a BMP one */
    }

    /* no further actions on bit-mapped image required                     */
    convert_to_color(&bci->timg.pimg, FALSE);              /* define color */
    bci->timg.typ = P6;                                      /* set format */
    bci->img_changed = TRUE;
    fmt = xplab_image_loaded();
    is_original = (fmt == P1 || fmt == P4);       /* original format found */
    break;

  case STATS_HISTO_SLOPE:                         /* no broadcast key used */
    if (xctrl.set.th_hist.hst.cnt == 0) {        /* global slope histogram */
      break;                                            /* histogram empty */
    }

    /* no further actions on slope histogram required                      */
    bci->timg.pimg = histo_to_image(xctrl.set.th_hist, PICT_MAX+1, 
				    XPLAB_STATISTIC_HISTO_HEIGHT);

    bci->timg.typ = P6;                                      /* set format */
    bci->img_changed = TRUE;
    break;

  case STATS_HISTO_GREY:                          /* no broadcast key used */
    if (xctrl.set.grey_hist.hst.cnt == 0) {/* global grey-scaled histogram */
      break;                                            /* histogram empty */
    }

    /* no further actions on grey-scaled histogram required                */
    bci->timg.pimg = histo_to_image(xctrl.set.grey_hist, PICT_MAX+1,
				    XPLAB_STATISTIC_HISTO_HEIGHT);

    bci->timg.typ = P6;                                      /* set format */
    bci->img_changed = TRUE;
    break;

  case STATS_HISTO_COLOR:
    format_stats_histo_color(bci, bc_widget, request_comment, keep_request);
    break;

  case STATS_HISTO_BMP:
    format_stats_histo_bmp(bci, bc_widget, request_comment, keep_request);
    break;

  case STATS_HISTO_RBS:                           /* no broadcast key used */
    format_stats_histo_rbs(bci);
    break;

  default:
    break;
  }

  return is_original;
}



/* Releases image for statistics.

Revision: 15.01.2005 E.Sack

Input : - pointer to broadcast format
        - flag (TRUE: inhibit function, else perform)
Output: - image released
*/
void
stats_release_image(bc_img *bci, gboolean keep_request)
{
  PNM_FORMAT pnm_format = PNM_FORMAT_UNDEF;

  if (keep_request == TRUE) {
    return;
  }

  switch (bci->dfmt) {
  case STATS_IMAGE_COLOR: 
  case STATS_IMAGE_GREY:
  case STATS_HISTO_COLOR: 
    pnm_format = PNM_FORMAT_ORIGINAL;
    break;

  case STATS_IMAGE_BMP: 
  case STATS_HISTO_BMP: 
    pnm_format = P1;
    break;

  case STATS_HISTO_RBS:                                /* nothing to release */
  case STATS_HISTO_GREY: 
  case STATS_HISTO_SLOPE:
    break;

  case STATS_IMAGE_ORIGINAL:                /* no direct access - no release */
  default:
    assert(0);
    break;
  }

  xplab_release_image(pnm_format, &bci->key, 0);/* reset key and broadcasting */
}



/* Redraws the original image window.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
gboolean
cb_expose(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN)
{
  bc_img    *image_data_path;
  GtkWidget *top_level;

  STSFY_GCC(event);

  assert(user_data==NULL);
  assert(widget);

  if (GTK_IS_WIDGET(widget) == FALSE) {
    return FALSE;
  }

  top_level = gtk_widget_get_toplevel(widget);

  if (GTK_IS_OBJECT(top_level) == FALSE) {
    return FALSE;
  }

  image_data_path = (bc_img *)g_object_get_data(G_OBJECT(top_level), 
						XPLAB_ASSN_KEY_0);

  if (image_data_path == NULL) {
    return FALSE;
  }

  simage_to_state(image_data_path, image_data_path->magnitude > 0.0);

  return FALSE;
}



/* Redraws the original image window.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
*/
static gboolean
cb_expose_icon(GTK_CONVENTION_DIALOG_EXPOSE_RTS_GBOOLEAN)
{
  int    x_start_to_render = 0, y_start_to_render = 0;
  bc_img *image_data_path;

  STSFY_GCC(event);

  assert(user_data==NULL);
  assert(widget);

  if (GTK_IS_WIDGET(widget) == FALSE) {
    return FALSE;
  }

  image_data_path = (bc_img *)g_object_get_data(G_OBJECT(widget),
						XPLAB_ASSN_KEY_0);

  if (image_data_path == NULL) {
    return FALSE;
  }

  assert(GDK_IS_DRAWABLE(image_data_path->preview->window) == TRUE);

  gdk_draw_rgb_image(image_data_path->preview->window,  /* image destination */
		     image_data_path->preview->style->fg_gc[GTK_STATE_NORMAL],
		     x_start_to_render, y_start_to_render, 
		     image_data_path->bicon.ar.width, 
		     image_data_path->bicon.ar.height,
		     GDK_RGB_DITHER_NONE,
		     (guchar *)image_data_path->bicon.col,   /* image source */
		     image_data_path->bicon.ar.width * 3);

  return FALSE;
}



/* Creates backing image according to state.

Revision: 11.12.2004 E.Sack

Input : - pointer to image data
        - flag (0: don't reallocate image buffer, 1: reallocate)
Output: - image buffer and window updated
*/
static void
simage_to_state(bc_img *image_data_path, gint redraw_buf)
{
  gint force;

  if (GTK_IS_WIDGET(image_data_path->topwin) == FALSE) {
    return;
  }

  if (image_data_path->timg.pimg.col == NULL) {
    return;
  }

  if (redraw_buf) {
    force = image_data_path->img_changed == TRUE ||
      image_data_path->magnitude_old != image_data_path->magnitude;

    /* allocate image buffer */
    magnitude_allocation(image_data_path->timg.pimg.col, 
			 image_data_path->timg.pimg.ar, 
			 image_data_path->scroll.h_adj->value, 
			 image_data_path->scroll.v_adj->value, 
			 image_data_path->magnitude,
			 &image_data_path->magn_img,
			 image_data_path->scroll.h_adj->page_size, 
			 image_data_path->scroll.v_adj->page_size, 
			 image_data_path->scroll.sdrawing->allocation.width,
			 image_data_path->scroll.sdrawing->allocation.height, 
			 force);

    image_data_path->magnitude_old = image_data_path->magnitude;
    image_data_path->img_changed = FALSE;
  }

  xplab_draw_image(image_data_path->timg.pimg.col,
		   image_data_path->timg.pimg.ar,
		   image_data_path->magn_img,
		   image_data_path->magnitude,
		   image_data_path->scroll.h_adj,
		   image_data_path->scroll.v_adj,
		   image_data_path->scroll.sdrawing);
}



/* Deletes top window.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - propagation flag
Usage:  - the delete callback sometimes needs user data because of widget
          deletion and initialization problem. That means the normal
          widget pointer delivered by the GTK interface cannot be set to NULL
          because of copy-by-value. Use user data instead or (as here) 
          data bound to widget.
*/
gboolean
cb_delete(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  bc_img    *image_data_path = NULL;
  GtkWidget *top_level;

  STSFY_GCC(widget);
  STSFY_GCC(event);
  STSFY_GCC(user_data);

  if (GTK_IS_WIDGET(widget) == FALSE) {
    return FALSE;
  }

  top_level = gtk_widget_get_toplevel(widget);
  assert(GTK_IS_OBJECT(top_level) == TRUE);

  image_data_path = (bc_img *)g_object_get_data(G_OBJECT(top_level),
						XPLAB_ASSN_KEY_0);

  assert(image_data_path);                  /* important for delete-strategy */
  assert(image_data_path->topwin == top_level);
  assert(image_data_path->topwin == widget);
  assert(GTK_IS_WIDGET(image_data_path->topwin) == TRUE);

  xplab_widget_destroy(&image_data_path->topwin);      /* destroy top widget */
  xplab_widget_destroy(&image_data_path->poproot);  /*destroy unbound widget */

  free_img_typ(&image_data_path->timg);          /* free other relevant data */

  return FALSE;
}



/* Creates a color icon from an image. If the image is empty a default 
   icon is created.

Revision: 20.12.2004 E.Sack

Input : - original image
        - icon sizes
Output: - icon image
*/
static pic_def
create_icon(pic_def icon_img, int icon_size_x, int icon_size_y, 
	    int icon_border, gboolean has_aspect, gboolean max_contrast)
{
  int     len;
  pic_def icon, spare;
  col_def icon_bgnd = XPLAB_STATISTIC_ICON_BGND;

  init_pic_def(&icon);
  init_pic_def(&spare);

  assert(icon_size_x>0);
  assert(icon_size_y>0);
  assert(icon_border>=0);

  spare.ar.width = icon_size_x;                          /* without border */
  spare.ar.height = icon_size_y;

  len = AREA_SIZE(spare.ar);

  CALLOC(col_def, len, spare.col);             /* background written first */

  while (len--) {                                       /* icon background */
    spare.col[len]= icon_bgnd;
  }

  if (icon_img.col) {                               /* may be non-existent */
    pic_def  itmp;
    area_def max_ar;

    init_pic_def(&itmp);

    max_ar.width = icon_size_x;                          /* without border */
    max_ar.height = icon_size_y;

    itmp = scale_down_image(icon_img, max_ar, has_aspect);

    if (max_contrast == TRUE) {
      max_contrast_rgb(itmp);         /* contrast enhancement for the icon */
    }

    copy_color(itmp, spare, itmp.ar, 0, 0, 0, 0);

    free_pic_def(&itmp);
  }

  /* add icon border */
  icon.ar.width = icon_size_x + 2*icon_border;          /* centered border */
  icon.ar.height = icon_size_y + 2*icon_border;

  len = AREA_SIZE(icon.ar);

  CALLOC(col_def, len, icon.col);                   /* icon always created */

  while (len--) {                                       /* icon background */
    icon.col[len]= icon_bgnd;
  }

  copy_color(spare, icon, spare.ar, 0, 0, icon_border, icon_border);

  free_pic_def(&spare);

  return icon;
}



/* Attaches an icon entry to table.

Revision: 05.08.2005 E.Sack

Input : - pointer to table
        - table attach parameters
        - icon sizes
        - icon border size
        - callback function
        - pointer to broadcast widget
        - pointer to image data
        - pointer to icon name string
        - icon image format
        - pointer to flag (0: don't display original image, 1: display)
Output: - table updated
        - image windows displayed by preferences
        - flag updated
*/
void
attach_icon(GtkWidget *table, guint left_attach, guint right_attach, 
	    guint top_attach, guint bottom_attach, int icon_size_x, 
	    int icon_size_y, int icon_border, GtkSignalFunc button_cb, 
	    GtkSignalFunc enter_cb, GtkSignalFunc leave_cb, 
	    GtkWidget **bc_widget, bc_img *image_data_path, 
	    gboolean has_aspect, gboolean max_contrast, char *preview_name,
	    STATS_FORMAT fmt, stats_win *do_original)
{
  gboolean  is_original;
  GtkWidget *icon_box, *icon_label;

  assert(bc_widget);
  assert(image_data_path);
  assert(do_original);
  assert(left_attach<right_attach);
  assert(top_attach<bottom_attach);

  if (GTK_IS_WIDGET(table) == FALSE) {
    return;
  }

  image_data_path->dfmt = fmt;                                 /* set format */

  is_original = stats_request_formats(image_data_path, bc_widget, 
				      XPLAB_STATISTIC_TITLE_ITEM);

  if (do_original->display_active == TRUE) {
    if (is_original == TRUE) {
      do_original->display_active = FALSE;          /* original format found */

      if (image_data_path->pwin.display_active == FALSE) {
	image_data_path->pwin.wsize = do_original->wsize;
	image_data_path->pwin.display_active = TRUE; /* window size strategy */
      }
    }
  }

  icon_box = gtk_vbox_new(FALSE, 0);             /* pack icon into event box */

  gtk_table_attach(GTK_TABLE(table), icon_box,
		   left_attach, right_attach, top_attach, bottom_attach, 
		   GTK_FILL, GTK_FILL, 0, 0);

  preview_icon(icon_box, icon_size_x, icon_size_y, icon_border, button_cb, 
	       enter_cb, leave_cb, image_data_path, preview_name, has_aspect, 
	       max_contrast);

  icon_label = gtk_label_new(preview_name);      /* pack icon label into box */

  gtk_box_pack_start(GTK_BOX(icon_box), icon_label, FALSE, FALSE, 0);

  prefs_window(image_data_path->preview);                /* automatic window */
}



/* Creates statistic windows using preferences settings.

Revision: 05.08.2005 E.Sack

Input : - pointer to widget
Output: - -
Usage:  - widget should be initialised with statistic data (data path)
          and is normally the icon contained in the top statistic window.
        - the indirection for getting the information over the preview 
          widget makes this function usable for callback handlers.
*/
static void
prefs_window(GtkWidget *preview)
{
  bc_img   *image_data;
  gpointer image_data_path;
  const gchar *wname;

  if (GTK_IS_WIDGET(preview) && GTK_IS_OBJECT(preview)) {
    image_data_path = g_object_get_data(G_OBJECT(preview), XPLAB_ASSN_KEY_0);

    image_data = (bc_img *)image_data_path;
    assert(image_data);

    if (image_data->pwin.display_active == TRUE) {
      wname = gtk_widget_get_name(GTK_WIDGET(preview));
      stats_window(image_data, wname, FALSE, image_data->pwin.wsize);
      image_data->pwin.display_active = FALSE;      /* reset prefs requested */
    }
  }
}



/* Creates a color icon and sets up further image data.

Revision: 05.08.2005 E.Sack

Input : - pointer to icon container
        - icon sizes
        - icon border size
        - pointer to callback functions
        - pointer to image data
        - pointer to icon name string
        - flags (aspect, contrast)
Output: - image data updated
Usage:  - icon is derived from the image data and fit to the icon sizes.
        - signal function used as handler for mouse button events on icon,
          normally to set up a top level window.
        - the image data path is rooted to the created icon drawing and 
          can be used in different top-level-windows. 
          Consequently the root of the path must not be destroyed. That
          means, destroying the statistics window after destroying all other 
          windows else hide the statistics window.
*/
static void
preview_icon(GtkWidget *icon_box, int icon_size_x, int icon_size_y, 
	     int icon_border, GtkSignalFunc button_cb, GtkSignalFunc enter_cb, 
	     GtkSignalFunc leave_cb, bc_img *image_data_path, 
	     char *preview_name, gboolean has_aspect, gboolean max_contrast)
{
  int       err;
  GtkWidget *preview;

  assert(image_data_path);
  assert(icon_size_x>0);
  assert(icon_size_y>0);

  free_pic_def(&image_data_path->bicon);

  image_data_path->bicon =                           /* needs static buffer */
    create_icon(image_data_path->timg.pimg, icon_size_x, icon_size_y, 
		icon_border, has_aspect, max_contrast);

  preview = gtk_drawing_area_new();

  g_object_set_data(G_OBJECT(preview), XPLAB_ASSN_KEY_0, 
		    (gpointer)image_data_path);
  gtk_widget_show(preview);

  err = check_name(preview_name);

  if (!err) {
    gtk_widget_set_name(preview, preview_name);
  }
  else {
    gtk_widget_set_name(preview, "unknown");
  }

  if (button_cb) {                                /* events must be set ASAP */
    gtk_widget_add_events(preview, GDK_BUTTON_PRESS_MASK);

    g_signal_connect(G_OBJECT(preview), "button-press-event", 
		     G_CALLBACK(button_cb), NULL);
  }

  if (enter_cb || leave_cb) {
    assert(enter_cb);                                 /* reasonably both CBs */
    assert(leave_cb);

    gtk_widget_add_events(preview, GDK_ENTER_NOTIFY_MASK | 
                                   GDK_LEAVE_NOTIFY_MASK);

    g_signal_connect(G_OBJECT(preview), "enter-notify-event", 
		     G_CALLBACK(enter_cb), (gpointer)&image_data_path->dfmt);
    g_signal_connect(G_OBJECT(preview), "leave-notify-event", 
		     G_CALLBACK(leave_cb), (gpointer)&image_data_path->dfmt);
    g_signal_connect(G_OBJECT(preview), "expose_event", 
		     G_CALLBACK(cb_expose_icon), NULL);
  }

  gtk_widget_set_size_request(GTK_WIDGET(preview), 
			      image_data_path->bicon.ar.width, 
			      image_data_path->bicon.ar.height);

  gtk_box_pack_start(GTK_BOX(icon_box), preview, FALSE, FALSE, 0);
  gtk_widget_show_all(icon_box);

  /* drawable must be realized                                               */
  assert(GDK_IS_DRAWABLE(preview->window) == TRUE);

  image_data_path->preview = preview;
}
