/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <stdarg.h>
#include <float.h>
#include <time.h>
#include <gtk/gtk.h>
#include <libgen.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_statistic_fe.h"
#include "xplab_statistic_be.h"
#include "xplab_main_be.h"
#include "xplab_main_fe.h"
#include "xplab_preferences_fe.h"
#include "xplab_training_be.h"
#include "xplab_algorithm.h"
#include "xplab_image.h"
#include "xplab_report.h"

/* local definitions *******************************************************/
#define XPLAB_STATISTIC_TITLE_COLOR         "I1"
#define XPLAB_STATISTIC_TITLE_GREY          "I2"
#define XPLAB_STATISTIC_TITLE_BITMAP        "I3"
#define XPLAB_STATISTIC_TITLE_HISTO_COLOR   "H1"
#define XPLAB_STATISTIC_TITLE_HISTO_GREY    "H2"
#define XPLAB_STATISTIC_TITLE_HISTO_BMP     "H3"
#define XPLAB_STATISTIC_TITLE_HISTO_RBS     "H4"
#define XPLAB_STATISTIC_TITLE_HISTO_SLOPE   "H5"
#define XPLAB_STATISTIC_WINDOW_WIDTH        511
#define XPLAB_STATISTIC_WINDOW_HEIGHT       342
#define XPLAB_STATISTIC_ICON_WIDTH          37
#define XPLAB_STATISTIC_ICON_HEIGHT         48
#define XPLAB_STATISTIC_ICON_BORDER         1
#define XPLAB_STATISTIC_TABLE_BORDER        20
#define XPLAB_STATISTIC_CONTEXT_DESCRIPTION "statusbar statistic"

typedef struct {
  GtkWidget *top_statistic;
  GtkWidget *refresh_button, **caller, *status_bar, *tblcont, *table;
  gint      context_id, message_id, basic_message_id;
  stats_win iorig, icon_contrast_max;
  bc_img    bc_image_color, bc_image_grey, bc_image_bmp, bc_histo_slope;
  bc_img    bc_histo_color, bc_histo_grey, bc_histo_bmp, bc_histo_rbs;
  hst_def   bmp_hist;                              /* histogram of BMP image */
  hst_def   rbs_hist;                                /* histogram of RB size */
  hst_def   red_hist, green_hist, blue_hist;    /* histograms of color image */
} statistic_def;

/* local prototypes ********************************************************/
static void init_statistic_def(statistic_def *pp);
static void free_statistic_def(statistic_def *pp);
static void statistic_menu_bar(GtkWidget *vbox);
static void statistic_menu_entry_file(GtkWidget *menu_bar);
static void statistic_menu_entry_refresh(GtkWidget *menu_bar);
static void statistic_file_menu_item_quit(GtkWidget *file_menu);
static void statistic_status_bar(GtkWidget *vbox);
static void release_images_all(void);
static void free_images_all(void);
static gboolean request_stats_prefs(void);
static void statistic_table(GtkWidget *vbox);
static void statistic_window(void);
static void image_actions(GdkEventButton *event, bc_img *image_data);
static void statistic_refresh(void);
static void statistic_map(void);

static void cb_statistic_hide(GTK_CONVENTION_MENUITEM_RTS_VOID);
static gboolean cb_statistic_menu_event(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN);
static void cb_statistic_refresh(GTK_CONVENTION_MENUITEM_RTS_VOID);
static gboolean cb_statistic_refresh_event(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN);
static gboolean cb_icon_button(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN);
static gboolean cb_icon_enter(GTK_CONVENTION_ENTER_NOTIFY_RTS_GBOOLEAN);
static gboolean cb_icon_leave(GTK_CONVENTION_LEAVE_NOTIFY_RTS_GBOOLEAN);

/* local data **************************************************************/

static statistic_def st;

/* Functions **************************************************************/

static void
init_statistic_def(statistic_def *pp)
{
  assert(pp);

  pp->top_statistic = NULL;                                  /* zero pointer */
  pp->refresh_button = NULL;                                 /* zero pointer */
  pp->caller = NULL;                                         /* zero pointer */
  pp->status_bar = NULL;                                     /* zero pointer */
  pp->tblcont = NULL;                                        /* zero pointer */
  pp->table = NULL;                                          /* zero pointer */

  pp->context_id = pp->message_id = pp->basic_message_id = 0;

  init_stats_win(&pp->iorig);
  init_stats_win(&pp->icon_contrast_max);

  init_bc_img(&pp->bc_image_color);
  init_bc_img(&pp->bc_image_grey);
  init_bc_img(&pp->bc_image_bmp);
  init_bc_img(&pp->bc_histo_color);
  init_bc_img(&pp->bc_histo_slope);
  init_bc_img(&pp->bc_histo_grey);
  init_bc_img(&pp->bc_histo_bmp);
  init_bc_img(&pp->bc_histo_rbs);

  init_hst_def(&pp->bmp_hist);
  init_hst_def(&pp->rbs_hist);
  init_hst_def(&pp->red_hist);
  init_hst_def(&pp->green_hist);
  init_hst_def(&pp->blue_hist);
}


static void
free_statistic_def(statistic_def *pp)
{
  assert(pp);
                           /* don't free GTK stuff */

  free_stats_win(&pp->iorig);
  free_stats_win(&pp->icon_contrast_max);

  free_bc_img(&pp->bc_image_color);
  free_bc_img(&pp->bc_image_grey);
  free_bc_img(&pp->bc_image_bmp);
  free_bc_img(&pp->bc_histo_color);
  free_bc_img(&pp->bc_histo_slope);
  free_bc_img(&pp->bc_histo_grey);
  free_bc_img(&pp->bc_histo_bmp);
  free_bc_img(&pp->bc_histo_rbs);

  free_hst_def(&pp->bmp_hist);
  free_hst_def(&pp->rbs_hist);
  free_hst_def(&pp->red_hist);
  free_hst_def(&pp->green_hist);
  free_hst_def(&pp->blue_hist);

  init_statistic_def(pp);
}



/* Initializes the statistic interface.

Revision: 30.10.2004 E.Sack

Input : - -
Output: - -
*/
void
xplab_init_statistic(void)
{
  init_statistic_def(&st);
}



/* Frees the statistic interface.

Revision: 30.10.2004 E.Sack

Input : - -
Output: - -
*/
void
xplab_free_statistic(void)
{
  free_statistic_def(&st);

  xplab_init_statistic();
}



/* Builds an entry for a menu bar.

Revision: 23.08.2005 E.Sack

Input : - menu bar
Output: - pointer to menu bar caller
*/
void
xplab_statistic_menu(GtkWidget *menu_bar, GtkWidget **caller)
{
  gboolean  prefs_created;
  GtkWidget *statistic_entry;

  st.caller = caller;

  statistic_entry = gtk_menu_item_new_with_label(XPLAB_STATISTIC_TITLE_ITEM);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), statistic_entry);

  g_signal_connect(G_OBJECT(statistic_entry), "event", 
		   G_CALLBACK(cb_statistic_menu_event), NULL);

  prefs_created = request_stats_prefs();              /* request preferences */

  if (prefs_created == TRUE) {
    statistic_window();              /* statistic top window without mapping */
  }
}



/* Frees all unused image buffers.

Revision: 15.12.2004 E.Sack

Input : - -
Output: - -
Usage:  - to be called after all calculations are made and if there is no 
          more need for images. As there are still statistics depending on
          image data, the requests for broadcasting is left untouched.
*/
static void
free_images_all(void)
{
  if (GTK_IS_WIDGET(st.bc_image_color.topwin) == FALSE) {
    free_img_typ(&st.bc_image_color.timg);
    free_pic_def(&st.bc_image_color.magn_img);
  }

  if (GTK_IS_WIDGET(st.bc_image_grey.topwin) == FALSE) {
    free_img_typ(&st.bc_image_grey.timg);
    free_pic_def(&st.bc_image_grey.magn_img);
  }

  if (GTK_IS_WIDGET(st.bc_image_bmp.topwin) == FALSE) {
    free_img_typ(&st.bc_image_bmp.timg);
    free_pic_def(&st.bc_image_bmp.magn_img);
  }

  if (GTK_IS_WIDGET(st.bc_histo_slope.topwin) == FALSE) {
    free_img_typ(&st.bc_histo_slope.timg);
    free_pic_def(&st.bc_histo_slope.magn_img);
  }

  if (GTK_IS_WIDGET(st.bc_histo_grey.topwin) == FALSE) {
    free_img_typ(&st.bc_histo_grey.timg);
    free_pic_def(&st.bc_histo_grey.magn_img);
  }

  if (GTK_IS_WIDGET(st.bc_histo_color.topwin) == FALSE) {
    free_img_typ(&st.bc_histo_color.timg);
    free_pic_def(&st.bc_histo_color.magn_img);
    free_hst_def(&st.red_hist);
    free_hst_def(&st.green_hist);
    free_hst_def(&st.blue_hist);
  }

  if (GTK_IS_WIDGET(st.bc_histo_bmp.topwin) == FALSE) {
    free_img_typ(&st.bc_histo_bmp.timg);
    free_pic_def(&st.bc_histo_bmp.magn_img);
    free_hst_def(&st.bmp_hist);
  }

  if (GTK_IS_WIDGET(st.bc_histo_rbs.topwin) == FALSE) {
    free_img_typ(&st.bc_histo_rbs.timg);
    free_pic_def(&st.bc_histo_rbs.magn_img);
    free_hst_def(&st.rbs_hist);
  }
}



/* Contains all image releases.

Revision: 05.11.2004 E.Sack

Input : - -
Output: - -
Usage:  - releasing an image in statistics package means above all
          switching off the broadcasting signal.
*/
static void
release_images_all(void)
{
  stats_release_image(&st.bc_image_color, FALSE);
  stats_release_image(&st.bc_image_grey, FALSE);
  stats_release_image(&st.bc_image_bmp, FALSE);
  stats_release_image(&st.bc_histo_color, FALSE);
  stats_release_image(&st.bc_histo_slope, FALSE);
  stats_release_image(&st.bc_histo_grey, FALSE);
  stats_release_image(&st.bc_histo_bmp, FALSE);
  stats_release_image(&st.bc_histo_rbs, FALSE);
}



/* Creates top window for statistic.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - flag (0: no window creation triggered from preferences,
                1: window creation triggered from preferences)
Postcondition: - if flag is unset, statistic top window is displayed,
                 else statistic top window is created however hidden.
Usage: - whenever any window is created (automatic by preference settings
         or by the user) the statistic top level window is built. Its 
         icons hold all the statistic data.
*/
static void
statistic_window(void)
{
  GtkWidget *vbox, *sep;

  /* set the training top level widget                                       */
  st.top_statistic = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(st.top_statistic), XPLAB_STATISTIC_TITLE_ITEM);
  gtk_widget_realize(st.top_statistic);              /* because of drawables */

  vbox = gtk_vbox_new(FALSE, 0); /* put into top level widget a vertical box */
  gtk_container_add(GTK_CONTAINER(st.top_statistic), vbox);

  statistic_menu_bar(vbox);               /* put into vertical box a menubar */

  sep = gtk_hseparator_new();           /* put into vertical box a separator */
  gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 0);

  /* put into vertical box a drawable                                        */
  st.tblcont = gtk_vbox_new(FALSE, 0);                    /* table container */
  gtk_container_set_border_width(GTK_CONTAINER(st.tblcont),
				 XPLAB_STATISTIC_TABLE_BORDER);
  gtk_box_pack_start(GTK_BOX(vbox), st.tblcont, TRUE, TRUE, 0);

  statistic_table(st.tblcont);                           /* thumbnails image */

  statistic_status_bar(vbox);          /* put into vertical box a status bar */

  g_signal_connect(G_OBJECT(st.top_statistic), "delete-event", 
		   G_CALLBACK(cb_statistic_delete_all), NULL);

  xplab_set_window_icon(st.top_statistic);                    /* window icon */

  free_images_all();                                    /* free unused space */
}



/* Builds a status bar.

Revision: 30.10.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
statistic_status_bar(GtkWidget *vbox)
{
  int err;

  /* put into container one status bar */
  st.status_bar = gtk_statusbar_new();
  gtk_box_pack_start(GTK_BOX (vbox), st.status_bar, FALSE, FALSE, 0);

  /* put into the status bar a message */
  st.context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(st.status_bar),
					       XPLAB_STATISTIC_CONTEXT_DESCRIPTION);

  err = check_name(xctrl.set.image_filename);

  /* basic status message */
  st.basic_message_id = 
  gtk_statusbar_push(GTK_STATUSBAR(st.status_bar), st.context_id, 
		     err ? XPLAB_NO_IMAGE_MSG : 
		     basename(xctrl.set.image_filename));
}



/* Builds a menu bar.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
statistic_menu_bar(GtkWidget *vbox)
{
  GtkWidget *menu_bar;

  menu_bar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menu_bar, FALSE, FALSE, 0);

  statistic_menu_entry_file(menu_bar);                        /* menu file */
  statistic_menu_entry_refresh(menu_bar);                  /* menu refresh */
  xplab_menu_help(menu_bar, &st.top_statistic);               /* menu help */

  gtk_widget_show_all(menu_bar);
}



/* Builds an entry for a menu bar.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu bar
Output: - menu bar updated
*/
static void
statistic_menu_entry_file(GtkWidget *menu_bar)
{
  GtkWidget *file_entry, *submenu, *sep;

  file_entry = gtk_menu_item_new_with_label(MENU_BUTTON_FILE);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), file_entry);

  submenu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(file_entry), submenu);

  xplab_preferences_menu(submenu, &st.top_statistic);       /* preferences */

  sep = gtk_menu_item_new();
  gtk_menu_shell_append(GTK_MENU_SHELL(submenu), sep);

  statistic_file_menu_item_quit(submenu);                     /* item quit */
}



/* Builds an entry for a menu bar.

Revision: 05.08.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static void
statistic_menu_entry_refresh(GtkWidget *menu_bar)
{
  st.refresh_button = gtk_menu_item_new_with_label(SINGLE_BUTTON_REFRESH);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), st.refresh_button);

  g_signal_connect(G_OBJECT(st.refresh_button), "event", 
		   G_CALLBACK(cb_statistic_refresh_event), NULL);

  g_signal_connect(G_OBJECT(st.refresh_button), "activate", 
		   G_CALLBACK(cb_statistic_refresh), NULL);
}



/* Refreshes statistics derived from original image.

Revision: 17.08.2005 E.Sack

Input : - GTK convention
Output: - -
Usage : Handler for events.
*/
static gboolean
cb_statistic_refresh_event(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN)
{
  gboolean filtered;

  STSFY_GCC(widget);
  STSFY_GCC(user_data);

  filtered = filter_event_type(event, GDK_BUTTON_PRESS);

  if (filtered == FALSE) {
    return FALSE;                      /* propagate further event processing */
  }

  statistic_refresh();

  return TRUE;                      /* prevent from further event processing */
}



/* Refreshes statistics derived from original image.

Revision: 17.08.2005 E.Sack

Input : - GTK convention
Output: - -
Usage : Handler for signals.
*/
static void
cb_statistic_refresh(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(menuitem);
  STSFY_GCC(user_data);

  statistic_refresh();
}



/* Refreshes statistics derived from original image.

Revision: 17.08.2005 E.Sack

Input : - -
Output: - statistics refreshed
*/
static void
statistic_refresh(void)
{
  gboolean err;

  if (GTK_IS_WIDGET(st.top_statistic) == TRUE) {  /* validate broadcasting */
    st.icon_contrast_max = xplab_request_stats_prefs(STATS_ICON_CONTRAST_MAX);

    statistic_table(st.tblcont);                         /* new thumbnails */

    rebuild_drawing(&st.bc_image_color,                      /* new images */
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
    rebuild_drawing(&st.bc_image_grey,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
    rebuild_drawing(&st.bc_image_bmp,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
    rebuild_drawing(&st.bc_histo_color,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
    rebuild_drawing(&st.bc_histo_slope,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
    rebuild_drawing(&st.bc_histo_grey,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
    rebuild_drawing(&st.bc_histo_bmp,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));
    rebuild_drawing(&st.bc_histo_rbs,
		    GTK_SIGNAL_FUNC(cb_image), GTK_SIGNAL_FUNC(cb_expose));

    free_images_all();                                /* free unused space */

    err = check_name(xctrl.set.image_filename);

    if (!err) {                                      /* refresh status bar */
      gtk_statusbar_remove(GTK_STATUSBAR(st.status_bar), st.context_id,
			   st.basic_message_id);

      st.basic_message_id = 
	gtk_statusbar_push(GTK_STATUSBAR(st.status_bar), st.context_id, 
			   basename(xctrl.set.image_filename));
    }
  }
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
Usage: - don't really quits but hides the statistic top window.
*/
static void
statistic_file_menu_item_quit(GtkWidget *file_menu)
{
  GtkWidget *file_menu_quit;

  file_menu_quit = gtk_menu_item_new_with_label(DIALOG_BUTTON_QUIT);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_quit);

  g_signal_connect(G_OBJECT(file_menu_quit), "activate", 
		   G_CALLBACK(cb_statistic_hide), (gpointer)st.caller);
}



/* Starts statistic interface by events.

Revision: 23.08.2005 E.Sack

Input : - GTK convention
Output: - GTK convention
*/
static gboolean
cb_statistic_menu_event(GTK_CONVENTION_WIDGET_EVENT_RTS_GBOOLEAN)
{
  gboolean filtered;

  STSFY_GCC(widget);
  STSFY_GCC(user_data);

  filtered = filter_event_type(event, GDK_BUTTON_PRESS);

  if (filtered == FALSE) {
    return FALSE;                      /* propagate further event processing */
  }

  statistic_map();

  return TRUE;                      /* prevent from further event processing */
}



/* Maps statistic window.

Revision: 23.08.2005 E.Sack

Input : - -
Output: - statistic window mapped
*/
static void
statistic_map(void)
{
  gboolean raised;

  raised = xplab_window_raise(&st.top_statistic, 1);

  if (raised == FALSE) {
    statistic_window();                              /* statistic top window */
    gtk_widget_show_all(st.top_statistic);
  }
  else {
    if (!GTK_WIDGET_MAPPED(st.top_statistic)) {
      gtk_widget_show_all(st.top_statistic);            /* map hidden window */
    }
  }
}



/* Fetches preferences settings.

Revision: 13.11.2004 E.Sack

Input : - GTK convention
Output: - -
Usage:  - settings used for automatic image display after start-up.
*/
static gboolean
request_stats_prefs(void)
{
  int sum = 0;

  st.icon_contrast_max = xplab_request_stats_prefs(STATS_ICON_CONTRAST_MAX);

  st.iorig = xplab_request_stats_prefs(STATS_IMAGE_ORIGINAL);
  sum += st.iorig.display_active == TRUE ? 1 : 0;

  st.bc_image_color.pwin = xplab_request_stats_prefs(STATS_IMAGE_COLOR);
  sum += st.bc_image_color.pwin.display_active == TRUE ? 1 : 0;

  st.bc_image_grey.pwin = xplab_request_stats_prefs(STATS_IMAGE_GREY);
  sum += st.bc_image_grey.pwin.display_active == TRUE ? 1 : 0;

  st.bc_image_bmp.pwin = xplab_request_stats_prefs(STATS_IMAGE_BMP);
  sum += st.bc_image_bmp.pwin.display_active == TRUE ? 1 : 0;

  st.bc_histo_color.pwin = xplab_request_stats_prefs(STATS_HISTO_COLOR);
  sum += st.bc_histo_color.pwin.display_active == TRUE ? 1 : 0;

  st.bc_histo_grey.pwin = xplab_request_stats_prefs(STATS_HISTO_GREY);
  sum += st.bc_histo_grey.pwin.display_active == TRUE ? 1 : 0;

  st.bc_histo_bmp.pwin = xplab_request_stats_prefs(STATS_HISTO_BMP);
  sum += st.bc_histo_bmp.pwin.display_active == TRUE ? 1 : 0;

  st.bc_histo_rbs.pwin = xplab_request_stats_prefs(STATS_HISTO_RBS);
  sum += st.bc_histo_rbs.pwin.display_active == TRUE ? 1 : 0;

  st.bc_histo_slope.pwin = xplab_request_stats_prefs(STATS_HISTO_SLOPE);
  sum += st.bc_histo_slope.pwin.display_active == TRUE ? 1 : 0;

  return sum ? TRUE : FALSE;
}



/* Deletes top window.

Revision: 07.11.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
gboolean
cb_statistic_delete_all(GTK_CONVENTION_WINDOW_DELETE_RTS_GBOOLEAN)
{
  STSFY_GCC(user_data);

  assert(event);
  assert(widget);

  cb_statistic_quit_all(NULL, NULL);

  return FALSE;
}



/* Quits GTK main loop.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
Usage:  - calling this function only should be made if the statistics package
          really shall be shoot down. All related windows are destroyed.
          The function shall be called only when finishing XPLAB.
*/
void
cb_statistic_quit_all(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(menuitem);

  if (GTK_IS_WIDGET(st.top_statistic) == TRUE) {
    release_images_all();      /* broadcasting switched off only made here */

    cb_delete(st.bc_image_color.topwin, NULL, NULL); /* delete top-windows */
    cb_delete(st.bc_image_grey.topwin, NULL, NULL);
    cb_delete(st.bc_image_bmp.topwin, NULL, NULL);
    cb_delete(st.bc_histo_color.topwin, NULL, NULL);
    cb_delete(st.bc_histo_slope.topwin, NULL, NULL);
    cb_delete(st.bc_histo_grey.topwin, NULL, NULL);
    cb_delete(st.bc_histo_bmp.topwin, NULL, NULL);
    cb_delete(st.bc_histo_rbs.topwin, NULL, NULL);

    xplab_widget_destroy(&st.table);        /* performs better for security */
    xplab_widget_destroy(&st.top_statistic);

    free_statistic_def(&st);
  }
}



/* Hides statistic dialog.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_statistic_hide(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  STSFY_GCC(menuitem);

  if (GTK_IS_WIDGET(st.top_statistic) == TRUE) {
    xplab_window_raise(user_data, xctrl.cfg.raise);
    gtk_widget_hide(st.top_statistic);
  }
}



/* Creates the icons table.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
statistic_table(GtkWidget *vbox)
{
  int      rows = 3, cols = 4, icon_size_x = XPLAB_STATISTIC_ICON_WIDTH;
  int      icon_size_y = XPLAB_STATISTIC_ICON_HEIGHT;
  int      icon_border = XPLAB_STATISTIC_ICON_BORDER;
  gboolean has_aspect;

  if (st.table) {         /* delete old table and associated data (previews) */
    xplab_widget_destroy(&st.table);
  }

  st.table = gtk_table_new(rows, cols, FALSE);   /* table for the thumbnails */
  gtk_table_set_col_spacings(GTK_TABLE(st.table), XPLAB_STATISTIC_TABLE_BORDER);
  gtk_table_set_row_spacings(GTK_TABLE(st.table), XPLAB_STATISTIC_TABLE_BORDER);
  gtk_box_pack_start(GTK_BOX(vbox), st.table, TRUE, TRUE, 0);

  /* icons respecting aspect                                                 */
  has_aspect = TRUE;
  attach_icon(st.table, 0, 1, 0, 1, icon_size_x, icon_size_y, icon_border, 
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button, 
	      &st.bc_image_color, has_aspect, 
	      st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_COLOR, STATS_IMAGE_COLOR, &st.iorig);

  attach_icon(st.table, 1, 2, 0, 1, icon_size_x, icon_size_y, icon_border,
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button, 
	      &st.bc_image_grey, has_aspect, st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_GREY, STATS_IMAGE_GREY, &st.iorig);

  attach_icon(st.table, 2, 3, 0, 1, icon_size_x, icon_size_y, icon_border,
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button, 
	      &st.bc_image_bmp, has_aspect, st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_BITMAP, STATS_IMAGE_BMP, &st.iorig);

  /* icons not respecting aspect                                             */
  has_aspect = FALSE;
  attach_icon(st.table, 0, 1, 1, 2, icon_size_x, icon_size_y, icon_border, 
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button, 
	      &st.bc_histo_color, has_aspect, st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_HISTO_COLOR, STATS_HISTO_COLOR, &st.iorig);

  attach_icon(st.table, 1, 2, 1, 2, icon_size_x, icon_size_y, icon_border, 
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button, 
	      &st.bc_histo_grey, has_aspect, st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_HISTO_GREY, STATS_HISTO_GREY, &st.iorig);

  attach_icon(st.table, 2, 3, 1, 2, icon_size_x, icon_size_y, icon_border, 
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button, 
	      &st.bc_histo_bmp, has_aspect, st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_HISTO_BMP, STATS_HISTO_BMP, &st.iorig);

  attach_icon(st.table, 3, 4, 1, 2, icon_size_x, icon_size_y, icon_border, 
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button,
	      &st.bc_histo_rbs, has_aspect, st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_HISTO_RBS, STATS_HISTO_RBS, &st.iorig);

  attach_icon(st.table, 1, 2, 2, 3, icon_size_x, icon_size_y, icon_border, 
	      GTK_SIGNAL_FUNC(cb_icon_button), GTK_SIGNAL_FUNC(cb_icon_enter),
	      GTK_SIGNAL_FUNC(cb_icon_leave), &st.refresh_button, 
	      &st.bc_histo_slope, has_aspect, st.icon_contrast_max.display_active,
	      XPLAB_STATISTIC_TITLE_HISTO_SLOPE, STATS_HISTO_SLOPE, &st.iorig);

  gtk_widget_show_all(st.table);
}



/* Provides mouse actions on window.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - propagation flag
Usage.  - CB meant for image windows.
*/
gboolean
cb_image(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN)
{
  area_def  far;
  gpointer  image_data_path = NULL;
  GtkWidget *top_level;

  STSFY_GCC(user_data);

  init_area_def(&far);

  if (GTK_IS_WIDGET(widget) == FALSE) {
    return TRUE;
  }

  if (event->type != GDK_BUTTON_PRESS) {
    return TRUE;
  }

  switch (event->button) {
  case 1: 
    top_level = gtk_widget_get_toplevel(widget);

    if (GTK_IS_OBJECT(top_level) == FALSE) {
      return FALSE;
    }

    image_data_path = g_object_get_data(G_OBJECT(top_level), XPLAB_ASSN_KEY_0);

    if (image_data_path == NULL) {
      return TRUE;
    }

    image_actions(event, (bc_img *)image_data_path);

    break;

  case 3: 
    if (GTK_IS_WIDGET(st.top_statistic) == TRUE) {
      if (GTK_WIDGET_MAPPED(st.top_statistic)) {            /* toggle ON/OFF */
	gtk_widget_hide(st.top_statistic);
      }
      else {
	gtk_widget_show_all(st.top_statistic);
      }
    }
    else {
      statistic_map();                                             /* create */
    }

    break;

  default: break;
  }

  return TRUE;
}



/* Creates a pop-up menu for an image window.

Revision: 18.11.2004 E.Sack

Input : - pointer to mouse button event
        - pointer to imaga data path
Output: - -
Usage:  - in order to keep an image window as small as possible, only pop-up
          menus instead of selection bars are used.
*/
static void
image_actions(GdkEventButton *event, bc_img *image_data)
{
  assert(event);
  assert(image_data);

  gtk_menu_popup(GTK_MENU(image_data->popmenu), NULL, NULL, NULL, NULL,
		 event->button, event->time);
}



/* Creates an image window.

Revision: 05.08.2005 E.Sack

Input : - pointer to image data
        - pointer to window name
        - flag (0: if window exists, don't delete it, 
                1: if window exists, delete it)
        - window size        
Output: - propagation flag
Postcondition: - the image data is bound to the created window.
*/
void
stats_window(bc_img *image_data_path, const gchar *wname, gboolean del,
	     area_def far)
{
  gboolean raised;

  assert(image_data_path);
  assert(wname);

  raised = xplab_window_raise(&image_data_path->topwin, 1);

  if (raised == FALSE) {
    if (!image_data_path->timg.pimg.col) {
      stats_request_formats(image_data_path, &st.refresh_button, 
			    XPLAB_STATISTIC_TITLE_ITEM);

      if (!image_data_path->timg.pimg.col) {
	return;                    /* still no image - don't create a window */
      }
    }

    assert(image_data_path->topwin == NULL);                /* must be freed */

    image_data_path->topwin = 
      xplab_image_window(wname, NULL, NULL, 
			 &image_data_path->scroll, image_data_path->timg.pimg.ar,
			 GTK_SIGNAL_FUNC(cb_image),
			 NULL,
			 GTK_SIGNAL_FUNC(cb_expose),
			 GTK_SIGNAL_FUNC(cb_delete), 0, far);

    g_object_set_data(G_OBJECT(image_data_path->topwin), XPLAB_ASSN_KEY_0, 
		      (gpointer)image_data_path);

    gtk_widget_show_all(image_data_path->topwin);

    image_data_path->poproot = popup_root(image_data_path);

    image_data_path->magnitude = XPLAB_MAGN_X1_VALUE;/* consequence of new window */
  }
  else if (del == TRUE) {
    cb_delete(image_data_path->topwin, NULL, NULL);
  }
}



/* Resets status bar.

Revision: 20.12.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_icon_enter(GTK_CONVENTION_ENTER_NOTIFY_RTS_GBOOLEAN)
{
  int          err;
  char         *txt = NULL;
  STATS_FORMAT fmt = *(STATS_FORMAT *)user_data;

  STSFY_GCC(widget);
  STSFY_GCC(event);

  assert(user_data);

  if (st.message_id) {
    gtk_statusbar_remove(GTK_STATUSBAR(st.status_bar), st.context_id,
			 st.message_id);
  }

  switch (fmt) {
  case STATS_IMAGE_COLOR: 
    add_string(STATS_ICOL_REM, &txt);
    break;
  case STATS_IMAGE_GREY:
    add_string(STATS_IGRY_REM, &txt);
    break;
  case STATS_IMAGE_BMP: 
    add_string(STATS_IBMP_REM, &txt);
    break;
  case STATS_HISTO_COLOR: 
    add_string(STATS_HCOL_REM, &txt);
    break;
  case STATS_HISTO_GREY: 
    add_string(STATS_HGRY_REM, &txt);
    break;
  case STATS_HISTO_BMP: 
    add_string(STATS_HBMP_REM, &txt);
    break;
  case STATS_HISTO_RBS:
    add_string(STATS_HRBS_REM, &txt);
    break;
  case STATS_HISTO_SLOPE: 
    add_string(STATS_HSLOPE_REM, &txt);
    break;
  case STATS_IMAGE_ORIGINAL:
  default:
    assert(0);
    break;
  }

  err = check_name(txt);

  if (!err) {
    st.message_id = gtk_statusbar_push(GTK_STATUSBAR(st.status_bar), 
				       st.context_id, txt);
  }

  FREEMEM(txt);

  return FALSE;
}



/* Resets status bar.

Revision: 20.12.2004 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_icon_leave(GTK_CONVENTION_LEAVE_NOTIFY_RTS_GBOOLEAN)
{
  STSFY_GCC(widget);
  STSFY_GCC(event);
  STSFY_GCC(user_data);

  if (st.message_id) {
    gtk_statusbar_remove(GTK_STATUSBAR(st.status_bar), st.context_id,
			 st.message_id);
    st.message_id = 0;
  }

  return FALSE;
}



/* Turns image windows ON or OFF.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - propagation flag
Usage.  - CB meant for icons.
*/
static gboolean
cb_icon_button(GTK_CONVENTION_BUTTON_PRESS_RTS_GBOOLEAN)
{
  area_def far;
  bc_img   *image_data;
  gpointer image_data_path;
  const gchar *wname;

  STSFY_GCC(user_data);

  init_area_def(&far);

  if (GTK_IS_WIDGET(widget) == FALSE || GTK_IS_OBJECT(widget) == FALSE) {
    return TRUE;
  }

  if (event->type != GDK_BUTTON_PRESS &&
      event->type != GDK_2BUTTON_PRESS &&
      event->type != GDK_3BUTTON_PRESS) {
    return TRUE;
  }

  switch (event->button) {
  case 1: 
    wname = gtk_widget_get_name(widget);

    image_data_path = g_object_get_data(G_OBJECT(widget), XPLAB_ASSN_KEY_0);

    if (image_data_path == NULL) {
      return TRUE;
    }

    image_data = (bc_img *)image_data_path;

    stats_window(image_data, wname, TRUE, far);

    break;

  default: break;
  }

  return TRUE;
}



/* Dumps histograms to hard disc (gnuplot format).

Revision: 22.12.2004 E.Sack

Input : - GTK convention
Output: - -
*/
void
cb_dump_histo(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  char    *name = NULL;
  char    histo_comment[] = "# XPLAB Statistic\n";          /* for gnuplot */
  bc_img  *image_data_path;
  hst_def histo;

  STSFY_GCC(menuitem);

  init_hst_def(&histo);

  image_data_path = (bc_img *)user_data;

  if (image_data_path == NULL) {
    return;                                                     /* no data */
  }

  switch (image_data_path->dfmt) {               /* only histogram formats */
  case STATS_HISTO_SLOPE: 
    add_string(STATS_HSLOPE_KEY, &name);
    histo = xctrl.set.th_hist;
    break;
  case STATS_HISTO_GREY: 
    add_string(STATS_HGRY_KEY, &name);
    histo = xctrl.set.grey_hist;
    break;
  case STATS_HISTO_COLOR: 
    add_string(STATS_HCOL_KEY, &name);
    histo = st.red_hist;             /* special case of several histograms */
    add_string(".histo_red", &name);
    write_histo(histo, 0, name, histo_comment);

    FREEMEM(name);
    add_string(STATS_HCOL_KEY, &name);
    histo = st.green_hist;
    add_string(".histo_green", &name);
    write_histo(histo, 0, name, histo_comment);

    FREEMEM(name);
    add_string(STATS_HCOL_KEY, &name);
    histo = st.blue_hist;
    add_string(".histo_blue", &name);
    write_histo(histo, 0, name, histo_comment);

    FREEMEM(name);
    return;
    break;
  case STATS_HISTO_BMP: 
    add_string(STATS_HBMP_KEY, &name);
    histo = st.bmp_hist;
    break;
  case STATS_HISTO_RBS:
    add_string(STATS_HRBS_KEY, &name);
    histo = st.rbs_hist;
    break;
  default:
    assert(0);
    break;
  }

  add_string(".histo", &name);

  write_histo(histo, 0, name, histo_comment);

  FREEMEM(name);
}



/* Requests format for histogram of bit-mapped image.

Revision: 15.01.2005 E.Sack

Input : - pointer to broadcasted format
        - pointer to broadcast widget
        - pointer to request comment
Output: - format request updated
Postcondition: - internal buffers are updated for further usage.
*/
void
format_stats_histo_bmp(bc_img *bci, GtkWidget **bc_widget, 
		       char *request_comment, gboolean keep_request)
{
  free_hst_def(&st.bmp_hist);

  bci->timg = xplab_request_image(P1, request_comment, &bci->key, 0, bc_widget);

  if (bci->timg.pimg.mem == NULL) {
    stats_release_image(bci, keep_request);
    free_img_typ(&bci->timg);
    return;                                       /* image is not a BMP one */
  }

  /* actions on bit-mapped image required: histogram */
  st.bmp_hist = histogram_8bit(bci->timg.pimg.mem, bci->timg.pimg.ar);         
  free_img_typ(&bci->timg);                /* temporary buffer no more used */

  bci->timg.pimg = histo_to_image(st.bmp_hist, PICT_MAX+1,
		 	 	  XPLAB_STATISTIC_HISTO_HEIGHT);

  bci->timg.typ = P6;                                         /* set format */
  bci->img_changed = TRUE;
}



/* Requests format for histogram of a RB list (RB size).

Revision: 22.12.2004 E.Sack

Input : - pointer to format
Output: - format updated
Postcondition: - internal buffers are updated for further usage.
*/
void
format_stats_histo_rbs(bc_img *bci)
{
  int i, *rb_size = NULL;

  free_hst_def(&st.rbs_hist);

  if (xctrl.set.rblst.cnt == 0) {                        /* global RB list */
    return;                                        /* global RB list empty */
  }

  /* actions on RB list required: histogram */
  CALLOC(int, xctrl.set.rblst.cnt, rb_size);

  for (i = 0; i < xctrl.set.rblst.cnt; i++) {
    rb_size[i] = SECTOR_SIZE(xctrl.set.rblst.rb[i].sc);
  }

  st.rbs_hist = histogram_auto(rb_size, xctrl.set.rblst.cnt);
  FREEMEM(rb_size);                       /* temporary buffer no more used */

  bci->timg.pimg = histo_to_image(st.rbs_hist, PICT_MAX+1,
				 XPLAB_STATISTIC_HISTO_HEIGHT);

  bci->timg.typ = P6;                                        /* set format */
  bci->img_changed = TRUE;
}



/* Requests format for histogram of colors from an image.

Revision: 15.01.2005 E.Sack

Input : - pointer to broadcasted format
        - pointer to broadcast widget
        - pointer to request comment
Output: - format request updated
Postcondition: - internal buffers are updated for further usage.
*/
void
format_stats_histo_color(bc_img *bci, GtkWidget **bc_widget, 
			 char *request_comment, gboolean keep_request)
{
  col_def bgnd = {PICT_MAX, PICT_MAX, PICT_MAX};       /* white background */
  col_def fgnd = {0, 0, 0};                            /* black foreground */

  free_hst_def(&st.red_hist);
  free_hst_def(&st.green_hist);
  free_hst_def(&st.blue_hist);

  bci->timg = xplab_request_image(PNM_FORMAT_ORIGINAL, request_comment, 
				  &bci->key, 0, bc_widget);

  if (bci->timg.pimg.col == NULL) {
    stats_release_image(bci, keep_request);
    free_img_typ(&bci->timg);
    return;                         /* original image is not a colored one */
  }

  /* actions on colored image required: histogram */
  assert(bci->timg.typ == P6 || bci->timg.typ == P3);      /* check format */
  histogram_rgb(bci->timg.pimg.col, bci->timg.pimg.ar,
		&st.red_hist, &st.green_hist, &st.blue_hist);         

  free_img_typ(&bci->timg);               /* temporary buffer no more used */

  bci->timg.pimg = histos_to_image(st.red_hist, st.green_hist, st.blue_hist,
				   bgnd, fgnd, XPLAB_STATISTIC_HISTO_HEIGHT);

  bci->timg.typ = P6;                                        /* set format */
  bci->img_changed = TRUE;
}
