/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <stdarg.h>
#include <time.h>
#include <float.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_toolkit.h"
#include "xplab_image.h"
#include "xplab_preferences_fe.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_report.h"

/* local definitions *******************************************************/
#define XPLAB_TOOLKIT_TITLE_ITEM "Toolkit"
#define TOOLPIC_DEFAULT "toolpic."        /* output toolkit image filename */
#define RBLIST_STRING_HDD   "dump RB list images to disc"
#define RBLIST_STRING_LOG   "dump RB list to log"
#define DT_STRING           "dump distance transform image to disc"

GtkWidget *dump_rblist_log, *dump_rblist_hdd, *dump_dt;

/* local prototypes ********************************************************/
static void menu_item_dump_rblist_hdd(GtkWidget *toolkit_menu);
static void menu_item_dump_rblist_log(GtkWidget *toolkit_menu);
static void menu_item_distance_transform(GtkWidget *toolkit_menu);
static void image_rb_lst(rb_lst rbl, pic_def rpic, char *filename);
static pic_def distance_transform(pic_def orig);
static void cb_toolkit_menu(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_dump_rblist_hdd(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_dump_rblist_log(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_distance_transform(GTK_CONVENTION_MENUITEM_RTS_VOID);

/* Functions **************************************************************/

/* Initializes the toolkit.

Revision: 04.04.2004 E.Sack

Input : - -
Output: - default setting
*/
void
xplab_init_toolkit(void)
{
}



/* Frees the toolkit.

Revision: 04.04.2004 E.Sack

Input : - -
Output: - toolkit freed
*/
void
xplab_free_toolkit(void)
{
}



/* Builds an entry for a menu bar.

Revision: 05.08.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
void
xplab_toolkit_menu(GtkWidget *menu_bar)
{
  GtkWidget *toolkit_entry, *toolkit_menu;

  toolkit_entry = gtk_menu_item_new_with_label(XPLAB_TOOLKIT_TITLE_ITEM);
  gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), toolkit_entry);

  toolkit_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(toolkit_entry), toolkit_menu);

  menu_item_dump_rblist_log(toolkit_menu);                    /* item dump */
  menu_item_dump_rblist_hdd(toolkit_menu);                    /* item dump */
  menu_item_distance_transform(toolkit_menu);                 /* item DT   */

  g_signal_connect(G_OBJECT(toolkit_entry), "activate",
		   G_CALLBACK(cb_toolkit_menu), NULL);
}



/* Updates toolkit menu strings and sensitiveness.

Revision: 09.01.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_toolkit_menu(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  char *buf = NULL;

  STSFY_GCC(user_data);

  assert(menuitem);

  int_to_string(RBLIST_STRING_LOG " (", xctrl.set.rblst.cnt, ")", &buf);
  xplab_label_set_text(GTK_BIN(dump_rblist_log)->child, buf);
  gtk_widget_set_sensitive(GTK_WIDGET(dump_rblist_log), 
			   xctrl.set.rblst.cnt ? TRUE : FALSE);

  FREEMEM(buf);

  int_to_string(RBLIST_STRING_HDD " (", xctrl.set.rblst.cnt, ")", &buf);
  xplab_label_set_text(GTK_BIN(dump_rblist_hdd)->child, buf);
  gtk_widget_set_sensitive(GTK_WIDGET(dump_rblist_hdd), 
			   xctrl.set.rblst.cnt ? TRUE : FALSE);

  FREEMEM(buf);

  gtk_widget_set_sensitive(GTK_WIDGET(dump_dt), 
			   (xplab_image_loaded() == PNM_FORMAT_UNDEF) ? 
			   FALSE : TRUE);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
menu_item_dump_rblist_hdd(GtkWidget *toolkit_menu)
{
  dump_rblist_hdd = gtk_menu_item_new_with_label(RBLIST_STRING_HDD);
  gtk_menu_shell_append(GTK_MENU_SHELL(toolkit_menu), dump_rblist_hdd);

  g_signal_connect(G_OBJECT(dump_rblist_hdd), "activate",
		   G_CALLBACK(cb_dump_rblist_hdd), NULL);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
menu_item_dump_rblist_log(GtkWidget *toolkit_menu)
{
  dump_rblist_log = gtk_menu_item_new_with_label(RBLIST_STRING_LOG);
  gtk_menu_shell_append(GTK_MENU_SHELL(toolkit_menu), dump_rblist_log);

  g_signal_connect(G_OBJECT(dump_rblist_log), "activate",
		   G_CALLBACK(cb_dump_rblist_log), NULL);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - pointer to menu
Output: - menu updated
*/
static void
menu_item_distance_transform(GtkWidget *toolkit_menu)
{
  dump_dt = gtk_menu_item_new_with_label(DT_STRING);
  gtk_menu_shell_append(GTK_MENU_SHELL(toolkit_menu), dump_dt);

  g_signal_connect(G_OBJECT(dump_dt), "activate",
		   G_CALLBACK(cb_distance_transform), NULL);
}



/* Dumps RB list as images.

Revision: 04.04.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dump_rblist_hdd(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  pic_def tmp_img;

  STSFY_GCC(user_data);

  assert(menuitem);

  tmp_img = xplab_request_image(P1, NULL, NULL, 0, NULL).pimg; /* image copy */
  image_rb_lst(xctrl.set.rblst, tmp_img, TOOLPIC_DEFAULT);

  free_pic_def(&tmp_img);
}



/* Dumps RB list to log.

Revision: 09.01.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dump_rblist_log(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  char *buf = NULL;

  STSFY_GCC(user_data);

  assert(menuitem);

  buf = print_report_rbs(xctrl.set.rblst);

  string_to_log(main_log, "", "", buf);

  FREEMEM(buf);
}



/* Calculates the distance transform of the current image.

Revision: 23.10.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_distance_transform(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  pic_def tmp_img, dt_img;

  STSFY_GCC(user_data);

  assert(menuitem);

  tmp_img = xplab_request_image(P1, NULL, NULL, 0, NULL).pimg; /* image copy */

  dt_img = distance_transform(tmp_img);
  free_pic_def(&tmp_img);

  if (dt_img.mem) {
    dump_pic_def(dt_img, TOOLPIC_DEFAULT "dt", dt_img.ar, 0, 0, 1,
		 PNM_FORMAT_UNDEF);
  }

  free_pic_def(&dt_img);
}



/* Calculates the distance transform image of a binary input image.
   The gray-valued DT image is normalized to 1-255. The input image 
   pixels are set to 0.

   The (grassfire) algorithm consists mainly of two passes over the 
   whole image. The first pass calculates some pixels not to their
   true value. These are corrected in the second pass.

Revision: 29.10.2004 E.Sack

Input : - image
Output: - distance transform image
Usage:  - structuring element SE is 3x3.
*/
static pic_def
distance_transform(pic_def orig)
{
  int     x, y;
  double  min, pix;
  pic_def dt_img, dt_out;
  pic_dbl dt_dbl;

#define oO(X,Y)  dt_img.mem[(X)   +  (Y)   *dt_img.ar.width]
#define oN(X,Y)  dt_img.mem[(X)   + ((Y)-1)*dt_img.ar.width]
#define oE(X,Y)  dt_img.mem[(X)+1 +  (Y)   *dt_img.ar.width]
#define oS(X,Y)  dt_img.mem[(X)   + ((Y)+1)*dt_img.ar.width]
#define oW(X,Y)  dt_img.mem[(X)-1 +  (Y)   *dt_img.ar.width]
#define oNE(X,Y) dt_img.mem[(X)+1 + ((Y)-1)*dt_img.ar.width]
#define oSE(X,Y) dt_img.mem[(X)+1 + ((Y)+1)*dt_img.ar.width]
#define oSW(X,Y) dt_img.mem[(X)-1 + ((Y)+1)*dt_img.ar.width]
#define oNW(X,Y) dt_img.mem[(X)-1 + ((Y)-1)*dt_img.ar.width]

#define dO(X,Y)  dt_dbl.mem[(X)   +  (Y)   *dt_dbl.ar.width]
#define dN(X,Y)  dt_dbl.mem[(X)   + ((Y)-1)*dt_dbl.ar.width]
#define dE(X,Y)  dt_dbl.mem[(X)+1 +  (Y)   *dt_dbl.ar.width]
#define dS(X,Y)  dt_dbl.mem[(X)   + ((Y)+1)*dt_dbl.ar.width]
#define dW(X,Y)  dt_dbl.mem[(X)-1 +  (Y)   *dt_dbl.ar.width]
#define dNE(X,Y) dt_dbl.mem[(X)+1 + ((Y)-1)*dt_dbl.ar.width]
#define dSE(X,Y) dt_dbl.mem[(X)+1 + ((Y)+1)*dt_dbl.ar.width]
#define dSW(X,Y) dt_dbl.mem[(X)-1 + ((Y)+1)*dt_dbl.ar.width]
#define dNW(X,Y) dt_dbl.mem[(X)-1 + ((Y)-1)*dt_dbl.ar.width]

#define dSEDOWNMAX(X,Y) XPMAX(XPMAX(dW(X,Y),dN(X,Y)),XPMAX(dNW(X,Y),dNE(X,Y)))
#define dSEUPMAX(X,Y)   XPMAX(XPMAX(dE(X,Y),dS(X,Y)),XPMAX(dSW(X,Y),dSE(X,Y)))

#define oSEDOWNFGND(X,Y) ((oW(X,Y)==FOREGROUND) || (oN(X,Y)==FOREGROUND) || (oNW(X,Y)==FOREGROUND) || (oNE(X,Y)==FOREGROUND))
#define oSEUPFGND(X,Y)   ((oE(X,Y)==FOREGROUND) || (oS(X,Y)==FOREGROUND) || (oSW(X,Y)==FOREGROUND) || (oSE(X,Y)==FOREGROUND))

  init_pic_def(&dt_img);
  init_pic_def(&dt_out);
  init_pic_dbl(&dt_dbl);

  if (orig.mem == NULL) {
    return dt_out;
  }

  dt_img.ar = orig.ar;
  dt_img.ar.width += 2;/* enlarge by one pixel to every side because of SE */
  dt_img.ar.height += 2;
  CALLOC(PICT, AREA_SIZE(dt_img.ar), dt_img.mem);   /* 0-initialised image */
  copy_bmps(orig, dt_img, orig.ar, 0, 0, 1, 1);/* copy original image centered*/

  dt_dbl.ar = orig.ar;
  dt_dbl.ar.width += 2;/* enlarge by one pixel to every side because of SE */
  dt_dbl.ar.height += 2;
  CALLOC(double, AREA_SIZE(dt_dbl.ar), dt_dbl.mem); /* 0-initialised image */

  /* keep in mind that the contour pixels are only for supporting the SE   */
  for (y = 1; y <= orig.ar.height; y++) {        /* first pass top-down... */
    for (x = 1; x <= orig.ar.width; x++) {       /* ...left-right          */
      if (oO(x, y) == FOREGROUND) {                     /* pixel on region */
	continue;
      }

      if (oSEDOWNFGND(x, y)) {                     /* neighbours on region */
	dO(x, y) = 1.0;                                /* initial distance */
      }
      else {                            
	min = dSEDOWNMAX(x, y);                      /* initialise minimum */

	if (min == 0.0) {
	  continue;                        /* neighbours all uninitialised */
	}

	if ((pix = dW(x,y))) {           /* only initialised SE neighbours */
	  min = XPMIN(min, pix);
	}

	if ((pix = dN(x,y))) {
	  min = XPMIN(min, pix);
	}

	if ((pix = dNW(x,y))) {
	  min = XPMIN(min, pix);
	}

	if ((pix = dNE(x,y))) {
	  min = XPMIN(min, pix);
	}

	dO(x, y) = min + 1.0;            /* incrementing distance 1st pass */
      }
    }
  }

  for (y = orig.ar.height; y >= 1; y--) {      /* second pass bottom-up... */
    for (x = orig.ar.width; x >= 1; x--) {     /* ...right-left            */
      if (oO(x, y) == FOREGROUND) {                     /* pixel on region */
	continue;
      }

      if (oSEUPFGND(x, y)) {                       /* neighbours on region */
	dO(x, y) = 1.0;                                /* initial distance */
      }
      else {                            
	min = dSEUPMAX(x, y);                        /* initialise minimum */

	if (min == 0.0) {
	  continue;                        /* neighbours all uninitialised */
	}

	if ((pix = dE(x,y))) {           /* only initialised SE neighbours */
	  min = XPMIN(min, pix);
	}

	if ((pix = dS(x,y))) {
	  min = XPMIN(min, pix);
	}

	if ((pix = dSW(x,y))) {
	  min = XPMIN(min, pix);
	}

	if ((pix = dSE(x,y))) {
	  min = XPMIN(min, pix);
	}

	min += 1.0;                      /* incrementing distance 2nd pass */

	if ((pix = dO(x,y))) {       /* if pixel is written in 1st pass... */
	  min = XPMIN(min, pix);              /* ...correct it in 2nd pass */
	}

	dO(x, y) = min;                     /* total incrementing distance */
      }
    }
  }

  free_pic_def(&dt_img);             /* free temporary original image copy */

  dt_img = convert_pic_dbl(dt_dbl);  /* convert DT image for visualization */
  free_pic_dbl(&dt_dbl);

  dt_out.ar = orig.ar;                                    /* original size */

  CALLOC(PICT, AREA_SIZE(dt_out.ar), dt_out.mem);     /* initialised image */

  copy_bmps(dt_img, dt_out, dt_out.ar, 1, 1, 0, 0);       /* copy DT image */

  free_pic_def(&dt_img);

  return dt_out;                   /* normalized DT image in original size */

#undef oO
#undef oN
#undef oNE
#undef oE
#undef oSE
#undef oS
#undef oSW
#undef oW
#undef oNW
#undef oSEDOWNFGND
#undef oSEUPFGND
#undef dO
#undef dN
#undef dNE
#undef dE
#undef dSE
#undef dS
#undef dSW
#undef dW
#undef dNW
#undef dSEDOWNMIN
#undef dSEUPMIN
}



/* Dumps the RB list in single images.

Revision: 24.09.2003 E.Sack

Input : - RB list
        - image
        - file name base
Output: - images dumped
Precondition: - list and image must be consistent.
*/
static void
image_rb_lst(rb_lst rbl, pic_def rpic, char *filename)
{
  int      rbn;
  char     *fname = NULL;
  area_def iar;

  for (rbn = 0; rbn < rbl.cnt; rbn++) {
    int_to_string(filename, rbn, "", &fname);

    SECTOR_TO_AREA(rbl.rb[rbn].sc, iar);

    dump_pic_def(rpic, fname, iar, rbl.rb[rbn].sc.column.start,
		 rbl.rb[rbn].sc.line.start, 
		 xctrl.cfg.verbose > VERBOSE_DEFAULT, PNM_FORMAT_UNDEF);

    FREEMEM(fname);
  }

  if (!rbl.cnt) {
    string_to_log(main_log, "TOOLKIT: empty RB list, nothing dumped.\n","","");
  }
  else {
    int_to_log(main_log, "TOOLKIT: ", rbl.cnt,
	       rbl.cnt==1?" image dumped\n":" images dumped\n");
  }
}
