/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <assert.h>
#include <math.h>
#include <unistd.h>
#include <time.h>
#include <libgen.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_training_be.h"
#include "xplab_preferences_fe.h"
#include "xplab_preferences_be.h"
#include "xplab_dchannel.h"
#include "xplab_algorithm.h"
#include "xplab_image.h"
#include "xplab_report.h"
#include "xplab_dbio.h"
#include "xplab_maintenance_fe.h"
#include "xplab_maintenance_be.h"
#include "xplab_recall_be.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/
#define XPLAB_PREFERENCES_TITLE_ITEM "preferences"
#define XPLAB_PREFS_PAGE_GENERAL     "General"
#define XPLAB_PREFS_PAGE_PREPROCESSING  "Preprocessing"
#define XPLAB_PREFS_PAGE_MATCHING "Matching"
#define XPLAB_PREFS_PAGE_CORRELATION "Correlation"
#define XPLAB_PREFS_PAGE_CORRELATION_GENERAL "general"
#define XPLAB_PREFS_PAGE_CORRELATION_CLUSTER "clustering"
#define XPLAB_PREFS_PAGE_STATISTIC "Statistic"
#define XPLAB_PREFS_PAGE_DC "DocChannel"
#define XPLAB_PREFS_PAGE_DC_GENERAL "general"
#define XPLAB_PREFS_PAGE_DC_OPTION_1D "1-D"
#define XPLAB_PREFS_PAGE_DC_OPTION_2D "2-D"
#define XPLAB_PREFS_PAGE_DC_1D "1-D flags"
#define XPLAB_PREFS_PAGE_DC_2DI "2-D flags I"
#define XPLAB_PREFS_PAGE_DC_2DII "2-D flags II"
#define XPLAB_PREFS_PAGE_DC_1_LIST_EMPTY "D1 list empty"
#define XPLAB_PREFS_PAGE_DC_2_LIST_EMPTY "D2 list empty"
#define XPLAB_PREFS_PAGE_DC_LIST_MAX 98
#define XPLAB_PREFS_ENTRY_CHARS_DC  15
#define XPLAB_PREFS_ENTRY_LEN_DC  115
#define XPLAB_PREFS_PAGE_IO " I/O "
#define XPLAB_PREFS_PAGE_IO_GENERAL "general"
#define XPLAB_PREFS_PAGE_IO_REPORT "report"
#define XPLAB_PREFS_PAGE_IO_IMAGE "image"
#define XPLAB_PREFS_PAGE_IO_HISTO "histogram"
#define XPLAB_PREFS_PAGE_IO_PHASE "phase"
#define XPLAB_PREFS_PAGE_MATCHING_GENERAL "general"
#define XPLAB_PREFS_PAGE_MATCHING_RULES "rules"
#define XPLAB_PREFS_SUBSTITUTION_CHARS 10
#define XPLAB_PREFS_VERBOSE_FORMAT_PREFIX  " "
#define XPLAB_PREFS_VERBOSE_FORMAT_SUFFIX  " "
#define XPLAB_PREFS_FLOAT_FORMAT    "%.3f"
#define XPLAB_PREFS_ENTRY_CHARS_PATH 1024

typedef enum {
  FILE_UNDEFINED,
  FILE_NOT_EXIST,
  FILE_NOT_READABLE,
  FILE_NOT_CLOSEABLE,
  FILE_READ,
  FILE_QUIET
} XFILE_STAT;

typedef struct {
  GtkWidget *parent, *verbose_combo, *raise_button;
  GtkWidget *rb_chain_tiles_button;
  GtkWidget *rb_chain_use_button, *rb_chain_color_button;
  GtkWidget *batchjob_void_button, *batchjob_matching_button;
  GtkWidget *batchjob_cluster_button, *batchjob_rbchain_button;
  GtkWidget *batchlog_no_button, *batchlog_ovwrt_button;
  GtkWidget *batchlog_appnd_button, *tips_button;
} general_def;

typedef struct {
  GtkWidget *parent, *renew_method_button, *single_line_button;
  GtkWidget *line_method_combo, *duration_button, *table_general;
  GtkWidget *single_nb_button, *broken_nb_button;
  GtkEntry  *maskweight_entry, *mborder_entry, *nborder_entry;
  GtkEntry  *nnbmatch_entry, *interleave_entry, *overlap_entry;
  GtkEntry  *nb_lap_entry, *symbolspacing_entry;
  GtkEntry  *subst_rb_entry, *subst_nb_entry;
} matching_def;

typedef struct {
  GtkWidget *parent, *domain_button;
  GtkWidget *cluster_force_button, *cluster_write_button;
  GtkWidget *cluster_nb_new_button, *cluster_nb_old_button;
  GtkEntry  *cluster_level_entry, *corrlevel_entry, *corrbroken_entry;
  GtkEntry  *cluster_step_entry, *cluster_thold_entry;
  GtkEntry  *ckern_entry[3][3];
} correlation_def;

typedef struct {
  GtkWidget *pos_ccf_button, *pos_pad_button, *tfin_button, *tflag_button;
  GtkWidget *pos_button, *pos_hphase_button;
} d1_def;

typedef struct {
  GtkWidget *bmp_button, *gv_button, *dot_button, *clean_button, *org_button;
  GtkWidget *pest_button, *pest_max_button;
} d2i_def;

typedef struct {
  GtkWidget *nb_good_button, *rb_all_button, *rb_all_inv_button;
  GtkWidget *rb_good_button, *rb_good_inv_button, *rb_fail_button;
  GtkWidget *rb_fail_inv_button, *rb_frame_bmp_button, *rb_frame_rgb_button;
  GtkWidget *mrb_frame_bmp_button, *mrb_frame_rgb_button;
  GtkWidget *hline_bmp_button, *hline_rgb_button, *sline_bmp_button;
  GtkWidget *sline_rgb_button;
} d2ii_def;

typedef struct {
  GtkWidget *parent;
  GtkEntry  *store_entry, *d1_entry, *d2i_entry, *d2ii_entry;
  int2_lst  dc1l, dc2l;
  int       d1_idx, d2_idx;
  d1_def    d1;
  d2i_def   d2i;
  d2ii_def  d2ii;
} dc_def;

typedef struct {
  GtkWidget *parent, *rawbit_button, *overwrite_image_button;
  GtkWidget *report_terse_button, *report_rbs_button, *histo_write_button;
  GtkWidget *replay_write_button, *timestamp_button, *report_suppress;
  GtkWidget *training_write_button, *writeauto_button;
  GtkEntry  *nb_border_x_entry, *nb_border_y_entry;
} io_def;

typedef struct {
  GtkWidget *parent, *support_dots_button, *support_dots_auto;
  GtkEntry *rgb_norm_b_entry, *rgb_norm_total_entry;
  GtkEntry  *histo_thold_entry, *rgb_norm_r_entry, *rgb_norm_g_entry;
} pre_def;

typedef struct {
  GtkWidget   *sparent;
  stats_prefs owin, cwin, gwin, bwin, rhst, chst, ghst, bhst, shst, contrast;
} stats_def;

typedef struct {
  general_def     general;
  pre_def         preprocessing;
  matching_def    matching;
  correlation_def correlation;
  stats_def       statistic;
  dc_def          dc;
  io_def          io;
  GtkWidget       *notebook, *page_label, *save_button, *apply_button, *revert_button;
} pr_pages_def;

typedef struct {
  GtkWidget    *preferences_dialog, *tpsid_item, *tpsid_menu, **caller;
  pr_pages_def page;
  XFILE_STAT   cfg_file_available;
  char         *xerrors, *rpt_store, *report_filename;
  key_lst      original_image_keys, image_keys, db_keys;
  img_typ      ximg, simg, oimg;
  db_def       xplab_db;
} pref_def;

static pref_def  pr;                                   /* preferences dialog */
static cfg_def   cfg_mirror;      /* configuration mirror used for reverting */
static prog_ctrl ctrl_apply;         /* preferences used for current session */

prog_ctrl xctrl;                        /* global preferences working buffer */

/* local prototypes ********************************************************/

static void init_stats_def(stats_def *sd);
static void init_pref_def(pref_def *pd);
static void default_pref_def(pref_def *pd);
static void init_general_def(general_def *gd);
static void init_matching_def(matching_def *md);
static void init_correlation_def(correlation_def *cd);
static void init_dc_def(dc_def *dd);
static void init_pre_def(pre_def *pd);
static void init_io_def(io_def *id);
static void init_pr_pages_def(pr_pages_def *ppd);
static void init_d1_def(d1_def *d1);
static void init_d2i_def(d2i_def *d2i);
static void init_d2ii_def(d2ii_def *d2ii);
static void free_stats_def(stats_def *sd);
static void free_pref_def(pref_def *pd);
static void free_general_def(general_def *gd);
static void free_matching_def(matching_def *md);
static void free_correlation_def(correlation_def *cd);
static void free_dc_def(dc_def *dd);
static void free_d1_def(d1_def *d1);
static void free_d2i_def(d2i_def *d2i);
static void free_d2ii_def(d2ii_def *d2ii);
static void free_pre_def(pre_def *pd);
static void free_io_def(io_def *id);
static void free_pr_pages_def(pr_pages_def *ppd);

/* Functions assembling widgets */
static void xplab_preferences(void);
static char *xplab_getcwd(void);
static void xplab_absolute_filespec(char **filespec);
static void read_cfg(char *fspec, char **gtk_rc, cfg_def *cfg, stats_def *sfg);
static void preferences_dialog_upper(GtkBox *vbox_in);
static void preferences_dialog_lower(GtkContainer *inbox);
static GtkWidget *preferences_dialog_page_general(void);
static GtkWidget *preferences_dialog_page_preprocessing(void);
static GtkWidget *preferences_dialog_page_matching(void);
static GtkWidget *preferences_dialog_page_correlation(void);
static GtkWidget* preferences_dialog_page_statistic(void);
static GtkWidget *preferences_dialog_page_dc(void);
static GtkWidget *preferences_dialog_page_io(void);
static void dialog_page_general_part1(GtkWidget *vbox);
static void dialog_page_general_part2(GtkWidget *vbox);
static void mborder_setting(GtkWidget *cbox);
static void nborder_setting(GtkWidget *cbox);
static void dc_store_setting(GtkWidget *cbox);
static d1_def dc_d1_setting(GtkWidget *cbox);
static d2i_def dc_d2i_setting(GtkWidget *fbox1);
static d2ii_def dc_d2ii_setting(GtkWidget *fbox2);
static void dc_option_list(GtkWidget *vbox_in, GtkEntry **entry, gpointer gpo, char *def_text);
static void maskweight_setting(GtkWidget *cbox);
static void nnbmatch_setting(GtkWidget *cbox);
static void interleave_setting(GtkWidget *cbox);
static void overlap_setting(GtkWidget *cbox);
static void symbolspacing_setting(GtkWidget *cbox);
static void nb_lap_setting(GtkWidget *cbox);
static void histo_thold_setting(GtkWidget *cbox);
static void cluster_force_setting(GtkWidget *cbox);
static void cluster_nb_setting(GtkWidget *cbox);
static void cluster_write_setting(GtkWidget *cbox);
static void timestamp_setting(GtkWidget **abox, GtkWidget **bbox);
static void writeauto_setting(GtkWidget *vbox);
static void domain_setting(GtkWidget *cbox);
static void line_method_setting(GtkWidget **abox, GtkWidget **bbox);
static void tooltips_setting(GtkWidget *cbox);
static void verbose_setting(GtkWidget *cbox);
static void raise_setting(GtkWidget *cbox);
static void rawbit_setting(GtkWidget **abox, GtkWidget **bbox);
static void image_filename_setting(GtkWidget **abox);
static void report_filename_setting(GtkWidget *vbox);
static void report_write_setting(GtkWidget **abox, GtkWidget **bbox, GtkWidget **cbox, GtkWidget **dbox);
static void writeauto_filename_setting(GtkWidget *vbox);
static void rb_chain_setting(GtkWidget *cbox);
static void cluster_nb_setting(GtkWidget *cbox);
static void batchjob_setting(GtkWidget *cbox);
static void batchlog_setting(GtkWidget *cbox);
static void separation_histo_filename_setting(GtkWidget *cbox);
static void separation_rules_filename_setting(GtkWidget *cbox);
static void histo_write_setting(GtkWidget *cbox);
static void histo_setting(GtkWidget *cbox);
static void duration_setting(GtkWidget **abox, GtkWidget **bbox);
static void support_dots_setting(GtkWidget *cbox);
static void overwrite_image_setting(GtkWidget **abox, GtkWidget **bbox);
static void training_session_filename(GtkWidget *cbox);
static void training_write_setting(GtkWidget *cbox);
static void replay_setting(GtkWidget *cbox);
static void training_setting(GtkWidget *cbox);
static void phases_setting(GtkWidget *vbox);
static void maintenance_setting(GtkWidget *fbox);
static void replay_filename_setting(GtkWidget *cbox);
static void replay_write_setting(GtkWidget *cbox);
static void single_line_setting(GtkWidget **abox, GtkWidget **bbox);
static void single_nb_setting(GtkWidget **abox, GtkWidget **bbox);
static void broken_nb_setting(GtkWidget **abox, GtkWidget **bbox);
static void renew_method_setting(GtkWidget **abox);
static void subst_rb_setting(GtkWidget **abox, GtkWidget **bbox);
static void subst_nb_setting(GtkWidget **abox, GtkWidget **bbox);
static void rgb_norm_setting(GtkWidget *cbox);
static void ckernel_setting(GtkWidget *cbox);
static void correlation_setting(GtkWidget *cbox);
static void corrlevel_setting(GtkWidget *cbox);
static void corrbroken_setting(GtkWidget *cbox);
static void clustering_setting(GtkWidget *cbox);
static void cluster_level_setting(GtkWidget *cbox);
static void cluster_step_setting(GtkWidget *cbox);
static void cluster_thold_setting(GtkWidget *cbox);
static int errors_setting(int esum);
static void batchname_setting(GtkWidget *cbox);
static void cfgname_setting(GtkWidget *cbox);
static void logname_setting(GtkWidget *cbox);
static void tps_id_setting(GtkWidget *cbox);
static void general_matching_setting(GtkWidget *cbox);
static void matching_rules_setting(GtkWidget *cbox);
static void general_io_setting(GtkWidget *cbox);
static void report_setting(GtkWidget *cbox);
static void report_file_table(GtkWidget *vbox);
static void prg_setting(GtkWidget *cbox);
static void rc_setting(GtkWidget *cbox);
static void cwd_setting(GtkWidget *cbox);
static void db_setting(GtkWidget *cbox);
static void image_setting(GtkWidget *cbox);
static void setting_integer_limits_no_default(GtkWidget *cbox, int min, int max);
static void setting_double_limits(GtkWidget *cbox, double def, double min, double max);
static GtkWidget *preference_menu_verbose(void);
static GtkWidget *combo_verbose(void);
static GtkWidget *preference_menu_line_method(void);
static GtkWidget *combo_line_method(void);
static img_typ evaluate_image_data_input(db_def idb, char *image_name_in, char **image_name_out, rb_lst *rblst, hst_def *grey_hist, hst_def *th_hist, cfg_def cfg, img_typ *porig, img_typ *tses);
static void evaluate_db_input(char *db_name, char **tps_id, int mborder, int maskweight, int verbose);
static void preferences_tpsid_menu_item_id(GtkWidget *tpsid_menu);
static char *dc_option_list_format(int idx, int cnt, int val);
static void update_dc_entries(char *user_data, int incr);
static void dc_option_entry_from_list(GtkEntry *entry, int idx, int2_lst dlst, char *def);
static void dc_option_entry_from_list_all(void);
static char *get_line(FILE *fin, int *line);
static int check_vals(char *fname, char *key, int  line, char *fmt, ...);
static int get_key(char *lstr, const char *key, const char *fmt, ...);
static void close_rpt_file(char *report_filename, set_def set, int verbose);
static char *open_rpt_file(prog_ctrl *ctrl, char *db_name, int verbose);
static gboolean emit_broadcast(key_lst klst);
static void emit_broadcast_images(void);
static void dump_session_image(pic_def simg, char *filename);

/* Handlers/Callbacks */
static void cb_preferences_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_preferences_page_switch(GTK_CONVENTION_NOTEBOOK_RTS_VOID);
static gboolean cb_preferences_delete(GTK_CONVENTION_DIALOG_RTS_GBOOLEAN);
static void cb_preferences_save_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_preferences_apply_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_preferences_revert_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_preferences_cancel_button(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_mborder_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_nborder_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_nb_border_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_dc_store_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_maskweight_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_nnbmatch_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_interleave_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_overlap_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_symbolspacing_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_nb_lap_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_histo_thold_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_rgb_norm_r_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_rgb_norm_g_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_rgb_norm_b_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_rgb_norm_total_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_subst_rb_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_subst_nb_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_mborder_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nborder_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dc_store_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nb_border_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_maskweight_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nnbmatch_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_interleave_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_overlap_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_symbolspacing_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nb_lap_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_histo_thold_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_r_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_g_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_b_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_total_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_subst_rb_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_subst_nb_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_subst_rb_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_subst_nb_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_histo_thold_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_r_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_g_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_b_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_rgb_norm_total_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_corrlevel_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_corrbroken_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_ckernel_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_cluster_level_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_cluster_step_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_cluster_thold_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_mborder_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nborder_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dc_store_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nb_border_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_maskweight_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nnbmatch_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_interleave_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_overlap_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_symbolspacing_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_nb_lap_revert(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_line_method_entry(GTK_CONVENTION_COMBO_CHANGED_RTS_VOID);
static void cb_ckernel_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_corrlevel_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_corrbroken_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_cluster_level_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_cluster_step_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_cluster_thold_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID);
static void cb_corrlevel_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_corrbroken_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_ckernel_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_cluster_level_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_cluster_step_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_cluster_thold_default(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_toggle_tpsid(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID);
static void cb_preferences_unload(GTK_CONVENTION_MENUITEM_RTS_VOID);
static void cb_dc_down(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dc_up(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dc_new(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dc_remove(GTK_CONVENTION_BUTTON_RTS_VOID);
static void cb_dc_button_to_entry(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);
static void cb_dots_support(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);
static void cb_cluster_nb(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID);

/* Functions assembling handlers/callbacks */
static int check_mborder_setting(void);
static int check_nborder_setting(void);
static int check_dc_store_setting(void);
static int check_nb_border_setting(GtkEntry *nb_border_entry, int nb_border_default);
static int check_maskweight_setting(void);
static int check_nnbmatch_setting(void);
static double check_interleave_setting(void);
static double check_overlap_setting(void);
static double check_symbolspacing_setting(void);
static double check_nb_lap_setting(void);
static double check_rgb_norm_r_setting(int adapt);
static double check_rgb_norm_g_setting(int adapt);
static double check_rgb_norm_b_setting(int adapt);
static void adapt_rgb_norm_total_setting(void);
static void check_rgb_norm_total_setting(void);
static char *check_subst_rb_setting(void);
static char *check_subst_nb_setting(void);
static double check_corrlevel_setting(void);
static double check_corrbroken_setting(void);
static int check_ckernel_setting(GtkEntry *entry, int row, int col);
static void enroll_dc_1d_to_list(void);
static void enroll_dc_2di_to_list(void);
static void enroll_dc_2dii_to_list(void);
static void apply_all_ckernel(kern_def ikern);
static void revert_statistic_page(stats_def *stats);
static int apply_domain(void);
static batch_def apply_batch(void);
static void apply_dc_1d(int2_lst *d1_lst);
static void apply_dc_2d(int2_lst *d2_lst);
static int apply_timestamp(void);
static int apply_writeauto(void);
static int apply_report_write(void);
static int apply_rb_chain(void);
static CLUSTER_NB_BIT apply_cluster_nb(void);
static int apply_histo_write(void);
static int apply_duration(void);
static int apply_support_dots(void);
static int apply_support_dots_auto(void);
static int apply_tooltips(void);
static int apply_raise(void);
static int apply_overwrite_image(void);
static int apply_replay_write(void);
static int apply_training_write(void);
static int apply_single_line(void);
static int apply_single_nb(void);
static MATCHING_METHOD apply_broken_nb(void);
static int apply_renew_method(void);
static int apply_cluster_force(void);
static int apply_cluster_write(void);
static int apply_rawbit(void);
static void apply_statistic_page(stats_def *stats);
static void set_all_ckernel(kern_def ikern);
static double check_cluster_level_setting(void);
static double check_cluster_step_setting(void);
static int check_cluster_thold_setting(void);
static int check_verbose_setting(void);
static INLINE_METHOD check_line_method_setting(void);
static void set_sensitivity_renew_method(void);
static void update_all(cfg_def cfg);
static void update_statistic_page(stats_def stats);
static void update_preprocessing_page(pre_def preprocessing, cfg_def cfg);
static void update_matching_page(matching_def matching, cfg_def cfg);
static void update_mborder(matching_def matching, cfg_def cfg);
static void update_nborder(matching_def matching, cfg_def cfg);
static void update_maskweight(matching_def matching, cfg_def cfg);
static void update_nnbmatch(matching_def matching, cfg_def cfg);
static void update_interleave(matching_def matching, cfg_def cfg);
static void update_overlap(matching_def matching, cfg_def cfg);
static void update_symbolspacing(matching_def matching, cfg_def cfg);
static void update_nb_lap(matching_def matching, cfg_def cfg);
static void update_line_method(matching_def matching, cfg_def cfg);
static void update_single_line(matching_def matching, cfg_def cfg);
static void update_single_nb(matching_def matching, cfg_def cfg);
static void update_broken_nb(matching_def matching, cfg_def cfg);
static void update_renew_method(matching_def matching, cfg_def cfg);
static void update_general_page(general_def general, cfg_def cfg);
static void update_rgb_norm(pre_def preprocessing, rgb_coef rgb_norm);
static void update_histo_thold(pre_def preprocessing, cfg_def cfg);
static void update_support_dots(pre_def preprocessing, cfg_def cfg);
static void update_support_dots_auto(pre_def preprocessing, cfg_def cfg);
static void update_verbose(general_def general, cfg_def cfg);
static void update_batch(general_def general, cfg_def cfg);
static void update_tooltips(general_def general, cfg_def cfg);
static void update_raise(general_def general, cfg_def cfg);
static void update_subst_rb(matching_def matching, cfg_def cfg);
static void update_subst_nb(matching_def matching, cfg_def cfg);
static void update_nb_border(io_def iod, cfg_def cfg);
static void update_timestamp(io_def iod, cfg_def cfg);
static void update_writeauto(io_def iod, cfg_def cfg);
static void update_rb_chain(general_def general, cfg_def cfg);
static void update_duration(matching_def general, cfg_def cfg);
static void update_overwrite_image(io_def iod, cfg_def cfg);
static void update_replay_write(io_def iod, cfg_def cfg);
static void update_training_write(io_def iod, cfg_def cfg);
static void update_dc_1d(d1_def d1, int opt1);
static void update_dc_2di(d2i_def d2, int opt2i);
static void update_dc_2dii(d2ii_def d2, int opt2ii);
static void update_rawbit(io_def iod, cfg_def cfg);
static void update_report_write(io_def iod, cfg_def cfg);
static void update_histo_write(io_def iod, cfg_def cfg);
static void update_domain(correlation_def correlation, cfg_def cfg);
static void update_correlation_page(correlation_def correlation, cfg_def cfg);
static void update_corrlevel(correlation_def correlation, cfg_def cfg);
static void update_corrbroken(correlation_def correlation, cfg_def cfg);
static void update_cluster_level(correlation_def correlation, cfg_def cfg);
static void update_cluster_step(correlation_def correlation, cfg_def cfg);
static void update_cluster_thold(correlation_def correlation, cfg_def cfg);
static void update_cluster_force(correlation_def correlation, cfg_def cfg);
static void update_cluster_write(correlation_def correlation, cfg_def cfg);
static void update_cluster_nb(correlation_def correlation, cfg_def cfg);
static void update_dc_page(cfg_def cfg);
static void update_io_page(io_def io, cfg_def cfg);
static void update_hidden_list(int2_lst *src, int2_lst *dst, int *dst_idx);
static void update_dc_buttons_from_list(void);
static int read_ckernel_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_gtkrc_key(char *fspec, char *cbuf, int line, char **gtk_rc);
static int read_rgb_norm_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_subst_rb_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_subst_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_rawbit_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_report_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_rb_chain_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_cluster_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_histo_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_duration_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_support_dots_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_support_dots_auto_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_overwrite_image_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_replay_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_training_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_single_line_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_single_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_broken_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_renew_method_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_cluster_force_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_cluster_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_timestamp_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_writeauto_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_domain_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_dc_1d_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_dc_2d_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_batchjob_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_batchlog_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_verbose_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_tooltips_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_raise_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_line_method_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_mborder_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_maskweight_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_nborder_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_dc_store_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_nb_border_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_nnbmatch_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_interleave_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_overlap_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_symbolspacing_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_nb_lap_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_histo_thold_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_corrlevel_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_corrbroken_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_cluster_level_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_cluster_step_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_cluster_thold_key(char *fspec, char *cbuf, int line, cfg_def *cfg);
static int read_stats_key(char *fspec, char *cbuf, int line, char *keyname, stats_prefs *cfg);
static int read_stats_contrast(char *fspec, char *cbuf, int line, char *keyname, stats_prefs *sfg);
static void xplab_errors_txt(char *key_string, int line, char *cfg_name, int howmuch);
static void write_cfg(set_def set, cfg_def cfg, stats_def sfg);
static void write_stats_prefs(FILE *file, char *keyname, stats_win swin, char *sepn);
static void write_gtk_rc_prefs(FILE *file, char *gtk_rc, char *sepn);
char *string_user_options(int argc, char *argv[]);
static cli_def read_cli(int argc, char *argv[], char **xprog, char **image_name, char **db_name, char **options, char **cfg_filespec, int *batch_enable);
static db_def setup_db(char *db_name, char **tps_id, int mborder, int maskweight, int verbose);
static int force_id(char **ident, tps_lst tpl);
static char *create_db_attributes(char *db_name, char **db_version);
static void xplab_unload_db(void);

/* Functions ***************************************************************/

static void
init_general_def(general_def *gd)
{
  assert(gd);
                           /* don't init GTK stuff */
}


static void
init_matching_def(matching_def *md)
{
  assert(md);
                           /* don't init GTK stuff */
}


static void
init_correlation_def(correlation_def *cd)
{
  assert(cd);
                           /* don't init GTK stuff */
}


static void
init_dc_def(dc_def *dd)
{
  assert(dd);
                           /* don't init GTK stuff */

  dd->d1_idx = dd->d2_idx = 0;
  init_d1_def(&dd->d1);
  init_d2i_def(&dd->d2i);
  init_d2ii_def(&dd->d2ii);
}


static void
init_pre_def(pre_def *pd)
{
  assert(pd);
                           /* don't init GTK stuff */
}

static void
default_stats_def(stats_def *sd)
{
  assert(sd);

  default_stats_prefs(&sd->owin);
  default_stats_prefs(&sd->cwin);
  default_stats_prefs(&sd->gwin);
  default_stats_prefs(&sd->bwin);
  default_stats_prefs(&sd->rhst);
  default_stats_prefs(&sd->chst);
  default_stats_prefs(&sd->ghst);
  default_stats_prefs(&sd->bhst);
  default_stats_prefs(&sd->shst);
  default_stats_icon_contrast(&sd->contrast);
}

static void
init_stats_def(stats_def *sd)
{
  assert(sd);
                           /* don't init GTK stuff */
  init_stats_prefs(&sd->owin);
  init_stats_prefs(&sd->cwin);
  init_stats_prefs(&sd->gwin);
  init_stats_prefs(&sd->bwin);
  init_stats_prefs(&sd->rhst);
  init_stats_prefs(&sd->chst);
  init_stats_prefs(&sd->ghst);
  init_stats_prefs(&sd->bhst);
  init_stats_prefs(&sd->shst);
  init_stats_prefs(&sd->contrast);
}


static void
init_io_def(io_def *id)
{
  assert(id);
                           /* don't init GTK stuff */
}


static void
init_d1_def(d1_def *d1)
{
  assert(d1);
                           /* don't init GTK stuff */
}


static void
init_d2i_def(d2i_def *d2i)
{
  assert(d2i);
                           /* don't init GTK stuff */
}


static void
init_d2ii_def(d2ii_def *d2ii)
{
  assert(d2ii);
                           /* don't init GTK stuff */
}

static void
default_pr_pages_def(pr_pages_def *ppd)
{
  assert(ppd);

  default_stats_def(&ppd->statistic);
}

static void
init_pr_pages_def(pr_pages_def *ppd)
{
  assert(ppd);
                           /* don't init GTK stuff */

  init_general_def(&ppd->general);
  init_pre_def(&ppd->preprocessing);
  init_matching_def(&ppd->matching);
  init_correlation_def(&ppd->correlation);
  init_stats_def(&ppd->statistic);
  init_dc_def(&ppd->dc);
  init_io_def(&ppd->io);
}


static void
default_pref_def(pref_def *pd)
{
  assert(pd);

  default_pr_pages_def(&pd->page);
}

static void
init_pref_def(pref_def *pd)
{
  assert(pd);

  pd->preferences_dialog = NULL;                           /* zero pointer */
  pd->caller = NULL;                                       /* zero pointer */
  pd->xerrors = NULL;                                      /* zero pointer */
  pd->rpt_store = NULL;                                    /* zero pointer */
  pd->report_filename = NULL;                              /* zero pointer */
  pd->tpsid_item = NULL;                                   /* zero pointer */
  pd->tpsid_menu = NULL;                                   /* zero pointer */

  pd->cfg_file_available = FILE_UNDEFINED;

  init_pr_pages_def(&pd->page);
  init_key_lst(&pd->original_image_keys);
  init_key_lst(&pd->image_keys);
  init_key_lst(&pd->db_keys);
  init_img_typ(&pd->ximg);
  init_img_typ(&pd->simg);
  init_img_typ(&pd->oimg);
  init_db_def(&pd->xplab_db);
}


void
xplab_free_preferences(void)
{
  close_rpt_file(pr.report_filename, xctrl.set, xctrl.cfg.verbose > BE_QUIET);

  free_pref_def(&pr);                      /* free local preferences data  */
  free_cfg_def(&cfg_mirror);               /* free local configuration data*/
  free_prog_ctrl(&ctrl_apply);             /* free local preferences data  */
  free_prog_ctrl(&xctrl);                  /* free global preferences data */
}


static void
free_general_def(general_def *gd)
{
  assert(gd);
                           /* don't free GTK stuff */

  init_general_def(gd);
}


static void
free_matching_def(matching_def *md)
{
  assert(md);
                           /* don't free GTK stuff */

  init_matching_def(md);
}


static void
free_correlation_def(correlation_def *cd)
{
  assert(cd);
                           /* don't free GTK stuff */

  init_correlation_def(cd);
}


static void
free_d1_def(d1_def *d1)
{
  assert(d1);
                           /* don't free GTK stuff */
}


static void
free_d2i_def(d2i_def *d2i)
{
  assert(d2i);
                           /* don't free GTK stuff */
}


static void
free_d2ii_def(d2ii_def *d2ii)
{
  assert(d2ii);
                           /* don't free GTK stuff */
}


static void
free_dc_def(dc_def *dd)
{
  assert(dd);
                           /* don't free GTK stuff */

  free_int2_lst(&dd->dc1l);
  free_int2_lst(&dd->dc2l);
  free_d1_def(&dd->d1);
  free_d2i_def(&dd->d2i);
  free_d2ii_def(&dd->d2ii);

  init_dc_def(dd);
}


static void
free_pre_def(pre_def *pd)
{
  assert(pd);
                           /* don't free GTK stuff */

  init_pre_def(pd);
}


static void
free_stats_def(stats_def *sd)
{
  assert(sd);
                           /* don't free GTK stuff */

  free_stats_prefs(&sd->owin);
  free_stats_prefs(&sd->cwin);
  free_stats_prefs(&sd->gwin);
  free_stats_prefs(&sd->bwin);
  free_stats_prefs(&sd->rhst);
  free_stats_prefs(&sd->chst);
  free_stats_prefs(&sd->ghst);
  free_stats_prefs(&sd->bhst);
  free_stats_prefs(&sd->shst);
  free_stats_prefs(&sd->contrast);

  init_stats_def(sd);
}


static void
free_io_def(io_def *id)
{
  assert(id);
                           /* don't free GTK stuff */

  init_io_def(id);
}


static void
free_pref_def(pref_def *pd) 
{
  assert(pd);
                           /* don't free GTK stuff */

  FREEMEM(pd->xerrors);
  FREEMEM(pd->rpt_store);
  FREEMEM(pd->report_filename);

  free_pr_pages_def(&pd->page);
  free_key_lst(&pd->original_image_keys);
  free_key_lst(&pd->image_keys);
  free_key_lst(&pd->db_keys);
  free_img_typ(&pd->ximg);
  free_img_typ(&pd->simg);
  free_img_typ(&pd->oimg);
  free_db_def(&pd->xplab_db);

  init_pref_def(pd); 
}


static void
free_pr_pages_def(pr_pages_def *ppd)
{
  assert(ppd);
                           /* don't free GTK stuff */

  free_general_def(&ppd->general);
  free_pre_def(&ppd->preprocessing);
  free_matching_def(&ppd->matching);
  free_correlation_def(&ppd->correlation);
  free_stats_def(&ppd->statistic);
  free_dc_def(&ppd->dc);
  free_io_def(&ppd->io);
}


/* Evaluates all input and initializes data at start-up.

Revision: 05.12.2004 E.Sack

Input : - -
Output: - default setting
*/
cli_def
xplab_init_preferences(int argc, char *argv[])
{
  char    *image_name = NULL, *db_name = NULL;
  cli_def clid;

  init_pref_def(&pr);                       /* init local preferences data */
  init_cfg_def(&cfg_mirror);              /* init local configuration data */
  init_prog_ctrl(&ctrl_apply);              /* init local preferences data */
  init_prog_ctrl(&xctrl);                  /* init public preferences data */

  default_pref_def(&pr);        /* default setup not initialised elsewhere */
  xdefault_prog_ctrl(&xctrl);      /* default setup - get data public asap */

  add_string(DB_FNAME_DEFAULT, &db_name);         /* init default filename */
  clid = read_cli(argc, argv, &xctrl.set.xprog, &image_name, &db_name, 
		  &xctrl.set.user_options, &xctrl.set.config_filespec, 
		  &xctrl.set.dobatch);

  read_cfg(xctrl.set.config_filespec, &xctrl.set.gtk_rc, &xctrl.cfg,
	   &pr.page.statistic);                      /* scan configuration */

  /* database evaluation has side effects on image evaluation              */
  evaluate_db_input(db_name, &xctrl.cfg.tps_id, xctrl.cfg.mad.mborder, 
		    xctrl.cfg.mad.maskweight, xctrl.cfg.verbose > BE_QUIET);
  FREEMEM(db_name);

  /* image evaluation depends on database (called after DB evaluation)     */
  pr.ximg = evaluate_image_data_input(pr.xplab_db, image_name, 
				      &xctrl.set.image_filename, 
				      &xctrl.set.rblst, &xctrl.set.grey_hist,
				      &xctrl.set.th_hist, xctrl.cfg, 
				      &pr.oimg, &pr.simg);
  FREEMEM(image_name);

  copy_prog_ctrl(&xctrl, &ctrl_apply);                   /* local <- local */
  copy_cfg_def(&xctrl.cfg, &cfg_mirror);                /* local <- global */

  pr.report_filename = open_rpt_file(&xctrl, pr.xplab_db.db_fname, 
				     xctrl.cfg.verbose > BE_QUIET);
  return clid;
}



/* Evaluates database input and sets according flags.

Revision: 21.11.2003 E.Sack

Input : - pointer to DB filename
        - pointer to TPS ID
        - mask border
        - verbose flag
Output: - local DB updated
Precondition: - prevent using DB items in interface because of deallocation
*/
static void
evaluate_db_input(char *db_name, char **tps_id, int mborder, int maskweight,
		  int verbose)
{
  assert(tps_id);

  free_db_def(&pr.xplab_db);                              /* delete old DB */

  pr.xplab_db = setup_db(db_name, tps_id, mborder, maskweight, verbose);
}



/* Reads and evaluates image data. Also image preprocessing will be done.

Revision: 04.11.2004 E.Sack

Input : - database
        - pointer to input image name
        - pointer to pointer to output image name
        - pointer to RB list
        - settings
        - pointer to original image
        - pointer to session image
Output: - binary image
        - settings updated
Precondition:  - if database is empty, some features cannot be calculated.
               - pointer to pointer to output image name must be supported.
               - pointer to RB list may be NULL
Postcondition: - if pointer of original image is NULL, original image will
                 be lost (e.g. if there is no need for them).
               - if pointer of session image is NULL, session image will
                 be lost (e.g. if there is no need for them).
               - input and output image name path is splitted to simplify
                 their handling outside.
*/
static img_typ
evaluate_image_data_input(db_def idb, char *image_name_in, 
			  char **image_name_out, rb_lst *rblst, 
			  hst_def *grey_hist, hst_def *th_hist, cfg_def cfg,
			  img_typ *porig, img_typ *tses)
{
  int     err;
  img_typ dimg;

  assert(image_name_out);
  assert(rblst);

  init_img_typ(&dimg);

  err = check_name(image_name_in);

  if (!err) {
    FREEMEM(*image_name_out);
    add_string(image_name_in, image_name_out);

    dimg = setup_input_image(idb, *image_name_out, cfg, rblst, grey_hist, 
			     th_hist, porig, tses);
  }

  if (!AREA_SIZE(dimg.pimg.ar)) {
    xplab_text_insert(main_log, XPLAB_NO_IMAGE_MSG "\n");
  }

  return dimg;
}



/* Broadcasts "images-have-changed signal" for all images.

Revision: 01.11.2004 E.Sack

Input : - -
Output: - signals emitted
Precondition: - assumed image processing hierarchy: if the original image
                has been changed by some reason, all other images may have
                been changed as well, because they depend on the original one.
                However, there is not an inverse dependency.
*/
static void
emit_broadcast_images(void)
{
  gboolean emitted;

  emitted = emit_broadcast(pr.original_image_keys);

  if (emitted == FALSE) {                      /* known processing hierarchy */
    emit_broadcast(pr.image_keys);/* may have changed, e.g. changed settings */
  }
}



/* Broadcasts "source-have-changed signal" for a key list.

Revision: 15.12.2004 E.Sack

Input : - key list
Output: - signal emitted as flagged
        - flag (set if a signal is emitted else unset)
Usage:  - this function is fitted to a key list which points to a button 
          widget as entry. Don't use the function if that is not true.
        - a filter is used to send a certain signal only one time.
        - Caution: as the widget is used for calling back a signal handler,
          which in turn has priority over this function, the list may be
          destroyed by that handler. Some protection has been implemented
          to avoid this race condition.
*/
static gboolean
emit_broadcast(key_lst klst)
{
  int        i, len = klst.cnt, filter = 0;
  GtkWidget  *bc_sig_new;
  widget_lst bc_sig_old;
  key_lst    cplst;
  gboolean   emitted = FALSE;

  init_key_lst(&cplst);
  init_widget_lst(&bc_sig_old);

  copy_key_lst(&klst, &cplst);          /* copy list list before raising CBs */

  while (len--) {
    if (cplst.kar[len].aux == NULL) {        /* cannot use undefined pointer */
      continue;
    }

    bc_sig_new = *(GtkWidget**)cplst.kar[len].aux;

    if (bc_sig_new == NULL) {                 /* cannot use NULL for signals */
      continue;
    }

    for (i = filter = 0; i < bc_sig_old.cnt; i++) {
      if (bc_sig_new == bc_sig_old.wid[i]) {
	filter = 1;
	break;
      }
    }

    if (filter == 1) {
      continue;                                       /* use signal one time */
    }

    put_widget(&bc_sig_old, bc_sig_new);               /* put signal to list */

    if (GTK_IS_BUTTON(bc_sig_new) == TRUE) {                  /* signal type */
      gtk_button_clicked(GTK_BUTTON(bc_sig_new));
      emitted = TRUE;
    }
    else if (GTK_IS_MENU_ITEM(bc_sig_new) == TRUE) {
      gtk_menu_item_activate(GTK_MENU_ITEM(bc_sig_new));
      emitted = TRUE;
    }
  }

  free_key_lst(&cplst);
  free_widget_lst(&bc_sig_old);

  return emitted;
}



/* Gives back the preference settings for statistic.

Revision: 16.12.2004 E.Sack

Input : - requested format
Output: - image window settings
Usage:  - access function for data hiding.
*/
stats_win 
xplab_request_stats_prefs(STATS_FORMAT fmt)
{
  stats_win pwin;

  init_stats_win(&pwin);

  switch (fmt) {
  case STATS_IMAGE_ORIGINAL:
    pwin = pr.page.statistic.owin.sapply;
    break;

  case STATS_IMAGE_COLOR:
    pwin = pr.page.statistic.cwin.sapply;
    break;

  case STATS_IMAGE_GREY:
    pwin = pr.page.statistic.gwin.sapply;
    break;

  case STATS_IMAGE_BMP:
    pwin = pr.page.statistic.bwin.sapply;
    break;

  case STATS_HISTO_RBS:
    pwin = pr.page.statistic.rhst.sapply;
    break;

  case STATS_HISTO_SLOPE:
    pwin = pr.page.statistic.shst.sapply;
    break;

  case STATS_HISTO_COLOR:
    pwin = pr.page.statistic.chst.sapply;
    break;

  case STATS_HISTO_GREY:
    pwin = pr.page.statistic.ghst.sapply;
    break;

  case STATS_HISTO_BMP:
    pwin = pr.page.statistic.bhst.sapply;
    break;

  case STATS_ICON_CONTRAST_MAX:
    pwin = pr.page.statistic.contrast.sapply;
    break;

  default: 
    break;
  }

  return pwin;
}



/* Gives back an image copy providing a locking mechanism. The image can be
   requested by more than one caller at the same time.

Revision: 10.12.2004 E.Sack

Input : - requested format
        - pointer to comment string
        - pointer to key
        - access flag (0: read only, else: read & write)
        - pointer to broadcast handler
Output: - image
        - key updated
Postcondition: - requested/returned image may be empty.
*/
img_typ
xplab_request_image(PNM_FORMAT fmt, char *caller_string, int *ikey, 
		    int write_access, void *bc_ptr)
{
  int     err, d, idx;
  char    *msg = NULL;
  img_typ image_copy, *image_source = NULL;
  key_lst *klst = NULL;

  init_img_typ(&image_copy);

  switch (fmt) {                                          /* request setup */
  case PNM_FORMAT_ORIGINAL:                              /* original image */
    klst = &pr.original_image_keys;
    add_string("original ", &msg);
    image_source = &pr.oimg;                      /* image bound to format */

    break;

  case P1: case P4:                                        /* binary image */
    klst = &pr.image_keys;
    add_string("binary ", &msg);
    image_source = &pr.ximg;                      /* image bound to format */

    break;

  case PXPLAB2: case PXPLAB5:                             /* session image */
    klst = &pr.image_keys;    /* session image is a binary image dependant */
    add_string("session ", &msg);
    image_source = AREA_SIZE(pr.simg.pimg.ar) ? 
                           &pr.simg :   /* session image history available */
               	           &pr.ximg; /* no session image history available */
    break;

  default: 
    break;                                        /* requested image empty */
  }

  if (write_access && klst) {        /* write access if key list available */
    idx = request_key_lst(*klst);

    if (idx >= 0) {
      string_to_log(main_log, msg, "image locked", "");

      d = check_name(klst->kar[idx].cmt);

      if (d) {
	string_to_log(main_log, "\n", "", "");
      }
      else {
	string_to_log(main_log, " by '", klst->kar[idx].cmt, "'\n");
      }

      FREEMEM(msg);
      return image_copy;
    }

    string_to_log(main_log, msg, "image locked", "");

    d = check_name(caller_string);

    if (d) {
      string_to_log(main_log, "\n", "", "");
    }
    else {
      string_to_log(main_log, " by '", caller_string, "'\n");
    }
  }

  /* activate locking mechanism even if image is empty                     */
  err = put_key(klst, ikey, caller_string, write_access, bc_ptr);
  assert(!err);                                     /* likely not released */

  if (image_source) {
    image_copy = xdup_img_typ(*image_source);                /* image copy */
  }

  FREEMEM(msg);

  return image_copy;                       /* requested image may be empty */
}



/* Gives back an image copy providing a locking mechanism.

Revision: 01.11.2004 E.Sack

Input : - key
Output: - flag (0: no error, 1: error detected)
*/
int
xplab_release_image(PNM_FORMAT fmt, int *ikey, int verbose)
{
  int     err;
  char    *msg = NULL, *cmt = NULL;
  key_lst *klst = NULL;

  switch (fmt) {                                          /* request setup */
  case PNM_FORMAT_ORIGINAL:                              /* original image */
    klst = &pr.original_image_keys;
    add_string("original ", &msg);

    break;

  case P1: case P4:                                        /* binary image */
    klst = &pr.image_keys;
    add_string("binary ", &msg);

    break;

  case PXPLAB2: case PXPLAB5:                             /* session image */
    klst = &pr.image_keys;    /* session image is a binary image dependant */
    add_string("session ", &msg);

    break;

  default: 
    return 0;                                       /* no image to release */
    break;
  }

  err = erase_key(klst, ikey, &cmt);

  if (!err && verbose) {
    string_to_log(main_log, msg, "image released", "");

    if (cmt) {
      string_to_log(main_log, " by '", cmt, "'\n");
    }
    else {
      string_to_log(main_log, "\n", "", "");
    }
  }

  FREEMEM(msg);
  FREEMEM(cmt);
  return err;
}



/* Releases a DB.

Revision: 01.06.2004 E.Sack

Input : - key
Output: - flag (0: no error, 1: error detected)
*/
int
xplab_release_db(int *key, int verbose)
{
  int  err;
  char *cmt = NULL;

  assert(key);

  err = erase_key(&pr.db_keys, key, &cmt);

  if (!err && verbose) {
    string_to_log(main_log, "DB released", "", "");

    if (cmt) {
      string_to_log(main_log, " by '", cmt, "'\n");
    }
    else {
      string_to_log(main_log, "\n", "", "");
    }
  }

  FREEMEM(cmt);

  return err;
}



/* Gives back a database copy providing a locking mechanism.

Revision: 18.07.2004 E.Sack

Input : - pointer to comment string
        - pointer to key
        - access flag
Output: - DB
        - key updated
*/
db_def
xplab_request_db(char *caller_string, int *key, XPLAB_DB db_access, 
		 void *bc_ptr)
{
  int    err, d, idx, rw_access = 0;
  db_def db_copy;

  init_db_def(&db_copy);

  if (db_access == XPLAB_DB_RW) {
    rw_access = 1;
    idx = request_key_lst(pr.db_keys);

    if (idx >= 0) {                         /* only one locking key possible */
      string_to_log(main_log, "DB locked", "", "");

      d = check_name(pr.db_keys.kar[idx].cmt);

      if (d) {
	string_to_log(main_log, "\n", "", "");
      }
      else {
	string_to_log(main_log, " by '", pr.db_keys.kar[idx].cmt, "'\n");
      }

      return db_copy;                             /* return empty DB on fail */
    }

    string_to_log(main_log, "DB locked", "", "");

    d = check_name(caller_string);

    if (d) {
      string_to_log(main_log, "\n", "", "");
    }
    else {
      string_to_log(main_log, " by '", caller_string, "'\n");
    }
  }

  err = put_key(&pr.db_keys, key, caller_string, rw_access, bc_ptr);/*activate locking */
  assert(!err);

  copy_db_def(&pr.xplab_db, &db_copy, FALSE);                     /* DB copy */

  if (db_copy.tpl.cnt == 0) {                 /* deliver a default empty TPS */
    int  tps_idx;
    char *buf = NULL;

    tps_idx = create_tps_def(&db_copy.tpl, &buf);        /* create empty TPS */
    assert(tps_idx >= 0);

    xdefault_media_att(&db_copy.tpl.tps[tps_idx].media);   /* default values */

    FREEMEM(buf);
  }

  return db_copy;
}



/* Returns the type of the current loaded original image.

Revision: 29.10.2004 E.Sack

Input : - -
Output: - original image type
Usage:  - the function is used to ask for the current image type avoiding 
          explicit request.
*/
PNM_FORMAT
xplab_image_loaded(void)
{
  return pr.oimg.typ;                               /* original image type */
}



/* Inputs a new original image controlled by a locking mechanism.

Revision: 04.11.2004 E.Sack

Input : - image file name
Output: - image copied as flagged
        - flag (0: image copied, 1: image not copied)
*/
int
xplab_input_image(char *fname)
{
  int d, idx, err = 1;

  idx = request_key_lst(pr.image_keys);

  if (idx >= 0) {                   /* locking activated by some application */
    string_to_log(main_log, "image locked", "", "");

    d = check_name(pr.image_keys.kar[idx].cmt);

    if (d) {
      string_to_log(main_log, "\n", "", "");
    }
    else {
      string_to_log(main_log, " by '", pr.image_keys.kar[idx].cmt, "'\n");
    }
  }
  else {                       /* image ready for loading, reset all hooks */
    xplab_dc_exit();
    xplab_dc_init();
    free_img_typ(&pr.ximg);                   /* destroy old binary image  */
    free_img_typ(&pr.simg);                   /* destroy old session image */
    pr.ximg = evaluate_image_data_input(pr.xplab_db, fname, 
					&xctrl.set.image_filename, 
					&xctrl.set.rblst, 
					&xctrl.set.grey_hist, 
					&xctrl.set.th_hist, 
					xctrl.cfg, &pr.oimg, &pr.simg);
    copy_set_def(&xctrl.set, &ctrl_apply.set);    /* config data unchanged */

    emit_broadcast_images();                        /* images changed hook */

    err = 0;
  }

  return err;
}



/* Updates and writes DB to disc, controlled by a locking mechanism.

Revision: 16.08.2005 E.Sack

Input : - DB
        - key
Output: - DB copied as flagged
        - flag (0: DB updated, 1: DB not updated)
Usage : - The function should be the only way to write the current DB to HDD
          by the different phases of XPLAB. It ensures that all program parts 
          are working on the same DB.
*/
int
xplab_update_db(db_def in_db, int key)
{
  int err, idx;

  idx = request_key_lst(pr.db_keys);

  if ((idx >= 0) &&                                             /* is locked */
      (key != pr.db_keys.kar[idx].key)) {            /* lockable only by one */
    string_to_log(main_log, "DB locked", "", "");

    err = check_name(pr.db_keys.kar[idx].cmt);

    if (err) {
      string_to_log(main_log, "\n", "", "");
    }
    else {
      string_to_log(main_log, " by '", pr.db_keys.kar[idx].cmt, "'\n");
    }

    return 1;                               /* DB is not locked by input key */
  }

  err = check_name(in_db.db_fname);
  assert(!err);

  copy_db_def(&in_db, &pr.xplab_db, FALSE);                       /* DB copy */
  write_db_file(in_db);                                   /* dump DB to disc */

  force_id(&xctrl.cfg.tps_id, pr.xplab_db.tpl);               /* TPS ID hook */

  FREEMEM(ctrl_apply.cfg.tps_id);                               /* fit hooks */
  add_string(xctrl.cfg.tps_id, &ctrl_apply.cfg.tps_id);

  emit_broadcast(pr.db_keys);                             /* DB changed hook */

  return 0;
}



/* Updates and writes session image to disc, controlled by a locking
   mechanism.

Revision: 01.11.2004 E.Sack

Input : - session image
        - key
        - pointer to session filename
Output: - image copied as flagged
        - flag (0: image updated, 1: image not updated)
*/
int
xplab_update_session(img_typ simg, int key, char *sfilename)
{
  int err, idx;

  assert(sfilename);

  idx = request_key_lst(pr.image_keys);

  if ((idx >= 0) &&                                           /* is locked */
      (key != pr.image_keys.kar[idx].key)) {       /* lockable only by one */
    string_to_log(main_log, "session image locked", "", "");

    err = check_name(pr.image_keys.kar[idx].cmt);

    if (err) {
      string_to_log(main_log, "\n", "", "");
    }
    else {
      string_to_log(main_log, " by '", pr.image_keys.kar[idx].cmt, "'\n");
    }

    return 1;                          /* image is not locked by input key */
  }

  assert(AREA_SIZE(simg.pimg.ar));

  copy_img_typ(&simg, &pr.simg);                             /* image copy */
  dump_session_image(simg.pimg, sfilename);  /* dump session image to disc */

  return 0;                                               /* image updated */
}



/* Dumps a session image to disc.

Revision: 23.07.2004 E.Sack

Input : - DB file name
Output: - DB copied as flagged
        - flag (0: DB copied, 1: DB not copied)
*/
static void
dump_session_image(pic_def simg, char *filename)
{
  int        found;
  char       *session_filename = NULL;
  sector_def msec;

  AREA_TO_SECTOR(simg.ar, msec);

  found = set_pixel_masked(msec, XPLAB_TRAINED_PIXEL, XPLAB_TRAINED_PIXEL, 
			   simg, FALSE);

  if (!found) {                        /* only if there are trained pixels */
    return;
  }

  session_filename = default_session_filename(filename);

  back_up_file(session_filename);                         /* make a backup */

  dump_pic_def(simg, session_filename, simg.ar, 0, 0, 1, 
	       xctrl.cfg.rawbit?PXPLAB5:PXPLAB2);

  FREEMEM(session_filename);
}



/* Inputs a new database controlled by a locking mechanism.

Revision: 16.08.2005 E.Sack

Input : - pointer to DB file name string
Output: - DB copied as flagged
        - flag (0: DB copied, 1: DB not copied)
Postcondition: - if pointer is NULL,the database is emptied.
*/
int
xplab_input_db(gchar *dname)
{
  int d, idx, err = 1;

  idx = request_key_lst(pr.db_keys);

  if (idx >= 0) {                   /* locking activated by some application */
    string_to_log(main_log, "DB locked", "", "");

    d = check_name(pr.db_keys.kar[idx].cmt);

    if (d) {
      string_to_log(main_log, "\n", "", "");
    }
    else {
      string_to_log(main_log, " by '", pr.db_keys.kar[idx].cmt, "'\n");
    }
  }  
  else {                                                  /* reset all hooks */
    err = check_name(dname);

    if (!err) {                /* overwrite only if DB filename string is ok */
      evaluate_db_input(dname, &xctrl.cfg.tps_id, xctrl.cfg.mad.mborder,
			xctrl.cfg.mad.maskweight, xctrl.cfg.verbose>BE_QUIET);
      FREEMEM(ctrl_apply.cfg.tps_id);                           /* fit hooks */
      add_string(xctrl.cfg.tps_id, &ctrl_apply.cfg.tps_id);

      emit_broadcast(pr.db_keys);                         /* DB changed hook */
    }
    else if (dname == NULL) {
      xplab_unload_db();
      err = 0;
    }
  }

  return err;
}



/* Unloads the database.

Revision: 18.08.2005 E.Sack

Input : - -
Output: - local DB exchanged with an empty default DB
*/
static void
xplab_unload_db(void)
{
  free_db_def(&pr.xplab_db);

  FREEMEM(xctrl.cfg.tps_id);                                /* force default */

  pr.xplab_db = setup_db(NULL, &xctrl.cfg.tps_id, xctrl.cfg.mad.mborder, 
			 xctrl.cfg.mad.maskweight, 
			 xctrl.cfg.verbose>BE_QUIET);

  FREEMEM(ctrl_apply.cfg.tps_id);                               /* fit hooks */
  add_string(xctrl.cfg.tps_id, &ctrl_apply.cfg.tps_id);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_histo_thold_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_entry_limits_integer(pr.page.preprocessing.histo_thold_entry, 
			    THOLD_MIN, THOLD_MAX, xctrl.cfg.histo_thold);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_r_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_rgb_norm_r_setting(1);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_g_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_rgb_norm_g_setting(1);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_b_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_rgb_norm_b_setting(1);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_total_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_rgb_norm_total_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_subst_rb_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  char *tmp;

  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  tmp = check_subst_rb_setting();
  FREEMEM(tmp);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_subst_nb_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  char *tmp;

  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  tmp = check_subst_nb_setting();
  FREEMEM(tmp);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_corrlevel_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_corrlevel_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_corrbroken_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_corrbroken_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_line_method_entry(GTK_CONVENTION_COMBO_CHANGED_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(combo_box);

  set_sensitivity_renew_method();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_ckernel_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  gpointer prow, pcol;

  STSFY_GCC(user_data);

  assert(entry);

  prow = g_object_get_data(G_OBJECT(entry), XPLAB_ASSN_KEY_ROW);
  pcol = g_object_get_data(G_OBJECT(entry), XPLAB_ASSN_KEY_COL);

  check_ckernel_setting(entry, GPOINTER_TO_INT(prow), GPOINTER_TO_INT(pcol));
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_toggle_tpsid(GTK_CONVENTION_CHECKMENUITEM_RTS_VOID)
{
  char *adata;

  assert(checkmenuitem);
  assert(user_data);

  adata = (char *)user_data;

  /* called twice, checked once                                            */
  if (GTK_CHECK_MENU_ITEM(checkmenuitem)->active == TRUE) {
    assert(adata);

    FREEMEM(ctrl_apply.cfg.tps_id);                    /* apply new TPS ID */
    add_string(adata, &ctrl_apply.cfg.tps_id);

    FREEMEM(xctrl.cfg.tps_id);                         /* apply new TPS ID */
    add_string(adata, &xctrl.cfg.tps_id);
  }
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_level_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_cluster_level_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_step_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_cluster_step_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_thold_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_cluster_thold_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_mborder_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_mborder_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_rgb_norm_r_setting(int adapt)
{
  gint    d, err;
  gdouble rgb_norm_r;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.preprocessing.rgb_norm_r_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    rgb_norm_r = xctrl.cfg.rgb_norm.r;                  /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &rgb_norm_r);

    if (!d || d == EOF) {                             /* test string content */
      rgb_norm_r = xctrl.cfg.rgb_norm.r;                /* fall back on fail */
    }
    else {
      rgb_norm_r = XPMIN(rgb_norm_r, RGB_NORM_MAX);
      rgb_norm_r = XPMAX(rgb_norm_r, RGB_NORM_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", rgb_norm_r, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_r_entry, buf);

  FREEMEM(buf);

  if (adapt) {
    adapt_rgb_norm_total_setting();
  }

  return rgb_norm_r;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_rgb_norm_g_setting(int adapt)
{
  gint    d, err;
  gdouble rgb_norm_g;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.preprocessing.rgb_norm_g_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    rgb_norm_g = xctrl.cfg.rgb_norm.g;                  /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &rgb_norm_g);

    if (!d || d == EOF) {                             /* test string content */
      rgb_norm_g = xctrl.cfg.rgb_norm.r;                /* fall back on fail */
    }
    else {
      rgb_norm_g = XPMIN(rgb_norm_g, RGB_NORM_MAX);
      rgb_norm_g = XPMAX(rgb_norm_g, RGB_NORM_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", rgb_norm_g, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_g_entry, buf);

  FREEMEM(buf);

  if (adapt) {
    adapt_rgb_norm_total_setting();
  }

  return rgb_norm_g;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_rgb_norm_b_setting(int adapt)
{
  gint    d, err;
  gdouble rgb_norm_b;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.preprocessing.rgb_norm_b_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    rgb_norm_b = xctrl.cfg.rgb_norm.b;                  /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &rgb_norm_b);

    if (!d || d == EOF) {                             /* test string content */
      rgb_norm_b = xctrl.cfg.rgb_norm.r;                /* fall back on fail */
    }
    else {
      rgb_norm_b = XPMIN(rgb_norm_b, RGB_NORM_MAX);
      rgb_norm_b = XPMAX(rgb_norm_b, RGB_NORM_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", rgb_norm_b, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_b_entry, buf);

  FREEMEM(buf);

  if (adapt) {
    adapt_rgb_norm_total_setting();
  }

  return rgb_norm_b;
}



/* Sets the total rgb sum.

Revision: 11.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static void
adapt_rgb_norm_total_setting(void)
{
  gchar  *buf = NULL;
  double red, green, blue;

  red   = check_rgb_norm_r_setting(0);
  green = check_rgb_norm_g_setting(0);
  blue  = check_rgb_norm_b_setting(0);

  double_to_string("", red + green + blue, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_total_entry, buf);
  FREEMEM(buf);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static void
check_rgb_norm_total_setting(void)
{
  gint     d, err;
  gdouble  rgb_norm_total, rgb_sum, fac;
  gchar    *buf = NULL;
  rgb_coef rgb;

  buf = xplab_entry_get_text(pr.page.preprocessing.rgb_norm_total_entry);

  rgb.r = check_rgb_norm_r_setting(0);       /* correct/fetch current values */
  rgb.g = check_rgb_norm_g_setting(0);
  rgb.b = check_rgb_norm_b_setting(0);

  rgb_sum = rgb.r + rgb.g + rgb.b;
  assert(rgb_sum > 0.0);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    rgb_norm_total = rgb_sum;                           /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &rgb_norm_total);

    if (!d || d == EOF) {                             /* test string content */
      rgb_norm_total = rgb_sum;                         /* fall back on fail */
    }
    else {
      fac = rgb_norm_total/rgb_sum;                         /* weight factor */

      rgb.r *= fac;                         /* weight new normalizing values */
      rgb.g *= fac;
      rgb.b *= fac;

      update_rgb_norm(pr.page.preprocessing, rgb);

      rgb.r = check_rgb_norm_r_setting(0);       /* correct/fetch new values */
      rgb.g = check_rgb_norm_g_setting(0);
      rgb.b = check_rgb_norm_b_setting(0);

      rgb_norm_total = rgb.r + rgb.g + rgb.b;         /* adapted total value */
    }
  }

  FREEMEM(buf);

  double_to_string("", rgb_norm_total, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_total_entry, buf);

  FREEMEM(buf);
}



/* Checks setting entry.

Revision: 22.08.2004 E.Sack

Input : - -
Output: - entry updated
*/
static char *
check_subst_rb_setting(void)
{
  gint  err, d, len;
  gchar *subst_rb = NULL, *buf = NULL;

  buf = xplab_entry_get_text(pr.page.matching.subst_rb_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    FREEMEM(subst_rb);
    add_string(xctrl.cfg.subst_rb, &subst_rb);          /* fall back on fail */
  }
  else {
    len = strlen(buf);

    CALLOC(char, len+1, subst_rb);

    d = sscanf(buf, "%s", subst_rb);

    if (!d || d == EOF) {                             /* test string content */
      FREEMEM(subst_rb);
      add_string(xctrl.cfg.subst_rb, &subst_rb);        /* fall back on fail */
    }
  }

  FREEMEM(buf);

  xplab_entry_set_text(pr.page.matching.subst_rb_entry, subst_rb);

  return subst_rb;
}



/* Checks setting entry.

Revision: 22.08.2004 E.Sack

Input : - -
Output: - entry updated
*/
static char *
check_subst_nb_setting(void)
{
  gint  err, d, len;
  gchar *subst_nb = NULL, *buf = NULL;

  buf = xplab_entry_get_text(pr.page.matching.subst_nb_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    FREEMEM(subst_nb);
    add_string(xctrl.cfg.subst_nb, &subst_nb);          /* fall back on fail */
  }
  else {
    len = strlen(buf);

    CALLOC(char, len+1, subst_nb);

    d = sscanf(buf, "%s", subst_nb);

    if (!d || d == EOF) {                             /* test string content */
      FREEMEM(subst_nb);
      add_string(xctrl.cfg.subst_nb, &subst_nb);        /* fall back on fail */
    }
  }

  FREEMEM(buf);

  xplab_entry_set_text(pr.page.matching.subst_nb_entry, subst_nb);

  return subst_nb;
}



/* Checks setting entry.

Revision: 05.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_mborder_setting(void)
{
  gint  err, d, mborder;
  gchar *buf = NULL;

  buf = xplab_entry_get_text(pr.page.matching.mborder_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    mborder = xctrl.cfg.mad.mborder;                    /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &mborder);

    if (!d || d == EOF) {                             /* test string content */
      mborder = xctrl.cfg.mad.mborder;                  /* fall back on fail */
    }
    else {
      mborder = XPMIN(mborder, MBORDER_MAX);
      mborder = XPMAX(mborder, MBORDER_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", mborder, "", &buf);
  xplab_entry_set_text(pr.page.matching.mborder_entry, buf);

  FREEMEM(buf);

  return mborder;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_corrlevel_setting(void)
{
  gint    err, d;
  gdouble corrlevel;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.correlation.corrlevel_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    corrlevel = xctrl.cfg.mad.corrlevel;                /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &corrlevel);

    if (!d || d == EOF) {                             /* test string content */
      corrlevel = xctrl.cfg.mad.corrlevel;              /* fall back on fail */
    }
    else {
      corrlevel = XPMIN(corrlevel, LEVEL_MAX);
      corrlevel = XPMAX(corrlevel, LEVEL_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", corrlevel, "", &buf);
  xplab_entry_set_text(pr.page.correlation.corrlevel_entry, buf);

  FREEMEM(buf);

  return corrlevel;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_corrbroken_setting(void)
{
  gint    err, d;
  gdouble corrbroken;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.correlation.corrbroken_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    corrbroken = xctrl.cfg.mad.corrbroken;              /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &corrbroken);

    if (!d || d == EOF) {                             /* test string content */
      corrbroken = xctrl.cfg.mad.corrbroken;            /* fall back on fail */
    }
    else {
      corrbroken = XPMIN(corrbroken, BROKEN_LEVEL_MAX);
      corrbroken = XPMAX(corrbroken, BROKEN_LEVEL_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", corrbroken, "", &buf);
  xplab_entry_set_text(pr.page.correlation.corrbroken_entry, buf);

  FREEMEM(buf);

  return corrbroken;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_cluster_level_setting(void)
{
  gint    err, d;
  gdouble cluster_level;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.correlation.cluster_level_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    cluster_level = xctrl.cfg.cluster_level;            /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &cluster_level);

    if (!d || d == EOF) {                             /* test string content */
      cluster_level = xctrl.cfg.cluster_level;          /* fall back on fail */
    }
    else {
      cluster_level = XPMIN(cluster_level, CLUSTER_LVL_MAX);
      cluster_level = XPMAX(cluster_level, CLUSTER_LVL_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", cluster_level, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_level_entry, buf);

  FREEMEM(buf);

  return cluster_level;
}



/* Checks setting entry.

Revision: 12.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_ckernel_setting(GtkEntry *entry, int row, int col)
{
  gint  err, d, cval;
  gchar *buf = NULL;

  buf = xplab_entry_get_text(entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    cval = xctrl.cfg.ckernel[col + row*CKERNEL_W];      /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &cval);

    if (!d || d == EOF) {                             /* test string content */
      cval = xctrl.cfg.ckernel[col + row*CKERNEL_W];    /* fall back on fail */
    }
    else {
      cval = XPMIN(cval, CKERNEL_MAX);
      cval = XPMAX(cval, CKERNEL_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", cval, "", &buf);
  xplab_entry_set_text(entry, buf);

  FREEMEM(buf);

  return cval;
}



/* Reads kernel settings.

Revision: 20.08.2005 E.Sack

Input : - kernel
Output: - kernel updated
*/
static void
apply_all_ckernel(kern_def ikern)
{
  int rr, row, col;

  for (row = 0; row < CKERNEL_H; row++) {     /* query raster input of table */
    rr = row*CKERNEL_W;

    for (col = 0; col < CKERNEL_W; col++) {
      ikern[col + rr] = 
	check_ckernel_setting(pr.page.correlation.ckern_entry[row][col], 
			      row, col);
    }
  }
}



/* Returns rawbit setting.

Revision: 17.10.2003 E.Sack

Input : - -
Output: - rawbit setting
*/
static int
apply_rawbit(void)
{
  int rawbit;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.rawbit_button)) == TRUE) {
    rawbit = RAWBIT_DEFAULT-1;
  }
  else {
    rawbit = RAWBIT_DEFAULT;
  }

  return rawbit;
}



/* Returns timestamp setting.

Revision: 17.10.2003 E.Sack

Input : - -
Output: - timestamp setting
*/
static int
apply_timestamp(void)
{
  int timestamp;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.timestamp_button)) == TRUE) {
    timestamp = TIMESTAMP_DEFAULT;
  }
  else {
    timestamp = TIMESTAMP_DEFAULT-1;
  }

  return timestamp;
}



/* Returns write-auto setting.

Revision: 08.10.2004 E.Sack

Input : - -
Output: - write-auto setting
*/
static int
apply_writeauto(void)
{
  int writeauto;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.writeauto_button)) == FALSE) {
    writeauto = WRITEAUTO_DEFAULT;
  }
  else {
    writeauto = WRITEAUTO_DEFAULT+1;
  }

  return writeauto;
}



/* Returns domain setting.

Revision: 17.10.2003 E.Sack

Input : - -
Output: - domain setting
*/
static int
apply_domain(void)
{
  int domain;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.correlation.domain_button)) == TRUE) {
    domain = DOMAIN_DEFAULT+1;
  }
  else {
    domain = DOMAIN_DEFAULT;
  }

  return domain;
}



/* Returns batch job setting.

Revision: 03.04.2004 E.Sack

Input : - -
Output: - batch job setting
*/
static batch_def
apply_batch(void)
{
  batch_def batch;

  init_batch_def(&batch);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.batchjob_matching_button)) == TRUE) {
    batch.phase = BATCH_PHASE_MATCHING;
  }
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.batchjob_rbchain_button)) == TRUE) {
    batch.phase = BATCH_PHASE_RB_CHAIN;
  }
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.batchjob_cluster_button)) == TRUE) {
    batch.phase = BATCH_PHASE_CLUSTER;
  }
  else {
    batch.phase = BATCH_PHASE_NO;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.batchlog_no_button)) == TRUE) {
    batch.blog = BATCH_LOG_NO;
  }
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.batchlog_ovwrt_button)) == TRUE) {
    batch.blog = BATCH_LOG_OVWRT;
  }
  else {
    batch.blog = BATCH_LOG_APPND;
  }

  return batch;
}



/* Returns documentation channel setting.

Revision: 20.10.2003 E.Sack

Input : - D1 buttons
Output: - D1 setting
*/
static void
apply_dc_1d(int2_lst *d1_lst)
{
  copy_int2_lst(&pr.page.dc.dc1l, d1_lst);             /* copy hidden list */
}



/* Enrolls documentation channel setting in hidden list.

Revision: 10.04.2004 E.Sack

Input : - -
Output: - hidden list updated
*/
static void
enroll_dc_1d_to_list(void)
{
  int dc_1d = 0;

  if (!pr.page.dc.dc1l.cnt) {
    return;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.dc.d1.pos_ccf_button)) == TRUE) {
    dc_1d |= DC_BIT_POS_CCF;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.dc.d1.pos_pad_button)) == TRUE) {
    dc_1d |= DC_BIT_POS_PAD;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.dc.d1.tfin_button)) == TRUE) {
    dc_1d |= DC_BIT_TFIN;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.dc.d1.tflag_button)) == TRUE) {
    dc_1d |= DC_BIT_TFLAG;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.dc.d1.pos_button)) == TRUE) {
    dc_1d |= DC_BIT_POS;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.dc.d1.pos_hphase_button)) == TRUE) {
    dc_1d |= DC_BIT_POS_HPHASE;
  }

  assert(pr.page.dc.d1_idx >= 0 && pr.page.dc.d1_idx < pr.page.dc.dc1l.cnt);
  pr.page.dc.dc1l.n[pr.page.dc.d1_idx].nr = dc_1d;
}



/* Returns documentation channel setting.

Revision: 22.10.2003 E.Sack

Input : - D2 buttons
Output: - D2 setting
*/
static void
apply_dc_2d(int2_lst *d2_lst)
{
  copy_int2_lst(&pr.page.dc.dc2l, d2_lst);             /* copy hidden list */
}



#define ENROLLD(button,bit)                                                \
  do {                                                                     \
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button)) == TRUE) { \
      dc_2d |= bit;                                                        \
    }                                                                      \
    mask |= bit;                                                           \
  } while (0)

/* Enrolls documentation channel setting in hidden list.

Revision: 10.04.2004 E.Sack

Input : - -
Output: - hidden list updated
*/
static void
enroll_dc_2di_to_list(void)
{
  int dc_2d = 0, mask = 0;

  if (!pr.page.dc.dc2l.cnt) {
    return;
  }

  ENROLLD(pr.page.dc.d2i.bmp_button, DC_BIT_BMP);
  ENROLLD(pr.page.dc.d2i.org_button, DC_BIT_ORG);
  ENROLLD(pr.page.dc.d2i.gv_button, DC_BIT_GV);
  ENROLLD(pr.page.dc.d2i.dot_button, DC_BIT_DOT);
  ENROLLD(pr.page.dc.d2i.clean_button, DC_BIT_CLEAN);
  ENROLLD(pr.page.dc.d2i.pest_button, DC_BIT_PEST);
  ENROLLD(pr.page.dc.d2i.pest_max_button, DC_BIT_PEST_MAX);

  assert(pr.page.dc.d2_idx >= 0 && pr.page.dc.d2_idx < pr.page.dc.dc2l.cnt);

  pr.page.dc.dc2l.n[pr.page.dc.d2_idx].nr &= ~mask;          /* reset bits */
  pr.page.dc.dc2l.n[pr.page.dc.d2_idx].nr |= dc_2d;            /* set bits */
}



/* Enrolls documentation channel setting in hidden list.

Revision: 10.04.2004 E.Sack

Input : - -
Output: - hidden list updated
*/
static void
enroll_dc_2dii_to_list(void)
{
  int dc_2d = 0, mask = 0;

  if (!pr.page.dc.dc2l.cnt) {
    return;
  }

  ENROLLD(pr.page.dc.d2ii.nb_good_button, DC_BIT_NB_GOOD);
  ENROLLD(pr.page.dc.d2ii.rb_all_button, DC_BIT_RB_ALL);
  ENROLLD(pr.page.dc.d2ii.rb_all_inv_button, DC_BIT_RB_ALL_INV);
  ENROLLD(pr.page.dc.d2ii.rb_good_button, DC_BIT_RB_GOOD);
  ENROLLD(pr.page.dc.d2ii.rb_good_inv_button, DC_BIT_RB_GOOD_INV);
  ENROLLD(pr.page.dc.d2ii.rb_fail_button, DC_BIT_RB_FAIL);
  ENROLLD(pr.page.dc.d2ii.rb_fail_inv_button, DC_BIT_RB_FAIL_INV);
  ENROLLD(pr.page.dc.d2ii.rb_frame_bmp_button, DC_BIT_RB_FRAME_BMP);
  ENROLLD(pr.page.dc.d2ii.rb_frame_rgb_button, DC_BIT_RB_FRAME_RGB);
  ENROLLD(pr.page.dc.d2ii.mrb_frame_bmp_button, DC_BIT_MRB_FRAME_BMP);
  ENROLLD(pr.page.dc.d2ii.mrb_frame_rgb_button, DC_BIT_MRB_FRAME_RGB);
  ENROLLD(pr.page.dc.d2ii.hline_bmp_button, DC_BIT_HLINE_BMP);
  ENROLLD(pr.page.dc.d2ii.hline_rgb_button, DC_BIT_HLINE_RGB);
  ENROLLD(pr.page.dc.d2ii.sline_bmp_button, DC_BIT_SLINE_BMP);
  ENROLLD(pr.page.dc.d2ii.sline_rgb_button, DC_BIT_SLINE_RGB);

  assert(pr.page.dc.d2_idx >= 0 && pr.page.dc.d2_idx < pr.page.dc.dc2l.cnt);

  pr.page.dc.dc2l.n[pr.page.dc.d2_idx].nr &= ~mask;          /* reset bits */
  pr.page.dc.dc2l.n[pr.page.dc.d2_idx].nr |= dc_2d;            /* set bits */
}

#undef ENROLLD


/* Returns report setting.

Revision: 17.10.2003 E.Sack

Input : - -
Output: - report setting
*/
static int
apply_report_write(void)
{
  int report = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.report_terse_button)) == TRUE) {
    report |= REPORT_WRITE_TERSE;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.report_rbs_button)) == TRUE) {
    report |= REPORT_WRITE_RBS;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.report_suppress)) == TRUE) {
    report |= REPORT_WRITE_NOT;
  }

  return report;
}



/* Returns RB chain setting.

Revision: 19.10.2003 E.Sack

Input : - -
Output: - report setting
*/
static int
apply_rb_chain(void)
{
  int rb_chain = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.rb_chain_use_button)) == TRUE) {
    rb_chain |= RB_CHAIN_USE;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.rb_chain_color_button)) == TRUE) {
    rb_chain |= RB_CHAIN_DUMP_COLOR;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.rb_chain_tiles_button)) == TRUE) {
    rb_chain |= RB_CHAIN_DUMP_TILES;
  }

  return rb_chain;
}



/* Returns cluster DB update setting.

Revision: 12.08.2005 E.Sack

Input : - -
Output: - cluster DB update setting
*/
static CLUSTER_NB_BIT
apply_cluster_nb(void)
{
  CLUSTER_NB_BIT cluster_nb = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_new_button)) == TRUE) {
    cluster_nb |= CLUSTER_NB_NEW;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_old_button)) == TRUE) {
    cluster_nb |= CLUSTER_NB_OLD;
  }

  if (!cluster_nb) {          /* fall back: at least one button shall be set */
    cluster_nb = CLUSTER_NB_NEW;

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_new_button), 
				 TRUE);
  }

  return cluster_nb;
}



/* Returns histogram setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - histogram setting
*/
static int
apply_histo_write(void)
{
  int histo = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.histo_write_button)) == TRUE) {
    histo = HISTO_WRITE_DEFAULT+1;
  }
  else {
    histo = HISTO_WRITE_DEFAULT;
  }

  return histo;
}



/* Returns duration setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - duration setting
*/
static int
apply_duration(void)
{
  int duration = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.matching.duration_button)) == TRUE) {
    duration = DURATION_DEFAULT+1;
  }
  else {
    duration = DURATION_DEFAULT;
  }

  return duration;
}



/* Returns support dots setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - support dots setting
*/
static int
apply_support_dots(void)
{
  int support_dots = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.preprocessing.support_dots_button)) == TRUE) {
    support_dots = SUPPORT_DOTS_DEFAULT+1;
  }
  else {
    support_dots = SUPPORT_DOTS_DEFAULT;
  }

  return support_dots;
}



/* Returns support dots setting.

Revision: 19.10.2004 E.Sack

Input : - -
Output: - support dots setting auto
*/
static int
apply_support_dots_auto(void)
{
  int support_dots_auto = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.preprocessing.support_dots_auto)) == TRUE) {
    support_dots_auto = SUPPORT_DOTS_AUTO_DEFAULT+1;
  }
  else {
    support_dots_auto = SUPPORT_DOTS_AUTO_DEFAULT;
  }

  return support_dots_auto;
}



/* Returns tooltips setting.

Revision: 06.05.2004 E.Sack

Input : - -
Output: - tooltips setting
*/
static int
apply_tooltips(void)
{
  int tips = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.tips_button)) == TRUE) {
    tips = TIPS_DEFAULT;
  }
  else {
    tips = TIPS_DEFAULT-1;
  }

  return tips;
}



/* Returns raise windows setting.

Revision: 23.11.2003 E.Sack

Input : - -
Output: - raise windows setting
*/
static int
apply_raise(void)
{
  int raise = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.general.raise_button)) == TRUE) {
    raise = RAISE_DEFAULT+1;
  }
  else {
    raise = RAISE_DEFAULT;
  }

  return raise;
}



/* Returns overwrite image setting.

Revision: 25.10.2003 E.Sack

Input : - -
Output: - overwrite_image setting
*/
static int
apply_overwrite_image(void)
{
  int overwrite_image = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.overwrite_image_button)) == TRUE) {
    overwrite_image = OVERWRITE_IMAGE_DEFAULT+1;
  }
  else {
    overwrite_image = OVERWRITE_IMAGE_DEFAULT;
  }

  return overwrite_image;
}



/* Returns replay write setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - replay write setting
*/
static int
apply_replay_write(void)
{
  int replay_write = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.replay_write_button)) == TRUE) {
    replay_write = REPLAY_WRITE_DEFAULT+1;
  }
  else {
    replay_write = REPLAY_WRITE_DEFAULT;
  }

  return replay_write;
}



/* Returns training write setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - training write setting
*/
static int
apply_training_write(void)
{
  int training_write = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.io.training_write_button)) == TRUE) {
    training_write = TRAINING_WRITE_DEFAULT+1;
  }
  else {
    training_write = TRAINING_WRITE_DEFAULT;
  }

  return training_write;
}



/* Returns single line setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - single line setting
*/
static int
apply_single_line(void)
{
  int single = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.matching.single_line_button)) == TRUE) {
    single = SINGLE_LINE_DEFAULT+1;
  }
  else {
    single = SINGLE_LINE_DEFAULT;
  }

  return single;
}



/* Returns single NB setting.

Revision: 16.10.2004 E.Sack

Input : - -
Output: - single NB setting
*/
static int
apply_single_nb(void)
{
  int single = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.matching.single_nb_button)) == TRUE) {
    single = SINGLE_NB_DEFAULT+1;
  }
  else {
    single = SINGLE_NB_DEFAULT;
  }

  return single;
}



/* Returns broken NB setting.

Revision: 03.01.2005 E.Sack

Input : - -
Output: - method
*/
static MATCHING_METHOD
apply_broken_nb(void)
{
  MATCHING_METHOD mtd;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.matching.broken_nb_button)) == TRUE) {
    mtd = MATCHING_METHOD_2;
  }
  else {
    mtd = MATCHING_METHOD_1;
  }

  return mtd;
}



/* Returns renew method setting.

Revision: 27.12.2003 E.Sack

Input : - -
Output: - renew method setting
*/
static int
apply_renew_method(void)
{
  int renew = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.matching.renew_method_button)) == TRUE) {
    renew = RENEW_METHOD_DEFAULT+1;
  }
  else {
    renew = RENEW_METHOD_DEFAULT;
  }

  return renew;
}



/* Returns cluster force setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - cluster force setting
*/
static int
apply_cluster_force(void)
{
  int force = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_force_button)) == TRUE) {
    force = CLUSTER_FORCE_DEFAULT+1;
  }
  else {
    force = CLUSTER_FORCE_DEFAULT;
  }

  return force;
}



/* Returns cluster write setting.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - cluster write setting
*/
static int
apply_cluster_write(void)
{
  int write = 0;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_write_button)) == TRUE) {
    write = CLUSTER_WRITE_DEFAULT+1;
  }
  else {
    write = CLUSTER_WRITE_DEFAULT;
  }

  return write;
}



/* Sets kernel settings.

Revision: 12.10.2003 E.Sack

Input : - pointer to kernel
Output: - kernel updated
*/
static void
set_all_ckernel(kern_def ikern)
{
  gint   rr, row, col;
  gchar *buf = NULL;

  for (row = 0; row < CKERNEL_H; row++) {   /* default raster input of table */
    rr = row*CKERNEL_W;

    for (col = 0; col < CKERNEL_W; col++) {
      FREEMEM(buf);
      int_to_string("", ikern[col + rr], "", &buf);
      xplab_entry_set_text(pr.page.correlation.ckern_entry[row][col], buf);
    }
  }

  FREEMEM(buf);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_cluster_step_setting(void)
{
  gint    err, d;
  gdouble cluster_step;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.correlation.cluster_step_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    cluster_step = xctrl.cfg.cluster_step;              /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &cluster_step);

    if (!d || d == EOF) {                             /* test string content */
      cluster_step = xctrl.cfg.cluster_step;            /* fall back on fail */
    }
    else {
      cluster_step = XPMIN(cluster_step, CLUSTER_STEP_MAX);
      cluster_step = XPMAX(cluster_step, CLUSTER_STEP_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", cluster_step, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_step_entry, buf);

  FREEMEM(buf);

  return cluster_step;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_verbose_setting(void)
{
  int verbose;

  assert(GTK_IS_COMBO_BOX(pr.page.general.verbose_combo) == TRUE);
  verbose = 1 + gtk_combo_box_get_active(GTK_COMBO_BOX(pr.page.general.verbose_combo));

  verbose = XPMIN(verbose, VERBOSE_LAST+1);              /* check user input */
  verbose = XPMAX(verbose, VERBOSE_FIRST-1);

  return verbose;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static INLINE_METHOD
check_line_method_setting(void)
{
  int d;

  assert(GTK_IS_COMBO_BOX(pr.page.matching.line_method_combo) == TRUE);
  d = gtk_combo_box_get_active(GTK_COMBO_BOX(pr.page.matching.line_method_combo));

  d = XPMIN(d, METHOD_LAST-1);                           /* check user input */
  d = XPMAX(d, METHOD_FIRST);

  return (INLINE_METHOD)d;
}



/* Sets sensitivity for method renewing.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static void
set_sensitivity_renew_method(void)
{
  int d;

  if (pr.page.matching.line_method_combo == NULL) {
    if (pr.page.matching.renew_method_button != NULL) {
      assert(GTK_IS_BUTTON(pr.page.matching.renew_method_button) == TRUE);
      gtk_widget_set_sensitive(pr.page.matching.renew_method_button, FALSE);
    }

    return;
  }

  assert(GTK_IS_COMBO_BOX(pr.page.matching.line_method_combo) == TRUE);
  d = gtk_combo_box_get_active(GTK_COMBO_BOX(pr.page.matching.line_method_combo));

  if (d) {
    gtk_widget_set_sensitive(pr.page.matching.renew_method_button, TRUE);
  }
  else {
    gtk_widget_set_sensitive(pr.page.matching.renew_method_button, FALSE);
  }
}



/* Checks setting entry.

Revision: 08.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_cluster_thold_setting(void)
{
  gint  err, d, cluster_thold;
  gchar *buf = NULL;

  buf = xplab_entry_get_text(pr.page.correlation.cluster_thold_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    cluster_thold = xctrl.cfg.cluster_thold;            /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &cluster_thold);

    if (!d || d == EOF) {                             /* test string content */
      cluster_thold = xctrl.cfg.cluster_thold;          /* fall back on fail */
    }
    else {
      cluster_thold = XPMIN(cluster_thold, CHOLD_MAX);
      cluster_thold = XPMAX(cluster_thold, CHOLD_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", cluster_thold, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_thold_entry, buf);

  FREEMEM(buf);

  return cluster_thold;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nborder_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_nborder_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nb_border_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_nb_border_setting(pr.page.io.nb_border_x_entry, xctrl.cfg.nb_border_x);
  check_nb_border_setting(pr.page.io.nb_border_y_entry, xctrl.cfg.nb_border_y);
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_store_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_dc_store_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maskweight_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_maskweight_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nnbmatch_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_nnbmatch_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_interleave_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_interleave_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_overlap_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_overlap_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_symbolspacing_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_symbolspacing_setting();
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nb_lap_entry(GTK_CONVENTION_GTKENTRY_ACTIVATE_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(entry);

  check_nb_lap_setting();
}



/* Checks setting entry.

Revision: 05.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_nborder_setting(void)
{
  gint  err, d, nborder;
  gchar *buf = NULL;

  buf = xplab_entry_get_text(pr.page.matching.nborder_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    nborder = xctrl.cfg.mad.nborder;                    /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &nborder);

    if (!d || d == EOF) {                             /* test string content */
      nborder = xctrl.cfg.mad.nborder;                  /* fall back on fail */
    }
    else {
      nborder = XPMIN(nborder, NBORDER_MAX);
      nborder = XPMAX(nborder, NBORDER_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", nborder, "", &buf);
  xplab_entry_set_text(pr.page.matching.nborder_entry, buf);

  FREEMEM(buf);

  return nborder;
}



/* Checks setting entry.

Revision: 18.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_dc_store_setting(void)
{
  gint  err, d, dc_store;
  gchar *buf = NULL;

  buf = xplab_entry_get_text(pr.page.dc.store_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    dc_store = xctrl.cfg.dc_store;                      /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &dc_store);

    if (!d || d == EOF) {                             /* test string content */
      dc_store = xctrl.cfg.dc_store;                    /* fall back on fail */
    }
    else {
      dc_store = XPMIN(dc_store, DC_STORE_MAX);
      dc_store = XPMAX(dc_store, DC_STORE_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", dc_store, "", &buf);
  xplab_entry_set_text(pr.page.dc.store_entry, buf);

  FREEMEM(buf);

  return dc_store;
}



/* Checks setting entry.

Revision: 24.02.2004 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_nb_border_setting(GtkEntry *nb_border_entry, int nb_border_default)
{
  gint  err, d, nb_border;
  gchar *buf;

  buf = xplab_entry_get_text(nb_border_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    nb_border = nb_border_default;                      /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &nb_border);

    if (!d || d == EOF) {                             /* test string content */
      nb_border = nb_border_default;                    /* fall back on fail */
    }
    else {
      nb_border = XPMIN(nb_border, NB_BORDER_MAX);
      nb_border = XPMAX(nb_border, NB_BORDER_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", nb_border, "", &buf);
  xplab_entry_set_text(nb_border_entry, buf);

  FREEMEM(buf);

  return nb_border;
}



/* Checks setting entry.

Revision: 10.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_maskweight_setting(void)
{
  gint  err, d, maskweight;
  gchar *buf;

  buf = xplab_entry_get_text(pr.page.matching.maskweight_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    maskweight = xctrl.cfg.mad.maskweight;              /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &maskweight);

    if (!d || d == EOF) {                             /* test string content */
      maskweight = xctrl.cfg.mad.maskweight;            /* fall back on fail */
    }
    else {
      maskweight = XPMIN(maskweight, MASKWEIGHT_MAX);
      maskweight = XPMAX(maskweight, MASKWEIGHT_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", maskweight, "", &buf);
  xplab_entry_set_text(pr.page.matching.maskweight_entry, buf);

  FREEMEM(buf);

  return maskweight;
}



/* Checks setting entry.

Revision: 05.10.2003 E.Sack

Input : - -
Output: - entry updated
*/
static int
check_nnbmatch_setting(void)
{
  gint  err, d, nnbmatch;
  gchar *buf;

  buf = xplab_entry_get_text(pr.page.matching.nnbmatch_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    nnbmatch = xctrl.cfg.mad.nnbmatch;                  /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%d", &nnbmatch);

    if (!d || d == EOF) {                             /* test string content */
      nnbmatch = xctrl.cfg.mad.nnbmatch;                /* fall back on fail */
    }
    else {
      nnbmatch = XPMIN(nnbmatch, NNBMATCH_MAX);
      nnbmatch = XPMAX(nnbmatch, NNBMATCH_MIN);
    }
  }

  FREEMEM(buf);

  int_to_string("", nnbmatch, "", &buf);
  xplab_entry_set_text(pr.page.matching.nnbmatch_entry, buf);

  FREEMEM(buf);

  return nnbmatch;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_interleave_setting(void)
{
  gint    err, d;
  gdouble interleave;
  gchar   *buf = NULL;

  buf = xplab_entry_get_text(pr.page.matching.interleave_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    interleave = xctrl.cfg.mad.interleave;              /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &interleave);

    if (!d || d == EOF) {                             /* test string content */
      interleave = xctrl.cfg.mad.interleave;            /* fall back on fail */
    }
    else {
      interleave = XPMIN(interleave, INTERLEAVE_MAX);
      interleave = XPMAX(interleave, INTERLEAVE_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", interleave, "", &buf);
  xplab_entry_set_text(pr.page.matching.interleave_entry, buf);

  FREEMEM(buf);

  return interleave;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_overlap_setting(void)
{
  int    err, d;
  double overlap;
  gchar  *buf = NULL;

  buf = xplab_entry_get_text(pr.page.matching.overlap_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    overlap = xctrl.cfg.mad.overlap;                    /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &overlap);

    if (!d || d == EOF) {                             /* test string content */
      overlap = xctrl.cfg.mad.overlap;                  /* fall back on fail */
    }
    else {
      overlap = XPMIN(overlap, OVERLAP_MAX);
      overlap = XPMAX(overlap, OVERLAP_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", overlap, "", &buf);
  xplab_entry_set_text(pr.page.matching.overlap_entry, buf);

  FREEMEM(buf);

  return overlap;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_symbolspacing_setting(void)
{
  int    err, d;
  double symbolspacing;
  gchar  *buf;

  buf = xplab_entry_get_text(pr.page.matching.symbolspacing_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    symbolspacing = xctrl.cfg.syspc;                    /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &symbolspacing);

    if (!d || d == EOF) {                             /* test string content */
      symbolspacing = xctrl.cfg.syspc;                  /* fall back on fail */
    }
    else {
      symbolspacing = XPMIN(symbolspacing, SYSPC_MAX);
      symbolspacing = XPMAX(symbolspacing, SYSPC_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", symbolspacing, "", &buf);
  xplab_entry_set_text(pr.page.matching.symbolspacing_entry, buf);

  FREEMEM(buf);

  return symbolspacing;
}



/* Checks setting entry.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - entry updated
*/
static double
check_nb_lap_setting(void)
{
  int    err, d;
  double nb_lap;
  gchar  *buf;

  buf = xplab_entry_get_text(pr.page.matching.nb_lap_entry);

  err = check_name(buf);

  if (err) {                                             /* check user input */
    nb_lap = xctrl.cfg.mad.nb_lap;                      /* fall back on fail */
  }
  else {
    d = sscanf(buf, "%lf", &nb_lap);

    if (!d || d == EOF) {                             /* test string content */
      nb_lap = xctrl.cfg.mad.nb_lap;                    /* fall back on fail */
    }
    else {
      nb_lap = XPMIN(nb_lap, NB_LAP_MAX);
      nb_lap = XPMAX(nb_lap, NB_LAP_MIN);
    }
  }

  FREEMEM(buf);

  double_to_string("", nb_lap, "", &buf);
  xplab_entry_set_text(pr.page.matching.nb_lap_entry, buf);

  FREEMEM(buf);

  return nb_lap;
}



/* Sets a setting to default.

Revision: 05.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_mborder_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf1 = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", MBORDER_DEFAULT, "", &buf1);
  xplab_entry_set_text(pr.page.matching.mborder_entry, buf1);

  FREEMEM(buf1);
}



/* Sets a setting to default.

Revision: 05.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_histo_thold_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf1 = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", THOLD_DEFAULT, "", &buf1);
  xplab_entry_set_text(pr.page.preprocessing.histo_thold_entry, buf1);

  FREEMEM(buf1);
}



/* Sets a setting to default.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_r_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", RGB_NORM_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_r_entry, buf);

  FREEMEM(buf);

  adapt_rgb_norm_total_setting();
}



/* Sets a setting to default.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_g_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", RGB_NORM_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_g_entry, buf);

  FREEMEM(buf);

  adapt_rgb_norm_total_setting();
}



/* Sets a setting to default.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_b_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", RGB_NORM_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_b_entry, buf);

  FREEMEM(buf);

  adapt_rgb_norm_total_setting();
}



/* Sets a setting to default.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_total_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", 3.0*RGB_NORM_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_total_entry, buf);

  FREEMEM(buf);

  check_rgb_norm_total_setting();            /* weight current RGB norming */
}



/* Sets a setting to default.

Revision: 22.08.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_subst_rb_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf1 = NULL;

  STSFY_GCC(user_data);

  assert(button);

  buf1 = default_subst_rb();

  xplab_entry_set_text(pr.page.matching.subst_rb_entry, buf1);

  FREEMEM(buf1);
}



/* Sets a setting to default.

Revision: 10.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_subst_nb_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf1 = NULL;

  STSFY_GCC(user_data);

  assert(button);

  buf1 = default_subst_nb();

  xplab_entry_set_text(pr.page.matching.subst_nb_entry, buf1);

  FREEMEM(buf1);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_subst_rb_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  FREEMEM(xctrl.cfg.subst_rb);
  add_string(cfg_mirror.subst_rb, &xctrl.cfg.subst_rb);

  FREEMEM(ctrl_apply.cfg.subst_rb);
  add_string(cfg_mirror.subst_rb, &ctrl_apply.cfg.subst_rb);

  xplab_entry_set_text(pr.page.matching.subst_rb_entry, xctrl.cfg.subst_rb);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_subst_nb_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  FREEMEM(xctrl.cfg.subst_nb);
  add_string(cfg_mirror.subst_nb, &xctrl.cfg.subst_nb);

  FREEMEM(ctrl_apply.cfg.subst_nb);
  add_string(cfg_mirror.subst_nb, &ctrl_apply.cfg.subst_nb);

  xplab_entry_set_text(pr.page.matching.subst_nb_entry, xctrl.cfg.subst_nb);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_histo_thold_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf1 = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.histo_thold      = cfg_mirror.histo_thold;
  ctrl_apply.cfg.histo_thold = cfg_mirror.histo_thold;

  int_to_string("", xctrl.cfg.histo_thold, "", &buf1);
  xplab_entry_set_text(pr.page.preprocessing.histo_thold_entry, buf1);

  FREEMEM(buf1);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_corrlevel_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.corrlevel      = cfg_mirror.mad.corrlevel;
  ctrl_apply.cfg.mad.corrlevel = cfg_mirror.mad.corrlevel;

  double_to_string("", xctrl.cfg.mad.corrlevel, "", &buf);
  xplab_entry_set_text(pr.page.correlation.corrlevel_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 05.01.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_corrbroken_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.corrbroken      = cfg_mirror.mad.corrbroken;
  ctrl_apply.cfg.mad.corrbroken = cfg_mirror.mad.corrbroken;

  double_to_string("", xctrl.cfg.mad.corrbroken, "", &buf);
  xplab_entry_set_text(pr.page.correlation.corrbroken_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 12.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_ckernel_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  copy_kern_def(cfg_mirror.ckernel, xctrl.cfg.ckernel);
  copy_kern_def(cfg_mirror.ckernel, ctrl_apply.cfg.ckernel);

  set_all_ckernel(xctrl.cfg.ckernel);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_r_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);
  STSFY_GCC(button);

  xctrl.cfg.rgb_norm.r      = cfg_mirror.rgb_norm.r;
  ctrl_apply.cfg.rgb_norm.r = cfg_mirror.rgb_norm.r;

  double_to_string("", xctrl.cfg.rgb_norm.r, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_r_entry, buf);

  FREEMEM(buf);

  adapt_rgb_norm_total_setting();
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_g_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);
  STSFY_GCC(button);

  xctrl.cfg.rgb_norm.g      = cfg_mirror.rgb_norm.g;
  ctrl_apply.cfg.rgb_norm.g = cfg_mirror.rgb_norm.g;

  double_to_string("", xctrl.cfg.rgb_norm.g, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_g_entry, buf);

  FREEMEM(buf);

  adapt_rgb_norm_total_setting();
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_b_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);
  STSFY_GCC(button);

  xctrl.cfg.rgb_norm.b      = cfg_mirror.rgb_norm.b;
  ctrl_apply.cfg.rgb_norm.b = cfg_mirror.rgb_norm.b;

  double_to_string("", xctrl.cfg.rgb_norm.b, "", &buf);
  xplab_entry_set_text(pr.page.preprocessing.rgb_norm_b_entry, buf);

  FREEMEM(buf);

  adapt_rgb_norm_total_setting();
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_rgb_norm_total_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  cb_rgb_norm_r_revert(NULL, user_data);
  cb_rgb_norm_g_revert(NULL, user_data);
  cb_rgb_norm_b_revert(NULL, user_data);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_level_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.cluster_level      = cfg_mirror.cluster_level;
  ctrl_apply.cfg.cluster_level = cfg_mirror.cluster_level;

  double_to_string("", xctrl.cfg.cluster_level, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_level_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_step_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.cluster_step      = cfg_mirror.cluster_step;
  ctrl_apply.cfg.cluster_step = cfg_mirror.cluster_step;

  double_to_string("", xctrl.cfg.cluster_step, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_step_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_thold_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.cluster_thold      = cfg_mirror.cluster_thold;
  ctrl_apply.cfg.cluster_thold = cfg_mirror.cluster_thold;

  int_to_string("", xctrl.cfg.cluster_thold, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_thold_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_mborder_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.mborder      = cfg_mirror.mad.mborder;
  ctrl_apply.cfg.mad.mborder = cfg_mirror.mad.mborder;

  int_to_string("", xctrl.cfg.mad.mborder, "", &buf);
  xplab_entry_set_text(pr.page.matching.mborder_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nborder_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.nborder      = cfg_mirror.mad.nborder;
  ctrl_apply.cfg.mad.nborder = cfg_mirror.mad.nborder;

  int_to_string("", xctrl.cfg.mad.nborder, "", &buf);
  xplab_entry_set_text(pr.page.matching.nborder_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 18.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_store_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.dc_store      = cfg_mirror.dc_store;
  ctrl_apply.cfg.dc_store = cfg_mirror.dc_store;

  int_to_string("", xctrl.cfg.dc_store, "", &buf);
  xplab_entry_set_text(pr.page.dc.store_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 24.02.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nb_border_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.nb_border_x      = cfg_mirror.nb_border_x;
  xctrl.cfg.nb_border_y      = cfg_mirror.nb_border_y;
  ctrl_apply.cfg.nb_border_x = cfg_mirror.nb_border_x;
  ctrl_apply.cfg.nb_border_y = cfg_mirror.nb_border_y;

  int_to_string("", xctrl.cfg.nb_border_x, "", &buf);
  xplab_entry_set_text(pr.page.io.nb_border_x_entry, buf);

  FREEMEM(buf);

  int_to_string("", xctrl.cfg.nb_border_y, "", &buf);
  xplab_entry_set_text(pr.page.io.nb_border_y_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maskweight_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.maskweight      = cfg_mirror.mad.maskweight;
  ctrl_apply.cfg.mad.maskweight = cfg_mirror.mad.maskweight;

  int_to_string("", xctrl.cfg.mad.maskweight, "", &buf);
  xplab_entry_set_text(pr.page.matching.maskweight_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nnbmatch_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.nnbmatch      = cfg_mirror.mad.nnbmatch;
  ctrl_apply.cfg.mad.nnbmatch = cfg_mirror.mad.nnbmatch;

  int_to_string("", xctrl.cfg.mad.nnbmatch, "", &buf);
  xplab_entry_set_text(pr.page.matching.nnbmatch_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_interleave_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.interleave      = cfg_mirror.mad.interleave;
  ctrl_apply.cfg.mad.interleave = cfg_mirror.mad.interleave;

  double_to_string("", xctrl.cfg.mad.interleave, "", &buf);
  xplab_entry_set_text(pr.page.matching.interleave_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_overlap_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.overlap      = cfg_mirror.mad.overlap;
  ctrl_apply.cfg.mad.overlap = cfg_mirror.mad.overlap;

  double_to_string("", xctrl.cfg.mad.overlap, "", &buf);
  xplab_entry_set_text(pr.page.matching.overlap_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_symbolspacing_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.syspc      = cfg_mirror.syspc;
  ctrl_apply.cfg.syspc = cfg_mirror.syspc;

  double_to_string("", xctrl.cfg.syspc, "", &buf);
  xplab_entry_set_text(pr.page.matching.symbolspacing_entry, buf);

  FREEMEM(buf);
}



/* Reverts a setting.

Revision: 11.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nb_lap_revert(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  xctrl.cfg.mad.nb_lap      = cfg_mirror.mad.nb_lap;
  ctrl_apply.cfg.mad.nb_lap = cfg_mirror.mad.nb_lap;

  double_to_string("", xctrl.cfg.mad.nb_lap, "", &buf);
  xplab_entry_set_text(pr.page.matching.nb_lap_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 05.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_corrlevel_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", LEVEL_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.correlation.corrlevel_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 05.01.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_corrbroken_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", BROKEN_LEVEL_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.correlation.corrbroken_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 12.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_ckernel_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  kern_def kern;

  xdefault_kern_def(kern);

  STSFY_GCC(user_data);

  assert(button);

  set_all_ckernel(kern);
}



/* Sets a setting to default.

Revision: 05.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_level_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", CLUSTER_LVL_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_level_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 08.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_step_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", CLUSTER_STEP_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_step_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 08.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_thold_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", CHOLD_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.correlation.cluster_thold_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 05.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nborder_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", NBORDER_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.matching.nborder_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 18.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_store_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", DC_STORE_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.dc.store_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 24.02.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nb_border_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", NB_BORDER_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.io.nb_border_x_entry, buf);
  xplab_entry_set_text(pr.page.io.nb_border_y_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 05.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_maskweight_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", MASKWEIGHT_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.matching.maskweight_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 05.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nnbmatch_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  int_to_string("", NNBMATCH_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.matching.nnbmatch_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 10.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_interleave_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", INTERLEAVE_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.matching.interleave_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 10.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_overlap_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", OVERLAP_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.matching.overlap_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 10.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_symbolspacing_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", SYSPC_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.matching.symbolspacing_entry, buf);

  FREEMEM(buf);
}



/* Sets a setting to default.

Revision: 10.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_nb_lap_default(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  gchar *buf = NULL;

  STSFY_GCC(user_data);

  assert(button);

  double_to_string("", NB_LAP_DEFAULT, "", &buf);
  xplab_entry_set_text(pr.page.matching.nb_lap_entry, buf);

  FREEMEM(buf);
}



/* Updates all preferences buttons of the current displays.

Revision: 29.10.2003 E.Sack

Input : - -
Output: - preferences buttons updated
*/
static void
update_all(cfg_def cfg)
{
  update_general_page(pr.page.general, cfg);
  update_preprocessing_page(pr.page.preprocessing, cfg);
  update_matching_page(pr.page.matching, cfg);
  update_correlation_page(pr.page.correlation, cfg);
  update_io_page(pr.page.io, cfg);
  update_dc_page(cfg);
  update_statistic_page(pr.page.statistic);
}



/* Switches between preferences notebook pages.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_preferences_page_switch(GTK_CONVENTION_NOTEBOOK_RTS_VOID)
{
  GtkWidget *npage, *tab_label;
  const gchar *str;

  STSFY_GCC(user_data);

  assert(notebook);
  assert(page);

  npage = gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook), page_num);
  tab_label = gtk_notebook_get_tab_label(GTK_NOTEBOOK(notebook), npage);

  str = gtk_label_get_text(GTK_LABEL(tab_label));

  if (!strcmp(str, XPLAB_PREFS_PAGE_GENERAL)) {
    update_general_page(pr.page.general, xctrl.cfg);
  }
  else if (!strcmp(str, XPLAB_PREFS_PAGE_PREPROCESSING)) {
    update_preprocessing_page(pr.page.preprocessing, xctrl.cfg);
  }
  else if (!strcmp(str, XPLAB_PREFS_PAGE_MATCHING)) {
    update_matching_page(pr.page.matching, xctrl.cfg);
  }
  else if (!strcmp(str, XPLAB_PREFS_PAGE_CORRELATION)) {
    update_correlation_page(pr.page.correlation, xctrl.cfg);
  }
  else if (!strcmp(str, XPLAB_PREFS_PAGE_STATISTIC)) {
    update_statistic_page(pr.page.statistic);
  }
  else if (!strcmp(str, XPLAB_PREFS_PAGE_DC)) {
    update_dc_page(xctrl.cfg);
  }
  else if (!strcmp(str, XPLAB_PREFS_PAGE_IO)) {
    update_io_page(pr.page.io, xctrl.cfg);
  }
  else {
    assert(0);
  }
}



/* Updates general page.

Revision: 05.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_general_page(general_def general, cfg_def cfg)
{
  update_tooltips(general, cfg);
  update_verbose(general, cfg);
  update_raise(general, cfg);
  update_rb_chain(general, cfg);
  update_batch(general, cfg);

  adapt_rgb_norm_total_setting();
}



/* Updates RGB flags.

Revision: 22.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_rgb_norm(pre_def preprocessing, rgb_coef rgb_norm)
{
  gchar *buf = NULL;

  double_to_string("", rgb_norm.r, "", &buf);
  xplab_entry_set_text(preprocessing.rgb_norm_r_entry, buf);

  FREEMEM(buf);

  double_to_string("", rgb_norm.g, "", &buf);
  xplab_entry_set_text(preprocessing.rgb_norm_g_entry, buf);

  FREEMEM(buf);

  double_to_string("", rgb_norm.b, "", &buf);
  xplab_entry_set_text(preprocessing.rgb_norm_b_entry, buf);

  FREEMEM(buf);
}



/* Updates binarisation threshold.

Revision: 22.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_histo_thold(pre_def preprocessing, cfg_def cfg)
{
  gchar *buf = NULL;

  int_to_string("", cfg.histo_thold, "", &buf);
  xplab_entry_set_text(preprocessing.histo_thold_entry, buf);

  FREEMEM(buf);
}



/* Updates verbose flag.

Revision: 05.08.2005 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_verbose(general_def general, cfg_def cfg)
{
  assert(GTK_IS_COMBO_BOX(general.verbose_combo) == TRUE);
  gtk_combo_box_set_active(GTK_COMBO_BOX(general.verbose_combo), cfg.verbose-1);
}



/* Updates batch job flags.

Revision: 03.04.2004 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_batch(general_def general, cfg_def cfg)
{
  switch (cfg.batch.phase) {
  case BATCH_PHASE_RB_CHAIN:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.batchjob_rbchain_button),
				 TRUE);
    break;

  case BATCH_PHASE_MATCHING:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.batchjob_matching_button),
				 TRUE);
    break;

  case BATCH_PHASE_CLUSTER:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.batchjob_cluster_button),
				 TRUE);
    break;

  case BATCH_PHASE_NO:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.batchjob_void_button),
				 TRUE);
    break;

  default: assert(0); break;
  }

  switch (cfg.batch.blog) {
  case BATCH_LOG_OVWRT:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.batchlog_ovwrt_button),
				 TRUE);
    break;

  case BATCH_LOG_APPND:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.batchlog_appnd_button),
				 TRUE);
    break;

  case BATCH_LOG_NO:
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.batchlog_no_button),
				 TRUE);
    break;

  default: assert(0); break;
  }
}



/* Updates line detection method flag.

Revision: 05.08.2005 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_line_method(matching_def matching, cfg_def cfg)
{
  assert(GTK_IS_COMBO_BOX(matching.line_method_combo) == TRUE);
  gtk_combo_box_set_active(GTK_COMBO_BOX(matching.line_method_combo), 
			   cfg.line_method);
}



/* Updates RB substitution flag.

Revision: 22.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_subst_rb(matching_def matching, cfg_def cfg)
{
  xplab_entry_set_text(matching.subst_rb_entry, cfg.subst_rb);
}



/* Updates NB substitution flag.

Revision: 22.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_subst_nb(matching_def matching, cfg_def cfg)
{
  xplab_entry_set_text(matching.subst_nb_entry, cfg.subst_nb);
}



/* Updates NB border.

Revision: 24.02.2004 E.Sack

Input : - I/O page definition
        - preferences
Output: - page updated
*/
static void
update_nb_border(io_def iod, cfg_def cfg)
{
  gchar *buf = NULL;

  int_to_string("", cfg.nb_border_x, "", &buf);
  xplab_entry_set_text(iod.nb_border_x_entry, buf);

  FREEMEM(buf);

  int_to_string("", cfg.nb_border_y, "", &buf);
  xplab_entry_set_text(iod.nb_border_y_entry, buf);

  FREEMEM(buf);
}



/* Updates rawbit flag.

Revision: 28.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_rawbit(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.rawbit_button), 
			       cfg.rawbit==RAWBIT_DEFAULT?FALSE:TRUE);
}



/* Updates timestamp flag.

Revision: 31.10.2003 E.Sack

Input : - I/O page definition
        - preferences
Output: - I/O page updated
*/
static void
update_timestamp(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.timestamp_button), 
			       cfg.timestamp!=TIMESTAMP_DEFAULT?FALSE:TRUE);
}



/* Updates writeauto flag.

Revision: 08.10.2004 E.Sack

Input : - I/O page definition
        - preferences
Output: - I/O page updated
*/
static void
update_writeauto(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.writeauto_button), 
			       cfg.writeauto==WRITEAUTO_DEFAULT?FALSE:TRUE);
}



/* Updates domain flag.

Revision: 22.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_domain(correlation_def correlation, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(correlation.domain_button), 
			       cfg.cdomain==DOMAIN_DEFAULT?FALSE:TRUE);
}



/* Updates report flag.

Revision: 17.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_report_write(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.report_terse_button), 
			       cfg.report_write & REPORT_WRITE_TERSE ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.report_rbs_button), 
			       cfg.report_write & REPORT_WRITE_RBS ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.report_suppress), 
			       cfg.report_write & REPORT_WRITE_NOT ? TRUE:FALSE);
}



/* Updates RB chain flag.

Revision: 19.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_rb_chain(general_def general, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.rb_chain_use_button), 
			       cfg.rb_chain & RB_CHAIN_USE ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.rb_chain_color_button), 
			       cfg.rb_chain & RB_CHAIN_DUMP_COLOR ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.rb_chain_tiles_button), 
			       cfg.rb_chain & RB_CHAIN_DUMP_TILES ? TRUE:FALSE);
}



/* Updates cluster DB update flag.

Revision: 27.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_cluster_nb(correlation_def correlation, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(correlation.cluster_nb_new_button), 
			       cfg.cluster_nb & CLUSTER_NB_NEW ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(correlation.cluster_nb_old_button), 
			       cfg.cluster_nb & CLUSTER_NB_OLD ? TRUE:FALSE);
}



/* Updates write histogram flag.

Revision: 18.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_histo_write(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.histo_write_button), 
			       cfg.histo_write!=HISTO_WRITE_DEFAULT ? TRUE:FALSE);
}



/* Updates duration flag.

Revision: 18.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_duration(matching_def general, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.duration_button), 
			       cfg.duration!=DURATION_DEFAULT ? TRUE:FALSE);
}



/* Updates support_dots flag.

Revision: 18.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_support_dots(pre_def preprocessing, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(preprocessing.support_dots_button), 
			       cfg.support_dots!=SUPPORT_DOTS_DEFAULT ? TRUE:FALSE);
}



/* Updates support_dots_auto flag.

Revision: 19.10.2004 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_support_dots_auto(pre_def preprocessing, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(preprocessing.support_dots_auto), 
			       cfg.support_dots_auto!=SUPPORT_DOTS_AUTO_DEFAULT ? TRUE:FALSE);
}



/* Updates tips flag.

Revision: 06.05.2004 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_tooltips(general_def general, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.tips_button), 
			       cfg.tips==TIPS_DEFAULT ? TRUE:FALSE);
}



/* Updates raise windows flag.

Revision: 23.11.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_raise(general_def general, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(general.raise_button), 
			       cfg.raise!=RAISE_DEFAULT ? TRUE:FALSE);
}



/* Updates overwrite image flag.

Revision: 28.10.2003 E.Sack

Input : - io page definition
        - preferences
Output: - io page updated
*/
static void
update_overwrite_image(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.overwrite_image_button), 
			       cfg.overwrite_image!=OVERWRITE_IMAGE_DEFAULT ? TRUE:FALSE);
}



/* Updates replay write flag.

Revision: 18.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_replay_write(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.replay_write_button), 
			       cfg.replay_write!=REPLAY_WRITE_DEFAULT ? TRUE:FALSE);
}



/* Updates training write flag.

Revision: 09.07.2004 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_training_write(io_def iod, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(iod.training_write_button), 
			       cfg.training_write!=TRAINING_WRITE_DEFAULT ? TRUE:FALSE);
}



/* Updates single line flag.

Revision: 18.10.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_single_line(matching_def matching, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(matching.single_line_button), 
			       cfg.single_line!=SINGLE_LINE_DEFAULT ? TRUE:FALSE);
}



/* Updates single NB flag.

Revision: 16.10.2004 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_single_nb(matching_def matching, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(matching.single_nb_button), 
			       cfg.single_nb!=SINGLE_NB_DEFAULT ? TRUE:FALSE);
}



/* Updates broken NB flag.

Revision: 03.01.2005 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_broken_nb(matching_def matching, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(matching.broken_nb_button), 
			       (cfg.match_method == MATCHING_METHOD_2) ? 
                                                   TRUE:FALSE);
}



/* Updates renew method flag.

Revision: 27.12.2003 E.Sack

Input : - general page definition
        - preferences
Output: - general page updated
*/
static void
update_renew_method(matching_def matching, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(matching.renew_method_button), 
			       cfg.renew_method!=RENEW_METHOD_DEFAULT ? TRUE:FALSE);
}



/* Updates 1-D flags.

Revision: 10.04.2004 E.Sack

Input : - dc page definition
        - preferences
Output: - 1-D page updated 
*/
static void
update_dc_1d(d1_def d1, int opt1)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d1.pos_ccf_button), 
			       opt1 & DC_BIT_POS_CCF ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d1.pos_pad_button), 
			       opt1 & DC_BIT_POS_PAD ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d1.tfin_button), 
			       opt1 & DC_BIT_TFIN ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d1.tflag_button), 
			       opt1 & DC_BIT_TFLAG ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d1.pos_button), 
			       opt1 & DC_BIT_POS ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d1.pos_hphase_button), 
			       opt1 & DC_BIT_POS_HPHASE ? TRUE:FALSE);
}



/* Updates 2-D flags.

Revision: 22.10.2003 E.Sack

Input : - dc page definition
        - preferences
Output: - 2-D page updated
*/
static void
update_dc_2di(d2i_def d2i, int opt2i)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2i.bmp_button), 
			       opt2i & DC_BIT_BMP ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2i.org_button), 
			       opt2i & DC_BIT_ORG ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2i.gv_button), 
			       opt2i & DC_BIT_GV ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2i.dot_button), 
			       opt2i & DC_BIT_DOT ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2i.clean_button), 
			       opt2i & DC_BIT_CLEAN ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2i.pest_button), 
			       opt2i & DC_BIT_PEST ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2i.pest_max_button), 
			       opt2i & DC_BIT_PEST_MAX ? TRUE:FALSE);
}



/* Updates 2-D flags.

Revision: 22.10.2003 E.Sack

Input : - dc page definition
        - preferences
Output: - 2-D page updated
*/
static void
update_dc_2dii(d2ii_def d2ii, int opt2ii)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.nb_good_button), 
			       opt2ii & DC_BIT_NB_GOOD ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_all_button), 
			       opt2ii & DC_BIT_RB_ALL ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_all_inv_button), 
			       opt2ii & DC_BIT_RB_ALL_INV ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_good_button), 
			       opt2ii & DC_BIT_RB_GOOD ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_good_inv_button), 
			       opt2ii & DC_BIT_RB_GOOD_INV ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_fail_button), 
			       opt2ii & DC_BIT_RB_FAIL ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_fail_inv_button), 
			       opt2ii & DC_BIT_RB_FAIL_INV ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_frame_bmp_button), 
			       opt2ii & DC_BIT_RB_FRAME_BMP ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.rb_frame_rgb_button), 
			       opt2ii & DC_BIT_RB_FRAME_RGB ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.mrb_frame_bmp_button), 
			       opt2ii & DC_BIT_MRB_FRAME_BMP ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.mrb_frame_rgb_button), 
			       opt2ii & DC_BIT_MRB_FRAME_RGB ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.hline_bmp_button), 
			       opt2ii & DC_BIT_HLINE_BMP ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.hline_rgb_button), 
			       opt2ii & DC_BIT_HLINE_RGB ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.sline_bmp_button), 
			       opt2ii & DC_BIT_SLINE_BMP ? TRUE:FALSE);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d2ii.sline_rgb_button), 
			       opt2ii & DC_BIT_SLINE_RGB ? TRUE:FALSE);
}



/* Updates mask border.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_mborder(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  int_to_string("", cfg.mad.mborder, "", &buf);
  xplab_entry_set_text(matching.mborder_entry, buf);

  FREEMEM(buf);
}



/* Updates lag.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_nborder(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  int_to_string("", cfg.mad.nborder, "", &buf);
  xplab_entry_set_text(matching.nborder_entry, buf);

  FREEMEM(buf);
}



/* Updates mask weight.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_maskweight(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  int_to_string("", cfg.mad.maskweight, "", &buf);
  xplab_entry_set_text(matching.maskweight_entry, buf);

  FREEMEM(buf);
}



/* Updates NNB pixel matching degree.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_nnbmatch(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  int_to_string("", cfg.mad.nnbmatch, "", &buf);
  xplab_entry_set_text(matching.nnbmatch_entry, buf);

  FREEMEM(buf);
}



/* Updates step width.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_interleave(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.mad.interleave, "", &buf);
  xplab_entry_set_text(matching.interleave_entry, buf);

  FREEMEM(buf);
}



/* Updates symbol overlapping.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_overlap(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.mad.overlap, "", &buf);
  xplab_entry_set_text(matching.overlap_entry, buf);

  FREEMEM(buf);
}



/* Updates symbol spacing.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_symbolspacing(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.syspc, "", &buf);
  xplab_entry_set_text(matching.symbolspacing_entry, buf);

  FREEMEM(buf);
}



/* Updates NB overlapping.

Revision: 22.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_nb_lap(matching_def matching, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.mad.nb_lap, "", &buf);
  xplab_entry_set_text(matching.nb_lap_entry, buf);

  FREEMEM(buf);
}



/* Updates correlation level flag.

Revision: 22.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_corrlevel(correlation_def correlation, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.mad.corrlevel, "", &buf);
  xplab_entry_set_text(correlation.corrlevel_entry, buf);

  FREEMEM(buf);
}



/* Updates correlation level flag.

Revision: 05.01.2005 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_corrbroken(correlation_def correlation, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.mad.corrbroken, "", &buf);
  xplab_entry_set_text(correlation.corrbroken_entry, buf);

  FREEMEM(buf);
}



/* Updates cluster level flag.

Revision: 22.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_cluster_level(correlation_def correlation, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.cluster_level, "", &buf);
  xplab_entry_set_text(correlation.cluster_level_entry, buf);

  FREEMEM(buf);
}



/* Updates cluster degradation flag.

Revision: 22.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_cluster_step(correlation_def correlation, cfg_def cfg)
{
  gchar *buf = NULL;

  double_to_string("", cfg.cluster_step, "", &buf);
  xplab_entry_set_text(correlation.cluster_step_entry, buf);

  FREEMEM(buf);
}



/* Updates cluster threshold flag.

Revision: 22.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_cluster_thold(correlation_def correlation, cfg_def cfg)
{
  gchar *buf = NULL;

  int_to_string("", cfg.cluster_thold, "", &buf);
  xplab_entry_set_text(correlation.cluster_thold_entry, buf);

  FREEMEM(buf);
}



/* Updates cluster force flag.

Revision: 18.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_cluster_force(correlation_def correlation, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(correlation.cluster_force_button), 
			       cfg.cluster_force!=CLUSTER_FORCE_DEFAULT ? TRUE:FALSE);
}



/* Updates cluster write flag.

Revision: 18.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_cluster_write(correlation_def correlation, cfg_def cfg)
{
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(correlation.cluster_write_button), 
			       cfg.cluster_write!=CLUSTER_WRITE_DEFAULT ? TRUE:FALSE);
}



/* Updates the statistic page.

Revision: 16.12.2004 E.Sack

Input : - pointer to statistic settings
Output: - statistic settings updated
Usage: - Sets all entrys to the current working settings.
*/
static void
update_statistic_page(stats_def stats)
{
  update_statistic_entry(stats.owin);
  update_statistic_entry(stats.cwin);
  update_statistic_entry(stats.gwin);
  update_statistic_entry(stats.bwin);
  update_statistic_entry(stats.rhst);
  update_statistic_entry(stats.chst);
  update_statistic_entry(stats.ghst);
  update_statistic_entry(stats.bhst);
  update_statistic_entry(stats.shst);
  update_statistic_entry(stats.contrast);
}



/* Copies the user input from the statistic page into the current 
   working set.

Revision: 16.12.2004 E.Sack

Input : - statistic settings
Output: - statistic settings updated from entrys
*/
static void
apply_statistic_page(stats_def *stats)
{
  apply_statistic_entry(&stats->owin);
  apply_statistic_entry(&stats->cwin);
  apply_statistic_entry(&stats->gwin);
  apply_statistic_entry(&stats->bwin);
  apply_statistic_entry(&stats->rhst);
  apply_statistic_entry(&stats->chst);
  apply_statistic_entry(&stats->ghst);
  apply_statistic_entry(&stats->bhst);
  apply_statistic_entry(&stats->shst);
  apply_statistic_entry(&stats->contrast);
}



/* Copies the old statistic settings to the user input in the statistic page
   and into the current working set.

Revision: 16.12.2004 E.Sack

Input : - statistic settings
Output: - statistic settings updated from entrys
*/
static void
revert_statistic_page(stats_def *stats)
{
  stats_revert_all(&stats->owin);
  stats_revert_all(&stats->cwin);
  stats_revert_all(&stats->gwin);
  stats_revert_all(&stats->bwin);
  stats_revert_all(&stats->rhst);
  stats_revert_all(&stats->chst);
  stats_revert_all(&stats->ghst);
  stats_revert_all(&stats->bhst);
  stats_revert_all(&stats->shst);
  stats_revert_all(&stats->contrast);
}



/* Updates preprocessing page.

Revision: 28.10.2003 E.Sack

Input : - preprocessing page definition
        - preferences
Output: - preprocessing page updated
*/
static void
update_preprocessing_page(pre_def preprocessing, cfg_def cfg)
{
  update_histo_thold(preprocessing, cfg);
  update_rgb_norm(preprocessing, cfg.rgb_norm);
  update_support_dots(preprocessing, cfg);
  update_support_dots_auto(preprocessing, cfg);

  xplab_label_set_text(pr.page.page_label, "");
}



/* Updates matching page.

Revision: 10.10.2003 E.Sack

Input : - matching page definition
        - preferences
Output: - matching page updated
*/
static void
update_matching_page(matching_def matching, cfg_def cfg)
{
  update_mborder(matching, cfg);
  update_nborder(matching, cfg);
  update_maskweight(matching, cfg);
  update_nnbmatch(matching, cfg);
  update_interleave(matching, cfg);
  update_overlap(matching, cfg);
  update_symbolspacing(matching, cfg);
  update_nb_lap(matching, cfg);
  update_single_line(matching, cfg);
  update_single_nb(matching, cfg);
  update_broken_nb(matching, cfg);
  update_renew_method(matching, cfg);
  update_line_method(matching, cfg);
  update_duration(matching, cfg);
  update_subst_rb(matching, cfg);
  update_subst_nb(matching, cfg);

  xplab_label_set_text(pr.page.page_label, "");
}



/* Updates correlation page.

Revision: 06.10.2003 E.Sack

Input : - correlation page definition
        - preferences
Output: - correlation page updated
*/
static void
update_correlation_page(correlation_def correlation, cfg_def cfg)
{
  set_all_ckernel(cfg.ckernel);

  update_corrlevel(correlation, cfg);
  update_corrbroken(correlation, cfg);
  update_cluster_level(correlation, cfg);
  update_cluster_step(correlation, cfg);
  update_cluster_thold(correlation, cfg);
  update_cluster_force(correlation, cfg);
  update_cluster_write(correlation, cfg);
  update_cluster_nb(correlation, cfg);
  update_domain(correlation, cfg);

  xplab_label_set_text(pr.page.page_label, "");
}



/* Updates documentation channel page.

Revision: 11.04.2004 E.Sack

Input : - -
Output: - page label updated
*/
static void
update_dc_page(cfg_def cfg)
{
  int_to_entry(pr.page.dc.store_entry, "", cfg.dc_store, "");

  update_hidden_list(&cfg.dc_1d, &pr.page.dc.dc1l, &pr.page.dc.d1_idx);
  update_hidden_list(&cfg.dc_2d, &pr.page.dc.dc2l, &pr.page.dc.d2_idx);

  update_dc_buttons_from_list();
  dc_option_entry_from_list_all();

  xplab_label_set_text(pr.page.page_label, "");
}



/* Updates DC buttons from list.

Revision: 12.04.2004 E.Sack

Input : - -
Output: - DC buttons updated
*/
static void
update_dc_buttons_from_list(void)
{
  int opt = 0;

  if (pr.page.dc.dc1l.cnt) {
    opt = pr.page.dc.dc1l.n[pr.page.dc.d1_idx].nr;
  }

  update_dc_1d(pr.page.dc.d1, opt);                         /* set buttons */

  if (pr.page.dc.dc2l.cnt) {
    opt = pr.page.dc.dc2l.n[pr.page.dc.d2_idx].nr;
  }

  update_dc_2di(pr.page.dc.d2i, opt);
  update_dc_2dii(pr.page.dc.d2ii, opt);
}



/* Updates hidden list and index.

Revision: 11.04.2004 E.Sack

Input : - pointer to source list
        - pointer to destination list
        - pointer to index
Output: - destination list updated
        - index updated
*/
static void
update_hidden_list(int2_lst *src, int2_lst *dst, int *dst_idx)
{
  copy_int2_lst(src, dst);                           /* update hidden list */

  if (dst->cnt) {
    xwrap_idx(dst_idx, dst->cnt);                    /* update current idx */
  }
}



/* Updates I/O page.

Revision: 12.10.2003 E.Sack

Input : - -
Output: - page label updated
*/
static void
update_io_page(io_def io, cfg_def cfg)
{
  xplab_label_set_text(pr.page.page_label, "");

  update_nb_border(io, cfg);
  update_rawbit(io, cfg);
  update_overwrite_image(io, cfg);
  update_report_write(io, cfg);
  update_histo_write(io, cfg);
  update_replay_write(io, cfg);
  update_training_write(io, cfg);
  update_timestamp(io, cfg);
  update_writeauto(io, cfg);
}



/* Deletes preferences dialog.

Revision: 23.10.2003 E.Sack

Input : - GTK convention
Output: - propagation flag
*/
static gboolean
cb_preferences_delete(GTK_CONVENTION_DIALOG_RTS_GBOOLEAN)
{
  assert(event);
  assert(widget);

  cb_preferences_cancel_button(NULL, user_data);

  return FALSE;
}



/* Saves the preferences.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_preferences_save_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  assert(button);

  cb_preferences_apply_button(NULL, NULL);    /* user input mask -> global */

  copy_cfg_def(&ctrl_apply.cfg, &cfg_mirror);            /* local -> local */

  write_cfg(ctrl_apply.set, cfg_mirror, 
	    pr.page.statistic);                           /* local -> disc */
}



/* Applies the preferences.

Revision: 05.01.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_preferences_apply_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  xplab_label_set_text(pr.page.page_label, "");

  ctrl_apply.cfg.mad.mborder = check_mborder_setting();/* user input mask -> global */
  ctrl_apply.cfg.mad.nborder = check_nborder_setting();
  ctrl_apply.cfg.mad.interleave = check_interleave_setting();
  ctrl_apply.cfg.mad.overlap = check_overlap_setting();
  ctrl_apply.cfg.mad.nb_lap = check_nb_lap_setting();
  ctrl_apply.cfg.mad.corrlevel = check_corrlevel_setting();
  ctrl_apply.cfg.mad.corrbroken = check_corrbroken_setting();
  ctrl_apply.cfg.mad.nnbmatch = check_nnbmatch_setting();
  ctrl_apply.cfg.mad.maskweight = check_maskweight_setting();

  ctrl_apply.cfg.nb_border_x = check_nb_border_setting(pr.page.io.nb_border_x_entry, xctrl.cfg.nb_border_x);
  ctrl_apply.cfg.nb_border_y = check_nb_border_setting(pr.page.io.nb_border_y_entry, xctrl.cfg.nb_border_y);
  ctrl_apply.cfg.dc_store = check_dc_store_setting();
  ctrl_apply.cfg.syspc = check_symbolspacing_setting();
  ctrl_apply.cfg.histo_thold = check_entry_limits_integer(pr.page.preprocessing.histo_thold_entry, THOLD_MIN, THOLD_MAX, xctrl.cfg.histo_thold);
  ctrl_apply.cfg.cluster_level = check_cluster_level_setting();
  ctrl_apply.cfg.cluster_step = check_cluster_step_setting();
  ctrl_apply.cfg.cluster_thold = check_cluster_thold_setting();
  ctrl_apply.cfg.tips = apply_tooltips();
  ctrl_apply.cfg.verbose = check_verbose_setting();
  ctrl_apply.cfg.raise = apply_raise();
  ctrl_apply.cfg.line_method = check_line_method_setting();
  ctrl_apply.cfg.report_write = apply_report_write();
  ctrl_apply.cfg.rb_chain = apply_rb_chain();
  ctrl_apply.cfg.histo_write = apply_histo_write();
  ctrl_apply.cfg.duration = apply_duration();
  ctrl_apply.cfg.support_dots = apply_support_dots();
  ctrl_apply.cfg.support_dots_auto = apply_support_dots_auto();
  ctrl_apply.cfg.overwrite_image = apply_overwrite_image();
  ctrl_apply.cfg.replay_write = apply_replay_write();
  ctrl_apply.cfg.training_write = apply_training_write();
  ctrl_apply.cfg.single_line = apply_single_line();
  ctrl_apply.cfg.single_nb = apply_single_nb();
  ctrl_apply.cfg.match_method = apply_broken_nb();
  ctrl_apply.cfg.renew_method = apply_renew_method();
  ctrl_apply.cfg.cluster_force = apply_cluster_force();
  ctrl_apply.cfg.cluster_write = apply_cluster_write();
  ctrl_apply.cfg.cluster_nb = apply_cluster_nb();
  ctrl_apply.cfg.rawbit = apply_rawbit();
  ctrl_apply.cfg.timestamp = apply_timestamp();
  ctrl_apply.cfg.writeauto = apply_writeauto();
  ctrl_apply.cfg.cdomain = apply_domain();
  ctrl_apply.cfg.batch = apply_batch();

  apply_dc_1d(&ctrl_apply.cfg.dc_1d);
  apply_dc_2d(&ctrl_apply.cfg.dc_2d);

  ctrl_apply.cfg.rgb_norm.r = check_rgb_norm_r_setting(0);
  ctrl_apply.cfg.rgb_norm.g = check_rgb_norm_g_setting(0);
  ctrl_apply.cfg.rgb_norm.b = check_rgb_norm_b_setting(1);

  apply_all_ckernel(ctrl_apply.cfg.ckernel);

  FREEMEM(ctrl_apply.cfg.subst_rb);
  ctrl_apply.cfg.subst_rb = check_subst_rb_setting();

  FREEMEM(ctrl_apply.cfg.subst_nb);
  ctrl_apply.cfg.subst_nb = check_subst_nb_setting();

  copy_prog_ctrl(&ctrl_apply, &xctrl);                  /* local -> global */

  apply_statistic_page(&pr.page.statistic);      /* different apply policy */

  xplab_tooltips(xctrl.cfg.tips);                     /* event driven hook */

}



/* Reverts the preferences.

Revision: 29.10.2003 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_preferences_revert_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  assert(button);
  user_data=user_data;

  copy_cfg_def(&cfg_mirror, &xctrl.cfg);                /* local -> global */
  copy_cfg_def(&cfg_mirror, &ctrl_apply.cfg);            /* local -> local */

  revert_statistic_page(&pr.page.statistic);    /* revert set -> apply set */

  update_all(xctrl.cfg);                      /* set all entries to config */
  cb_preferences_apply_button(NULL, NULL);   /* force setting calculations */

  xplab_label_set_text(pr.page.page_label, "");
}



/* Unloads DB.

Revision: 18.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_preferences_unload(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  int err;

  STSFY_GCC(user_data);

  assert(menuitem);

  err = xplab_input_db(NULL);

  if (err) {
    string_to_log(main_log, "WARNING: couldn't unload DB\n", "", "");
  }
}



/* Cancels the preferences input.

Revision: 09.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_preferences_cancel_button(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(button);

  xplab_label_set_text(pr.page.page_label, "");

  copy_prog_ctrl(&ctrl_apply, &xctrl);                  /* local -> global */

  xplab_window_raise(pr.caller, xctrl.cfg.raise);         /* modal window! */
  xplab_widget_destroy(&pr.preferences_dialog);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - menu
Output: - menu item
*/
void
xplab_preferences_menu(GtkWidget *file_menu, GtkWidget **caller)
{
  GtkWidget *file_menu_prefs;

  file_menu_prefs = gtk_menu_item_new_with_label("preferences" XPLAB_MENU_EXT);
  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), file_menu_prefs);

  g_signal_connect(G_OBJECT(file_menu_prefs), "activate", 
		   G_CALLBACK(cb_preferences_dialog), (gpointer)caller);
}



/* Starts preferences window.

Revision: 13.06.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_preferences_dialog(GTK_CONVENTION_MENUITEM_RTS_VOID)
{
  gboolean raised;

  assert(user_data);
  assert(menuitem);

  raised = xplab_window_raise(&pr.preferences_dialog, 1);

  if (raised == FALSE) {
    copy_prog_ctrl(&ctrl_apply, &xctrl);              /* reset global data */

    pr.caller = (GtkWidget **)user_data;            /* last succeeded call */
    xplab_preferences();                      /* create preferences dialog */
  }
}



/* Creates a dialog for preferences.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - preferences dialog
*/
static void
xplab_preferences(void)
{
  pr.preferences_dialog = gtk_dialog_new();/* set the preferences dialog widget */

  /* set the preferences dialog title */
  gtk_window_set_title(GTK_WINDOW(pr.preferences_dialog), XPLAB_PREFERENCES_TITLE_ITEM);
  gtk_window_set_modal(GTK_WINDOW(pr.preferences_dialog), TRUE);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(GTK_DIALOG(pr.preferences_dialog)->action_area),
			    GTK_BUTTONBOX_SPREAD);

    /* set the preferences dialog input */
  preferences_dialog_upper(GTK_BOX(GTK_DIALOG(pr.preferences_dialog)->vbox));

  /* set the preferences dialog action */
  preferences_dialog_lower(GTK_CONTAINER(GTK_DIALOG(pr.preferences_dialog)->action_area));

  gtk_widget_show_all(pr.preferences_dialog);

  g_signal_connect(G_OBJECT(pr.preferences_dialog), "delete-event", 
		   G_CALLBACK(cb_preferences_delete), NULL);

  update_all(xctrl.cfg);

  xplab_label_set_text(pr.page.page_label, "");
}



/* Creates the preferences dialog upper part as a notebook with pages.

Revision: 05.08.2005 E.Sack

Input : - pointer to container box
Output: - content updated
*/
static void
preferences_dialog_upper(GtkBox *vbox_in)
{
  /* put into vertical box a notebook */
  pr.page.notebook = gtk_notebook_new();
  gtk_box_pack_start(vbox_in, pr.page.notebook, FALSE, FALSE, 0);

  /* put the first page into the notebook, automatically displayed         */
  pr.page.general.parent = preferences_dialog_page_general();
  gtk_notebook_append_page(GTK_NOTEBOOK(pr.page.notebook), pr.page.general.parent, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(pr.page.notebook), pr.page.general.parent, XPLAB_PREFS_PAGE_GENERAL);

  /* put the next page into the notebook */
  pr.page.preprocessing.parent = preferences_dialog_page_preprocessing();
  gtk_notebook_append_page(GTK_NOTEBOOK(pr.page.notebook), pr.page.preprocessing.parent, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(pr.page.notebook), pr.page.preprocessing.parent, XPLAB_PREFS_PAGE_PREPROCESSING);

  /* put the next page into the notebook */
  pr.page.matching.parent = preferences_dialog_page_matching();
  gtk_notebook_append_page(GTK_NOTEBOOK(pr.page.notebook), pr.page.matching.parent, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(pr.page.notebook), pr.page.matching.parent, XPLAB_PREFS_PAGE_MATCHING);

  /* put the next page into the notebook */
  pr.page.correlation.parent = preferences_dialog_page_correlation();
  gtk_notebook_append_page(GTK_NOTEBOOK(pr.page.notebook), pr.page.correlation.parent, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(pr.page.notebook), pr.page.correlation.parent, XPLAB_PREFS_PAGE_CORRELATION);

  /* put the next page into the notebook */
  pr.page.statistic.sparent = preferences_dialog_page_statistic();
  gtk_notebook_append_page(GTK_NOTEBOOK(pr.page.notebook), pr.page.statistic.sparent, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(pr.page.notebook), pr.page.statistic.sparent, XPLAB_PREFS_PAGE_STATISTIC);

  /* put the next page into the notebook */
  pr.page.dc.parent = preferences_dialog_page_dc();
  gtk_notebook_append_page(GTK_NOTEBOOK(pr.page.notebook), pr.page.dc.parent, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(pr.page.notebook), pr.page.dc.parent, XPLAB_PREFS_PAGE_DC);

  /* put the last page into the notebook */
  pr.page.io.parent = preferences_dialog_page_io();
  gtk_notebook_append_page(GTK_NOTEBOOK(pr.page.notebook), pr.page.io.parent, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(pr.page.notebook), pr.page.io.parent, XPLAB_PREFS_PAGE_IO);

  g_signal_connect(G_OBJECT(pr.page.notebook), "switch-page", 
		   G_CALLBACK(cb_preferences_page_switch), NULL);

  /* put into vertical box a label */
  pr.page.page_label = gtk_label_new(NULL);
  gtk_box_pack_start(vbox_in, pr.page.page_label, FALSE, FALSE, 0);
}



/* Creates the preferences page general.

Revision: 03.10.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget*
preferences_dialog_page_general(void)
{
  GtkWidget *vbox1, *vbox2, *frame, *sep;

  /* create a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox1), XPLAB_IMG_BORDER_SIZE);

  /* put into vertical box a frame */
  frame = gtk_frame_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox1), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  /* put into frame a vertical box */
  vbox2 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox2), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), vbox2);

  dialog_page_general_part1(vbox2);

  sep = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox2), sep, FALSE, FALSE, 0);

  dialog_page_general_part2(vbox2);

  return vbox1;
}



/* Creates a preference setting box.

Revision: 03.04.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
dialog_page_general_part1(GtkWidget *vbox)
{
  GtkWidget *hbox, *vbox3;

  /* put into vertical box a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a vertical box */
  vbox3 = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox3, FALSE, FALSE, 0);

  tooltips_setting(vbox3);
  verbose_setting(vbox3);
  raise_setting(vbox3);
  rb_chain_setting(vbox3);
}



/* Creates a preference setting box.

Revision: 03.04.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
dialog_page_general_part2(GtkWidget *vbox)
{
  GtkWidget *hbox, *vbox3;

  /* put into vertical box a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a vertical box */
  vbox3 = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox3, FALSE, FALSE, 0);

  batchjob_setting(vbox3);
  batchlog_setting(vbox3);
}



/* Creates a preference setting box.

Revision: 12.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
rawbit_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, RAWBIT_KEY, RAWBIT_REM, 0);

  pr.page.io.rawbit_button = gtk_check_button_new_with_label("ASCII");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.io.rawbit_button,
		     FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
image_filename_setting(GtkWidget **abox)
{
  gint       err;
  GtkWidget *label;
  GtkEntry  *entry;

  *abox = gtk_hbox_new(FALSE, 0);

  label = gtk_label_new("image file name: ");
  gtk_box_pack_start(GTK_BOX(*abox), label, FALSE, FALSE, 2);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);

  err = check_name(xctrl.set.image_filename);
  if (err) {
    xplab_entry_set_text(entry, NO_INPUT);
  }
  else {
    xplab_entry_set_text(entry, xctrl.set.image_filename);
  }

  gtk_box_pack_start(GTK_BOX(*abox), GTK_WIDGET(entry), TRUE, TRUE, 2);
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
separation_rules_filename_setting(GtkWidget *cbox)
{
  int       err;
  char      *buf = NULL;
  GtkWidget *hbox;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, "word separation rules", INFNAMD_REM, 0);

  err = check_name(xctrl.set.image_filename);

  if (err) {
    add_string(NO_INPUT, &buf);
  }
  else {
    char nam1[MAXCHAR];

    g_snprintf(nam1, sizeof(nam1), INFNAMD_FORMAT(0), xctrl.set.image_filename);
    add_string(nam1, &buf);
  }

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, buf);

  FREEMEM(buf);

  gtk_box_pack_start(GTK_BOX(cbox), GTK_WIDGET(entry), FALSE, FALSE, 2);
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
separation_histo_filename_setting(GtkWidget *cbox)
{
  int       err;
  char      *buf = NULL;
  GtkWidget *hbox;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, "word separation histogram", LINNAMH_REM, 0);

  err = check_name(xctrl.set.image_filename);

  if (err) {
    add_string(NO_INPUT, &buf);
  }
  else {
    char nam1[MAXCHAR];

    g_snprintf(nam1, sizeof(nam1), LINNAMH_FORMAT(0), xctrl.set.image_filename);
    add_string(nam1, &buf);
  }

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, buf);

  gtk_box_pack_start(GTK_BOX(cbox), GTK_WIDGET(entry), FALSE, FALSE, 2);

  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
report_filename_setting(GtkWidget *vbox)
{
  char      *report_filename = NULL;
  GtkWidget *abox, *label;
  GtkEntry  *entry;

  abox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), abox, FALSE, FALSE, 2);

  label = gtk_label_new("report file name: ");
  gtk_box_pack_start(GTK_BOX(abox), label, FALSE, FALSE, 2);

  report_filename = default_report_filename(xctrl.set.image_filename);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, report_filename);

  gtk_box_pack_start(GTK_BOX(abox), GTK_WIDGET(entry), TRUE, TRUE, 2);

  FREEMEM(report_filename);
}



/* Creates a preference setting box.

Revision: 08.10.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
writeauto_filename_setting(GtkWidget *vbox)
{
  char      *wauto_filename = NULL;
  GtkWidget *abox, *label;
  GtkEntry  *entry;

  abox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), abox, FALSE, FALSE, 2);

  label = gtk_label_new("auto file name: ");
  gtk_box_pack_start(GTK_BOX(abox), label, FALSE, FALSE, 2);

  wauto_filename = default_auto_filename(NULL);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, wauto_filename);

  gtk_box_pack_start(GTK_BOX(abox), GTK_WIDGET(entry), TRUE, TRUE, 2);

  FREEMEM(wauto_filename);
}



/* Creates a preference setting box.

Revision: 08.10.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
report_write_setting(GtkWidget **abox, GtkWidget **bbox, GtkWidget **cbox,
		     GtkWidget **dbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);
  *cbox = gtk_hbox_new(FALSE, 0);
  *dbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, REPORT_WRITE_KEY, REPORT_WRITE_REM, 0);

  pr.page.io.report_terse_button = gtk_check_button_new_with_label("terse");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.io.report_terse_button, FALSE, FALSE, 0);

  pr.page.io.report_rbs_button = gtk_check_button_new_with_label("rbs");
  gtk_box_pack_start(GTK_BOX(*cbox), pr.page.io.report_rbs_button, FALSE, FALSE, 0);

  pr.page.io.report_suppress = gtk_check_button_new_with_label("suppress");
  gtk_box_pack_start(GTK_BOX(*dbox), pr.page.io.report_suppress, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 19.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
rb_chain_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, RB_CHAIN_KEY, RB_CHAIN_REM, 0);

  pr.page.general.rb_chain_use_button = gtk_check_button_new_with_label("use");
  gtk_box_pack_end(GTK_BOX(hbox), pr.page.general.rb_chain_use_button, FALSE, FALSE, 0);

  pr.page.general.rb_chain_color_button = gtk_check_button_new_with_label("color");
  gtk_box_pack_end(GTK_BOX(hbox), pr.page.general.rb_chain_color_button, FALSE, FALSE, 0);

  pr.page.general.rb_chain_tiles_button = gtk_check_button_new_with_label("tiles");
  gtk_box_pack_end(GTK_BOX(hbox), pr.page.general.rb_chain_tiles_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 03.04.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
batchjob_setting(GtkWidget *cbox)
{
  GSList    *item_group = NULL;
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, BATCH_PHASE_KEY, BATCH_PHASE_REM, 0);

  pr.page.general.batchjob_void_button = 
    gtk_radio_button_new_with_label(item_group, "void");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.batchjob_void_button, FALSE, FALSE, 0);

  pr.page.general.batchjob_rbchain_button = 
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(pr.page.general.batchjob_void_button),
						"RB chains");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.batchjob_rbchain_button, FALSE, FALSE, 0);

  pr.page.general.batchjob_matching_button = 
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(pr.page.general.batchjob_rbchain_button),
						"matching");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.batchjob_matching_button, FALSE, FALSE, 0);

  pr.page.general.batchjob_cluster_button = 
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(pr.page.general.batchjob_matching_button),
						"clustering");
  gtk_box_pack_start(GTK_BOX(hbox), 
		     pr.page.general.batchjob_cluster_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 08.04.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
batchlog_setting(GtkWidget *cbox)
{
  GSList    *item_group = NULL;
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, BATCH_LOG_KEY, BATCH_LOG_REM, 0);

  pr.page.general.batchlog_no_button = 
    gtk_radio_button_new_with_label(item_group, "no");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.batchlog_no_button, FALSE, FALSE, 0);

  pr.page.general.batchlog_ovwrt_button = 
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(pr.page.general.batchlog_no_button),
						"overwrite");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.batchlog_ovwrt_button,
		     FALSE, FALSE, 0);

  pr.page.general.batchlog_appnd_button = 
    gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(pr.page.general.batchlog_ovwrt_button),
						"append");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.batchlog_appnd_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 27.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cluster_nb_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, CLUSTER_NB_KEY, CLUSTER_NB_REM, 0);

  pr.page.correlation.cluster_nb_old_button = gtk_check_button_new_with_label("old");
  gtk_box_pack_end(GTK_BOX(hbox), pr.page.correlation.cluster_nb_old_button, FALSE, FALSE, 0);

  pr.page.correlation.cluster_nb_new_button = gtk_check_button_new_with_label("new");
  gtk_box_pack_end(GTK_BOX(hbox), pr.page.correlation.cluster_nb_new_button, FALSE, FALSE, 0);

  g_signal_connect(G_OBJECT(pr.page.correlation.cluster_nb_old_button), "toggled",
		   G_CALLBACK(cb_cluster_nb), NULL);

  g_signal_connect(G_OBJECT(pr.page.correlation.cluster_nb_new_button), "toggled",
		   G_CALLBACK(cb_cluster_nb), NULL);
}



/* Creates a preference setting box.

Revision: 18.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
histo_write_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, HISTO_WRITE_KEY, HISTO_WRITE_REM, 0);

  pr.page.io.histo_write_button = gtk_check_button_new_with_label("write");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.io.histo_write_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 18.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
duration_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, DURATION_KEY, DURATION_REM, 0);

  pr.page.matching.duration_button = gtk_check_button_new_with_label("show");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.matching.duration_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
support_dots_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, SUPPORT_DOTS_KEY, SUPPORT_DOTS_REM, 0);

  pr.page.preprocessing.support_dots_button = gtk_check_button_new_with_label("support");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.preprocessing.support_dots_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(pr.page.preprocessing.support_dots_button), "toggled",
		   G_CALLBACK(cb_dots_support), NULL);

  setting_key(hbox, "  " SUPPORT_DOTS_AUTO_KEY, NULL, 0);

  pr.page.preprocessing.support_dots_auto = gtk_check_button_new_with_label("DB driven");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.preprocessing.support_dots_auto, FALSE, FALSE, 0);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pr.page.preprocessing.support_dots_button), 
			       SUPPORT_DOTS_DEFAULT!=0);     /* emits signal */
}



/* Creates a preference setting box.

Revision: 25.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
overwrite_image_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, OVERWRITE_IMAGE_KEY, OVERWRITE_IMAGE_REM, 0);

  pr.page.io.overwrite_image_button = gtk_check_button_new_with_label("overwrite image");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.io.overwrite_image_button, 
		     FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
        - D1 buttons definition
*/
static d1_def
dc_d1_setting(GtkWidget *fbox)
{
  d1_def    d1;
  GtkWidget *vbox, *vbox1, *hbox;

  init_d1_def(&d1);

  /* put into inbox1 a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(fbox), vbox);

  /* put into vertical box a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

  /* put into horizontal box a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox1), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_start(GTK_BOX(hbox), vbox1, FALSE, FALSE, 0);

  d1.pos_ccf_button = 
    gtk_check_button_new_with_label(DC_OPT_POS_CCF_DESC " (" DC_OPT_POS_CCF_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox1), d1.pos_ccf_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d1.pos_ccf_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d1.pos_pad_button = gtk_check_button_new_with_label(DC_OPT_POS_PAD_DESC " (" DC_OPT_POS_PAD_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox1), d1.pos_pad_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d1.pos_pad_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d1.tfin_button = gtk_check_button_new_with_label(DC_OPT_TFIN_DESC " (" DC_OPT_TFIN_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox1), d1.tfin_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d1.tfin_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d1.tflag_button = gtk_check_button_new_with_label(DC_OPT_TFLAG_DESC " (" DC_OPT_TFLAG_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox1), d1.tflag_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d1.tflag_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d1.pos_button = gtk_check_button_new_with_label(DC_OPT_POS_DESC " (" DC_OPT_POS_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox1), d1.pos_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d1.pos_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d1.pos_hphase_button = gtk_check_button_new_with_label(DC_OPT_POS_HPHASE_DESC " (" DC_OPT_POS_HPHASE_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox1), d1.pos_hphase_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d1.pos_hphase_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  dc_option_list(vbox, &pr.page.dc.d1_entry,
		 (gpointer)XPLAB_PREFS_PAGE_DC_OPTION_1D,
		 XPLAB_PREFS_PAGE_DC_1_LIST_EMPTY);

  return d1;
}



/* Updates option text entries and lists from buttons.

Revision: 10.04.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_button_to_entry(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(togglebutton);

  enroll_dc_1d_to_list();                              /* scan all buttons */
  enroll_dc_2di_to_list();
  enroll_dc_2dii_to_list();

  dc_option_entry_from_list_all();
}



/* Triggers actions on check button events.

Revision: 19.10.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dots_support(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);
  STSFY_GCC(togglebutton);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.preprocessing.support_dots_button)) == TRUE) {
    gtk_widget_set_sensitive(pr.page.preprocessing.support_dots_auto, FALSE);
  }
  else {
    gtk_widget_set_sensitive(pr.page.preprocessing.support_dots_auto, TRUE);
  }
}



/* Triggers actions on check button events.

Revision: 19.10.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_cluster_nb(GTK_CONVENTION_TOGGLEBUTTON_RTS_VOID)
{
  STSFY_GCC(user_data);

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_old_button)) == FALSE &&
      gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_new_button)) == FALSE) {

    /* at least one button shall be set                                      */
    if (GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_new_button) == togglebutton) {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_old_button),
				   TRUE);
    }
    else {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pr.page.correlation.cluster_nb_new_button),
				   TRUE);
    }
  }
}



/* Updates DC option entries from their lists.

Revision: 11.04.2004 E.Sack

Input : - -
Output: - entries updated
*/
static void
dc_option_entry_from_list_all(void)
{
  dc_option_entry_from_list(pr.page.dc.d1_entry, pr.page.dc.d1_idx, 
			    pr.page.dc.dc1l, 
			    XPLAB_PREFS_PAGE_DC_1_LIST_EMPTY);

  dc_option_entry_from_list(pr.page.dc.d2i_entry, pr.page.dc.d2_idx, 
			    pr.page.dc.dc2l,
			    XPLAB_PREFS_PAGE_DC_2_LIST_EMPTY);

  dc_option_entry_from_list(pr.page.dc.d2ii_entry, pr.page.dc.d2_idx, 
			    pr.page.dc.dc2l,
			    XPLAB_PREFS_PAGE_DC_2_LIST_EMPTY);
}



/* Updates DC option entry from list.

Revision: 11.04.2004 E.Sack

Input : - pointer to text entry widget
        - index
        - option list
        - pointer to default entry text
Output: - entry updated
*/
static void
dc_option_entry_from_list(GtkEntry *entry, int idx, int2_lst dlst, char *def)
{
  char *buf = NULL;

  assert(entry);
  assert(def);

  if (dlst.cnt) {
    assert(idx>=0 && idx<dlst.cnt);
    buf = dc_option_list_format(idx, dlst.cnt, dlst.n[idx].nr);
    xplab_entry_set_text(entry, buf);
  }
  else {
    xplab_entry_set_text(entry, def);
  }

  FREEMEM(buf);
}



/* Creates an option entry.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
        - pointer to entry
        - CB for increment and decrement
        - CB user data pointer
        - default entry text
Output: - container updated
        - D1 buttons definition
*/
static void
dc_option_list(GtkWidget *vbox_in, GtkEntry **entry, gpointer gpo, 
	       char *def_text)
{
  char      *buf = NULL;
  GtkWidget *hbox, *box, *label, *button;

  /* put into inbox a horizontal box */
  hbox = gtk_hbox_new(FALSE, 2);
  gtk_box_pack_end(GTK_BOX(vbox_in), hbox, FALSE, FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(hbox), 2);

  cat_string((char *)gpo, " option list ", "", &buf);
  label = gtk_label_new(buf);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
  FREEMEM(buf);

  button = gtk_button_new_with_label(SINGLE_BUTTON_NEW);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);

  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_dc_new), gpo);

  button = gtk_button_new_with_label(SINGLE_BUTTON_REMOVE);
  gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);

  g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(cb_dc_remove), gpo);

  box = pentry(entry, FALSE, NULL, def_text, XPLAB_PREFS_ENTRY_CHARS_DC, 
	       XPLAB_PREFS_ENTRY_LEN_DC, 
	       GTK_SIGNAL_FUNC(cb_dc_down), "<", 
	       GTK_SIGNAL_FUNC(cb_dc_up), ">", NULL, NULL, gpo);

  gtk_box_pack_start(GTK_BOX(hbox), box, FALSE, FALSE, 2);
}



/* Adds an option to the option list.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_new(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int cmp;

  assert(user_data);
  assert(button);

  cmp = strcmp(XPLAB_PREFS_PAGE_DC_OPTION_1D, user_data);

  if (!cmp) {
    if (pr.page.dc.d1_idx < XPLAB_PREFS_PAGE_DC_LIST_MAX) {
      pr.page.dc.d1_idx = pr.page.dc.dc1l.cnt++;

      if (pr.page.dc.d1_idx) {
	REALLOC(int2, pr.page.dc.dc1l.cnt, pr.page.dc.dc1l.n);
      }
      else {
	CALLOC(int2, pr.page.dc.dc1l.cnt, pr.page.dc.dc1l.n);
      }
    }
  }
  else {
    if (pr.page.dc.d2_idx < XPLAB_PREFS_PAGE_DC_LIST_MAX) {
      pr.page.dc.d2_idx = pr.page.dc.dc2l.cnt++;

      if (pr.page.dc.d2_idx) {
	REALLOC(int2, pr.page.dc.dc2l.cnt, pr.page.dc.dc2l.n);
      }
      else {
	CALLOC(int2, pr.page.dc.dc2l.cnt, pr.page.dc.dc2l.n);
      }
    }
  }

  /* setup new list entry from current settings                              */
  g_signal_emit_by_name((gpointer)pr.page.dc.d1.pos_ccf_button,
			"toggled");    /* use any button triggering right CB */
}



/* Removes an option from the option list.

Revision: 05.08.2005 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_remove(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  int cmp;

  assert(user_data);
  assert(button);

  cmp = strcmp(XPLAB_PREFS_PAGE_DC_OPTION_1D, user_data);

  if (!cmp) {
    if (pr.page.dc.dc1l.cnt) {
      delete_int2(&pr.page.dc.dc1l, pr.page.dc.d1_idx);
    }
  }
  else {
    if (pr.page.dc.dc2l.cnt) {
      delete_int2(&pr.page.dc.dc2l, pr.page.dc.d2_idx);
    }
  }

  update_dc_entries((char*)user_data, 0);
}



/* Decrements a setting.

Revision: 11.04.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_down(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  assert(user_data);
  assert(button);

  update_dc_entries((char*)user_data, -1);
}



/* Increments/Decrements the option list index and updates the entries
   and buttons accordingly.

Revision: 12.04.2004 E.Sack

Input : - pointer to option string ID
        - step
Output: - option list index updated
*/
static void
update_dc_entries(char *user_data, int incr)
{
  int cmp;

  cmp = strcmp(XPLAB_PREFS_PAGE_DC_OPTION_1D, user_data);

  if (!cmp) {
    if (pr.page.dc.dc1l.cnt) {
      pr.page.dc.d1_idx += incr;
      xwrap_idx(&pr.page.dc.d1_idx, pr.page.dc.dc1l.cnt);
    }
  }
  else {
    if (pr.page.dc.dc2l.cnt) {
      pr.page.dc.d2_idx += incr;
      xwrap_idx(&pr.page.dc.d2_idx, pr.page.dc.dc2l.cnt);
    }
  }

  update_dc_buttons_from_list();
  dc_option_entry_from_list_all();
}



/* Builds the format for the DC options list.

Revision: 05.08.2005 E.Sack

Input : - index
        - number of items
        - value
Output: - pointer to format string
Postcondition: index incremented by 1 [1, ...]
*/
static char *
dc_option_list_format(int idx, int cnt, int val)
{
  guint         len;
  char          num[256], *buf = NULL;
  unsigned char *ptr;

  g_snprintf(num, sizeof(num), "%08x", val);

  len = strlen(num);
  ptr = (unsigned char *)num;

  while (len--) {
    *ptr = toupper(*ptr);
    ptr++;
  }

  int_to_string("", idx+1, "/", &buf);              /* index starts with 1 */
  int_to_string("", cnt, ": 0x", &buf);
  add_string(num, &buf);

  return buf;
}



/* Increments a setting.

Revision: 11.04.2004 E.Sack

Input : - GTK convention
Output: - -
*/
static void
cb_dc_up(GTK_CONVENTION_BUTTON_RTS_VOID)
{
  assert(user_data);
  assert(button);

  update_dc_entries((char*)user_data, 1);
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
        - D2 buttons definition
*/
static d2i_def
dc_d2i_setting(GtkWidget *fbox1)
{
  d2i_def   d2i;
  GtkWidget *vbox, *hbox1, *vbox11, *vbox12;

  init_d2i_def(&d2i);

  /* put into inbox1 a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(fbox1), vbox);

  /* put into vertical box a horizontal box */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox1, FALSE, FALSE, 0);

  /* put into horizontal box1 a vertical box11 */
  vbox11 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox11), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_start(GTK_BOX(hbox1), vbox11, FALSE, FALSE, 0);

  /* put into horizontal box1 a vertical box12 */
  vbox12 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox12), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_end(GTK_BOX(hbox1), vbox12, FALSE, FALSE, 0);

  d2i.bmp_button = gtk_check_button_new_with_label(DC_OPT_BMP_DESC " (" DC_OPT_BMP_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox11), d2i.bmp_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2i.bmp_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2i.org_button = gtk_check_button_new_with_label(DC_OPT_ORG_DESC " (" DC_OPT_ORG_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox11), d2i.org_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2i.org_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2i.gv_button = gtk_check_button_new_with_label(DC_OPT_GV_DESC " (" DC_OPT_GV_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox11), d2i.gv_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2i.gv_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2i.dot_button = gtk_check_button_new_with_label(DC_OPT_DOT_DESC " (" DC_OPT_DOT_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox11), d2i.dot_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2i.dot_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2i.clean_button = gtk_check_button_new_with_label(DC_OPT_CLEAN_DESC " (" DC_OPT_CLEAN_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox12), d2i.clean_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2i.clean_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2i.pest_button = gtk_check_button_new_with_label(DC_OPT_PEST_DESC " (" DC_OPT_PEST_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox12), d2i.pest_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2i.pest_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2i.pest_max_button = gtk_check_button_new_with_label(DC_OPT_PEST_MAX_DESC " (" DC_OPT_PEST_MAX_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox12), d2i.pest_max_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2i.pest_max_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  dc_option_list(vbox, &pr.page.dc.d2i_entry, 
		 (gpointer)XPLAB_PREFS_PAGE_DC_OPTION_2D,
		 XPLAB_PREFS_PAGE_DC_2_LIST_EMPTY);

  return d2i;
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
        - D2 buttons definition
*/
static d2ii_def
dc_d2ii_setting(GtkWidget *fbox2)
{
  d2ii_def  d2ii;
  GtkWidget *vbox, *hbox2, *vbox21, *vbox22;

  init_d2ii_def(&d2ii);

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(fbox2), vbox);

  /* put into vertical box a horizontal box */
  hbox2 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox2, FALSE, FALSE, 0);

  /* put into horizontal box2 a vertical box21 */
  vbox21 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox21), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_start(GTK_BOX(hbox2), vbox21, FALSE, FALSE, 0);

  /* put into horizontal box2 a vertical box22 */
  vbox22 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox22), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_end(GTK_BOX(hbox2), vbox22, FALSE, FALSE, 0);

  d2ii.nb_good_button = gtk_check_button_new_with_label(DC_OPT_NB_GOOD_DESC " (" DC_OPT_NB_GOOD_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox21), d2ii.nb_good_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.nb_good_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_all_button = gtk_check_button_new_with_label(DC_OPT_RB_ALL_DESC " (" DC_OPT_RB_ALL_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox21), d2ii.rb_all_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_all_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_all_inv_button = gtk_check_button_new_with_label(DC_OPT_RB_ALL_INV_DESC " (" DC_OPT_RB_ALL_INV_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox21), d2ii.rb_all_inv_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_all_inv_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_good_button = gtk_check_button_new_with_label(DC_OPT_RB_GOOD_DESC " (" DC_OPT_RB_GOOD_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox21), d2ii.rb_good_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_good_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_good_inv_button = gtk_check_button_new_with_label(DC_OPT_RB_GOOD_INV_DESC " (" DC_OPT_RB_GOOD_INV_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox21), d2ii.rb_good_inv_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_good_inv_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_fail_button = gtk_check_button_new_with_label(DC_OPT_RB_FAIL_DESC " (" DC_OPT_RB_FAIL_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox21), d2ii.rb_fail_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_fail_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_fail_inv_button = gtk_check_button_new_with_label(DC_OPT_RB_FAIL_INV_DESC " (" DC_OPT_RB_FAIL_INV_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox21), d2ii.rb_fail_inv_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_fail_inv_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_frame_bmp_button = gtk_check_button_new_with_label(DC_OPT_RB_FRAME_BMP_DESC " (" DC_OPT_RB_FRAME_BMP_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.rb_frame_bmp_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_frame_bmp_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.rb_frame_rgb_button = gtk_check_button_new_with_label(DC_OPT_RB_FRAME_RGB_DESC " (" DC_OPT_RB_FRAME_RGB_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.rb_frame_rgb_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.rb_frame_rgb_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.mrb_frame_bmp_button = gtk_check_button_new_with_label(DC_OPT_MRB_FRAME_BMP_DESC " (" DC_OPT_MRB_FRAME_BMP_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.mrb_frame_bmp_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.mrb_frame_bmp_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.mrb_frame_rgb_button = gtk_check_button_new_with_label(DC_OPT_MRB_FRAME_RGB_DESC " (" DC_OPT_MRB_FRAME_RGB_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.mrb_frame_rgb_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.mrb_frame_rgb_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.hline_bmp_button = gtk_check_button_new_with_label(DC_OPT_HLINE_BMP_DESC " (" DC_OPT_HLINE_BMP_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.hline_bmp_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.hline_bmp_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.hline_rgb_button = gtk_check_button_new_with_label(DC_OPT_HLINE_RGB_DESC " (" DC_OPT_HLINE_RGB_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.hline_rgb_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.hline_rgb_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.sline_bmp_button = gtk_check_button_new_with_label(DC_OPT_SLINE_BMP_DESC " (" DC_OPT_SLINE_BMP_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.sline_bmp_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.sline_bmp_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  d2ii.sline_rgb_button = gtk_check_button_new_with_label(DC_OPT_SLINE_RGB_DESC " (" DC_OPT_SLINE_RGB_FILE ")");
  gtk_box_pack_start(GTK_BOX(vbox22), d2ii.sline_rgb_button, FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(d2ii.sline_rgb_button), "toggled", 
		   G_CALLBACK(cb_dc_button_to_entry), NULL);

  dc_option_list(vbox, &pr.page.dc.d2ii_entry, 
		 (gpointer)XPLAB_PREFS_PAGE_DC_OPTION_2D,
		 XPLAB_PREFS_PAGE_DC_2_LIST_EMPTY);

  return d2ii;
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
replay_filename_setting(GtkWidget *cbox)
{
  char      *replay_filename = NULL;
  GtkWidget *hbox, *label;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new("replay file name: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  replay_filename = default_replay_filename(xctrl.set.image_filename);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, replay_filename);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);

  FREEMEM(replay_filename);
}



/* Creates a preference setting box.

Revision: 18.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
replay_write_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, REPLAY_WRITE_KEY, REPLAY_WRITE_REM, 0);

  pr.page.io.replay_write_button = gtk_check_button_new_with_label("write");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.io.replay_write_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 17.10.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
single_line_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, SINGLE_LINE_KEY, SINGLE_LINE_REM, 0);

  pr.page.matching.single_line_button = gtk_check_button_new_with_label("single line");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.matching.single_line_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 17.10.2004 E.Sack

Input : - pointer to container a,b
Output: - container updated
*/
static void
single_nb_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, SINGLE_NB_KEY, SINGLE_NB_REM, 0);

  pr.page.matching.single_nb_button = gtk_check_button_new_with_label("single NB");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.matching.single_nb_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 03.01.2005 E.Sack

Input : - pointer to container a,b
Output: - container updated
*/
static void
broken_nb_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, BROKEN_NB_KEY, BROKEN_NB_REM, 0);

  pr.page.matching.broken_nb_button = gtk_check_button_new_with_label("broken NB");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.matching.broken_nb_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 27.12.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
renew_method_setting(GtkWidget **abox)
{
  *abox = gtk_hbox_new(FALSE, 0);

  pr.page.matching.renew_method_button = gtk_check_button_new_with_label("renew");
  gtk_box_pack_start(GTK_BOX(*abox), pr.page.matching.renew_method_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 18.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cluster_force_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, CLUSTER_FORCE_KEY, CLUSTER_FORCE_REM, 0);

  pr.page.correlation.cluster_force_button = gtk_check_button_new_with_label("force");
  gtk_box_pack_end(GTK_BOX(hbox), pr.page.correlation.cluster_force_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 18.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cluster_write_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, CLUSTER_WRITE_KEY, CLUSTER_WRITE_REM, 0);

  pr.page.correlation.cluster_write_button = gtk_check_button_new_with_label("write");
  gtk_box_pack_end(GTK_BOX(hbox), pr.page.correlation.cluster_write_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 12.10.2003 E.Sack

Input : - pointer to container widget A
        - pointer to container widget B
Output: - container updated
*/
static void
timestamp_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, TIMESTAMP_KEY, TIMESTAMP_REM, 0);

  pr.page.io.timestamp_button = gtk_check_button_new_with_label("use");
  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.io.timestamp_button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pr.page.io.timestamp_button), TRUE);
}



/* Creates a preference setting box.

Revision: 08.10.2004 E.Sack

Input : - pointer to container widget A
        - pointer to container widget B
Output: - container updated
*/
static void
writeauto_setting(GtkWidget *vbox)
{
  GtkWidget *abox;

  abox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), abox, FALSE, FALSE, 2);

  setting_key(abox, WRITEAUTO_KEY, WRITEAUTO_REM, 0);

  pr.page.io.writeauto_button = gtk_check_button_new_with_label("auto");
  gtk_box_pack_start(GTK_BOX(abox), pr.page.io.writeauto_button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pr.page.io.writeauto_button), FALSE);
}



/* Creates a preference setting box.

Revision: 13.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
domain_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, DOMAIN_KEY, DOMAIN_REM, 0);

  pr.page.correlation.domain_button = gtk_check_button_new_with_label("FFT");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.correlation.domain_button, FALSE, FALSE, 0);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pr.page.correlation.domain_button), TRUE);

  if (DOMAIN_SUPPRESS_OTHERS) {
    gtk_widget_set_sensitive(pr.page.correlation.domain_button, FALSE); 
  }
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
verbose_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, VERBOSE_KEY, VERBOSE_REM, 0);    /* needs horizontal box */

  pr.page.general.verbose_combo = preference_menu_verbose();
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.verbose_combo, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 05.05.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
tooltips_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, TIPS_KEY, TIPS_REM, 0);

  pr.page.general.tips_button = gtk_check_button_new_with_label("show");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.tips_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 23.11.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
raise_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, RAISE_KEY, RAISE_REM, 0);

  pr.page.general.raise_button = gtk_check_button_new_with_label("raise");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.general.raise_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
line_method_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, LINE_METHOD_KEY, LINE_METHOD_REM, 0);

  pr.page.matching.line_method_combo = preference_menu_line_method();

  gtk_box_pack_start(GTK_BOX(*bbox), pr.page.matching.line_method_combo, FALSE, FALSE, 0);
}



/* Builds a menu bar.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - menu bar
*/
static GtkWidget*
preference_menu_verbose(void)
{
  GtkWidget *combo_txt;

  combo_txt = combo_verbose();

  return combo_txt;
}



/* Builds a menu bar.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - menu bar
*/
static GtkWidget*
preference_menu_line_method(void)
{
  GtkWidget *combo_txt;

  combo_txt = combo_line_method();

  return combo_txt;
}



/* Builds an entry for a menu bar.

Revision: 05.08.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static GtkWidget *
combo_verbose(void)
{
  gint      itm;
  gchar     *gtxt = NULL;
  GtkWidget *combo_txt;

  assert(VERBOSE_LAST>VERBOSE_FIRST+1);

  combo_txt = gtk_combo_box_new_text();

  for (itm = VERBOSE_FIRST+1; itm < VERBOSE_LAST; itm++) {
    FREEMEM(gtxt);
    int_to_string(XPLAB_PREFS_VERBOSE_FORMAT_PREFIX, itm, 
		  XPLAB_PREFS_VERBOSE_FORMAT_SUFFIX, &gtxt);
    gtk_combo_box_append_text(GTK_COMBO_BOX(combo_txt), gtxt);
  }

  FREEMEM(gtxt);

  gtk_combo_box_set_active(GTK_COMBO_BOX(combo_txt),         /* emits signal */
			   VERBOSE_DEFAULT-1);

  return combo_txt;
}



/* Builds an entry for a menu bar.

Revision: 05.08.2005 E.Sack

Input : - menu bar
Output: - menu bar entry
*/
static GtkWidget*
combo_line_method(void)
{
  int       itm, idx = 0;
  char      *buf = NULL;
  GtkWidget *combo_txt;

  assert(METHOD_LAST>METHOD_FIRST);

  combo_txt = gtk_combo_box_new_text();

  for (itm = METHOD_FIRST; itm < METHOD_LAST; itm++, idx++) {
    inline_method_to_string(itm, &buf);

    gtk_combo_box_append_text(GTK_COMBO_BOX(combo_txt), buf);

    FREEMEM(buf);
  }

  gtk_combo_box_set_active(GTK_COMBO_BOX(combo_txt),         /* emits signal */ 
			   METHOD_FIRST);

  g_signal_connect(G_OBJECT(combo_txt), "changed", 
		   G_CALLBACK(cb_line_method_entry), NULL);

  return combo_txt;
}



/* Creates a preference setting box.

Revision: 10.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
subst_rb_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, SUBSTRB_KEY, SUBSTRB_REM, 0);

  *bbox = pentry(&pr.page.matching.subst_rb_entry, TRUE, 
		 GTK_SIGNAL_FUNC(cb_subst_rb_entry), xctrl.cfg.subst_rb, 
		 XPLAB_PREFS_SUBSTITUTION_CHARS, XPLAB_ENTRY_LEN_CHARS,
		 GTK_SIGNAL_FUNC(cb_subst_rb_revert), "R",
		 GTK_SIGNAL_FUNC(cb_subst_rb_default), "D", 
		 NULL, NULL, NULL);
}



/* Creates a preference setting box.

Revision: 10.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
subst_nb_setting(GtkWidget **abox, GtkWidget **bbox)
{
  *abox = gtk_hbox_new(FALSE, 0);
  *bbox = gtk_hbox_new(FALSE, 0);

  setting_key(*abox, SUBSTNB_KEY, SUBSTNB_REM, 0);

  *bbox = pentry(&pr.page.matching.subst_nb_entry, TRUE, 
		 GTK_SIGNAL_FUNC(cb_subst_nb_entry), xctrl.cfg.subst_nb, 
		 XPLAB_PREFS_SUBSTITUTION_CHARS, XPLAB_ENTRY_LEN_CHARS,
		 GTK_SIGNAL_FUNC(cb_subst_nb_revert), "R",
		 GTK_SIGNAL_FUNC(cb_subst_nb_default), "D", 
		 NULL, NULL, NULL);
}



/* Creates a preference setting box.

Revision: 11.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
rgb_norm_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox, *hbox1, *label, *vbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox1, FALSE, FALSE, 2);

  setting_key(hbox1, RGB_NORM_KEY, RGB_NORM_REM, 0);
  setting_double_limits(hbox1, RGB_NORM_DEFAULT, RGB_NORM_MIN, RGB_NORM_MAX);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_end(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

  double_to_string("", RGB_NORM_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);

  /* entry for red color */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 1);
  double_to_string("", xctrl.cfg.rgb_norm.r, "", &buf);
  box = pentry(&pr.page.preprocessing.rgb_norm_r_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_r_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_rgb_norm_r_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_r_default), "D", 
	       NULL, NULL, NULL);
  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  label = gtk_label_new("red");
  gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  FREEMEM(buf);

  /* entry for green color */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 1);
  double_to_string("", xctrl.cfg.rgb_norm.g, "", &buf);
  box = pentry(&pr.page.preprocessing.rgb_norm_g_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_g_entry), buf, len, XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_rgb_norm_g_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_g_default), "D", 
	       NULL, NULL, NULL);
  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  label = gtk_label_new("green");
  gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  FREEMEM(buf);

  /* entry for blue color */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 1);
  double_to_string("", xctrl.cfg.rgb_norm.b, "", &buf);
  box = pentry(&pr.page.preprocessing.rgb_norm_b_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_b_entry), buf, len, XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_rgb_norm_b_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_b_default), "D", 
	       NULL, NULL, NULL);
  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  label = gtk_label_new("blue");
  gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  FREEMEM(buf);

  /* entry for total */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 1);
  double_to_string("", xctrl.cfg.rgb_norm.r+xctrl.cfg.rgb_norm.g+xctrl.cfg.rgb_norm.b, "", &buf);
  box = pentry(&pr.page.preprocessing.rgb_norm_total_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_total_entry), buf, len, XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_rgb_norm_total_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_rgb_norm_total_default), "D", 
	       NULL, NULL, NULL);
  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  label = gtk_label_new("total");
  gtk_box_pack_end(GTK_BOX(hbox), label, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 06.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
histo_thold_setting(GtkWidget *cbox)
{
  int       len;
  char      *buf1 = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, THOLD_KEY, THOLD_REM, 0);
  setting_integer_limits(hbox, THOLD_DEFAULT, THOLD_MIN, THOLD_MAX);

  int_to_string("", THOLD_MAX, "", &buf1);
  len = strlen(buf1);
  FREEMEM(buf1);
  int_to_string("", xctrl.cfg.histo_thold, "", &buf1);

  box = pentry(&pr.page.preprocessing.histo_thold_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_histo_thold_entry), buf1, len, 
	       XPLAB_ENTRY_LEN_INT3,
	       GTK_SIGNAL_FUNC(cb_histo_thold_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_histo_thold_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf1);
}



/* Creates a preference setting box.

Revision: 29.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
corrlevel_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox1;

  /* put into inbox a horizontal box */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox1, FALSE, FALSE, 0);

  setting_key(hbox1, LEVEL_KEY, LEVEL_REM, 0);
  setting_double_limits(hbox1, LEVEL_DEFAULT, LEVEL_MIN, LEVEL_MAX);

  double_to_string("", LEVEL_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  double_to_string("", xctrl.cfg.mad.corrlevel, "", &buf);

  box = pentry(&pr.page.correlation.corrlevel_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_corrlevel_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_corrlevel_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_corrlevel_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox1), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 05.01.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
corrbroken_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox1;

  /* put into inbox a horizontal box */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox1, FALSE, FALSE, 0);

  setting_key(hbox1, BROKEN_LEVEL_KEY, BROKEN_LEVEL_REM, 0);
  setting_double_limits(hbox1, BROKEN_LEVEL_DEFAULT, BROKEN_LEVEL_MIN,
			BROKEN_LEVEL_MAX);

  double_to_string("", BROKEN_LEVEL_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  double_to_string("", xctrl.cfg.mad.corrbroken, "", &buf);

  box = pentry(&pr.page.correlation.corrbroken_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_corrbroken_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_corrbroken_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_corrbroken_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox1), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cfgname_setting(GtkWidget *cbox)
{
  GtkWidget *hbox, *label;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new("preferences filespec: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, xctrl.set.config_filespec);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
batchname_setting(GtkWidget *cbox)
{
  GtkWidget *hbox, *label;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new("batch log file name: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, BATCHNAME_DEFAULT);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
logname_setting(GtkWidget *cbox)
{
  guint         len;
  char          *buf = NULL;
  GtkWidget     *hbox;
  GtkEntry      *entry;
  GtkTextBuffer *buffer; 

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(main_log));/* fetch default buffer */
  len = gtk_text_buffer_get_char_count(GTK_TEXT_BUFFER(buffer));

  int_to_string("", len, "", &buf);
  cat_string(" char", len==1?"":"s", " in log", &buf);

  setting_key(hbox, "main log file name", buf, 0);

  FREEMEM(buf);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, XPLAB_MAIN_LOGNAME_DEFAULT);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);
}



/* Creates a preference setting box.

Revision: 17.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
tps_id_setting(GtkWidget *cbox)
{
  GtkWidget *hbox, *label;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new(TPSID_REM ":");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  label = gtk_label_new(xctrl.cfg.tps_id);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);
}



/* Creates a preference setting box.

Revision: 05.01.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
correlation_setting(GtkWidget *cbox)
{
  GtkWidget *vbox;

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(cbox), vbox);

  corrlevel_setting(vbox);
  corrbroken_setting(vbox);
  domain_setting(vbox);
}



/* Creates a preference setting box.

Revision: 29.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
clustering_setting(GtkWidget *cbox)
{
  GtkWidget *vbox;

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(cbox), vbox);

  cluster_level_setting(vbox);
  cluster_step_setting(vbox);
  cluster_thold_setting(vbox);
  ckernel_setting(vbox);
  cluster_force_setting(vbox);
  cluster_write_setting(vbox);
  cluster_nb_setting(vbox);
}



/* Creates a preference setting box.

Revision: 30.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
matching_rules_setting(GtkWidget *cbox)
{
  GtkWidget *vbox;

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(cbox), vbox);

  mborder_setting(vbox);
  nborder_setting(vbox);
  maskweight_setting(vbox);
  nnbmatch_setting(vbox);
  interleave_setting(vbox);
  overlap_setting(vbox);
  symbolspacing_setting(vbox);
  nb_lap_setting(vbox);
}



/* Creates a preference setting box.

Revision: 16.10.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
general_matching_setting(GtkWidget *cbox)
{
  int       rows = 7, cols = 3;
  GtkWidget *abox, *bbox;

  /* put into inbox a table */
  pr.page.matching.table_general = gtk_table_new(rows, cols, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(pr.page.matching.table_general),
				 XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(cbox), pr.page.matching.table_general);

  single_line_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), bbox, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

  single_nb_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), bbox, 1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 0);

  broken_nb_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), bbox, 1, 2, 2, 3, GTK_FILL, GTK_FILL, 0, 0);

  duration_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 0, 1, 3, 4, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), bbox, 1, 2, 3, 4, GTK_FILL, GTK_FILL, 0, 0);

  line_method_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 0, 1, 4, 5, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), bbox, 1, 2, 4, 5, GTK_FILL, GTK_FILL, 0, 0);

  renew_method_setting(&abox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 2, 3, 4, 5, GTK_FILL, GTK_FILL, 5, 0);

  subst_rb_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 0, 1, 5, 6, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), bbox, 1, 2, 5, 6, GTK_FILL, GTK_FILL, 0, 0);

  subst_nb_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), abox, 0, 1, 6, 7, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(pr.page.matching.table_general), bbox, 1, 3, 6, 7, GTK_FILL, GTK_FILL, 0, 0);

  set_sensitivity_renew_method();
}



/* Creates a preference setting box.

Revision: 14.01.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
general_io_setting(GtkWidget *cbox)
{
  GtkWidget *vbox, *sep;

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(cbox), vbox);

  prg_setting(vbox);
  rc_setting(vbox);

  sep = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 2);

  cwd_setting(vbox);
  db_setting(vbox);

  cfgname_setting(vbox);

  logname_setting(vbox);

  batchname_setting(vbox);

  sep = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox), sep, FALSE, FALSE, 2);

  tps_id_setting(vbox);
}



/* Creates a preference setting box.

Revision: 08.10.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
report_setting(GtkWidget *inbox)
{
  GtkWidget *vbox;

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(inbox), vbox);

  /* put into vertical box a text entry */
  report_filename_setting(vbox);

  /* put into vertical box a table */
  report_file_table(vbox);

  /* put into vertical box a text entry and a setting */
  writeauto_filename_setting(vbox);
  writeauto_setting(vbox);
}



/* Creates a preference setting box.

Revision: 08.10.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
report_file_table(GtkWidget *vbox)
{
  int       rows = 4, cols = 2;
  GtkWidget *abox, *bbox, *cbox, *dbox, *table;

  table = gtk_table_new(rows, cols, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), 0);
  gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 2);

  report_write_setting(&abox, &bbox, &cbox, &dbox);
  gtk_table_attach(GTK_TABLE(table), abox, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(table), bbox, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(table), cbox, 2, 3, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(table), dbox, 3, 4, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

  timestamp_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(table), abox, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(table), bbox, 1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
}



/* Creates a preference setting box.

Revision: 28.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
histo_setting(GtkWidget *cbox)
{
  GtkWidget *hdis, *vbox;

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(cbox), vbox);

  separation_rules_filename_setting(vbox);

  /* put into vertical box a horizontal distance */
  hdis = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hdis, FALSE, FALSE, XPLAB_IMG_BORDER_SIZE);

  separation_histo_filename_setting(vbox);

  /* put into vertical box a horizontal distance */
  hdis = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hdis, FALSE, FALSE, XPLAB_IMG_BORDER_SIZE);

  histo_write_setting(vbox);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
replay_setting(GtkWidget *fbox)
{
  GtkWidget *vbox, *frame;

  /* put into inbox a frame */
  frame = gtk_frame_new("replay");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_start(GTK_BOX(fbox), frame, FALSE, FALSE, 0);

  /* put into frame a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  replay_filename_setting(vbox);
  replay_write_setting(vbox);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
training_setting(GtkWidget *fbox)
{
  GtkWidget *vbox, *frame;

  /* put into inbox a frame */
  frame = gtk_frame_new("training");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_start(GTK_BOX(fbox), frame, FALSE, FALSE, 0);

  /* put into frame a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  training_session_filename(vbox);
  training_write_setting(vbox);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
training_session_filename(GtkWidget *cbox)
{
  char      *session_filename = NULL;
  GtkWidget *hbox, *label;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new("session file name: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  session_filename = default_session_filename(xctrl.set.image_filename);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, session_filename);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);

  FREEMEM(session_filename);
}



/* Creates a preference setting box.

Revision: 09.07.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
training_write_setting(GtkWidget *cbox)
{
  GtkWidget *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, TRAINING_WRITE_KEY, TRAINING_WRITE_REM, 0);

  pr.page.io.training_write_button = gtk_check_button_new_with_label("write");
  gtk_box_pack_start(GTK_BOX(hbox), pr.page.io.training_write_button, FALSE, FALSE, 0);
}



/* Creates a preference setting box.

Revision: 05.12.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
prg_setting(GtkWidget *cbox)
{
  int       err;
  char      *buf = NULL;
  GtkWidget *hbox, *label;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new("program: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  err = check_name(xctrl.set.xprog);

  if (err) {
    add_string("*** cannot get program path ***", &buf);
  }
  else {
    add_string(xctrl.set.xprog, &buf);
  }

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, buf);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);

  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 11.09.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
rc_setting(GtkWidget *cbox)
{
  int       err;
  char      *buf = NULL;
  GtkWidget *hbox;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, XPLAB_GTK2_RC_KEY, XPLAB_GTK2_RC_REM, 0);

  err = check_name(xctrl.set.gtk_rc);         /* file spec of Gtk resource */

  if (err) {
    add_string("*** Gtk resource file spec undefined ***", &buf);
  }
  else {
    add_string(xctrl.set.gtk_rc, &buf);
  }

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, buf);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);

  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 21.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cwd_setting(GtkWidget *cbox)
{
  char      *path;
  GtkWidget *hbox, *label;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new("CWD: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  path = xplab_getcwd();

  if (!path) {
    add_string("*** cannot get CWD ***", &path);
  }

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, path);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);

  FREEMEM(path);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
db_setting(GtkWidget *cbox)
{
  GtkWidget *hbox, *label;
  GtkEntry  *entry;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  label = gtk_label_new("DB file name: ");
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 2);

  entry = GTK_ENTRY(gtk_entry_new());
  gtk_editable_set_editable(GTK_EDITABLE(entry), FALSE);
  gtk_entry_set_max_length(entry, XPLAB_PREFS_ENTRY_CHARS_PATH);
  xplab_entry_set_text(entry, pr.xplab_db.db_fname);

  gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(entry), TRUE, TRUE, 2);
}



/* Creates a preference setting box.

Revision: 22.08.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
image_setting(GtkWidget *cbox)
{
  GtkWidget *abox, *bbox, *table, *vbox;

  /* put into inbox a vertical box */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(cbox), vbox);

  /* put into vertical box a label */
  image_filename_setting(&abox);
  gtk_box_pack_start(GTK_BOX(vbox), abox, FALSE, FALSE, 2);

  /* put into vertical box a table */
  table = gtk_table_new(2, 2, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), 0);
  gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 2);

  rawbit_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(table), abox, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(table), bbox, 1, 2, 0, 1, GTK_FILL, GTK_FILL, 0, 0);

  overwrite_image_setting(&abox, &bbox);
  gtk_table_attach(GTK_TABLE(table), abox, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
  gtk_table_attach(GTK_TABLE(table), bbox, 1, 2, 1, 2, GTK_FILL, GTK_FILL, 0, 0);
}



/* Creates a temporary text buffer.

Revision: 12.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static int
errors_setting(int esum)
{
  int err;

  err = check_name(pr.xerrors) && esum;

  if (err) {
    int_to_string("\n", esum, esum==1?" error":" errors", &pr.xerrors);

    cat_string(" found parsing preferences file '",
	       xctrl.set.config_filespec, "'\n\n", &pr.xerrors);

    xplab_text_insert(main_log, pr.xerrors);
  }

  FREEMEM(pr.xerrors);

  return err;
}



/* Creates a preference setting box.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
ckernel_setting(GtkWidget *cbox)
{
  gint      len, rr, row, col;
  char      *buf1 = NULL;
  GtkWidget *box, *hbox, *hbox1, *vbox, *kbin, *table;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox), hbox1, FALSE, FALSE, 2);

  setting_key(hbox1, CKERNEL_KEY, CKERNEL_REM, 0);
  setting_integer_limits_no_default(hbox1, CKERNEL_MIN, CKERNEL_MAX);

  int_to_string("", CKERNEL_MAX, "", &buf1);
  len = strlen(buf1) + 1;

  box = pentry(NULL, FALSE, NULL, NULL, 0, 0,
	       GTK_SIGNAL_FUNC(cb_ckernel_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_ckernel_default), "D", 
	       NULL, NULL, NULL);
  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);

  table = gtk_table_new(3, 3, TRUE);
  gtk_box_pack_end(GTK_BOX(hbox), table, FALSE, FALSE, 2);

  for (row = 0; row < CKERNEL_H; row++) {   /* create raster input for table */
    rr = row*CKERNEL_W;

    for (col = 0; col < CKERNEL_W; col++) {
      FREEMEM(buf1);
      int_to_string("", xctrl.cfg.ckernel[col + rr], "", &buf1);

      kbin = pentry(&pr.page.correlation.ckern_entry[row][col], TRUE, 
		    GTK_SIGNAL_FUNC(cb_ckernel_entry), buf1, len, 
		    XPLAB_ENTRY_LEN_INT1,
		    NULL, NULL, NULL, NULL, NULL, NULL, NULL);

      /* give the CB infos about its anonymous entry                      */
      g_object_set_data(G_OBJECT(pr.page.correlation.ckern_entry[row][col]), 
			XPLAB_ASSN_KEY_ROW, GINT_TO_POINTER(row));

      g_object_set_data(G_OBJECT(pr.page.correlation.ckern_entry[row][col]), 
			XPLAB_ASSN_KEY_COL, GINT_TO_POINTER(col));

      gtk_table_attach_defaults(GTK_TABLE(table), kbin, col, col+1, row, row+1);
    }
  }

  FREEMEM(buf1);
}



/* Creates a preference setting box.

Revision: 08.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cluster_level_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox;

  /* put into inbox a horizontal box */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, CLUSTER_LVL_KEY, CLUSTER_LVL_REM, 0);
  setting_double_limits(hbox, CLUSTER_LVL_DEFAULT, CLUSTER_LVL_MIN, CLUSTER_LVL_MAX);

  double_to_string("", CLUSTER_LVL_MAX, "", &buf);
  len = strlen(buf);                                 /* check maximal length */

  FREEMEM(buf);

  double_to_string("", xctrl.cfg.cluster_level, "", &buf);

  box = pentry(&pr.page.correlation.cluster_level_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_cluster_level_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_cluster_level_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_cluster_level_default), "D", 
	       NULL, NULL, NULL);

  FREEMEM(buf);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
}



/* Creates a preference setting box.

Revision: 08.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cluster_step_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, CLUSTER_STEP_KEY, CLUSTER_STEP_REM, 0);
  setting_double_limits(hbox, CLUSTER_STEP_DEFAULT, CLUSTER_STEP_MIN, CLUSTER_STEP_MAX);

  double_to_string("", CLUSTER_STEP_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  double_to_string("", xctrl.cfg.cluster_step, "", &buf);

  box = pentry(&pr.page.correlation.cluster_step_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_cluster_step_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_cluster_step_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_cluster_step_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 08.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
cluster_thold_setting(GtkWidget *cbox)
{
  int       len;
  char      *buf1 = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, CHOLD_KEY, CHOLD_REM, 0);
  setting_integer_limits(hbox, CHOLD_DEFAULT, CHOLD_MIN, CHOLD_MAX);

  int_to_string("", CHOLD_MAX, "", &buf1);
  len = strlen(buf1);
  FREEMEM(buf1);
  int_to_string("", xctrl.cfg.cluster_thold, "", &buf1);

  box = pentry(&pr.page.correlation.cluster_thold_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_cluster_thold_entry), buf1, len, 
	       XPLAB_ENTRY_LEN_INT2,
	       GTK_SIGNAL_FUNC(cb_cluster_thold_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_cluster_thold_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf1);
}



/* Creates a preference setting box.

Revision: 05.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
mborder_setting(GtkWidget *cbox)
{
  int       len;
  char      *buf1 = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, MBORDER_KEY, MBORDER_REM, 0);
  setting_integer_limits(hbox, MBORDER_DEFAULT, MBORDER_MIN, MBORDER_MAX);

  int_to_string("", MBORDER_MAX, "", &buf1);
  len = strlen(buf1);
  FREEMEM(buf1);
  int_to_string("", xctrl.cfg.mad.mborder, "", &buf1);

  box = pentry(&pr.page.matching.mborder_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_mborder_entry), buf1, len, 
	       XPLAB_ENTRY_LEN_INT2,
	       GTK_SIGNAL_FUNC(cb_mborder_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_mborder_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf1);
}



/* Creates a preference setting box.

Revision: 05.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
nborder_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, NBORDER_KEY, NBORDER_REM, 0);
  setting_integer_limits(hbox, NBORDER_DEFAULT, NBORDER_MIN, NBORDER_MAX);

  int_to_string("", NBORDER_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  int_to_string("", xctrl.cfg.mad.nborder, "", &buf);

  box = pentry(&pr.page.matching.nborder_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_nborder_entry), buf, len, 
	       XPLAB_ENTRY_LEN_INT2,
	       GTK_SIGNAL_FUNC(cb_nborder_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_nborder_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 29.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
dc_store_setting(GtkWidget *fbox)
{
  int       len;
  char      *buf1 = NULL;
  GtkWidget *box, *hbox1, *vbox1;

  /* put into inbox a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox1), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(fbox), vbox1);

  /* put into vertical box a horizontal box */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);

  setting_key(hbox1, DC_STORE_KEY, DC_STORE_REM, 0);
  setting_integer_limits(hbox1, DC_STORE_DEFAULT, DC_STORE_MIN, DC_STORE_MAX);

  int_to_string("", DC_STORE_MAX, "", &buf1);
  len = strlen(buf1);
  FREEMEM(buf1);
  int_to_string("", xctrl.cfg.dc_store, "", &buf1);

  box = pentry(&pr.page.dc.store_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_dc_store_entry), buf1, len, 
	       XPLAB_ENTRY_LEN_INT2,
	       GTK_SIGNAL_FUNC(cb_dc_store_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_dc_store_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox1), box, FALSE, FALSE, 2);
  FREEMEM(buf1);
}



/* Creates a preference setting box.

Revision: 09.07.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
phases_setting(GtkWidget *vbox)
{
  training_setting(vbox);
  replay_setting(vbox);
  maintenance_setting(vbox);
}



/* Creates a preference setting box.

Revision: 24.02.2004 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
maintenance_setting(GtkWidget *fbox)
{
  int       len;
  char      *buf1 = NULL;
  GtkWidget *frame, *box, *hbox1, *vbox1;

  /* put into inbox a frame */
  frame = gtk_frame_new("maintenance");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame), XPLAB_IMG_BORDER_SIZE);
  gtk_box_pack_start(GTK_BOX(fbox), frame, FALSE, FALSE, 0);

  /* put into frame a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox1), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), vbox1);

  /* put into vertical box a horizontal box */
  hbox1 = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);

  setting_key(hbox1, NB_BORDER_KEY, NB_BORDER_REM, 0);
  setting_integer_limits(hbox1, NB_BORDER_DEFAULT, NB_BORDER_MIN, NB_BORDER_MAX);

  int_to_string("", NB_BORDER_MAX, "", &buf1);
  len = strlen(buf1);
  FREEMEM(buf1);
  int_to_string("", xctrl.cfg.nb_border_y, "", &buf1);

  box = pentry(&pr.page.io.nb_border_y_entry, TRUE,
	       GTK_SIGNAL_FUNC(cb_nb_border_entry), buf1, len, 
	       XPLAB_ENTRY_LEN_INT2,
	       GTK_SIGNAL_FUNC(cb_nb_border_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_nb_border_default), "D", 
	       NULL, NULL, NULL);
  gtk_box_pack_end(GTK_BOX(hbox1), box, FALSE, FALSE, 2);

  FREEMEM(buf1);
  int_to_string("", xctrl.cfg.nb_border_x, "", &buf1);
  box = pentry(&pr.page.io.nb_border_x_entry, TRUE,
	       GTK_SIGNAL_FUNC(cb_nb_border_entry), buf1, len, 
	       XPLAB_ENTRY_LEN_INT2,
	       NULL, NULL, NULL, NULL, NULL, NULL, NULL);
  gtk_box_pack_end(GTK_BOX(hbox1), box, FALSE, FALSE, 2);
  FREEMEM(buf1);
}



/* Creates a preference setting box.

Revision: 05.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
maskweight_setting(GtkWidget *cbox)
{
  int       len;
  char      *buf1 = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, MASKWEIGHT_KEY, MASKWEIGHT_REM, 0);
  setting_integer_limits(hbox, MASKWEIGHT_DEFAULT, MASKWEIGHT_MIN, MASKWEIGHT_MAX);

  int_to_string("", MASKWEIGHT_MAX, "", &buf1);
  len = strlen(buf1);
  FREEMEM(buf1);
  int_to_string("", xctrl.cfg.mad.maskweight, "", &buf1);

  box = pentry(&pr.page.matching.maskweight_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_maskweight_entry), buf1, len, 
	       XPLAB_ENTRY_LEN_INT3,
	       GTK_SIGNAL_FUNC(cb_maskweight_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_maskweight_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf1);
}



/* Creates a preference setting box.

Revision: 10.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
nnbmatch_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, NNBMATCH_KEY, NNBMATCH_REM, 0);
  setting_integer_limits(hbox, NNBMATCH_DEFAULT, NNBMATCH_MIN, NNBMATCH_MAX);

  int_to_string("", NNBMATCH_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  int_to_string("", xctrl.cfg.mad.nnbmatch, "", &buf);

  box = pentry(&pr.page.matching.nnbmatch_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_nnbmatch_entry), buf, len, 
	       XPLAB_ENTRY_LEN_INT2,
	       GTK_SIGNAL_FUNC(cb_nnbmatch_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_nnbmatch_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 10.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
interleave_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, INTERLEAVE_KEY, INTERLEAVE_REM, 0);
  setting_double_limits(hbox, INTERLEAVE_DEFAULT, INTERLEAVE_MIN, INTERLEAVE_MAX);

  double_to_string("", INTERLEAVE_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  double_to_string("", xctrl.cfg.mad.interleave, "", &buf);

  box = pentry(&pr.page.matching.interleave_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_interleave_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_interleave_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_interleave_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 10.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
overlap_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, OVERLAP_KEY, OVERLAP_REM, 0);
  setting_double_limits(hbox, OVERLAP_DEFAULT, OVERLAP_MIN, OVERLAP_MAX);

  double_to_string("", OVERLAP_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  double_to_string("", xctrl.cfg.mad.overlap, "", &buf);

  box = pentry(&pr.page.matching.overlap_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_overlap_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_overlap_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_overlap_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 10.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
symbolspacing_setting(GtkWidget *cbox)
{
  int       len;
  char     *buf = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, SYSPC_KEY, SYSPC_REM, 0);
  setting_double_limits(hbox, SYSPC_DEFAULT, SYSPC_MIN, SYSPC_MAX);

  double_to_string("", SYSPC_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  double_to_string("", xctrl.cfg.syspc, "", &buf);

  box = pentry(&pr.page.matching.symbolspacing_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_symbolspacing_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_symbolspacing_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_symbolspacing_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference setting box.

Revision: 10.10.2003 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
nb_lap_setting(GtkWidget *cbox)
{
  int       len;
  char      *buf = NULL;
  GtkWidget *box, *hbox;

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(cbox), hbox, FALSE, FALSE, 2);

  setting_key(hbox, NB_LAP_KEY, NB_LAP_REM, 0);
  setting_double_limits(hbox, NB_LAP_DEFAULT, NB_LAP_MIN, NB_LAP_MAX);

  double_to_string("", NB_LAP_MAX, "", &buf);
  len = strlen(buf);
  FREEMEM(buf);
  double_to_string("", xctrl.cfg.mad.nb_lap, "", &buf);

  box = pentry(&pr.page.matching.nb_lap_entry, TRUE, 
	       GTK_SIGNAL_FUNC(cb_nb_lap_entry), buf, len, 
	       XPLAB_ENTRY_LEN_FLOAT,
	       GTK_SIGNAL_FUNC(cb_nb_lap_revert), "R", 
	       GTK_SIGNAL_FUNC(cb_nb_lap_default), "D", 
	       NULL, NULL, NULL);

  gtk_box_pack_end(GTK_BOX(hbox), box, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference label for integer limits.

Revision: 05.10.2003 E.Sack

Input : - pointer to container
        - default integer
        - minimum integer
        - maximum integer
Output: - container updated
*/
static void
setting_integer_limits_no_default(GtkWidget *cbox, int min, int max)
{
  char      *buf = NULL;
  GtkWidget *label;

  int_to_string("", min, "...", &buf);
  int_to_string("", max, "", &buf);

  label = gtk_label_new(buf);
  gtk_box_pack_start(GTK_BOX(cbox), label, FALSE, FALSE, 2);
  FREEMEM(buf);
}



/* Creates a preference label for double limits.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
        - default double
        - minimum double
        - maximum double
Output: - container updated
*/
static void
setting_double_limits(GtkWidget *cbox, double def, double min, double max)
{
  char      *buf = NULL, buf1[256];
  GtkWidget *label;

  g_snprintf(buf1, sizeof(buf1), XPLAB_PREFS_FLOAT_FORMAT, min);
  cat_string(buf1, "...", "", &buf);

  g_snprintf(buf1, sizeof(buf1), XPLAB_PREFS_FLOAT_FORMAT, max);
  cat_string(buf1, " [", "", &buf);

  g_snprintf(buf1, sizeof(buf1), XPLAB_PREFS_FLOAT_FORMAT, def);
  cat_string(buf1, "]", "", &buf);

  label = gtk_label_new(buf);
  gtk_box_pack_start(GTK_BOX(cbox), label, FALSE, FALSE, 2);

  FREEMEM(buf);
}



/* Creates the preferences page preprocessing.

Revision: 28.10.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget*
preferences_dialog_page_preprocessing(void)
{
  GtkWidget *vbox1, *vbox2, *frame;

  /* create a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox1), XPLAB_IMG_BORDER_SIZE);

  /* put into vertical box a frame */
  frame = gtk_frame_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox1), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  /* put into frame a vertical box */
  vbox2 = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox2), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), vbox2);

  histo_thold_setting(vbox2);
  rgb_norm_setting(vbox2);
  support_dots_setting(vbox2);

  return vbox1;
}



/* Creates the preferences page statistic.

Revision: 05.08.2005 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget*
preferences_dialog_page_statistic(void)
{
  GtkWidget *vbox1, *vbox2, *frame;

  /* create a vertical box */
  vbox1 = gtk_vbox_new(FALSE, 10);
  gtk_container_set_border_width(GTK_CONTAINER(vbox1), XPLAB_IMG_BORDER_SIZE);

  /* put into vertical box a frame */
  frame = gtk_frame_new(NULL);
  gtk_box_pack_start(GTK_BOX(vbox1), frame, TRUE, TRUE, 0);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);

  /* put into frame a vertical box */
  vbox2 = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox2), XPLAB_IMG_BORDER_SIZE);
  gtk_container_add(GTK_CONTAINER(frame), vbox2);

  stats_window_setting(vbox2, STATS_IORG_KEY, STATS_IORG_REM, &pr.page.statistic.owin);
  stats_window_setting(vbox2, STATS_ICOL_KEY, STATS_ICOL_REM, &pr.page.statistic.cwin);
  stats_window_setting(vbox2, STATS_IGRY_KEY, STATS_IGRY_REM, &pr.page.statistic.gwin);
  stats_window_setting(vbox2, STATS_IBMP_KEY, STATS_IBMP_REM, &pr.page.statistic.bwin);
  stats_window_setting(vbox2, STATS_HRBS_KEY, STATS_HRBS_REM, &pr.page.statistic.rhst);
  stats_window_setting(vbox2, STATS_HCOL_KEY, STATS_HCOL_REM, &pr.page.statistic.chst);
  stats_window_setting(vbox2, STATS_HGRY_KEY, STATS_HGRY_REM, &pr.page.statistic.ghst);
  stats_window_setting(vbox2, STATS_HBMP_KEY, STATS_HBMP_REM, &pr.page.statistic.bhst);
  stats_window_setting(vbox2, STATS_HSLOPE_KEY,STATS_HSLOPE_REM, &pr.page.statistic.shst);

  stats_icon_contrast_setting(vbox2, STATS_ICON_CONTRAST_KEY, STATS_ICON_CONTRAST_REM, &pr.page.statistic.contrast);

  return vbox1;
}



/* Creates the preferences page matching.

Revision: 03.10.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget*
preferences_dialog_page_matching(void)
{
  GtkWidget *notebook, *frame_gen, *frame_rules;

  /* create a notebook */
  notebook = gtk_notebook_new();

  /* put the first page into the notebook, automatically displayed         */
  frame_gen = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_gen), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_gen), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_gen, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_gen, XPLAB_PREFS_PAGE_MATCHING_GENERAL);

  /* put the last page into the notebook */
  frame_rules = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_rules), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_rules), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_rules, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_rules, XPLAB_PREFS_PAGE_MATCHING_RULES);

  general_matching_setting(frame_gen);
  matching_rules_setting(frame_rules);

  return notebook;
}



/* Creates the preferences page correlation.

Revision: 03.10.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget*
preferences_dialog_page_correlation(void)
{
  GtkWidget *notebook, *frame_cluster, *frame_gen;

  /* create a notebook */
  notebook = gtk_notebook_new();

  /* put the first page into the notebook, automatically displayed         */
  frame_gen = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_gen), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_gen), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_gen, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_gen, XPLAB_PREFS_PAGE_CORRELATION_GENERAL);

  /* put the second page into the notebook */
  frame_cluster = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_cluster), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_cluster), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_cluster, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_cluster, XPLAB_PREFS_PAGE_CORRELATION_CLUSTER);

  correlation_setting(frame_gen);
  clustering_setting(frame_cluster);

  return notebook;
}



/* Creates the preferences page documentation channel for images.

Revision: 18.10.2003 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget*
preferences_dialog_page_dc(void)
{
  GtkWidget *notebook, *frame_1d, *frame_2di, *frame_2dii, *frame_general;

  /* create a notebook */
  notebook = gtk_notebook_new();

  /* put the first page into the notebook, automatically displayed         */
  frame_general = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_general), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_general), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_general, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_general, XPLAB_PREFS_PAGE_DC_GENERAL);

  /* put the second page into the notebook */
  frame_1d = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_1d), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_1d), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_1d, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_1d, XPLAB_PREFS_PAGE_DC_1D);

  /* put the third page into the notebook */
  frame_2di = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_2di), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_2di), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_2di, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_2di, XPLAB_PREFS_PAGE_DC_2DI);

  /* put the last page into the notebook */
  frame_2dii = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_2dii), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_2dii), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_2dii, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_2dii, XPLAB_PREFS_PAGE_DC_2DII);

  dc_store_setting(frame_general);
  pr.page.dc.d1 = dc_d1_setting(frame_1d);
  pr.page.dc.d2i = dc_d2i_setting(frame_2di);
  pr.page.dc.d2ii = dc_d2ii_setting(frame_2dii);

  return notebook;
}



/* Creates the preferences page I/O.

Revision: 24.02.2004 E.Sack

Input : - pointer to container box
Output: - pointer to page widget
*/
static GtkWidget*
preferences_dialog_page_io(void)
{
  GtkWidget *notebook, *frame_report, *frame_image;
  GtkWidget *frame_histo, *frame_gen, *vbox;

  /* create a notebook */
  notebook = gtk_notebook_new();

  /* put the first page into the notebook, automatically displayed         */
  frame_gen = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_gen), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_gen), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_gen, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_gen, XPLAB_PREFS_PAGE_IO_GENERAL);

  /* put the second page into the notebook */
  frame_report = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_report), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_report), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_report, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_report, XPLAB_PREFS_PAGE_IO_REPORT);

  /* put the third page into the notebook */
  frame_image = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_image), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_image), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_image, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_image, XPLAB_PREFS_PAGE_IO_IMAGE);

  /* put the forth page into the notebook */
  frame_histo = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame_histo), GTK_SHADOW_ETCHED_IN);
  gtk_container_set_border_width(GTK_CONTAINER(frame_histo), XPLAB_IMG_BORDER_SIZE);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), frame_histo, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), frame_histo, XPLAB_PREFS_PAGE_IO_HISTO);

  /* put the last page into the notebook */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, NULL);
  gtk_notebook_set_tab_label_text(GTK_NOTEBOOK(notebook), vbox, XPLAB_PREFS_PAGE_IO_PHASE);

  general_io_setting(frame_gen);
  report_setting(frame_report);

  image_setting(frame_image);

  histo_setting(frame_histo);

  phases_setting(vbox);

  return notebook;
}



/* Creates the preferences action part.

Revision: 05.08.2005 E.Sack

Input : - pointer to container
Output: - container updated
*/
static void
preferences_dialog_lower(GtkContainer *inbox)
{
  GtkWidget *cancel_button;

  /* put into box a save button */
  pr.page.save_button = gtk_button_new_with_label(DIALOG_BUTTON_SAVE);
  gtk_container_add(inbox, pr.page.save_button);
  g_signal_connect(G_OBJECT(pr.page.save_button), "clicked", 
		   G_CALLBACK(cb_preferences_save_button), NULL);

  /* put into box an apply button */
  pr.page.apply_button = gtk_button_new_with_label(DIALOG_BUTTON_APPLY);
  gtk_container_add(inbox, pr.page.apply_button);
  g_signal_connect(G_OBJECT(pr.page.apply_button), "clicked", 
		   G_CALLBACK(cb_preferences_apply_button), NULL);

  /* put into box a revert button */
  pr.page.revert_button = gtk_button_new_with_label(DIALOG_BUTTON_REVERT);
  gtk_container_add(inbox, pr.page.revert_button);
  g_signal_connect(G_OBJECT(pr.page.revert_button), "clicked", 
		   G_CALLBACK(cb_preferences_revert_button), NULL);

  /* put into box a cancel button */
  cancel_button = gtk_button_new_with_label(DIALOG_BUTTON_CANCEL);
  gtk_container_add(inbox, cancel_button);
  g_signal_connect(G_OBJECT(cancel_button), "clicked", 
		   G_CALLBACK(cb_preferences_cancel_button), NULL);
}



/* Completes a filespec with the current working directory.

Revision: 05.04.2004 E.Sack

Input : - pointer to filespec
Output: - filespec updated
Postcondition: - if path of CWD can't be build filespec is left unchanged
*/
static void
xplab_absolute_filespec(char **filespec)
{
  char *path = NULL;

  assert(filespec);

  if (**filespec == UNIX_PATH_SEPARATOR) {          
    return;                        /* filespec is absolute - nothing to do */
  }

  path = xplab_getcwd();

  if (!path) {
    return;
  }

  add_character(UNIX_PATH_SEPARATOR, &path); /* separator not counted as part of the path */
  add_string(*filespec, &path);           /* complete to absolute filespec */
  FREEMEM(*filespec);                          /* free incomplete filespec */
  *filespec = path;                            /* return complete filespec */
}



/* Wrapper to get CWD.

Revision: 05.08.2005 E.Sack

Input : - -
Output: - pointer to string containing current working directory
        - message if unsuccessful
Postcondition: - pointer is NULL if path of CWD can't be get.
*/
static char *
xplab_getcwd(void)
{
  char *cwd, *path = NULL;

  CALLOC(char, XPLAB_PATH_LENGTH, path);

  cwd = getcwd(path, XPLAB_PATH_LENGTH);

  if (cwd == NULL) {
    xplab_text_insert(main_log,
		      "WARNING: cannot get path of current working directory\n");
    FREEMEM(path);
  }
  else {
    assert(*path == UNIX_PATH_SEPARATOR);                       /* what else */
  }

  assert(cwd==path);

  return path;
}



/* Reads the configuration file.

Revision: 09.01.2005 E.Sack

Input : - pointer to filename
        - pointer to settings
Output: - settings updated
Postcondition: - assumed initialized settings
Bug: - line index following continued lines is wrong
*/
static void
read_cfg(char *fspec, char **gtk_rc, cfg_def *cfg, stats_def *sfg)
{
  FILE *fin;
  int  line = FIRST_LINE_NBR, errsum = 0, clos;
  char *cbuf = NULL;

  pr.cfg_file_available = FILE_UNDEFINED;

  if (access(fspec, F_OK)) {
    pr.cfg_file_available = FILE_NOT_EXIST;
    cat_string("configuration file '", fspec,
	       "' not found, using defaults\n", &cbuf);
    xplab_text_insert(main_log, cbuf);
    FREEMEM(cbuf);
    return;
  }

  fin = xplab_open_file(fspec, "r");
  if (!fin) {
    pr.cfg_file_available = FILE_NOT_READABLE;
    cat_string("configuration file '", fspec,
	       "' not readable, using defaults\n", &cbuf);
    xplab_text_insert(main_log, cbuf);
    FREEMEM(cbuf);
    return;
  }

  string_to_log(main_log, "read configuration file '", fspec, "' ... ");

  while (1) {
    FREEMEM(cbuf);
    cbuf = get_line(fin, &line);       /* returns next line number to read */

    if (!cbuf) {
      break;
    }

    errsum += read_subst_rb_key(fspec, cbuf, line-1, cfg);
    errsum += read_subst_nb_key(fspec, cbuf, line-1, cfg);
    errsum += read_rawbit_key(fspec, cbuf, line-1, cfg);
    errsum += read_report_write_key(fspec, cbuf, line-1, cfg);
    errsum += read_rb_chain_key(fspec, cbuf, line-1, cfg);
    errsum += read_histo_write_key(fspec, cbuf, line-1, cfg);
    errsum += read_duration_key(fspec, cbuf, line-1, cfg);
    errsum += read_support_dots_key(fspec, cbuf, line-1, cfg);
    errsum += read_support_dots_auto_key(fspec, cbuf, line-1, cfg);
    errsum += read_overwrite_image_key(fspec, cbuf, line-1, cfg);
    errsum += read_replay_write_key(fspec, cbuf, line-1, cfg);
    errsum += read_training_write_key(fspec, cbuf, line-1, cfg);
    errsum += read_single_line_key(fspec, cbuf, line-1, cfg);
    errsum += read_single_nb_key(fspec, cbuf, line-1, cfg);
    errsum += read_broken_nb_key(fspec, cbuf, line-1, cfg);
    errsum += read_renew_method_key(fspec, cbuf, line-1, cfg);
    errsum += read_cluster_force_key(fspec, cbuf, line-1, cfg);
    errsum += read_cluster_write_key(fspec, cbuf, line-1, cfg);
    errsum += read_cluster_nb_key(fspec, cbuf, line-1, cfg);
    errsum += read_timestamp_key(fspec, cbuf, line-1, cfg);
    errsum += read_writeauto_key(fspec, cbuf, line-1, cfg);
    errsum += read_domain_key(fspec, cbuf, line-1, cfg);
    errsum += read_verbose_key(fspec, cbuf, line-1, cfg);
    errsum += read_tooltips_key(fspec, cbuf, line-1, cfg);
    errsum += read_raise_key(fspec, cbuf, line-1, cfg);
    errsum += read_batchjob_key(fspec, cbuf, line-1, cfg);
    errsum += read_batchlog_key(fspec, cbuf, line-1, cfg);
    errsum += read_line_method_key(fspec, cbuf, line-1, cfg);
    errsum += read_mborder_key(fspec, cbuf, line-1, cfg);
    errsum += read_maskweight_key(fspec, cbuf, line-1, cfg);
    errsum += read_nborder_key(fspec, cbuf, line-1, cfg);
    errsum += read_dc_store_key(fspec, cbuf, line-1, cfg);
    errsum += read_nb_border_key(fspec, cbuf, line-1, cfg);
    errsum += read_dc_1d_key(fspec, cbuf, line-1, cfg);
    errsum += read_dc_2d_key(fspec, cbuf, line-1, cfg);
    errsum += read_nnbmatch_key(fspec, cbuf, line-1, cfg);
    errsum += read_interleave_key(fspec, cbuf, line-1, cfg);
    errsum += read_overlap_key(fspec, cbuf, line-1, cfg);
    errsum += read_symbolspacing_key(fspec, cbuf, line-1, cfg);
    errsum += read_nb_lap_key(fspec, cbuf, line-1, cfg);
    errsum += read_histo_thold_key(fspec, cbuf, line-1, cfg);
    errsum += read_corrlevel_key(fspec, cbuf, line-1, cfg);
    errsum += read_corrbroken_key(fspec, cbuf, line-1, cfg);
    errsum += read_cluster_level_key(fspec, cbuf, line-1, cfg);
    errsum += read_cluster_step_key(fspec, cbuf, line-1, cfg);
    errsum += read_cluster_thold_key(fspec, cbuf, line-1, cfg);
    errsum += read_rgb_norm_key(fspec, cbuf, line-1, cfg);
    errsum += read_ckernel_key(fspec, cbuf, line-1, cfg);
    errsum += read_gtkrc_key(fspec, cbuf, line-1, gtk_rc);
    errsum += read_stats_contrast(fspec, cbuf, line-1, STATS_ICON_CONTRAST_KEY, &sfg->contrast);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_IORG_KEY, &sfg->owin);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_ICOL_KEY, &sfg->cwin);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_IGRY_KEY, &sfg->gwin);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_IBMP_KEY, &sfg->bwin);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_HRBS_KEY, &sfg->rhst);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_HCOL_KEY, &sfg->chst);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_HGRY_KEY, &sfg->ghst);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_HBMP_KEY, &sfg->bhst);
    errsum += read_stats_key(fspec, cbuf, line-1, STATS_HSLOPE_KEY, &sfg->shst);
  }

  clos = xplab_close_file(fin, fspec);

  if (clos) {
    pr.cfg_file_available = FILE_NOT_CLOSEABLE;
  }
  else {
    pr.cfg_file_available = FILE_READ;
  }

  errsum = errors_setting(errsum);

  if (!errsum) {
    xplab_text_insert(main_log, "ok\n");
  }

  if (line < FIRST_LINE_NBR) {
    string_to_log(main_log, "WARNING: empty configuration file '", 
		  fspec, "'\n");
  }

  if ((cfg->cdomain != DOMAIN_DEFAULT) && DOMAIN_SUPPRESS_OTHERS) {
    cfg->cdomain = DOMAIN_DEFAULT;
    string_to_log(main_log, "INFO: only spatial domain accepted in '", 
		  fspec, "'\n");
  }

  FREEMEM(cbuf);
}



/* Reads a key setting from file. 

Revision: 12.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_ckernel_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int      err = 0, get, len;
  kern_def kern;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = 
    get_key(cbuf, CKERNEL_KEY, "%d %d %d %d %d %d %d %d %d",
	    &kern[0 + 0*CKERNEL_W], &kern[1 + 0*CKERNEL_W], &kern[2 + 0*CKERNEL_W],
	    &kern[0 + 1*CKERNEL_W], &kern[1 + 1*CKERNEL_W], &kern[2 + 1*CKERNEL_W],
	    &kern[0 + 2*CKERNEL_W], &kern[1 + 2*CKERNEL_W], &kern[2 + 2*CKERNEL_W]);

  switch (get) {
  case -1: 
    return err;
    break;

  case 9: err = check_vals(fspec, CKERNEL_KEY, line, "iiiiiiiiiii",
			   CKERNEL_MIN, CKERNEL_MAX,
			   kern[0 + 0*CKERNEL_W], kern[1 + 0*CKERNEL_W], kern[2 + 0*CKERNEL_W],
			   kern[0 + 1*CKERNEL_W], kern[1 + 1*CKERNEL_W], kern[2 + 1*CKERNEL_W],
			   kern[0 + 2*CKERNEL_W], kern[1 + 2*CKERNEL_W], kern[2 + 2*CKERNEL_W]);
  break;

  default:
    xplab_errors_txt(CKERNEL_KEY, line, fspec, 2);
    err = 1;
    break;
  }

  if (!err) {
    copy_kern_def(kern, cfg->ckernel);
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 11.09.2005 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_gtkrc_key(char *fspec, char *cbuf, int line, char **gtk_rc)
{
  int  err = 0, get, len;
  char *rc_fname = NULL;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    FREEMEM(rc_fname);
    return err;
  }

  CALLOC(char, len+1, rc_fname);                        /* preserve buffer */

  get = get_key(cbuf, XPLAB_GTK2_RC_KEY, "%s", rc_fname);

  switch (get) {
  case -1: 
    FREEMEM(rc_fname);
    return err;
    break;

  case 1: err = check_name(rc_fname);

  break;

  default:
    xplab_errors_txt(XPLAB_GTK2_RC_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (!err) {
    add_string(rc_fname, gtk_rc);
  }

  *cbuf = 0;

  FREEMEM(rc_fname);
  return err;
}



/* Reads a key setting from file. 

Revision: 11.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_rgb_norm_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, RGB_NORM_KEY, "%lf %lf %lf",
		&cfg->rgb_norm.r, &cfg->rgb_norm.g, 
		&cfg->rgb_norm.b);

  switch (get) {
  case -1: 
    return err;
    break;

  case 3:  err = check_vals(fspec, RGB_NORM_KEY, line, "ddddd",
			    RGB_NORM_MIN, RGB_NORM_MAX, 
			    cfg->rgb_norm.r, cfg->rgb_norm.g, 
			    cfg->rgb_norm.b);
  break;

  default:
    xplab_errors_txt(RGB_NORM_KEY, line, fspec, 2);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->rgb_norm.r = RGB_NORM_DEFAULT;
    cfg->rgb_norm.g = RGB_NORM_DEFAULT;
    cfg->rgb_norm.b = RGB_NORM_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 21.12.2004 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_stats_contrast(char *fspec, char *cbuf, int line, char *keyname, 
		    stats_prefs *sfg)
{
  int err, get;

  err = check_name(cbuf);

  if (err) {
    return 0;                                           /* nothing to scan */
  }

  get = get_key(cbuf, keyname, "%d", &sfg->srevert.display_active);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, keyname, line, "b", 
			   sfg->srevert.display_active);
    break;

  default:
    xplab_errors_txt(keyname, line, fspec, 1);
    err = 1;
    break;
  }

  if (!err) {
    stats_revert_all(sfg);
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 12.11.2004 E.Sack

Input : - pointer to file name string
        - pointer to line string
        - line index
        - pointer to key string
        - pointer to statistic settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_stats_key(char *fspec, char *cbuf, int line, char *keyname,
	       stats_prefs *sfg)
{
  int err, get;

  err = check_name(cbuf);

  if (err) {
    return 0;                                           /* nothing to scan */
  }

  get = get_key(cbuf, keyname, "%d %d %d",
		&sfg->srevert.display_active, 
		&sfg->srevert.wsize.width, &sfg->srevert.wsize.height);

  switch (get) {
  case -1: 
    return err;
    break;

  case 3:
    err = check_vals(fspec, keyname, line, "b",
		     sfg->srevert.display_active);

    if (err) {
      break;
    }

    err = check_vals(fspec, keyname, line, "iiii",
		     STATS_WIN_MIN, STATS_WIN_MAX, 
		     sfg->srevert.wsize.width, sfg->srevert.wsize.height);

  break;

  default:
    xplab_errors_txt(keyname, line, fspec, 2);
    err = 1;
    break;
  }

  if (!err) {                     /* assuming default settings initialised */
    stats_revert_all(sfg);
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1)
        - settings updated
*/
static int
read_subst_rb_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int  err = 0, get, len;
  char *buf = NULL;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  CALLOC(char, len+1, buf);

  get = get_key(cbuf, SUBSTRB_KEY, "%s", buf);

  switch (get) {
  case -1: 
    FREEMEM(buf);
    return err;
    break;

  case 1: err = check_vals(fspec, SUBSTRB_KEY, line, "p", buf);
    break;

  default:
    xplab_errors_txt(SUBSTRB_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  FREEMEM(cfg->subst_rb);

  if (err > 0) {
    cfg->subst_rb = default_subst_rb();
  }
  else {
    add_string(buf, &cfg->subst_rb);
  }

  FREEMEM(buf);

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1)
        - settings updated
*/
static int
read_subst_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int  err = 0, get, len;
  char *buf = NULL;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  CALLOC(char, len+1, buf);

  get = get_key(cbuf, SUBSTNB_KEY, "%s", buf);

  switch (get) {
  case -1: 
    FREEMEM(buf);
    return err;
    break;

  case 1: err = check_vals(fspec, SUBSTNB_KEY, line, "p", buf);
    break;

  default:
    xplab_errors_txt(SUBSTNB_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  FREEMEM(cfg->subst_nb);

  if (err > 0) {
    cfg->subst_nb = default_subst_nb();
  }
  else {
    add_string(buf, &cfg->subst_nb);
  }

  FREEMEM(buf);

  *cbuf = 0;

  return err;
}


 
/* Reads a key setting from file. 

Revision: 09.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_rawbit_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, RAWBIT_KEY, "%d", &cfg->rawbit);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, RAWBIT_KEY, line, "b", cfg->rawbit);
    break;

  default:
    xplab_errors_txt(RAWBIT_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->rawbit = RAWBIT_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 17.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_report_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, REPORT_WRITE_KEY, "%d", &cfg->report_write);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, REPORT_WRITE_KEY, line, "i", 
			   cfg->report_write);
    break;

  default:
    xplab_errors_txt(REPORT_WRITE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->report_write = REPORT_WRITE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 19.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_rb_chain_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, RB_CHAIN_KEY, "%d", &cfg->rb_chain);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, RB_CHAIN_KEY, line, "i", cfg->rb_chain);
    break;

  default:
    xplab_errors_txt(RB_CHAIN_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->rb_chain = RB_CHAIN_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 27.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_cluster_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, CLUSTER_NB_KEY, "%d", &cfg->cluster_nb);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, CLUSTER_NB_KEY, line, "i", cfg->cluster_nb);
    break;

  default:
    xplab_errors_txt(CLUSTER_NB_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0 || !cfg->cluster_nb) {
    cfg->cluster_nb = CLUSTER_NB_NEW;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_histo_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, HISTO_WRITE_KEY, "%d", &cfg->histo_write);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, HISTO_WRITE_KEY, line, "b", 
			   cfg->histo_write);
    break;

  default:
    xplab_errors_txt(HISTO_WRITE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->histo_write = HISTO_WRITE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_duration_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, DURATION_KEY, "%d", &cfg->duration);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, DURATION_KEY, line, "b", cfg->duration);
    break;

  default:
    xplab_errors_txt(DURATION_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->duration = DURATION_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_support_dots_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, SUPPORT_DOTS_KEY, "%d", &cfg->support_dots);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, SUPPORT_DOTS_KEY, line, "b", 
			   cfg->support_dots);
    break;

  default:
    xplab_errors_txt(SUPPORT_DOTS_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->support_dots = SUPPORT_DOTS_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 19.10.2004 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_support_dots_auto_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, SUPPORT_DOTS_AUTO_KEY, "%d", &cfg->support_dots_auto);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, SUPPORT_DOTS_AUTO_KEY, line, "b", 
			   cfg->support_dots_auto);
    break;

  default:
    xplab_errors_txt(SUPPORT_DOTS_AUTO_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->support_dots_auto = SUPPORT_DOTS_AUTO_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 06.05.2004 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_tooltips_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, TIPS_KEY, "%d", &cfg->tips);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, TIPS_KEY, line, "b", cfg->tips);
    break;

  default:
    xplab_errors_txt(TIPS_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->tips = TIPS_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 23.11.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_raise_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, RAISE_KEY, "%d", &cfg->raise);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, RAISE_KEY, line, "b", cfg->raise);
    break;

  default:
    xplab_errors_txt(RAISE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->raise = RAISE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 25.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_overwrite_image_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, OVERWRITE_IMAGE_KEY, "%d", &cfg->overwrite_image);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, OVERWRITE_IMAGE_KEY, line, "b", 
			   cfg->overwrite_image);
    break;

  default:
    xplab_errors_txt(OVERWRITE_IMAGE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->overwrite_image = OVERWRITE_IMAGE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_replay_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, REPLAY_WRITE_KEY, "%d", &cfg->replay_write);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, REPLAY_WRITE_KEY, line, "b", 
			   cfg->replay_write);
    break;

  default:
    xplab_errors_txt(REPLAY_WRITE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->replay_write = REPLAY_WRITE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 09.07.2004 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_training_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, TRAINING_WRITE_KEY, "%d", &cfg->training_write);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, TRAINING_WRITE_KEY, line, "b", 
			   cfg->training_write);
    break;

  default:
    xplab_errors_txt(TRAINING_WRITE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->training_write = TRAINING_WRITE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_single_line_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, SINGLE_LINE_KEY, "%d", &cfg->single_line);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, SINGLE_LINE_KEY, line, "b", 
			   cfg->single_line);
    break;

  default:
    xplab_errors_txt(SINGLE_LINE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->single_line = SINGLE_LINE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 16.10.2004 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_single_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, SINGLE_NB_KEY, "%d", &cfg->single_nb);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, SINGLE_NB_KEY, line, "b", 
			   cfg->single_nb);
    break;

  default:
    xplab_errors_txt(SINGLE_NB_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->single_nb = SINGLE_NB_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 02.01.2005 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_broken_nb_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len, broken_nb;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, BROKEN_NB_KEY, "%d", &broken_nb);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, BROKEN_NB_KEY, line, "b", broken_nb);
    break;

  default:
    xplab_errors_txt(BROKEN_NB_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    broken_nb = BROKEN_NB_DEFAULT;
  }

  *cbuf = 0;

  if (broken_nb) {
    cfg->match_method = MATCHING_METHOD_2;
  }

  return err;
}



/* Reads a key setting from file. 

Revision: 27.12.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_renew_method_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, RENEW_METHOD_KEY, "%d", &cfg->renew_method);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, RENEW_METHOD_KEY, line, "i", 
			   cfg->renew_method);
    break;

  default:
    xplab_errors_txt(RENEW_METHOD_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->renew_method = RENEW_METHOD_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_cluster_force_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, CLUSTER_FORCE_KEY, "%d", &cfg->cluster_force);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, CLUSTER_FORCE_KEY, line, "b", 
			   cfg->cluster_force);
    break;

  default:
    xplab_errors_txt(CLUSTER_FORCE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->cluster_force = CLUSTER_FORCE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_cluster_write_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, CLUSTER_WRITE_KEY, "%d", &cfg->cluster_write);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, CLUSTER_WRITE_KEY, line, "b", 
			   cfg->cluster_write);
    break;

  default:
    xplab_errors_txt(CLUSTER_WRITE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->cluster_write = CLUSTER_WRITE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 12.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_timestamp_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, TIMESTAMP_KEY, "%d", &cfg->timestamp);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, TIMESTAMP_KEY, line, "b", cfg->timestamp);
    break;

  default:
    xplab_errors_txt(TIMESTAMP_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->timestamp = TIMESTAMP_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 08.10.2004 E.Sack

Input : - pointer to filename string
        - pointer to textline string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_writeauto_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, WRITEAUTO_KEY, "%d", &cfg->writeauto);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, WRITEAUTO_KEY, line, "b", cfg->writeauto);
    break;

  default:
    xplab_errors_txt(WRITEAUTO_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->writeauto = WRITEAUTO_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 13.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_domain_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, DOMAIN_KEY, "%d", &cfg->cdomain);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, DOMAIN_KEY, line, "b", cfg->cdomain);
    break;

  default:
    xplab_errors_txt(DOMAIN_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->cdomain = DOMAIN_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 20.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_dc_1d_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len, dc_1d, idx;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, DC_1D_KEY, "%d", &dc_1d);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, DC_1D_KEY, line, "i", dc_1d);
    break;

  default:
    xplab_errors_txt(DC_1D_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    return err;
  }

  idx = cfg->dc_1d.cnt++;
  if (idx) {
    REALLOC(int2, cfg->dc_1d.cnt, cfg->dc_1d.n);
  }
  else {
    CALLOC(int2, cfg->dc_1d.cnt, cfg->dc_1d.n);
  }

  cfg->dc_1d.n[idx].nr = dc_1d;

  return err;
}



/* Reads a key setting from file. 

Revision: 22.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_dc_2d_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len, dc_2d, idx;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, DC_2D_KEY, "%d", &dc_2d);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, DC_2D_KEY, line, "i", dc_2d);
    break;

  default:
    xplab_errors_txt(DC_2D_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    return err;
  }
 
  idx = cfg->dc_2d.cnt++;
  if (idx) {
    REALLOC(int2, cfg->dc_2d.cnt, cfg->dc_2d.n);
  }
  else {
    CALLOC(int2, cfg->dc_2d.cnt, cfg->dc_2d.n);
  }

  cfg->dc_2d.n[idx].nr = dc_2d;

  return err;
}



/* Reads a key setting from file. 

Revision: 14.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_verbose_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, VERBOSE_KEY, "%d", &cfg->verbose);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, VERBOSE_KEY, line, "iii", BE_QUIET, 
			   VERBOSE_LOT, cfg->verbose);
    break;

  default:
    xplab_errors_txt(VERBOSE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->verbose = VERBOSE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 03.04.2004 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_batchjob_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, BATCH_PHASE_KEY, "%d", &cfg->batch.phase);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, BATCH_PHASE_KEY, line, "iii", 
			   BATCH_PHASE_NO, BATCH_PHASE_RB_CHAIN, 
			   cfg->batch.phase);
    break;

  default:
    xplab_errors_txt(BATCH_PHASE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->batch.phase = BATCH_PHASE_NO;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 08.04.2004 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_batchlog_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, BATCH_LOG_KEY, "%d", &cfg->batch.blog);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, BATCH_LOG_KEY, line, "iii", BATCH_LOG_NO, 
			   BATCH_LOG_APPND, cfg->batch.blog);
    break;

  default:
    xplab_errors_txt(BATCH_LOG_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->batch.blog = BATCH_LOG_NO;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_line_method_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, LINE_METHOD_KEY, "%d", &cfg->line_method);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, LINE_METHOD_KEY, line, "i", 
			   cfg->line_method);
    break;

  default:
    xplab_errors_txt(LINE_METHOD_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->line_method = METHOD_FIRST;              /* don't force a method */
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_nnbmatch_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, NNBMATCH_KEY, "%d", &cfg->mad.nnbmatch);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, NNBMATCH_KEY, line, "iii", 
			   NNBMATCH_MIN, NNBMATCH_MAX, cfg->mad.nnbmatch);
  break;

  default:
    xplab_errors_txt(NNBMATCH_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.nnbmatch = NNBMATCH_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_interleave_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, INTERLEAVE_KEY, "%lf", &cfg->mad.interleave);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, INTERLEAVE_KEY, line, "ddd",
			   INTERLEAVE_MIN, INTERLEAVE_MAX, 
			   cfg->mad.interleave);
  break;

  default:
    xplab_errors_txt(INTERLEAVE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.interleave = INTERLEAVE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_overlap_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, OVERLAP_KEY, "%lf", &cfg->mad.overlap);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, OVERLAP_KEY, line, "ddd", 
			   OVERLAP_MIN, OVERLAP_MAX, cfg->mad.overlap);
  break;

  default:
    xplab_errors_txt(OVERLAP_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.overlap = OVERLAP_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_symbolspacing_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, SYSPC_KEY, "%lf", &cfg->syspc);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1:  err = check_vals(fspec, SYSPC_KEY, line, "ddd",
			    SYSPC_MIN, SYSPC_MAX, cfg->syspc);
  break;

  default:
    xplab_errors_txt(SYSPC_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->syspc = SYSPC_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_nb_lap_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, NB_LAP_KEY, "%lf", &cfg->mad.nb_lap);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, NB_LAP_KEY, line, "ddd",
			   NB_LAP_MIN, NB_LAP_MAX, cfg->mad.nb_lap);
  break;

  default:
    xplab_errors_txt(NB_LAP_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.nb_lap = NB_LAP_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 05.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_mborder_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, MBORDER_KEY, "%d", &cfg->mad.mborder);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, MBORDER_KEY, line, "iii", 
			   MBORDER_MIN, MBORDER_MAX, cfg->mad.mborder);
  break;

  default:
    xplab_errors_txt(MBORDER_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.mborder = MBORDER_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 10.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_maskweight_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, MASKWEIGHT_KEY, "%d", &cfg->mad.maskweight);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, MASKWEIGHT_KEY, line, "iii", 
			   MASKWEIGHT_MIN, MASKWEIGHT_MAX, 
			   cfg->mad.maskweight);
  break;

  default:
    xplab_errors_txt(MASKWEIGHT_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.maskweight = MASKWEIGHT_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 06.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_nborder_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, NBORDER_KEY, "%d", &cfg->mad.nborder);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, NBORDER_KEY, line, "iii", 
			   NBORDER_MIN, NBORDER_MAX, cfg->mad.nborder);
  break;

  default:
    xplab_errors_txt(NBORDER_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.nborder = NBORDER_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_dc_store_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, DC_STORE_KEY, "%d", &cfg->dc_store);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, DC_STORE_KEY, line, "iii", 
			   DC_STORE_MIN, DC_STORE_MAX, cfg->dc_store);
  break;

  default:
    xplab_errors_txt(DC_STORE_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->dc_store = DC_STORE_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 18.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_nb_border_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, NB_BORDER_KEY, "%d %d", &cfg->nb_border_x, 
		&cfg->nb_border_y);

  switch (get) {
  case -1: 
    return err;
    break;

  case 2: err = check_vals(fspec, NB_BORDER_KEY, line, "iiii",
			   NB_BORDER_MIN, NB_BORDER_MAX, 
			   cfg->nb_border_x, 
			   cfg->nb_border_y);
  break;

  default:
    xplab_errors_txt(NB_BORDER_KEY, line, fspec, 2);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->nb_border_x = NB_BORDER_DEFAULT;
    cfg->nb_border_y = NB_BORDER_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 06.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_histo_thold_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, THOLD_KEY, "%d", &cfg->histo_thold);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, THOLD_KEY, line, "iii",
			   THOLD_MIN, THOLD_MAX, cfg->histo_thold);
  break;

  default:
    xplab_errors_txt(THOLD_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->histo_thold = THOLD_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 06.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_corrlevel_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, LEVEL_KEY, "%lf", &cfg->mad.corrlevel);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, LEVEL_KEY, line, "ddd",
		           LEVEL_MIN, LEVEL_MAX, cfg->mad.corrlevel);
  break;

  default:
    xplab_errors_txt(LEVEL_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.corrlevel = LEVEL_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 05.01.2005 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_corrbroken_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, BROKEN_LEVEL_KEY, "%lf", &cfg->mad.corrbroken);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, BROKEN_LEVEL_KEY, line, "ddd",
		           BROKEN_LEVEL_MIN, BROKEN_LEVEL_MAX, 
			   cfg->mad.corrbroken);
  break;

  default:
    xplab_errors_txt(BROKEN_LEVEL_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->mad.corrbroken = BROKEN_LEVEL_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 08.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_cluster_level_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, CLUSTER_LVL_KEY, "%lf", &cfg->cluster_level);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, CLUSTER_LVL_KEY, line,
			   "ddd", CLUSTER_LVL_MIN, CLUSTER_LVL_MAX,
			   cfg->cluster_level);
  break;

  default:
    xplab_errors_txt(CLUSTER_LVL_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->cluster_level = CLUSTER_LVL_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 08.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_cluster_step_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, CLUSTER_STEP_KEY, "%lf", &cfg->cluster_step);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, CLUSTER_STEP_KEY, line,
			   "ddd", CLUSTER_STEP_MIN, CLUSTER_STEP_MAX,
			   cfg->cluster_step);
  break;

  default:
    xplab_errors_txt(CLUSTER_STEP_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->cluster_step = CLUSTER_STEP_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Reads a key setting from file. 

Revision: 08.10.2003 E.Sack

Input : - pointer to string
        - line index
        - pointer to settings
Output: - error flag (0: no error, 1: error)
        - settings updated
*/
static int
read_cluster_thold_key(char *fspec, char *cbuf, int line, cfg_def *cfg)
{
  int err = 0, get, len;

  assert(cbuf);

  len = strlen(cbuf);

  if (!len) {
    return err;
  }

  get = get_key(cbuf, CHOLD_KEY, "%d", &cfg->cluster_thold);

  switch (get) {
  case -1: 
    return err;
    break;

  case 1: err = check_vals(fspec, CHOLD_KEY, line, "iii",
			   CHOLD_MIN, CHOLD_MAX, cfg->cluster_thold);
  break;

  default:
    xplab_errors_txt(CHOLD_KEY, line, fspec, 1);
    err = 1;
    break;
  }

  if (err > 0) {
    cfg->cluster_thold = CHOLD_DEFAULT;
  }

  *cbuf = 0;

  return err;
}



/* Assembles standard error message text string. 

Revision: 16.12.2004 E.Sack

Input : - pointer to key string
        - file line number
        - pointer to filename
        - flag (1: error about one value, 2: more than one value)
Output: - errors text updated
*/
static void
xplab_errors_txt(char *key_string, int line, char *cfg_name, int howmuch)
{
  assert(howmuch == 1 || howmuch == 2);

  cat_string(howmuch == 1?
	     "WARNING: no/wrong Value for ":
	     "WARNING: no complete parameter set for ", 
	     key_string, "", &pr.xerrors);

  int_to_string(" in line ", line, " of ", &pr.xerrors);
  cat_string(cfg_name, " found\n", "", &pr.xerrors);
}



/* Writes configuration file to disc. 

Revision: 14.01.2005 E.Sack

Input : - filename
Output: - configuration file
*/
static void
write_cfg(set_def set, cfg_def cfg, stats_def stats) 
{
  int      d, i;
  char     *sepn = NULL, *buf = NULL;
  FILE     *file;
  kern_def kern;

  xdefault_kern_def(kern);

  add_string("   ", &sepn);
  add_character(COMMENT_SIGN, &sepn);
  add_string(" ", &sepn);

  file = xplab_open_file(set.config_filespec, "w");
  if (!file) {
    cat_string("could not write configuration file '", set.config_filespec, 
	       "'", &buf);
    xplab_label_set_text(pr.page.page_label, buf);
    FREEMEM(buf);
    FREEMEM(sepn);
    return;
  }

  cat_string("write configuration file '", set.config_filespec, " ' ... ", &buf);
  xplab_label_set_text(pr.page.page_label, buf);

  d = fprintf(file, "%-12s %d%s%s\n", RAWBIT_KEY, cfg.rawbit, sepn, 
	      RAWBIT_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", REPORT_WRITE_KEY, cfg.report_write, sepn,
	      REPORT_WRITE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", RB_CHAIN_KEY, cfg.rb_chain, sepn,
	      RB_CHAIN_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", HISTO_WRITE_KEY, cfg.histo_write, sepn,
	      HISTO_WRITE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", DURATION_KEY, cfg.duration, sepn,
	      DURATION_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", SUPPORT_DOTS_KEY, cfg.support_dots, sepn,
	      SUPPORT_DOTS_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", SUPPORT_DOTS_AUTO_KEY, cfg.support_dots_auto, sepn,
	      SUPPORT_DOTS_AUTO_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", OVERWRITE_IMAGE_KEY, cfg.overwrite_image, sepn,
	      OVERWRITE_IMAGE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", REPLAY_WRITE_KEY, cfg.replay_write, sepn,
	      REPLAY_WRITE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", TRAINING_WRITE_KEY, cfg.training_write, sepn,
	      TRAINING_WRITE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", SINGLE_LINE_KEY, cfg.single_line, sepn,
	      SINGLE_LINE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", SINGLE_NB_KEY, cfg.single_nb, sepn,
	      SINGLE_NB_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", BROKEN_NB_KEY, 
	      (cfg.match_method == MATCHING_METHOD_2) ? 1 : 0,
	      sepn, BROKEN_NB_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", RENEW_METHOD_KEY, cfg.renew_method, sepn,
	      RENEW_METHOD_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", CLUSTER_FORCE_KEY, cfg.cluster_force, sepn,
	      CLUSTER_FORCE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", CLUSTER_WRITE_KEY, cfg.cluster_write, sepn,
	      CLUSTER_WRITE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", CLUSTER_NB_KEY, cfg.cluster_nb, sepn,
	      CLUSTER_NB_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", TIMESTAMP_KEY, cfg.timestamp, sepn, 
	      TIMESTAMP_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", WRITEAUTO_KEY, cfg.writeauto, sepn, 
	      WRITEAUTO_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", DOMAIN_KEY, cfg.cdomain, sepn, 
	      DOMAIN_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", VERBOSE_KEY, cfg.verbose, sepn, 
	      VERBOSE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", BATCH_PHASE_KEY, cfg.batch.phase, sepn,
	      BATCH_PHASE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", BATCH_LOG_KEY, cfg.batch.blog, sepn,
	      BATCH_LOG_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", TIPS_KEY, cfg.tips, sepn,
	      TIPS_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", RAISE_KEY, cfg.raise, sepn,
	      RAISE_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%s\n", LINE_METHOD_KEY, cfg.line_method, sepn, 
	      LINE_METHOD_REM);
  assert(d > 0);

  d = fprintf(file, "%-12s %s%s%s\n", SUBSTRB_KEY, cfg.subst_rb, sepn,
	      SUBSTRB_DEFAULT);
  assert(d > 0);

  d = fprintf(file, "%-12s %s%s%s\n", SUBSTNB_KEY, cfg.subst_nb, sepn, 
	      SUBSTNB_DEFAULT);  
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", LEVEL_KEY, cfg.mad.corrlevel, 
	      sepn, LEVEL_DEFAULT, LEVEL_MIN, LEVEL_MAX);    
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", BROKEN_LEVEL_KEY, cfg.mad.corrbroken,
	      sepn, BROKEN_LEVEL_DEFAULT, BROKEN_LEVEL_MIN, BROKEN_LEVEL_MAX);
  assert(d > 0);

  d = fprintf(file,"%-12s %d%s%d, %d-%d\n", NNBMATCH_KEY, cfg.mad.nnbmatch,
              sepn, NNBMATCH_DEFAULT, NNBMATCH_MIN, NNBMATCH_MAX);  
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", INTERLEAVE_KEY, cfg.mad.interleave,
              sepn, INTERLEAVE_DEFAULT, INTERLEAVE_MIN, INTERLEAVE_MAX);   
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", OVERLAP_KEY, cfg.mad.overlap, sepn,
              OVERLAP_DEFAULT, OVERLAP_MIN, OVERLAP_MAX);     
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", NB_LAP_KEY, cfg.mad.nb_lap,
	      sepn, NB_LAP_DEFAULT, NB_LAP_MIN, NB_LAP_MAX);   
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", SYSPC_KEY, cfg.syspc, sepn,
              SYSPC_DEFAULT, SYSPC_MIN, SYSPC_MAX);     
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%d, %d-%d\n", MASKWEIGHT_KEY, cfg.mad.maskweight,
	      sepn, MASKWEIGHT_DEFAULT, MASKWEIGHT_MIN, MASKWEIGHT_MAX);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%d, %d-%d\n", MBORDER_KEY, cfg.mad.mborder,
              sepn, MBORDER_DEFAULT, MBORDER_MIN, MBORDER_MAX); 
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%d, %d-%d\n", NBORDER_KEY, cfg.mad.nborder,
              sepn, NBORDER_DEFAULT, NBORDER_MIN, NBORDER_MAX); 
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%d, %d-%d\n", DC_STORE_KEY, cfg.dc_store,
              sepn, DC_STORE_DEFAULT, DC_STORE_MIN, DC_STORE_MAX); 
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%d, %d-%d (%d %s)\n", THOLD_KEY, 
	      cfg.histo_thold, sepn, THOLD_DEFAULT, THOLD_MIN, THOLD_MAX, 
	      THOLD_MIN, "automatic");  
  assert(d > 0);

  d = fprintf(file, "%-12s %d %d%s%d %d, %d-%d (%s)\n", NB_BORDER_KEY, 
	      cfg.nb_border_x, cfg.nb_border_y,
              sepn, NB_BORDER_DEFAULT, NB_BORDER_DEFAULT, 
	      NB_BORDER_MIN, NB_BORDER_MAX, NB_BORDER_REM); 
  assert(d > 0);

  d = fprintf(file, "%-12s %f %f %f%s%f %f %f, %f-%f\n", RGB_NORM_KEY, 
	      cfg.rgb_norm.r, cfg.rgb_norm.g, cfg.rgb_norm.b, sepn, 
	      RGB_NORM_DEFAULT, RGB_NORM_DEFAULT, RGB_NORM_DEFAULT, 
	      RGB_NORM_MIN, RGB_NORM_MAX); 
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", CLUSTER_LVL_KEY, 
	      cfg.cluster_level, sepn, CLUSTER_LVL_DEFAULT,
	      CLUSTER_LVL_MIN, CLUSTER_LVL_MAX);
  assert(d > 0);

  d = fprintf(file, "%-12s %f%s%f, %f-%f\n", CLUSTER_STEP_KEY, 
	      cfg.cluster_step, sepn, CLUSTER_STEP_DEFAULT, 
	      CLUSTER_STEP_MIN, CLUSTER_STEP_MAX);
  assert(d > 0);

  d = fprintf(file, "%-12s %d %d %d %c%s%d %d %d, %d-%d\n", CKERNEL_KEY, 
	      cfg.ckernel[0 + 0*CKERNEL_W], cfg.ckernel[1 + 0*CKERNEL_W], cfg.ckernel[2 + 0*CKERNEL_W], 
	      LINE_CONT_SIGN, sepn, 
	      kern[0 + 0*CKERNEL_W], kern[1 + 0*CKERNEL_W], kern[2 + 0*CKERNEL_W], 
	      CKERNEL_MIN, CKERNEL_MAX); 
  assert(d > 0);

  d = fprintf(file, "%-12s %d %d %d %c%s%d %d %d\n", "",
	      cfg.ckernel[0 + 1*CKERNEL_W], cfg.ckernel[1 + 1*CKERNEL_W], cfg.ckernel[2 + 1*CKERNEL_W], 
	      LINE_CONT_SIGN, sepn, 
	      kern[0 + 1*CKERNEL_W], kern[1 + 1*CKERNEL_W], kern[2 + 1*CKERNEL_W]);
  assert(d > 0);

  d = fprintf(file, "%-12s %d %d %d  %s%d %d %d\n", "",
	      cfg.ckernel[0 + 2*CKERNEL_W], cfg.ckernel[1 + 2*CKERNEL_W], cfg.ckernel[2 + 2*CKERNEL_W], 
	      sepn, 
	      kern[0 + 2*CKERNEL_W], kern[1 + 2*CKERNEL_W], kern[2 + 2*CKERNEL_W]);
  assert(d > 0);

  d = fprintf(file, "%-12s %d%s%d, %d-%d\n", CHOLD_KEY, cfg.cluster_thold, 
	      sepn, CHOLD_DEFAULT, CHOLD_MIN, CHOLD_MAX);
  assert(d > 0);

  write_stats_prefs(file, STATS_IORG_KEY, stats.owin.sapply, sepn);
  write_stats_prefs(file, STATS_ICOL_KEY, stats.cwin.sapply, sepn);
  write_stats_prefs(file, STATS_IGRY_KEY, stats.gwin.sapply, sepn);
  write_stats_prefs(file, STATS_IBMP_KEY, stats.bwin.sapply, sepn);
  write_stats_prefs(file, STATS_HRBS_KEY, stats.rhst.sapply, sepn);
  write_stats_prefs(file, STATS_HCOL_KEY, stats.chst.sapply, sepn);
  write_stats_prefs(file, STATS_HGRY_KEY, stats.ghst.sapply, sepn);
  write_stats_prefs(file, STATS_HBMP_KEY, stats.bhst.sapply, sepn);
  write_stats_prefs(file, STATS_HSLOPE_KEY, stats.shst.sapply, sepn);

  d = fprintf(file, "%-12s %d%s%s\n", STATS_ICON_CONTRAST_KEY, 
	      stats.contrast.sapply.display_active==TRUE,
	      sepn, STATS_ICON_CONTRAST_REM);
  assert(d > 0);

  write_gtk_rc_prefs(file, set.gtk_rc, sepn);

  for (i = 0; i < cfg.dc_1d.cnt; i++) {
    d = fprintf(file, "%-12s %d%s%s\n", DC_1D_KEY, cfg.dc_1d.n[i].nr, sepn, 
		DC_1D_REM);
    assert(d > 0);
  }

  for (i = 0; i < cfg.dc_2d.cnt; i++) {
    d = fprintf(file, "%-12s %d%s%s\n", DC_2D_KEY, cfg.dc_2d.n[i].nr, sepn, 
		DC_2D_REM);
    assert(d > 0);
  }

  d = xplab_close_file(file, set.config_filespec);

  if (!d) {
    add_string("ok", &buf);
  }

  xplab_label_set_text(pr.page.page_label, buf);
  FREEMEM(buf);
  FREEMEM(sepn);
}



/* Writes Gtk resource configuration to disc. 

Revision: 11.09.2005 E.Sack

Input : - pointer to stream
        - pointer to resource file name
        - pointer to separation string
Output: - stream updated
Postcondition: - if no file name is given an inactive template is written.
                 This template is to activate by the user manually. Else
                 the file name input is written with an active key.
*/
static void
write_gtk_rc_prefs(FILE *file, char *gtk_rc, char *sepn)
{
  int d;

  d = fprintf(file, gtk_rc ? "%-12s %s%s%s\n" : "# %-12s %s%s%s\n", 
	      XPLAB_GTK2_RC_KEY, 
	      gtk_rc ? gtk_rc : XPLAB_GTK2_RC_DEF,
	      sepn, XPLAB_GTK2_RC_REM);

  assert(d > 0);
}



/* Writes statistic configuration to disc. 

Revision: 12.11.2004 E.Sack

Input : - pointer to stream
        - pointer to key string
        - statistic data
        - pointer to separation string
Output: - stream updated
*/
static void
write_stats_prefs(FILE *file, char *keyname, stats_win swin, char *sepn)
{
  int d;

  assert(file);
  assert(sepn);

  d = fprintf(file, "%-12s %d %d %d%s%d, %d x %d (%d-%d)\n", keyname, 
	      swin.display_active, swin.wsize.width, swin.wsize.height, sepn, 
	      STATS_WIN_DEFACTV!=FALSE, STATS_WIN_DEFAULT, STATS_WIN_DEFAULT, 
	      STATS_WIN_MIN, STATS_WIN_MAX); 
  assert(d > 0);
}



/* Variable value limits checking.

Revision: 11.07.2004 E.Sack

Input : - pointer to filename
        - pointer to key name
        - line number
        - pointer to format variable
        - variable argument
Output: - flag (0: no limit violation, 1: limit violation)
*/
static int 
check_vals(char *fname, char *key, int  line, char *fmt, ...)
{
  va_list ap;
  int     i, mini = 0, maxi = 0, cd = 0, ci = 0, err = 0, bool;
  double  d, mind = 0.0, maxd = 0.0;
  char    *s;

  va_start(ap, fmt);

  switch(*fmt++) {                                   /* first control char */
  case 'b':                                                     /* boolean */
    bool = va_arg(ap, int);
    if (bool != 0 && bool != 1) {
      int_to_log(main_log, "\nWARNING: Value '", bool, "' for '");
      string_to_log(main_log, key, "' in line ", "");
      int_to_log(main_log, "", line, " of '");
      string_to_log(main_log, fname, "' is not a boolean\n", "");
      err = 1;
    }
    break;

  case 'd':                                                      /* double */
    mind = va_arg(ap, double);
    cd++;
    break;

  case 'i':                                                         /* int */
    mini = va_arg(ap, int);
    ci++;
    break;

  case 'a':                                                       /* alpha */
    s = va_arg(ap, char *);
    i = isalpha(s[0]);                    /* simple test: only first entry */
    if (!i) {
      string_to_log(main_log, "\nWARNING: Value '", s, "' for '");
      string_to_log(main_log, key, "' in line ", "");
      int_to_log(main_log, "", line, " of '");
      string_to_log(main_log, fname, "' is not an alpha\n", "");
      err = 1;
    }
    break;

  case 'p':                                                   /* printable */
    s = va_arg(ap, char *);
    i = isprint(s[0]);                    /* simple test: only first entry */
    if (!i) {
      string_to_log(main_log, "\nWARNING: Value '", s, "' for '");
      string_to_log(main_log, key, "' in line ", "");
      int_to_log(main_log, "", line, " of '");
      string_to_log(main_log, fname, "' is not printable\n", "");
      err = 1;
    }
    break;

  default: assert(0);
    break;
  }

  if (*fmt) {
    switch(*fmt++) {                                /* second control char */
    case 'd':                                                    /* double */
      maxd = va_arg(ap, double);
      assert(cd);
      assert(mind <= maxd);
      break;

    case 'i':                                                       /* int */
      maxi = va_arg(ap, int);
      assert(ci);
      assert(mini <= maxi);
      break;

    default: assert(0);
      break;
    }
  }

  while (*fmt)
    switch(*fmt++) {                                 /* next control chars */
    case 'd':                                                    /* double */
      d = va_arg(ap, double);
      if (mind > d || d > maxd) {
	double_to_log("\nWARNING: Value ", d, " for '");
	string_to_log(main_log, key, "' in line ", "");
	int_to_log(main_log, "", line, " of '");
	string_to_log(main_log, fname, "' is out of range\n", "");

	double_to_log("valid range is ", mind, " <= ");
	string_to_log(main_log, key, " <= ", "");
	double_to_log("", maxd, "\n");
	err= 1;
      }
      break;

    case 'D':                                                    /* double */
      d = va_arg(ap, double);
      if (mind >= d || d >= maxd) {
	double_to_log("\nWARNING: Value ", d, " for '");
	string_to_log(main_log, key, "' in line ", "");
	int_to_log(main_log, "", line, " of '");
	string_to_log(main_log, fname, "' is out of range\n", "");

	double_to_log("valid range is ", mind, " <= ");
	string_to_log(main_log, key, " <= ", "");
	double_to_log("", maxd, "\n");
	err= 1;
      }
      break;

    case 'i':                                                       /* int */
      i = va_arg(ap, int);
      if (mini > i || i > maxi) {
	int_to_log(main_log, "\nWARNING: Value ", i, " for '");
	string_to_log(main_log, key, "' in line ", "");
	int_to_log(main_log, "", line, " of '");
	string_to_log(main_log, fname, "' is out of range\n", "");

	int_to_log(main_log, "valid range is ", mini, " <= ");
	string_to_log(main_log, key, " <= ", "");
	int_to_log(main_log, "", maxi, "\n");
	err= 1;
      }
      break;

    default: assert(0);
      break;
    }

  va_end(ap);

  return err;
}



/* Searches a key in a string and if success converts the belonging value. 

Revision: 05.08.2005 E.Sack

Input : - pointer to string
        - pointer to key
        - pointer to format
        - variable parameter list
Output: - parameter updated
        - number of conversions (-1: no key found, 
                                  0: key found, no conversion made,
                                 >0: key found, conversion made)
*/
static int 
get_key(char *lstr, 
        const char *key, 
        const char *fmt, 
        ...)
{
  int     cnt;
  char    *p = lstr, buf[MAXCHAR];
  va_list ap;

  va_start(ap, fmt);

  g_snprintf(buf, sizeof(buf), "%s ", key);         /* white spaces required */

  if (strncmp(p, buf, strlen(buf))) {
    return -1;                                           /* no key in string */
  }

  p = lstr + strlen(buf);                                   /* jump over key */

  cnt = vsscanf(p, fmt, ap);                          /* scan rest of string */

  va_end(ap);

  if (cnt < 0) {
    cnt = 0;                                                /* no conversion */
  }

  return cnt;
}



/* Read one line of a text file, comments are stripped.

Revision: 07.04.2004 E.Sack

Input : - pointer to file descriptor
        - pointer to current line number
Output: - pointer to string (NULL: no more lines)
        - line number updated
Precondition: - lines must be terminated with LF '\n'
              - lines can have any length
Postcondition: - if line number returned is less than input, no line was read,
                 else next line number to read is returned.
*/
char *
get_line(FILE *fin, int *line)
{
  int  line_continued;
  char ibuf[MAXCHAR], *cbuf = NULL, *cptr;

  while(1) {
    cptr = fgets(ibuf, sizeof(ibuf), fin);  /* read current line from file */

    if (cptr == NULL) {                                             /* EOF */
      (*line)--;                                         /* last line read */
      break;
    }

    line_continued = 0;

    cptr = strchr(ibuf, '\n');                      /* is there a newline? */

    if (cptr) {                                           
      (*line)++;                                       /* next line number */
      *cptr = '\0';                                       /* strip newline */
    }

    cptr = strchr(ibuf, COMMENT_SIGN);              /* is there a comment? */

    if (cptr) {                                           
      *cptr = '\0';                                       /* strip comment */
    }

    if (!strlen(ibuf)) {
      continue;                                /* rule: ignore empty lines */
    }

    cptr = strchr(ibuf, LINE_CONT_SIGN);             /* is line continued? */

    if (cptr) {                                 
      *cptr = '\0';                             /* strip line continuation */
      line_continued = 1;
    }

    add_string(ibuf, &cbuf);

    if (line_continued) {
      continue;                                          /* round out line */
    }

    break;                                               /* line completed */
  }

  return cbuf;
}



/* Converts command line arguments to XPLAB arguments.

Revision: 05.12.2004 E.Sack

Input : - CLI argument count
        - CLI argument string array
        - pointer to image file name pointer
        - pointer to DB file name pointer
        - pointer to options string pointer
        - pointer to configuration filespec pointer
        - pointer to batch flag
Output: - CLI parameter
*/
static cli_def
read_cli(int argc, char *argv[], char **xprog, char **image_name, 
	 char **db_name, char **options, char **cfg_filespec, 
	 int *batch_enable) 
{
  cli_def clid;
  int     c, err;

  assert(image_name);
  assert(db_name);
  assert(cfg_filespec);

  init_cli_def(&clid);

  add_string(argv[0], xprog);         /* fetch filespec of called program */

  while (1) {
    c = getopt(argc, argv, GUI_OPTION_LIST BATCH_OPTION_LIST);

    if (c == -1) {
      break;
    }

    switch (c) {
    case 'h': clid.help    = 1; break;
    case 'v': clid.license = 1; break;
    case 'Y': *batch_enable = TRUE; break;

    case 'c':
      err = check_name(optarg);

      if (!err) {
	FREEMEM(*cfg_filespec);
	add_string(optarg, cfg_filespec);
	xplab_absolute_filespec(cfg_filespec);          /* more convenient */
      }
      break;

    case 'D': 
      err = check_name(optarg);

      if (!err) {
	FREEMEM(*db_name);
	add_string(optarg, db_name);
      }
      break;

    default: break;
    }
  }

  while (optind < argc) {
    err = check_name(argv[optind]);

    if (!err) {
      add_string(argv[optind], image_name);
    }

    optind++;
  }

  FREEMEM(*options);
  *options = string_user_options(argc, argv);      /* sorted argument list */

  return clid;
}



/* Assembles input options to a string.

Revision: 19.01.2003 E.Sack

Input : - CLI argument count
        - CLI argument vector
Output: - pointer to string
*/
char *
string_user_options(int argc, char *argv[])
{
  int  i;
  char *user_options = 0;

  add_string("user options: ", &user_options);

  if (argc > 1) {
    for (i = 1; i < argc; i++) {
      cat_string(argv[i], " ", "", &user_options);
    }
  }
  else {
    add_string("*** no valid options ***", &user_options);
  }

  return user_options;
}



/* Sets up DB and selects the default TPS ID.

Revision: 03.10.2004 E.Sack

Input : - pointer to DB file name string 
        - pointer to TPS ID string
        - pointer to TPS ID index
        - mask border
Output: - DB file name updated
        - TPS ID updated
        - TPS ID index updated
*/
static db_def
setup_db(char *db_name, char **tps_id, int mborder, int maskweight, 
	 int verbose)
{
  int    err, tps_idx, serious_error;
  db_def db;                                                   /* database */

  init_db_def(&db);

  err = check_name(db_name);

  if (!err) {
    db = xplab_read_db(db_name, 1, verbose, &serious_error);    /* read DB */

    if (serious_error) {
      db_name = NULL;            /* something went wrong with the filename */
    }
  }

  if (db.db_fname == NULL) {                   /* create new DB attributes */
    db.db_fname = create_db_attributes(db_name, &db.db_version);

    tps_idx = create_tps_def(&db.tpl, tps_id);         /* create empty TPS */

    if (tps_idx >= 0) {
      xdefault_media_att(&db.tpl.tps[tps_idx].media);    /* default values */
    }

    if (verbose) {
      string_to_log(main_log, "using empty DB '", db.db_fname, "'\n");
    }
  }

  force_id(tps_id, db.tpl);

  /* generate database features not stored in file ...                     */

  for (tps_idx = 0; tps_idx < db.tpl.cnt; tps_idx++) {          /* NB refs */
    gen_class_references(db.tpl.tps[tps_idx].cl, mborder, maskweight);
  }

  return db;
}



/* Searching the DB for a certain TPS ID and returns ID if there is any.

Revision: 28.11.2003 E.Sack

Input : - pointer to TPS ID string
        - TPS list
Output: - TPS index for TPS ID (-1: no ID found)
        - TPS ID string updated
Postcondition: - values returned are valid and contained in DB. 
Usage : The function shall secure a valid TPS ID (and its index), which may
        have been changed. The function also can be used to return the first
        TPS ID and index of the DB if pointer input is NULL.
*/
static int
force_id(char **ident, tps_lst tpl)
{
  int tps_idx, err;

  tps_idx = search_tps_id(ident, tpl);        /* look for the input TPS ID */

  if (tps_idx < 0) {                             /* input TPS ID not found */
    char *tmp = NULL;

    tps_idx = search_tps_id(&tmp, tpl);       /* look for the first TPS ID */

    if (tps_idx >= 0) {                            /* another TPS ID found */
      err = check_name(tmp);
      assert(!err);

      FREEMEM(*ident);                  /* empty unsuccessful input TPS ID */
      *ident = tmp;                                          /* new TPS ID */
      return tps_idx;
    }

    FREEMEM(tmp);
  }

  return tps_idx;
}



/* Initialize DB file name.

Revision: 21.08.2004 E.Sack

Input : - pointer to DB
        - pointer to DB file name
Output: - DB file name updated
Postcondition: - DB file name without path
*/
static char*
create_db_attributes(char *db_name, char **db_version)
{
  int  err;
  char *new_name = NULL, *tmp_name = NULL, *bname;

  assert(db_version);                                   /* must be defined */
  FREEMEM(*db_version);
  add_string(DB_VERSION_CURRENT, db_version);

  err = check_name(db_name);

  if (err) {
    err = check_name(DB_FNAME_DEFAULT);

    if (err) {                                              /* logic error */
      xplab_text_insert(main_log, "ERROR: new DB file name undefined\n");
    }
    else {
      add_string(DB_FNAME_DEFAULT, &new_name);
    }
  }
  else {
    add_string(db_name, &new_name);
  }

  bname = basename(new_name);                      /* see "man 3 basename" */
  add_string(bname, &tmp_name);
  FREEMEM(new_name);                                   /* use with caution */

  return tmp_name;
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - menu
Output: - menu item
*/
void
xplab_preferences_menu_unload(GtkWidget *file_menu)
{
  GtkWidget *menu_unload;

  menu_unload = gtk_menu_item_new_with_label("unload DB");

  gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), menu_unload);

  g_signal_connect(G_OBJECT(menu_unload), "activate", 
		   G_CALLBACK(cb_preferences_unload), NULL);
}



/* Adds a dynamic item to a menu.

Revision: 05.08.2005 E.Sack

Input : - menu
Output: - menu item
*/
void
xplab_preferences_menu_tpsid(GtkWidget *file_menu)
{
  if (GTK_IS_WIDGET(pr.tpsid_item) == FALSE) {
    assert(file_menu);
    pr.tpsid_item = gtk_menu_item_new_with_label("TPS ID");
    gtk_menu_shell_append(GTK_MENU_SHELL(file_menu), pr.tpsid_item);
  }

  if (GTK_IS_WIDGET(pr.tpsid_menu) == TRUE) {
    xplab_widget_destroy(&pr.tpsid_menu);            /* destroy for new menu */
  }

  pr.tpsid_menu = gtk_menu_new();
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(pr.tpsid_item), pr.tpsid_menu);

  preferences_tpsid_menu_item_id(pr.tpsid_menu);
  gtk_widget_show_all(pr.tpsid_menu);
}



/* Adds an item to a menu.

Revision: 05.08.2005 E.Sack

Input : - menu
Output: - menu item
*/
static void
preferences_tpsid_menu_item_id(GtkWidget *tpsid_menu)
{
  int       i, d, found = 0;
  GtkWidget *wid;
  GSList    *item_group = NULL;

  assert(tpsid_menu);

  if (!pr.xplab_db.tpl.cnt) {
    gtk_widget_set_sensitive(pr.tpsid_item, FALSE); 
  }
  else {
    gtk_widget_set_sensitive(pr.tpsid_item, TRUE);
 
    d = check_name(xctrl.cfg.tps_id);                 /* check consistency */
    assert(!d);

    for (i = 0; i < pr.xplab_db.tpl.cnt; i++) {
      d = strcmp(xctrl.cfg.tps_id,              /* consider ID concurrency */
		 pr.xplab_db.tpl.tps[i].ident);

      if (!d) {
	found++;

	if (found > 1) {
	  continue;              /* inconsistencies found: identic TPS IDs */
	}
      }

      wid = gtk_radio_menu_item_new_with_label(item_group, 
					       pr.xplab_db.tpl.tps[i].ident);

      item_group = gtk_radio_menu_item_get_group(GTK_RADIO_MENU_ITEM(wid));

      gtk_menu_shell_append(GTK_MENU_SHELL(tpsid_menu), wid);

      if (!d) {                          /* only one certain ID set active */
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(wid), TRUE);
      }

      g_signal_connect(G_OBJECT(wid), "toggled",
		       G_CALLBACK(cb_toggle_tpsid),
		       (gpointer)pr.xplab_db.tpl.tps[i].ident);
    }
  }
}



/* Wrapper for vsnprintf() to avoid fixed allocation size.

Revision: 17.10.2004 E.Sack

Input : - format string
        - argument list
Output: - pointer to string
Postcondition: - string to be freed.
*/
char *
xplab_vsnprintf(const char *fmt, ...)
{
  char *tmp = NULL;

  XPLAB_VA_LIST(fmt, tmp);     /* produce string from variable format list */

  return tmp;
}



/* Writes text to report file.

Revision: 17.10.2004 E.Sack

Input : - format string
        - argument list
Output: - report file stream updated
        - flag (0: no output errors, 1: output errors)
*/
int
xreport(const char *fmt, ...)
{
  int  err = 0, d;
  char *tmp = NULL;

  XPLAB_VA_LIST(fmt, tmp);     /* produce string from variable format list */

  string_to_log(main_log, tmp, "", "");      /* try to write to GUI always */

  if (xctrl.set.dobatch == TRUE) {
    if (xctrl.set.report_stream) {
      d = fprintf(xctrl.set.report_stream, "%s", tmp);

      if (d <= 0) {
	fprintf(stdout, "\nERROR: cannot write to report file stream\n\n");
	err = 1;
      }
    }
    else {
      add_string(tmp, &pr.rpt_store);      /* fill auxiliary report store */
    }
  }

  FREEMEM(tmp);

  return err;
}



/* Opens report file.

Revision: 21.06.2004 E.Sack

Input : - pointer to settings
        - pointer to DB name
        - verbose flag
Output: - report opened
        - pointer to report file name string (may be NULL)
*/
static char *
open_rpt_file(prog_ctrl *ctrl, char *db_name, int verbose)
{
  int  err;
  char *report_filename = NULL;

  if (ctrl->set.dobatch == FALSE) {
    return NULL;                                    /* only for batch jobs */
  }

  if (ctrl->cfg.report_write & REPORT_WRITE_NOT) {
    return NULL;                                       /* no report wanted */
  }

  err = check_name(ctrl->set.image_filename);
  if (err) {
    return NULL;                                    /* no report file name */
  }

  report_filename = default_report_filename(ctrl->set.image_filename);

  if (ctrl->set.report_stream != NULL) {
    assert(0);                                    /* open stream only once */
  }

  ctrl->set.report_stream = xplab_open_file(report_filename, "w");

  if (ctrl->set.report_stream) {
    char *settings = NULL;

    if (verbose) {
      fprintf(stdout, "report file '%s' opened\n", report_filename);
    }

    settings = print_settings(*ctrl, RPT_HEADER, db_name);
    xreport("%s", settings);                       /* write to report file */

    FREEMEM(settings);

    if (pr.rpt_store) {              /* empty general auxiliary text store */
      xreport("%s", pr.rpt_store);                 /* get rid of old stuff */

      FREEMEM(pr.rpt_store);
    }
  }

  return report_filename;
}



/* Closes the report.

Revision: 01.11.2003 E.Sack

Input : - setting
        - verbose flag
Output: - report stream closed 
*/
static void
close_rpt_file(char *report_filename, set_def set, int verbose)
{
  int  err, clos;
  char *buf = NULL;

  err = check_name(report_filename);
  if (err) {
    cat_string("<", "unknown", ">", &buf);
  }
  else {
    cat_string("'", report_filename, "'", &buf);
  }

  if (set.report_stream) {
    clos = xplab_close_file(set.report_stream, buf);
    set.report_stream = NULL;

    if (verbose) {
      if (!clos) {
	fprintf(stdout, "report file %s closed\n", buf);
      }
      else {
	fprintf(stdout, "WARNING: cannot close report file %s\n", buf);
      }
    }
  }

  FREEMEM(buf);
}
