/*****************************************************************************
This is a Pattern Recognition program
Copyright (C) 2006  Elmar Sack

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Elmar.Sack@mnet-online.de
*****************************************************************************/

/*#define NDEBUG*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <float.h>
#include <stdarg.h>
#include <time.h>
#include <gtk/gtk.h>
#include "xplab_init_defs.h"
#include "xplab_init_types.h"
#include "xplab_gtk_types.h"
#include "xplab_debug.h"
#include "xplab_main_fe.h"
#include "xplab_main_be.h"
#include "xplab_algorithm.h"
#include "xplab_recall_be.h"
#include "xplab_output.h"
#include "xplab_report.h"
#include "xplab_image.h"
#include "xplab_dchannel.h"
#include "xplab_tprint.h"
#include "xplab_training_be.h"
#include "xplab_preferences_fe.h"
#include "xplab_preferences_be.h"
#include "xplab_batch.h"

/* local definitions *******************************************************/
typedef struct {
  int        uniq, uniq_old, rut_old;
  pic_def    *rpic, *wpic;
  cand_lst   *cdl;
  sector_def *sec;
} support_def; 

/* local prototypes ********************************************************/
static void init_support_def(support_def *sd);
static void free_support_def(support_def *sd);
static cand_lst matching_method_1(pic_def *rpic, match_def mat, 
				  int single_line, rb_lst rblst,
				  cls_lst clst, int fdomain, run_def *crun,
				  int single_nb, MATCHING_METHOD mtd);
static void matching_method_2(pic_def *rpic, cand_lst *hit_lst, match_def mat,
			      rb_lst rblst, cls_lst clst, run_def *crun,
			      int single_nb);
static int candidate_list_horizontal_fd(match_def mat, cand_lst hlist, 
					int x, sector_def rbs, 
					int r, rb_lst rblst, cls_lst cl, 
					rb_ref rref, rc_lst *rlist, 
					run_def *crun, int single_nb, 
					MATCHING_METHOD mtd);
static area_def max_nb_area(nb_lst nbl);
static int contour_fill_all(pic_def pic, sector_def fr, PICT search, 
			    PICT fill);
static void crop(pic_def src, pic_def dst, sector_def clr, pic_def ref, int dx, 
		 int dy, int refx, int refy, int cpy);
static int stepping(pic_def src, sector_def roi, int step_increment, 
		    double interleave);
static int lookahead(cand_lst klist, cls_def *cls, int step_increment, 
		     double interleave);
static nb_ref gen_nb_ref(pic_def pat, int mask_border, int maskweight);
static void delete_non_competitors(cand_lst *klist, cls_def *cls, double nb_lap);
static void sort_out(int jitter, double corrlevel, cand_lst hlist,
		     cls_def *cls, cand_lst *list, pic_def pic, int single);
static int line_rule(ival_def nb_row, ival_def rb_row, ALIGMNT zeroline);
static sector_def check_cnb(sector_def rbs, area_def nnb_ar, int x, int y, 
			    area_def cnb_ar, sector_def rb_sec, area_def par);
static sector_def check_broken(sector_def rbs_in, area_def nnb_ar, 
			       area_def par, int mborder);
static void pattern_fragments_i(int_lst *brb, sector_def rbs_broken, int rut,
			        rb_lst rblst);
static void pattern_fragments_ii(cand_lst *hit_lst, rb_lst rblst);
static void check_hit_sd(match_def mat, int rut, int n, int c, cls_def *cls,
			 sector_def rbs, sector_def rbse, rb_ref rref, 
			 cand_lst *klist, sector_def rbs_org, rb_lst rblst,
			 MATCHING_METHOD mtd);
static void check_hit_fd(match_def mat, int r, int n, int c, cls_def *cls,
			 sector_def rbs, sector_def rbse, rb_ref rref, 
			 cand_lst *klist, sector_def rbs_old, rb_lst rblst, 
			 cv_def_i kmark);
static void trans_lag(ival_def org, ival_def *rbs, ival_def *rbse, int nb_len, 
		      int pic_len, int mask_border, int noise_border, int *lag);
static int lookup_rc_lst(rc_lst rlist, sector_def hdx);
static int create_rlist(cand_def kr, double kwert, cand_lst *zlist, rc_lst *rlist, 
			int gap);
static double mass_of_image(pic_def pic);
static double mass_of_masked_image(pic_def mask, pic_def pic, int mbv);
static void sifting_gap(cand_lst vlist, int *idx, int *yet_created, rc_lst rlist,
			sector_def rbs, int noise_border);
static int sifting_lap_h(cand_def cand, int idx, int yet_created, rc_lst rlist,
			 cls_lst cl, double overlap);
static void trim_vertical_list(cand_lst *vlist, int start);
static int gap_sum(int idx, int yet_created, rc_lst rlist, ival_def cand, 
		   ival_def rbs, int *ref_end);
static int add_substring(cand_lst vlist, int *idx, int *yet_created, rc_lst *rlist,
			 sector_def rbs, cls_lst cl);
static int best_candidate(cand_lst vlist, int *idx, rc_lst rlist, cls_lst cl, 
			  ALIGMNT align_class, double *corr_max);
static void reduce_alignment_class(cand_lst vlist, int *idx, rc_lst rlist, 
				   cls_lst cl, int best);
static void reduce_alignment(cand_lst vlist, int *idx, rc_lst rlist, cls_lst cl,
			     int best, ALIGMNT align_class);
static int check_nb(sector_def *rbs, sector_def *rbse, int lhead, 
		    area_def nar, pic_def rpic, int mborder, int nborder, 
		    int single_nb, MATCHING_METHOD mtd);
static int candidate_list_horizontal_sd(match_def mat, cand_lst hlist,
					sector_def rbs, int rut, 
					rb_lst rblst, cls_lst cl,
					rb_ref *rref, rc_lst *rlist, 
					run_def *crun, int single_nb, 
					MATCHING_METHOD mtd);
static void candidate_list_vertical_sd(match_def mat, sector_def rbs, 
				       int lhead, int rut, rb_lst rblst,
				       int ndx, int cdx, cls_def *cls,
				       cand_lst *klist, rb_ref rref,
				       run_def *crun, int single_nb, 
				       MATCHING_METHOD mtd);
static void candidate_list_vertical_fd(match_def mat, cv_def_i kmark, 
				       int lhead, int r, rb_lst rblst,
				       int n, int c, cls_def *cls, 
				       cand_lst *klist, rb_ref rref, 
				       run_def *crun, int single_nb, 
				       MATCHING_METHOD mtd);
static cand_lst vertical_list_part_1_sd(match_def mat, sector_def *rbs, 
					int rut, rb_lst rblst,
					rb_ref rref, cls_lst cl, 
					run_def *crun, int single_nb, 
					MATCHING_METHOD mtd);
static cand_lst vertical_list_part_2_sd(match_def mat, int lhead, 
					sector_def rbs, int rut, 
					rb_lst rblst, rb_ref rref, 
					cls_lst cl, run_def *crun, 
					int single_nb, MATCHING_METHOD mtd);
static cand_lst vertical_list_part_1_fd(match_def mat, int *lx, int *start_x,
					int r, rb_lst rblst, rb_ref rref, 
					cls_lst cl, run_def *crun, 
					int single_nb, MATCHING_METHOD mtd);
static cand_lst vertical_list_part_2_fd(match_def mat, int lhead, int x,
					ival_def spalte, int r, 
					rb_lst rblst, rb_ref rref, 
					cls_lst cl, run_def *crun, 
					int single_nb, MATCHING_METHOD mtd);
static void check_hit_cnb(match_def mat, sector_def rbs_in, int lhead, 
			  int rut, rb_lst rblst, int n, int c, cls_def *cls, 
			  cand_lst *klist, rb_ref rref, run_def *crun, 
			  int single_nb, MATCHING_METHOD mtd);
static int_lst cnb_validation(int c, int n, cls_def *cls, sector_def rbs, 
			      int x, int y, rb_ref rref, int rut, rb_lst rblst,
			      int *fail, int nnbmatch, int mborder);
static int_lst cnb_validation_by_pixel(int c, int n, cls_def *cls, sector_def rbs, 
				       int lx, int ly, rb_ref rref, int rut, 
				       rb_lst rblst, int *fail, int nnbmatch,
				       int mborder);
static int cnb_validation_by_contour(int c, int n, cls_def *cls, sector_def rbs, 
				     int lx, int ly, pic_def pic);
static rb_ref gen_rb_references(match_def mat, pic_def *rpic, cls_lst cl, 
				rb_lst rblst, int fdomain, int single_nb,
				MATCHING_METHOD mtd);
static void gen_nb_references(nb_lst nbl, int mask_border, int maskweight);
static cand_lst compose_hit_list(tiles_def rcl);
static fft_def gen_fnb(pic_def pic, area_def ar);
static void part_frb(part_lst pl);
static kwh_lst gen_kwh(match_def mat, cls_lst cl, rb_lst rblst, pic_def rpic, 
		       int single_nb, MATCHING_METHOD mtd);
static pic_def xpic_def_filling(pic_def in);
static pic_def xpic_def_dilating(pic_def in, int border);
static area_def max_rb_area(rb_lst rbl);
static int min_bound(double corrlevel, double **cv, area_def ar, double mass_nb, double mass_dnb, 
		     area_def lub);
static void merge_kwh(lag_lst *ll, double **cvmin, double **cvmax, area_def ar, 
		      int cls, int nb, double mass_nb, double mass_dnb, 
		      area_def lub);
static int rule_nb_size(ival_def b_ivl, int r_nb, int max_lag, int single_nb);
static area_def gen_lub(sector_def bsec, area_def nar, int mask_border, 
			area_def noise_border);
static part_lst part_rb(sector_def rbs, pic_def rpic, area_def klen, int part_x, 
			int border);
static fft_def total_corr(part_lst rp, fft_def fnb, area_def nar, int part_x, 
			  int mask_border);
static int interval_overlap(ival_def iva, ival_def ivb, int jitter);
static void purge_rb_chains(rb_chn_lst rch);
static void merge_rb_chains(rb_chn_lst *rch, run_def *crun);
static void neighborhood_of_rb(cls_lst cls, rb_lst rblst, sector_def rsec);
static void neighborhood_of_nnb(cls_def *cls, rb_lst rblst, nb_def nnb, 
				sector_def in_sec);
static sector_def neighborhood_of_cnb(cls_def *cls, int cnb_c, int cnb_n, 
				      sector_def in_sec);
static void neighbors_of_rb(rb_lst rblst, sector_def nbh);
static void get_rb_chain(rb_lst rblst, rb_lst *rb_chain);
static sector_def chain_frame_comp(rb_lst rb_chain);
static void chain_frames_lst(rb_chn_lst rcl);
static int max_chain_sector(rb_chn_lst rch);
static void reset_tiles_def(rb_chn_lst rch);
static void set_tiles_def(rb_chn_lst rch, RB_STATUS from, RB_STATUS to);
static void reset_nb_aux(cls_lst cl);
static int inner_frame_rule(rb_chn_lst *rch, int *tr);
static int touch_frame_rule(rb_chn_lst *rch, int *tr, sector_def tsector, 
			    double pfac);
static int spanning_sector(rb_chn_lst rch, sector_def s_a, sector_def s_b);
static sector_def merge_chain_frame(sector_def sec_1, sector_def sec_2);
static rb_chn_lst chain_neighborhoods(cls_lst cls, rb_lst rblst, run_def *crun);
static void chain_collect_rule(rb_chn_lst *rch);
static void chain_extend_rule(rb_chn_lst *rch, area_def par, int tile_size);
static int extension_start(rb_chn_lst *rch, int *tr, double pfac, area_def par);
static int circulate_extension(rb_chn_lst *rch, int *tr, double pfac, area_def par);
static int extend_frame_rule(rb_chn_lst *rch, int tr, double pfac, area_def par);
static int extend_frame_left(rb_chn_lst *rch, int *tr, double pfac);
static int extend_frame_left_up(rb_chn_lst *rch, int *tr, double pfac);
static int extend_frame_up(rb_chn_lst *rch, int *tr, double pfac);
static int extend_frame_up_right(rb_chn_lst *rch, int *tr, double pfac, 
				 area_def par);
static int extend_frame_right(rb_chn_lst *rch, int *tr, double pfac, area_def par);
static int extend_frame_right_down(rb_chn_lst *rch, int *tr, double pfac, 
				   area_def par);
static int extend_frame_down(rb_chn_lst *rch, int *tr, double pfac, area_def par);
static int extend_frame_down_left(rb_chn_lst *rch, int *tr, double pfac, 
				  area_def par);
static double calc_tile_factor(area_def par, int tile);
static void calc_extension_distance(area_def par, sector_def dut, int *lw, int *uw, 
				    int *rw, int *dw);
static pic_def assemble_tile(pic_def rpic, rb_chn_def rcd, int border);
static int dump_chain_lst(tiles_def *rcl, pic_def *rpic, RB_CHAIN_BIT dflag,
			  int match, char *filename);
static void dump_colored_chain_lst(rb_chn_lst rch, pic_def rpic, RB_CHAIN_BIT dflag,
				   char *filename, int rawbit);
static void reset_rb_number(rb_chn_def rc);
static void reset_rb_sector(cand_lst hlst, int dx, int dy);
static void evaluate_rb_lst(rb_lst rblst, tiles_def rcl);
static pic_def minimum_of_colors(rb_chn_lst rch, pic_def rpic, int rawbit);
static pic_def maximum_of_colors(rb_chn_lst rch, pic_def rpic, int rawbit);
static pic_def random_colors(rb_chn_lst rch, pic_def rpic, int rawbit);
static int search_cnb_competitor(cand_lst hit_lst, int ci, int rb);
static void register_negative_result(reg_def vregister, int column, 
				     ival_def y_range, char *key);
static void register_positive_result(reg_def vregister, int column, 
				     ival_def y_range, char *key, cand_lst vlist);
static REGISTER_INDEX check_register(reg_def vregister, int column, 
				     ival_def y_range, char *key, int *positive_idx);
static void compose_key(int cnr, int nnr, char **key);
static void reduce_key(char **key);
static int check_inh_lst(inh_lst ilist, INH_FLAG iflag, sector_def sector);
static void update_inh_lst(inh_lst *il, INH_FLAG iflag, sector_def loc);
static int inhibition_sum(cls_lst clst);
static void inhibition_reset(cls_lst clst);
static void sort_out_inhibited_classes(cls_def *cls, cand_lst *vcl);
static void inhibit_classes_by_cnb(double corrlevel, cand_lst hlist, cls_def *cls, cand_lst *vcl, 
				   int jitter, pic_def pic);
static void inhibit_classes_by_nb_alignment(cls_def *cls, cand_lst *vcl, int jitter);
static void inhibit_classes_by_rb_alignment(cls_def *cls, cand_lst *vcl, int jitter);
static void sort_out_scions(cls_def *cls, cand_lst *vcl, int jitter);
static void sort_out_doublets(cls_def *cls, cand_lst *vcl, int jitter);
static sector_def max_frame(cand_lst cl);
static sector_def max_sector(sector_def sec_a, sector_def sec_b);
static void uncrop_cand_lst(pic_def rpic, pic_def wpic, cls_def *extern_cls, 
			    support_def *intern);
static char *crop_cand_lst(cand_lst in, pic_def rpic, pic_def wpic, 
			   cls_def *cls, int uniq, support_def *intern);
static int class_inhibition_loop(match_def mat, cand_lst hlist, 
				 sector_def rbs, int rut, 
				 rb_lst rblst, cls_lst clst, 
				 rb_ref *rref, rc_lst *rlist, run_def *crun, 
				 int single_nb, MATCHING_METHOD mtd);
static cand_lst single_line_strings(match_def mat, cand_lst hlist, 
				    sector_def rbs, int rut, rb_lst rblst, 
				    cls_lst clst, rb_ref *rref, 
				    rc_lst *rlist, int reset_uniq, 
				    run_def *crun, support_def *sup, 
				    int single_nb, MATCHING_METHOD mtd);
static ival_lst multi_line_separation(cand_lst hit_lst, sector_def multi_sc, 
				      cls_lst clst);
static REGISTER_INDEX compare_image_state(char *key_a, char *key_b);
static REGISTER_INDEX compare_range(ival_def range_a, ival_def range_b);
static int join_range_keys(ival_def range_a, ival_def *range_b);
static void restricted_nb_loop(match_def mat, sector_def rbs, int lhead, 
			       int rut, rb_lst rblst, cls_lst clst,
			       cand_lst *vcl_sum, rb_ref rref, 
			       cand_lst vcl_root, run_def *crun, 
			       int single_nb, MATCHING_METHOD mtd);
static void allocate_image_store(pic_def **expic, sector_def max_sec, 
				 int offset_x, int offset_y);
static int h_contour_fill(h_contour_lst ctr, PICT to, pic_def pic);
static v_contour_lst convert_h_contour_lst(h_contour_lst outer);
static int flood_fill_all(h_contour_lst outer, PICT mark, pic_def pic);
static int flood_fill_horizontal(h_contour_lst horizontal, PICT from, PICT to, 
				 pic_def pic);
static int flood_fill_vertical(v_contour_lst vertical, PICT from, PICT to, 
			       pic_def pic);
static int overlap_list(cand_lst hlist, cls_def *cls, cand_def cand, int jitter);
static void validate_rb_lst(rb_lst rblst, cls_lst cl);
static cand_lst multi_line_strings(match_def mat, int single_line, 
				   cand_lst hlist, int rut, rb_lst rblst, 
				   cls_lst clst, rb_ref *rref, 
				   rc_lst *rlist, run_def *crun, 
				   int single_nb, MATCHING_METHOD mtd);
static void check_cnb_competition(double corrlevel, rb_lst rblst, rb_ref rref, cand_lst *hit_lst, 
				  cls_def *cls);
static void check_gap(cand_lst *hlst, rb_lst rbl, int gap_criterion);
static cand_def substitution(int rb_nr, sector_def sec, int sub);
static pic_def xpic_def_weighting(pic_def in, int maskweight);
static rb_lst merge_crb(cand_lst hlist, rb_lst rbl);
static sector_def expand_sector(sector_def src, sector_def dst);

/* Functions ***************************************************************/

static void 
init_support_def(support_def *sd) 
{
  assert(sd);

  sd->uniq = sd->uniq_old = sd->rut_old = 0;

  sd->rpic = NULL;                                         /* zero pointer */
  sd->wpic = NULL;                                         /* zero pointer */
  sd->cdl = NULL;                                          /* zero pointer */
  sd->sec = NULL;                                          /* zero pointer */
}

static void 
free_support_def(support_def *sd) 
{
  assert(sd);

  if (sd->rpic) {
    free_pic_def(sd->rpic);
    FREEMEM(sd->rpic);
  }

  if (sd->wpic) {
    free_pic_def(sd->wpic);
    FREEMEM(sd->wpic);
  }

  if (sd->cdl) {
    free_cand_lst(sd->cdl);
    FREEMEM(sd->cdl);
  }

  if (sd->sec) {
    FREEMEM(sd->sec);
  }

  init_support_def(sd);
}



/* Wrapper for pattern matching RB chains. Instead of the whole document
   certain pieces, called tiles, which contains several RB chains, are
   matched against the database. All tiles together builds the document.
   This is mainly made due to memory optimization. If tiles are unsatisfying
   the document's RB list is used as a fallback.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - flag (0: multi line, 1: single line)
        - RB list
        - RB chain list
        - class list
        - pointer to RB image
        - pointer to TPS ID
        - pointer to input image filename
        - domain
        - chain control flag
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - candidate list
        - control structure updated
*/
cand_lst
xtiled_pattern_matching(match_def mat, int single_line, rb_lst rblst, 
			tiles_def rcl, cls_lst clst, img_typ *rpic, 
			char *ident, char *filename, int fdomain, 
		        RB_CHAIN_BIT rb_flag, run_def *crun, int single_nb,
			MATCHING_METHOD mtd)
{
  int      c, mgv, found, amount = AREA_SIZE(rpic->pimg.ar);
  cand_lst total;

  init_cand_lst(&total);

  if ((rcl.rch.cnt <= 1)  ||                          /* ignore single tile */
      (!(rb_flag & RB_CHAIN_USE))) {                            /* no tiles */
    if (xctrl.cfg.verbose > VERBOSE_DEFAULT) {
      char *buf = NULL;
      cat_string("pattern matching '", filename, "' using TPS ID '", &buf);
      cat_string(ident, "'\n", "", &buf);
      FREEMEM(buf);
    }

    total = xpattern_matching(mat, single_line, rblst, clst, &rpic->pimg, 
			      fdomain, crun, single_nb, mtd);
    return total;
  }

  for (c = found = 0; c < rcl.rch.cnt; c++) {  /* seeking first loaded tile */
    if (rcl.rch.rcd[c].caux != BOOKED) {
      continue;
    }

    int_to_log(main_log, "pattern matching tile no.", c, " using TPS ID ");
    string_to_log(main_log, "'", ident, "'\n");

    rcl.rch.rcd[c].bhclst = xpattern_matching(mat, single_line, 
					      rcl.rch.rcd[c].rb_chain, clst, 
					      &rpic->pimg, fdomain, crun, 
					      single_nb, mtd);
    found = 1;
    break;
  }
  assert(found);

  for (c = 0; c < rcl.rch.cnt; c++) {                      /* rest of tiles */
    if (rcl.rch.rcd[c].caux == BOOKED) {
      continue;
    }

    free_img_typ(rpic);

    assert(rcl.rch.rcd[c].caux == PIC_DUMPED);     /* reload tile from disk */
    *rpic = read_pnm_file(rcl.rch.rcd[c].chain_name, &mgv, 
			  xctrl.cfg.verbose >= VERBOSE_LOT);
    assert(rpic->typ==P4);                            /* must be raw Bitmap */
    if (!(rb_flag & RB_CHAIN_DUMP_TILES)) {
      remove(rcl.rch.rcd[c].chain_name);           /* remove file from disk */
    }
    rcl.rch.rcd[c].caux = BOOKED;                            /* tile loaded */
    amount = XPMAX(amount, AREA_SIZE(rpic->pimg.ar));

    int_to_log(main_log, "pattern matching tile no.", c, " using TPS ID ");
    string_to_log(main_log, "'", ident, "'\n");

    rcl.rch.rcd[c].bhclst = xpattern_matching(mat, single_line, 
					      rcl.rch.rcd[c].rb_chain, clst, 
					      &rpic->pimg, fdomain, crun,
					      single_nb, mtd);
    found++;
  }
  assert(found==rcl.rch.cnt);

  evaluate_rb_lst(rblst, rcl);                       /* mark backannotation */
  total = compose_hit_list(rcl);                     /* hit backannotation  */

  assert(rcl.doc.caux == PIC_DUMPED);
  free_img_typ(rpic);                               /* delete current image */
                                                         /* reload document */
  *rpic = read_pnm_file(rcl.doc.chain_name, &mgv, 
			xctrl.cfg.verbose>=VERBOSE_LOT);
  assert(rpic->typ==P4);                              /* must be raw Bitmap */
  remove(rcl.doc.chain_name);                      /* remove file from disk */
  rcl.doc.caux = BOOKED;

  double_to_log("internal memory usage ", 
		(double)(amount*100)/(double)(AREA_SIZE(rpic->pimg.ar)),"%\n");

  return total;
}



/* Main algorithm for pattern matching, assembles all strings from an image.

Revision: 02.01.2005 E.Sack

Input : - matching parameters
        - flag (0: multi line, 1: single line)
        - RB list
        - class list
        - pointer to RB image
        - domain
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - candidate list
        - control structure updated
Postcondition: - RBs in RB list marked as booked
Usage: There are several matching methodes available, however, the main
       matching method always is performed first.
*/
cand_lst
xpattern_matching(match_def mat, int single_line, rb_lst rblst, cls_lst clst, 
		  pic_def *rpic, int fdomain, run_def *crun, int single_nb, 
		  MATCHING_METHOD mtd)
{
  cand_lst hit_lst;

  assert(crun);

  validate_rb_lst(rblst, clst);                 /* invalidate too small RB */

  /* main pattern matching method first                                    */
  hit_lst = matching_method_1(rpic, mat, single_line, rblst, clst, fdomain,
			      crun, single_nb, MATCHING_METHOD_1);

  /* broken pattern matching method next                                   */
  if (mtd == MATCHING_METHOD_2) {
    matching_method_2(rpic, &hit_lst, mat, rblst, clst, crun, single_nb);
  }

  return hit_lst;
}



/* Matching algorithm for broken patterns.

Revision: 06.01.2005 E.Sack

Input : - pointer to image
        - pointer to candidate list
        - matching parameters
        - RB list
        - class list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - candidate list updated
        - control structure updated
Precondition:  - the hit list must be consistent to the RB list.
               - the hit list may be empty
               - RBs in RB list marked as booked
Postcondition: - RBs in RB list marked as booked
Usage:  - this method is a derivative from the main algorithm.
        - Since a broken pattern has a reduced detection probability,
          this method shall be used restrictively, hence the following rules
          shall be applied:
          R1: The method shall be used after the main algorithm.
          R2: Only completely unrecognized RBs shall be tested.
          R3: The test area of the RB shall be enlarged in that direction by
              the NBs size, which is smaller. Recognized RBs limit that area
              as well as the image size.
          R4: The RB contains only one text line (single line).
          R5: There are no invalid RBs.
          R6: The method shall not be used for
              - Training,
              - Clustering.
*/
static void
matching_method_2(pic_def *rpic, cand_lst *hit_lst1, match_def mat,
		  rb_lst rblst, cls_lst clst, run_def *crun, int single_nb)
{
  int      cd, rut, x, y;
  int      fdomain = 0;                        /* only spatial domain used */
  int      single_line = 1;                  /* rule: only one line per RB */
  pic_def  broken_pat;
  area_def bar;
  cand_lst hit_lst2;
  rb_lst   vrblst;

  assert(rpic);
  assert(hit_lst1);
  assert(crun);

  if (!hit_lst1->c) {                                 /* consistency check */
    assert(!hit_lst1->cnt);
  }
  else {
    assert(hit_lst1->cnt>0);
  }

  init_pic_def(&broken_pat);
  init_rb_lst(&vrblst);

  broken_pat.ar = rpic->ar;
  CALLOC(PICT, AREA_SIZE(broken_pat.ar), broken_pat.mem);

  for (cd = 0; cd < hit_lst1->cnt; cd++) {  /* search for unrecognized RBs */
    if (hit_lst1->c[cd].n != SUBSTRB_FLAG) {
      continue;                   /* rule: avoid (partially) recognized RB */
    }

    rut = hit_lst1->c[cd].rbn;                              /* index of RB */
    rblst.rb[rut].rld = RB_STATUS_INIT;                 /* reset RB status */

    SECTOR_TO_AREA(rblst.rb[rut].sc, bar);
    x = rblst.rb[rut].sc.column.start;
    y = rblst.rb[rut].sc.line.start;

    /* copy unrecognized RBs, no recognized RBs now                        */
    copy_bmps(*rpic, broken_pat, bar, x, y, x, y);     /* copy original RB */

    delete_cand_def(hit_lst1, cd--);      /* erase candidate from hit list */
  }

  copy_rb_lst(&rblst, &vrblst);                /* protect original RB list */

  for (rut = 0; rut < vrblst.cnt; rut++) {         /* rule: no invalid RBs */
    if (vrblst.rb[rut].rld != INVALID) {     /* search for invalidated RBs */
      continue;
    }

    vrblst.rb[rut].rld = RB_STATUS_INIT;                /* reset RB status */

    SECTOR_TO_AREA(vrblst.rb[rut].sc, bar);
    x = vrblst.rb[rut].sc.column.start;
    y = vrblst.rb[rut].sc.line.start;

    /* copy validated RBs */
    copy_bmps(*rpic, broken_pat, bar, x, y, x, y);     /* copy original RB */
  }

  hit_lst2 = matching_method_1(&broken_pat, mat, single_line, vrblst, clst,
			       fdomain, crun, single_nb, MATCHING_METHOD_2);

  cat_cand_lst(&hit_lst2, hit_lst1);                /* append new hit list */

  free_pic_def(&broken_pat);                     /* free temporary buffers */
  free_cand_lst(&hit_lst2);
  free_rb_lst(&vrblst);
}



/* Main algorithm for pattern matching, assembles all strings from an image.

Revision: 05.08.2005 E.Sack

Input : - pointer to image
        - matching parameters
        - flag (0: multi line, 1: single line)
        - RB list
        - class list
        - domain
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
        - method
Output: - candidate list
        - control structure updated
Postcondition: - RBs in RB list marked as booked
*/
static cand_lst
matching_method_1(pic_def *rpic, match_def mat, int single_line, 
		  rb_lst rblst, cls_lst clst, int fdomain, run_def *crun,
		  int single_nb, MATCHING_METHOD mtd)
{
  int        i, j, rut, idx;
  double     cv_threshold;
  rb_ref     rref;                                        /* RB references */
  rc_lst     rlist;                 /* recycling list contains all strings */
  cand_lst   hit_lst, hit_lst_bos, *hcl = NULL;
  sector_def rbs;                                        /* clipping of RB */

  assert(rpic);
  assert(crun);

  init_cand_lst(&hit_lst);
  init_cand_lst(&hit_lst_bos);
  init_rc_lst(&rlist);

  rref = gen_rb_references(mat, rpic, clst, rblst, fdomain, single_nb, mtd);

  cv_threshold = (mtd == MATCHING_METHOD_2) ? mat.corrbroken : mat.corrlevel;

  xplab_progress_configure(crun, "matching phase");
  xplab_progress_set_value(crun, 1.0, (gdouble)rblst.cnt);

  for (rut = 0; rut < rblst.cnt && *crun->pactive == TRUE; rut++) {/* RB under test */
    xplab_progress_set_value(crun, (gdouble)(rut+1), (gdouble)rblst.cnt);

    if ((rblst.rb[rut].rld == BOOKED) || (rblst.rb[rut].rld == INVALID)) {
      continue;
    }

    rbs = rblst.rb[rut].sc;                                   /* short cut */

    if (fdomain) {
      idx = candidate_list_horizontal_fd(mat, hit_lst, 0, rbs, rut, rblst, 
					 clst, rref, &rlist, crun, single_nb,
					 mtd);
      hcl = idx >= 0 ? &rlist.s[idx].r : NULL;
    }
    else {
      free_cand_lst(&hit_lst_bos);          /* contains multi-line strings */

      hit_lst_bos = multi_line_strings(mat, single_line, hit_lst, rut, rblst,
				       clst, &rref, &rlist, crun, single_nb,
				       mtd);
      hcl = hit_lst_bos.cnt ? &hit_lst_bos : NULL;/* optimal multi-line strings */
    }

    if (!hcl) {                   /* whole RB not recognized: substitution */
      cat_cand_def(substitution(rut, rbs, SUBSTRB_FLAG), &hit_lst);
    }
    else {
      for (i = 0; i < hcl->cnt; i++) {
	for (j = 0; j < hcl->c[i].crb.cnt; j++) {
	  rblst.rb[hcl->c[i].crb.nr[j]].rld = BOOKED;        /* CRB booked */
	}
      }

      cat_cand_lst(hcl, &hit_lst);                  /* append new hit list */
    }

    rblst.rb[rut].rld = BOOKED;                               /* RB booked */
  }                                                       /* all RB tested */

  check_cnb_competition(cv_threshold, rblst, rref, &hit_lst, clst.cls);
  check_gap(&hit_lst, rblst, mat.mborder+2*mat.nborder);
  pattern_fragments_ii(&hit_lst, rblst);

  free_cand_lst(&hit_lst_bos);
  free_rc_lst(&rlist);
  free_rb_ref(&rref);

  return hit_lst;
}



/* The RB marks of the tiles are copied back to the document's RB list.

Revision: 26.03.2002 E.Sack

Input : - RB list of the document
        - RB chain list
Output: - RB list updated
*/
static void
evaluate_rb_lst(rb_lst rblst, tiles_def rcl)
{
  int r, c, h;

  for (c = 0; c < rcl.rch.cnt; c++) {
    for (h = 0; h < rcl.rch.rcd[c].rb_chain.cnt; h++) {
      r = rcl.rch.rcd[c].rb_chain.rb[h].raux;           /* original RB no. */
      rblst.rb[r].rld = rcl.rch.rcd[c].rb_chain.rb[h].rld;/* backannotate mark */
    }
  }
}



/* The document's hitlist is composed from the RB chains of all tiles.

Revision: 26.03.2002 E.Sack

Input : - RB chain list
Output: - total candidate hit list
*/
static cand_lst
compose_hit_list(tiles_def rcl)
{
  int   c;
  cand_lst comp;

  init_cand_lst(&comp);

  for (c = 0; c < rcl.rch.cnt; c++) {
    reset_rb_number(rcl.rch.rcd[c]);                        /* reset RB no. */
    reset_rb_sector(rcl.rch.rcd[c].bhclst,                /* reset RB place */
		    rcl.rch.rcd[c].chain_frame.column.start-rcl.border, 
		    rcl.rch.rcd[c].chain_frame.line.start-rcl.border);

    cat_cand_lst(&rcl.rch.rcd[c].bhclst, &comp);       /* catenate hitlists */
  }

  return comp;
}



/* The RB sectors in the hitlist are calculated according to the document.

Revision: 26.03.2002 E.Sack

Input : - hit list
        - tile's offset to the document
Output: - hit list updated
*/
static void
reset_rb_sector(cand_lst hlst, int dx, int dy)
{
  int k;

  for (k = 0; k < hlst.cnt; k++) {                       /* hitlist entries */
    hlst.c[k].roi.column.start += dx;
    hlst.c[k].roi.column.end += dx;
    hlst.c[k].roi.line.start += dy;
    hlst.c[k].roi.line.end += dy;

    hlst.c[k].ext.column.start += dx;
    hlst.c[k].ext.column.end += dx;
    hlst.c[k].ext.line.start += dy;
    hlst.c[k].ext.line.end += dy;

    hlst.c[k].knl.column.start += dx;
    hlst.c[k].knl.column.end += dx;
    hlst.c[k].knl.line.start += dy;
    hlst.c[k].knl.line.end += dy;

    hlst.c[k].hdx.column.start += dx;
    hlst.c[k].hdx.column.end += dx;
    hlst.c[k].hdx.line.start += dy;
    hlst.c[k].hdx.line.end += dy;
  }
}



/* The RB numbers of the tiles are copied back to the document's RB numbers.

Revision: 26.03.2002 E.Sack

Input : - hit list
        - tile's offset to the document
Output: - hit list updated
*/
static void
reset_rb_number(rb_chn_def rc)
{
  int k, r, j;

  for (k = 0; k < rc.bhclst.cnt; k++) {                  /* hitlist entries */
    r = rc.bhclst.c[k].rbn;                              /* RB no. in chain */
    rc.bhclst.c[k].rbn = rc.rb_chain.rb[r].raux;            /* original no. */

    for (j = 0; j < rc.bhclst.c[k].crb.cnt; j++) {
      r = rc.bhclst.c[k].crb.nr[j];                      /* RB no. in chain */
      rc.bhclst.c[k].crb.nr[j] = rc.rb_chain.rb[r].raux; /* original RB no. */
    }
  }
}



/* Assembles all strings from a multi-line RB sector. Without any further
   Information a RB is assumed to contain several lines of text. This routine
   separates lines within a RB recursively.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - flag (0: multi line, 1: single line)
        - candidate list
        - RB under test index
        - RB list
        - class list
        - pointer to RB references
        - pointer to recycling list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - multi line candidates list
        - recycling list updated
        - control structure updated
*/
static cand_lst
multi_line_strings(match_def mat, int single_line, cand_lst hlist, int rut,
		   rb_lst rblst, cls_lst clst, rb_ref *rref, rc_lst *rlist, 
		   run_def *crun, int single_nb, MATCHING_METHOD mtd)
{
  int         i;
  sector_def  rbs;                                       /* clipping of RB */
  cand_lst    hit_lst_sos;                   /* sub-optimal candidate list */
  cand_lst    hit_lst_bos;                  /* best optimal candidate list */
  rc_lst      rtemp;
  support_def sup;

  assert(rref);
  assert(rlist);

  rbs = rblst.rb[rut].sc;                                /* copy important */

  init_support_def(&sup);
  init_rc_lst(&rtemp);
  init_cand_lst(&hit_lst_bos);

  copy_rc_lst(rlist, &rtemp);                /* protect old recycling list */

  free_ival_lst(&rblst.rb[rut].gridy);  /* generated after first substring */

  /* calculate an initial hit-list treating the RB as a single text line   */
  hit_lst_sos = single_line_strings(mat, hlist, rbs, rut, rblst, clst, 
				    rref, &rtemp, 0, crun, &sup, single_nb,
				    mtd);

  if (single_line) {                                      /* RB qualifying */
    rblst.rb[rut].gridy.cnt = 1;                      /* grid substitution */
    CALLOC(ival_def, 1, rblst.rb[rut].gridy.ival);
    init_ival_def(&rblst.rb[rut].gridy.ival[0]);
    rblst.rb[rut].gridy.ival[0] = rbs.line;
  }
  else {
    rblst.rb[rut].gridy = multi_line_separation(hit_lst_sos, rbs, clst);
  }

  copy_rc_lst(rlist, &rtemp);          /* reset sub-optimal recycling list */

  for (i = 0; i < rblst.rb[rut].gridy.cnt; i++) {/*recheck line candidates */
    free_cand_lst(&hit_lst_sos);

    rbs.line = rblst.rb[rut].gridy.ival[i];        /* RB confined to lines */

    hit_lst_sos = single_line_strings(mat, hlist, rbs, rut, rblst, clst, 
				      rref, &rtemp,
				      rblst.rb[rut].gridy.cnt==1, crun, &sup,
				      single_nb, mtd);
    cat_cand_lst(&hit_lst_sos,          /* gather best optimal strings BOS */
		 &hit_lst_bos);
  }

  free_cand_lst(&hit_lst_sos);

  /* non-overlapping RB separates inhibition location good enaugh          */
  inhibition_reset(clst);

  move_rc_lst(&rtemp, rlist);               /* save optimal recycling list */

  free_support_def(&sup);

  return hit_lst_bos;
}



/* Divides a multi-line RB in single-lines.
   An input candidate list is investigated. Using its sector, the candidates
   position is checked wether the candidates are lying in different text
   lines. If that is true, the sector is divided into horizontal sub-sectors.
   It is ensured, that the sub-sectors cover the original sector.

Revision: 06.12.2002 E.Sack

Input : - candidate list
        - sector
        - class list
Output: - interval list
Precondition: - candidate list and sector must be consistent
Postcondition: - the interval list divides the input sector horizontally.
Usage: - the function is not meant to be used iterative. If there are several
         text lines detected - even partly - , the whole input sector is
         divided.
*/
static ival_lst
multi_line_separation(cand_lst hlst, sector_def multi_sc, cls_lst clst)
{
  int        lnr, nr, cand, start, end, startn;
  int_lstm   iln;
  ival_lst   line;

  init_ival_lst(&line);

  if (!hlst.cnt) {
    assert(!hlst.c);
    return line;                                   /* candidate list empty */
  }

  iln = method_line_overlapping(hlst, clst);/*cand. numbers sorted to lines*/

  if (!iln.cnt) {
    assert(!iln.il);
    return line;                                          /* method failed */
  }

  line.cnt = iln.cnt;                       /* number to sector conversion */
  CALLOC(ival_def, line.cnt, line.ival);

  for (lnr = 0; lnr < iln.cnt; lnr++) { /* lines sorted from top to bottom */
    init_ival_def(&line.ival[lnr]);

    cand = iln.il[lnr].nr[0];
    start = hlst.c[cand].hdx.line.start;                           /* init */
    end   = hlst.c[cand].hdx.line.end;

    for (nr = 1; nr < iln.il[lnr].cnt; nr++) {           /* line start/end */
      cand = iln.il[lnr].nr[nr];
      start = XPMIN(start, hlst.c[cand].hdx.line.start);  /* limited to RB */
      start = XPMAX(start, multi_sc.line.start);
      end = XPMAX(end, hlst.c[cand].hdx.line.end);
      end = XPMIN(end, multi_sc.line.end);
    }

    line.ival[lnr].start = start;
    line.ival[lnr].end = end;
  }

  line.ival[0].start = multi_sc.line.start;             /* close to border */
  line.ival[line.cnt-1].end = multi_sc.line.end;
  end = line.ival[0].end;
  for (lnr = 1; lnr < line.cnt; lnr++) {
    startn = line.ival[lnr].start;

    if (end < startn) {                          
      line.ival[lnr-1].end = (startn+end)>>1;      /* mid-position of gaps */
      line.ival[lnr].start = line.ival[lnr-1].end+1;
    }

    end = line.ival[lnr].end;
  }

  free_int_lstm(&iln);

  return line;
}



/* Assembles all strings from a RB sector assuming single-line. However if
   the sector isn't a single line, the strings found may be sub-optimal.

Revision: 03.01.2005 E.Sack

Input : - matching parameters
        - candidate list
        - RB sector
        - RB under test index
        - RB list
        - class list
        - pointer to RB references
        - pointer to recycling list
        - flag (0: don't reset image state, 1: reset image state)
        - pointer to control structure
        - pointer to support
        - flag (0: multi NB, 1: single NB)
Output: - single line candidates list
        - recycling list updated
        - control structure updated
*/
static cand_lst
single_line_strings(match_def mat, cand_lst hlist, sector_def rbs, int rut,
		    rb_lst rblst, cls_lst clst, rb_ref *rref, rc_lst *rlist,
		    int reset_uniq, run_def *crun, support_def *sup, 
		    int single_nb, MATCHING_METHOD mtd)
{
  int      idx;
  cand_lst hit_lst;

  assert(sup);

  init_cand_lst(&hit_lst);

  if (sup->rut_old - rut) {
    sup->uniq = sup->uniq_old = 0;/*new RB under test, used for repeatability */
    sup->rut_old = rut;
  }
  else {
    if (reset_uniq) {
      sup->uniq = sup->uniq_old;
    }
    else {
      sup->uniq_old = sup->uniq;
    }
  }

  /* repeatedly erase recognized patterns from the image until no more found*/
  while (1) {
    idx = class_inhibition_loop(mat, hlist, rbs, rut, rblst, clst, rref, 
				rlist, crun, single_nb, mtd);

    if (idx < 0) {
      break;                                    /* no new strings detected */
    }

    cat_cand_lst(&rlist->s[idx].r, &hit_lst);      /* gather strings found */

    FREEMEM(rref->pic_state);                          /* new image status */
    rref->pic_state = crop_cand_lst(hit_lst, *rref->rpic, rref->wpic, 
				    clst.cls, sup->uniq++, sup);
  }

  FREEMEM(rref->pic_state);
  uncrop_cand_lst(*rref->rpic, rref->wpic, clst.cls, sup);/* uncrop all strings */

  return hit_lst;
}



/* Inhibits classes according to heuristics and delivers a candidate list
   which are ruled.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - candidate list
        - RB sector
        - RB under test index
        - RB list
        - class list
        - pointer to RB references
        - pointer to recycling list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - index of recycling list (<0: no entry)
        - recycling list updated
        - control structure updated
*/
static int
class_inhibition_loop(match_def mat, cand_lst hlist, sector_def rbs, int rut,
		      rb_lst rblst, cls_lst clst, rb_ref *rref, 
		      rc_lst *rlist, run_def *crun, int single_nb, 
		      MATCHING_METHOD mtd)
{
  int    idx, d, inhibited = 0, new;
  rc_lst rtemp;

  init_rc_lst(&rtemp);

  do {                                            /* class inhibition loop */
    copy_rc_lst(rlist, &rtemp);

    idx = candidate_list_horizontal_sd(mat, hlist, rbs, rut, rblst, clst, 
				       rref, &rtemp, crun, single_nb, mtd);
    d = inhibition_sum(clst);

    new = d - inhibited;
    assert(new>=0);

    inhibited = d;
  } while(new);                        /* recheck if new inhibited classes */

  copy_rc_lst(&rtemp, rlist);

  free_rc_lst(&rtemp);

  return idx;
}



/* Checks for CNB (fractional pattern) competition. This check is necessary
   because CNB competition over multiple RB is not handled properly enaugh 
   in the single RB recognition. The check can be done safely only with the
   complete candidate list at the end of the recognition phase.

   Pseudo code:
   - for all candidates Ci from the candidate list which are CNB
     - for every of the enlisted RBij
       - for every candidate Ck which shares the RBij
         - build the common pattern P by superpositioning Ci and Ck with the
           document.           .
         - (*) ascertain the correlation value Ki and Kj with P and erase the
           candidate which has the smaller correlation value from the list.
   
   (*) This algo may be of any fashion.

   The effect should be, that even a candidate will be erased, who superpose
   e.g., RBa with 100%, but RBb with 0% (a simple NB). However a CNB does
   not win forever. In that case, the CNB is just too weak and its class is
   supposed to be trained once more.

Revision: 06.04.2003 E.Sack

Input : - RB list
        - RB references
        - pointer to candidate list
        - pointer to classes
Output: - candidate list updated
Postcondition: - deletes some image memory in RB references
*/
static void
check_cnb_competition(double corrlevel, rb_lst rblst, rb_ref rref, cand_lst *hit_lst, 
		      cls_def *cls)
{
  int        l, ci, ck, rij, rij_cnt, cnri, nnri, cnrk, nnrk, xi, yi, xk, yk;
  int        rb, **pair = 0;
  double     ki, kk;
  sector_def sec, rbsi, rbsei, rbsk, rbsek;
  pic_def    *workbuf1 = &rref.shdw1, *workbuf2 = &rref.shdw2;

  assert(hit_lst);
  assert(cls);

  if (!hit_lst->c) {
    assert(!hit_lst->cnt);
    return;
  }
  else {
    assert(hit_lst->cnt>0);
  }

  CALLOC(int*, hit_lst->cnt, pair);          /* competitor pair index list */
  for (l = 0; l < hit_lst->cnt; l++) {
    CALLOC(int, hit_lst->cnt, pair[l]);
  }

  set_bitmap(AREA_SIZE(workbuf1->ar), workbuf1->mem, BACKGROUND);
  set_bitmap(AREA_SIZE(workbuf2->ar), workbuf2->mem, BACKGROUND);

  for (ci = 0; ci < hit_lst->cnt; ci++) {
    rij_cnt = hit_lst->c[ci].crb.cnt;

    if (!rij_cnt) {
      continue;                                     /* candidate not a CNB */
    }

    cnri  = hit_lst->c[ci].c;
    nnri  = hit_lst->c[ci].n;

    if (cnri < 0 || nnri < 0) {                       /* RB not recognized */
      continue;
    }

    rb = hit_lst->c[ci].rbn;
    ck = search_cnb_competitor(*hit_lst, ci, rb);                /* anchor */

    if (ck < 0) {
      for (rij = 0; rij < rij_cnt; rij++) {             /* RBij of the CRB */
	rb = hit_lst->c[ci].crb.nr[rij];
	ck = search_cnb_competitor(*hit_lst, ci, rb);

	if (ck >= 0) {                              /* competitor Ck found */
	  break;
	}
      }
    }

    if (ck < 0) {            /* no competitor for anchor or CNB list found */
      continue;
    }

    cnrk  = hit_lst->c[ck].c;
    nnrk  = hit_lst->c[ck].n;

    if (cnrk < 0 || nnrk < 0) {                       /* RB not recognized */
      continue;
    }

    if (pair[ci][ck]) {
      continue;              /* candidate competition pair already handled */
    }

    pair[ci][ck] = 1;              /* tested competition of candidate pair */

                                 /* Repeat correlation of Ci & Ck with     */
                                 /* (document-P). Results are correlation  */
    xi    = hit_lst->c[ci].m.x;  /* values Ki and Kk.                      */
    yi    = hit_lst->c[ci].m.y;  /* If both of the candidates have         */
    rbsi  = hit_lst->c[ci].roi;  /* a significant part of pattern in the   */
    rbsei = hit_lst->c[ci].ext;  /* competing area, the one, who has the   */
                                 /* biggest correlation value, wins.       */
                                 /* If at least one of Ci and Ck has not a */
    xk    = hit_lst->c[ck].m.x;  /* significant part of pattern in the     */
    yk    = hit_lst->c[ck].m.y;  /* competing area, competition is         */
    rbsk  = hit_lst->c[ck].roi;  /* negligible and both candidates are     */
    rbsek = hit_lst->c[ck].ext;  /* left in the list.                      */

    crop(*rref.rpic, *workbuf1, hit_lst->c[ci].knl,  /* P1 = Ci * document */
	 cls[cnri].nbl.nb[nnri].pat, 0, 0, 
	 hit_lst->c[ci].knl.column.start - hit_lst->c[ci].hdx.column.start,
	 hit_lst->c[ci].knl.line.start - hit_lst->c[ci].hdx.line.start,
	 1);

    crop(*workbuf1, *workbuf2, hit_lst->c[ck].knl,         /* P2 = Ck * P1 */
	 cls[cnrk].nbl.nb[nnrk].pat, 0, 0, 
	 hit_lst->c[ck].knl.column.start - hit_lst->c[ck].hdx.column.start,
	 hit_lst->c[ck].knl.line.start - hit_lst->c[ck].hdx.line.start,
	 1);

    set_pixel(rblst.rb[rb].sc, FOREGROUND,                  /* P = RB * P2 */
	      BACKGROUND, *workbuf2, 1);

    set_bitmap(AREA_SIZE(workbuf1->ar), workbuf1->mem, BACKGROUND);

    AREA_TO_SECTOR(workbuf2->ar, sec); 
    crop(rref.wpic, *workbuf1, sec, *workbuf2, 0, 0,       /* document - P */
	 0, 0, 0);

    ki = (double)wtd_correlation_ext(cls[cnri].nbl.nb[nnri].ref.wtd,
				     cls[cnri].nbl.nb[nnri].ref.msk, 
				     rbsi, rbsei, rref.wpic, xi, yi);

    ki /= cls[cnri].nbl.nb[nnri].ref.wtdm;/* normalize with mask area mass */

    if (ki < corrlevel) {     /* common part of pattern significant for Ci */
      kk = (double)wtd_correlation_ext(cls[cnrk].nbl.nb[nnrk].ref.wtd,
				       cls[cnrk].nbl.nb[nnrk].ref.msk, 
				       rbsk, rbsek, rref.wpic, xk, yk);

      kk /= cls[cnrk].nbl.nb[nnrk].ref.wtdm;   /* normalize with mask area */

      if (kk < corrlevel) {   /* common part of pattern significant for Cj */
	for (l = 0; l < hit_lst->cnt; l++) {       /* erase old index list */
	  FREEMEM(pair[l]);
	}
	FREEMEM(pair);

	if (kk < ki) {  /* erase candidate from list according correlation */
	  delete_cand_def(hit_lst, ck);                        /* erase Ck */
	}
	else {
	  delete_cand_def(hit_lst, ci);                        /* erase Ci */
	}

	ci = 0;                          /* restart changed candidate list */

	CALLOC(int*, hit_lst->cnt, pair);       /* allocate new index list */
	for (l = 0; l < hit_lst->cnt; l++) {
	  CALLOC(int, hit_lst->cnt, pair[l]);
	}
      }
    }

    crop(*workbuf1, rref.wpic, sec, *workbuf2, 0,    /* restore common pxl */
	 0, 0, 0, 0);
    set_bitmap(AREA_SIZE(workbuf2->ar), workbuf2->mem, BACKGROUND);
  }

  for (l = 0; l < hit_lst->cnt; l++) {
    FREEMEM(pair[l]);
  }
  FREEMEM(pair);

  set_bitmap(AREA_SIZE(workbuf1->ar), workbuf1->mem, BACKGROUND);
  set_bitmap(AREA_SIZE(workbuf2->ar), workbuf2->mem, BACKGROUND);
}



/* Searches for a candidate Ck which is a competitor to a candidate Ci.
   The competition is made by a common RB.

Revision: 26.10.2002 E.Sack

Input : - candidate list
        - number of candidate Ci
        - number of RB
Output: - number of candidate Ck (-1 if no candidate found)
*/
static int
search_cnb_competitor(cand_lst hit_lst, int ci, int rb)
{
  int ck, rkj, rkj_cnt;

  for (ck = 0; ck < hit_lst.cnt; ck++) {         /* search for competitors */
    if (ci == ck) {
      continue;                                     /* skip the originator */
    }

    if (rb == hit_lst.c[ck].rbn) {
      return ck;                              /* RB is a NB location of Ck */
    }
    else {
      rkj_cnt = hit_lst.c[ck].crb.cnt;

      for (rkj = 0; rkj < rkj_cnt; rkj++) {
	if (rb == hit_lst.c[ck].crb.nr[rkj]) {
	  return ck;                         /* RB is a NNB location of Ck */
	}
      }
    }
  }

  return -1;                                 /* Ck doesn't compete with Ci */
}



/* Builds the correlations KWH(RBi) of all RB with all NB partitioning 
   the RB. The partitions (RBp, FRBp) and the FNB exists temporarely.

Revision: 05.01.2005 E.Sack

Input : - matching parameters
        - class list
        - RB list
        - RB image
        - flag (0: multi NB, 1: single NB)
Output: - KWH list
*/
static kwh_lst
gen_kwh(match_def mat, cls_lst clst, rb_lst rblst, pic_def rpic, 
	int single_nb, MATCHING_METHOD mtd)
{
  int        r, c, n, found, part_x;
  double     cv_threshold;
  fft_def    fnb, fnbd, fmin, fmax;
  area_def   lub, nmax, klen, zar, border;
  sector_def rbs;
  kwh_lst    kl;
  part_lst   *frp = 0;

  cv_threshold = (mtd == MATCHING_METHOD_2) ? mat.corrbroken : mat.corrlevel;

  init_kwh_lst(&kl);
  kl.cnt = rblst.cnt;    /* number of 2-D correlation hills is number of RB */
  CALLOC(kwh_def, kl.cnt, kl.kwh);

  CALLOC(part_lst, rblst.cnt, frp);

  nmax = max_cls_area(clst);
  nmax.height = XPMAX(nmax.height, max_rb_area(rblst).height); /* no y part.*/

  XPOW2(2*nmax.width + 4*mat.mborder - 1, klen.width);
  XPOW2(nmax.height + 2*(mat.mborder+mat.nborder), klen.height);
  part_x = klen.width - nmax.width - 2*mat.mborder + 1;      /* x partition */

  zar.height = klen.height;
  for (r = 0; r < rblst.cnt; r++) {                           /* for all RB */
    frp[r] = part_rb(rblst.rb[r].sc, rpic, klen,      /* divide RB into RBp */
		     part_x, mat.mborder+mat.nborder);

    part_frb(frp[r]);                                      /* generate FRBp */

    zar.width = part_x*frp[r].cnt;      /* KWHi size need not be power of 2 */
    kl.kwh[r] = alloc_kwh_def(zar);
  }

  for (c = 0; c < clst.cnt; c++) {                           /* alle NB  */
    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      fnb = gen_fnb(clst.cls[c].nbl.nb[n].ref.pic, 
		    klen);                              /* FNB, with frame */
      rlft2(fnb.data, fnb.spec, (ulong)fnb.ar.height, (ulong)fnb.ar.width, 1);
      fnbd = gen_fnb(clst.cls[c].nbl.nb[n].ref.dlt, 
		     klen);                  /* dilated FNB, without frame */
      rlft2(fnbd.data,fnbd.spec,(ulong)fnbd.ar.height,(ulong)fnbd.ar.width,1);

      for (r = 0; r < rblst.cnt; r++) {              /* parallel FNB x FRBi */
	rbs = rblst.rb[r].sc;                                   /* shortcut */

	border.width = rule_nb_size(rbs.column, 
				    clst.cls[c].nbl.nb[n].pat.ar.width,
				    mat.nborder, single_nb);
	border.height = rule_nb_size(rbs.line,
				     clst.cls[c].nbl.nb[n].pat.ar.height,
				     mat.nborder, single_nb);

        if (border.width<0 || border.height<0) continue;     /* NB too big */

	lub = gen_lub(rbs, clst.cls[c].nbl.nb[n].ref.pic.ar, 
		      mat.mborder, border);

	fmin = total_corr(frp[r], fnb, clst.cls[c].nbl.nb[n].pat.ar, /* NB */
			  part_x, mat.mborder);

	found = min_bound(cv_threshold, fmin.data, fmin.ar, 
			  clst.cls[c].nbl.nb[n].ref.nbm, 
			  clst.cls[c].nbl.nb[n].ref.dltm, lub);

	if (found) {
	  fmax = total_corr(frp[r], fnbd,                        /*NB mask */
			    clst.cls[c].nbl.nb[n].pat.ar, part_x, 0);

 	  merge_kwh(&kl.kwh[r].ll[0][0], fmin.data, fmax.data, fmax.ar, c, n,
		    clst.cls[c].nbl.nb[n].ref.nbm,
		    clst.cls[c].nbl.nb[n].ref.dltm, lub);

	  free_fft_def(&fmax);
	}

        free_fft_def(&fmin);
      }

      free_fft_def(&fnb);
      free_fft_def(&fnbd);
    }
  }

  for (r = 0; r < rblst.cnt; r++) {
    free_part_lst(&frp[r]);
  }
  FREEMEM(frp);

  return kl;
}



/* Calculates maximum of width and height of all RB.

Revision: 27.07.2002 E.Sack

Input : - RB list
Output: - maximal size
*/
static area_def
max_rb_area(rb_lst rbl)
{
  int  r;
  area_def ar = {0, 0}, zar;

  for (r = 0; r < rbl.cnt; r++) {
    SECTOR_TO_AREA(rbl.rb[r].sc, zar);

    ar.width = XPMAX(ar.width, zar.width);
    ar.height = XPMAX(ar.height, zar.height);
  }

  return ar;
}



/* Calculates maximum of width and height of a class.

Revision: 15.02.2004 E.Sack

Input : - NB list
Output: - maximal size
*/
static area_def
max_nb_area(nb_lst nbl)
{
  int      i;
  area_def ar;

  init_area_def(&ar);

  for (i = 0; i < nbl.cnt; i++) {
    ar.width = XPMAX(nbl.nb[i].pat.ar.width, ar.width);
    ar.height = XPMAX(nbl.nb[i].pat.ar.height, ar.height);
  }

  return ar;
}



/* Calculates maximum of width and height of all classes.

Revision: 15.02.2004 E.Sack

Input : - class list
Output: - maxima
*/
area_def
max_cls_area(cls_lst clst)
{
  int      i;
  area_def ar, za;

  init_area_def(&ar);

  for (i = 0; i < clst.cnt; i++) {
    za = max_nb_area(clst.cls[i].nbl);

    ar.width = XPMAX(za.width, ar.width);
    ar.height = XPMAX(za.height, ar.height);
  }

  return ar;
}



/* Wrapper: Checks the matching resources.

Revision: 11.07.2004 E.Sack

Input : - DB
        - pointer to TPS ID
        - pointer to image file name string
        - image
        - RB list
Output: - TPS index (-1: error)
        - message output
*/
int
check_recall_resources(db_def db, char *tps_id, char *image_filename, 
		       pic_def rpic, rb_lst rbl)
{
  int tps_idx;

  tps_idx = check_training_resources(db, tps_id, image_filename, rpic, rbl);

  if (tps_idx < 0) {
    return -1;
  }

  if (db.tpl.tps[tps_idx].cl.cnt == 0) {    /* at least one class required */
    string_to_log(main_log, "empty class for TPS ID '", db.tpl.tps[tps_idx].ident, "'\n");
    return -1;
  }

  return tps_idx;
}



/* Aus den partitionierten RB (RBp) werden die FRBp gebildet, indem die RBp
   der FFT unterworfen werden.

Input : - RBp list
Output: - FRBp list
Postcondition: on place computation
*/ 
static void
part_frb(part_lst pl)
{
  int p;

  for (p = 0; p < pl.cnt; p++) {
    rlft2(pl.fp[p].data, pl.fp[p].spec,
	  (ulong)pl.fp[p].ar.height, (ulong)pl.fp[p].ar.width, 1);
  }
}



/* Das KWH aus den RBp und der Referenz erzeugen.

Input : - FRBp list
        - FNB
        - NB size
        - partition size
        - mask border size
Output: - total KWH
*/
static fft_def
total_corr(part_lst frp, fft_def fnb, area_def nar, int part_x, 
	   int mask_border)
{
  int     p, lxmax;
  int     lx, ly, lxu, lxv;                                 /* relative lag */
  int     lx1, lx2, lxw;                                    /* absolute lag */
  area_def    klen = frp.fp[0].ar;
  fft_def tot, fz;

  fz = alloc_fft_def(klen.width, klen.height);
  tot = alloc_fft_def(part_x*frp.cnt, klen.height);

  lxw = part_x - nar.width - mask_border + 1;/* impl.: mapping lag to index */
  lxmax = part_x - mask_border;                           /* implementation */
  lxu = klen.width - nar.width - mask_border + 1;         /* implementation */
  lxv = klen.width - mask_border;                         /* implementation */
  for (p = 0; p < frp.cnt; p++) {                    /* amount of partition */
    copy_fft_def(frp.fp[p], fz);/* on-place operation is here a bad solution*/
    corr2d(fz, fnb, 1, 1);                              /* KWHp <= FRBp,FNB */

    for (lx = 0, lx1 = p*part_x; lx < lxmax; lx++, lx1++) {   /* linear lag */
      for (ly = 0; ly <= klen.height; ly++) {     
	tot.data[ly][lx1] = fz.data[ly][lx];                   /* copy KWHp */
      }
    }

    if (!p) continue; /* implementation, no negative lag from 1st partition */

    for (lx = lxu, lx2 = lxw; lx < lxv; lx++, lx2++) {          /* wrap lag */
      for (ly = 0; ly <= klen.height; ly++) {     
	tot.data[ly][lx2] += fz.data[ly][lx];       /* add overlapping KWHp */
      }
    }

    for (lx = lxv; lx < klen.width; lx++, lx2++) {              /* wrap lag */
      for (ly = 0; ly <= klen.height; ly++) {     
	tot.data[ly][lx2] = fz.data[ly][lx];                   /* copy KWHp */
      }
    }

    lxw += part_x;
  }

  free_fft_def(&fz);

  return tot;
}




/* Redundante lags berechnen.

Input : - Bildgre
        - NB-Gre
        - mask border size
        - noise border size
Output: - upper bounds of lags
*/
static area_def
gen_lub(sector_def bsec, area_def nar, int mask_border, area_def noise_border)
{
  area_def ar, bar;
  int  w = nar.width-2*mask_border;   /* added NB mask border must be      */
  int  h = nar.height-2*mask_border;  /* compensated.                      */

  assert(w>0);
  assert(h>0);

  SECTOR_TO_AREA(bsec, bar);

  ar.width = bar.width + 2*noise_border.width - w;
  ar.height = bar.height + 2*noise_border.height - h;

  return ar;
}



/* Calculates maximal lag |p| of the reference r against the interval b. 
   The reference must superpose the interval at least with one pixel. 
   Coevally reference size rules are performed.

Revision: 16.10.2004 E.Sack

Input : - interval b
        - reference size
        - maximal allowed lag
        - flag (0: multi NB, 1: single NB)
Output: - maximal lag (-1: reference too big/small)
Usage:  - setting the single NB flag only has an effect if the interval 
          input is the original RB size and not a clipping fitted to the NB.
*/
static int
rule_nb_size(ival_def interval_b, int r_size, int lag_max, int single_nb)
{
  int b_size = INTERVAL_SIZE(interval_b);

  assert(lag_max >= 0);
  assert(r_size > 0);             /* reference contains at least one pixel */

  lag_max = XPMIN(lag_max, r_size-1);                   /* limit lag, Ch.9 */

  if (r_size > b_size + 2*lag_max) {
    return -1;                                        /* reference too big */
  }

  if (single_nb) {                                        /* one NB per RB */
    if (r_size < b_size - 2*lag_max) {
      return -1;                                    /* reference too small */
    }
  }

  return lag_max;
}



/* Eine 2-D Korrelationsliste wird aus den Korrelationsarrays
   gebildet. Fr jede Verschiebung in 2-D wird ein Array mit den diversen
   Korrelationswerten angelegt.

Input : - pointer to correlation list
        - pointer to correlation array dilated NB
        - pointer to correlation array original NB
        - array size
        - class number
        - NB number
Output: - updated correlation list
Precondition: arrays without gaps.  
*/
static void
merge_kwh(lag_lst *ll, double **cvmin, double **cvmax, area_def ar, int cls, 
	  int nb, double mass_nb, double mass_dnb, area_def lub)
{
  int     i;
  double  *nptr = &cvmin[0][0], *dptr = &cvmax[0][0];
  lag_lst *lptr = ll;

  assert(ll);

  for (i = 0; i < ar.height*ar.width; i++, nptr++, dptr++, lptr++) {
    if (*nptr == 0.0 || *dptr == 0.0) continue;        /* NB not recognized */

    if (lptr->cnt) {
      REALLOC(lag_def, lptr->cnt+1, lptr->ld);
    }
    else {
      CALLOC(lag_def, 1, lptr->ld);
    }

    lptr->ld[lptr->cnt].c = cls;
    lptr->ld[lptr->cnt].n = nb;
    lptr->ld[lptr->cnt].kwm =             /* correlation value of masked NB */
      (mass_dnb - mass_nb - *dptr + 2**nptr)/mass_dnb;
    lptr->ld[lptr->cnt].lub = lub;                  /* non redundant values */
    lptr->cnt++;
  }
}



/* Ein 2-D Korrelationsarray wird auf den unteren Grenzwert untersucht.

Input : - pointer to correlation array
        - array size
        - minimal correlation value
Output: - updated correlation array (0: fail, other: ok)
        - total (0: all values less than minimum, 
                 1: not all values less than minimum)
Precondition: array without gaps.
*/
static int
min_bound(double corrlevel, double **cv, area_def ar, double mass_nb, double mass_dnb, 
	  area_def lub)
{
  int    i, j, total;
  double mz = mass_nb - mass_dnb*(1.0-corrlevel);

  assert(&cv[ar.height-1][ar.width-1]-&cv[0][0]+1 == ar.height*ar.width);

  for (i = total = 0; i < ar.height; i++) {
    for (j = 0; j < ar.width; j++) {

      if (j > lub.width || i > lub.height) {
        cv[i][j] = 0.0;              /* sortiere redundante KWHi-Werte aus */
	continue;
      }

      if (cv[i][j] >= mz) {
	total = 1;
      }
      else {
        cv[i][j] = 0.0;               /* sortiere zu kleine KWHi-Werte aus */
      }
    }
  }

  return total;
}



/* A NB is prepared for a correlation.

Input : - NB image
        - minimal extension
Output: - prepared NB
Precondition: NB with frame.
*/
static fft_def
gen_fnb(pic_def pic, area_def ar) 
{
  fft_def fnb;

  init_fft_def(&fnb);

  fnb = alloc_fft_def(ar.width, ar.height);

  copy_bmp_to_matrix(pic, fnb.data, pic.ar, 0, 0, 0, 0);

  return fnb;
}



/* Calculates from a RB its partial RBp. The RBp are aligned to the 
   maximal NB (size), to fit the partitioning for all NB, in particular:
   The height of the partitioning is adjusted to the maximal height of the 
   NBi and RBi, that means, there is no partitioning in vertical direction.
   The width of the partitioning is adjusted to the maximal width of the NBi.

Revision: 05.08.2005 E.Sack

Input : - RB sector
        - RB image
        - max. NB size
        - NB size
        - border size
Output: - RBp-Liste
*/
static part_lst
part_rb(sector_def rbs, pic_def rpic, area_def klen, int part_x, int border)
{
  int        z_start, s_start, z_offset, s_offset, delta_x, end;
  area_def   rar;
  part_lst   frbp;
  sector_def par;

  init_part_lst(&frbp);

  end = XPMIN(rpic.ar.width-1, rbs.column.end+border);         /* shortcut */

  /* border punched out from image to keep consistency                     */
  z_start = XPMAX(0, rbs.line.start - border);
  z_offset = XPMAX(0, z_start - rbs.line.start + border);
  par.line.start = z_start;
  par.line.end = XPMIN(rpic.ar.height-1, rbs.line.end + border);

  delta_x = rbs.column.start - border;

  do {
    if (frbp.cnt) {
      REALLOC(fft_def, frbp.cnt+1, frbp.fp);
    }
    else {
      CALLOC(fft_def, frbp.cnt+1, frbp.fp);
    }

    frbp.fp[frbp.cnt] = alloc_fft_def(klen.width, klen.height);

    s_start = XPMAX(0, delta_x);
    s_offset = XPMAX(0, s_start - delta_x);
    par.column.start = s_start;
    par.column.end = XPMIN(end, delta_x+part_x-1);

    SECTOR_TO_AREA(par, rar);                                /* punch size */

    copy_bmp_to_matrix(rpic, frbp.fp[frbp.cnt].data, rar,
	       	                  s_start, z_start, s_offset, z_offset);

    delta_x += part_x;
    frbp.cnt++;
  }  while (delta_x == XPMIN(end, delta_x));

  return frbp;
}



/* Performs matching of a pattern template with sector RBS. The methods are
   - NB matching (all other methods are lead back to this method),
   - CNB matching after preparing the RBS,
   - broken pattern matching after preparing the RBS.

   There are two versions of the routine; one with the correlation made in 
   spatial domain and one with the correlation made in frequency domain.

Revision: 03.01.2005 E.Sack

Input : - matching parameters
        - RBS
        - lookahead
        - RB under test index
        - RB list
        - NB index
        - class index
        - pointer to class list
        - pointer to candidate list
        - RB references
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
        - flag (0: non-broken NB, 1: broken NB)
Output: - updated candidate list
        - control structure updated
*/
static void
candidate_list_vertical_sd(match_def mat, sector_def rbs_in, int lhead, 
			   int rut, rb_lst rblst, int ndx, int cdx, 
			   cls_def *cls, cand_lst *klist, rb_ref rref, 
			   run_def *crun, int single_nb, MATCHING_METHOD mtd)
{
  int        d;
  sector_def rbs, rbse;

  assert(cls);
  assert(klist);

  if (mtd == MATCHING_METHOD_2) {             /* enlarge RBS for broken NB */
    rbs_in = check_broken(rbs_in, cls[cdx].nbl.nb[ndx].pat.ar, rref.rpic->ar,
			  mat.mborder);
  }

  rbs = rbs_in;

  d = check_nb(&rbs, &rbse, lhead, cls[cdx].nbl.nb[ndx].pat.ar, *rref.rpic, 
	       mat.mborder, mat.nborder, single_nb, mtd);

  if (!d) {
    return;                                              /* NB test failed */
  }

  check_hit_sd(mat, rut, ndx, cdx, cls, rbs, rbse, rref, klist, rbs_in, 
	       rblst, mtd);

  lhead = lhead?lhead:1;                    /* enable complete correlation */

  check_hit_cnb(mat, rbs_in, lhead, rut, rblst, ndx, cdx, cls, klist, rref,
		crun, single_nb, mtd);
}



/* Matching of a NB with sector RBS will be prepared by some rules. Further 
   if the NB is a NNB the CNB will be investigated, too.

   There exists two versions of the routine; one with the correlation made
   in spatial domain and one with the correlation made in frequency domain.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - correlation features
        - lookahead
        - RB number
        - RB list
        - NB number
        - class number
        - pointer to class list
        - pointer to candidate list
        - RB references
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - updated candidate list
        - control structure updated
Usage:  - not designed for broken NB.
*/
static void
candidate_list_vertical_fd(match_def mat, cv_def_i kmark, int lhead, int r,
			   rb_lst rblst, int n, int c, cls_def *cls,
			   cand_lst *klist, rb_ref rref, run_def *crun, 
			   int single_nb, MATCHING_METHOD mtd)
{
  sector_def rbs, rbse, org = rblst.rb[r].sc;

  trans_lag(org.column, &rbs.column, &rbse.column, 
	    cls[c].nbl.nb[n].pat.ar.width, rref.rpic->ar.width, 
	    mat.mborder, mat.nborder, &kmark.x);

  trans_lag(org.line, &rbs.line, &rbse.line, 
	    cls[c].nbl.nb[n].pat.ar.height, rref.rpic->ar.height, 
	    mat.mborder, mat.nborder, &kmark.y);

  if (line_rule(rbse.line, org.line, cls[c].att.alignment)) {
    return;                                               /* Regeln testen */ 
  }

  check_hit_fd(mat, r, n, c, cls, rbs, rbse, rref, klist, org, rblst, kmark);

  lhead = lhead?lhead:1;            /* vollstndige Korrelation einschalten */

  check_hit_cnb(mat, rbs, lhead, r, rblst, n, c, cls, klist, rref, crun, 
		single_nb, mtd);
}



/* Investigation of CNB when a NNB is detected. It is necessary to test all
   CNB members, because they causes different CNB frames.                    

Revision: 05.08.2005 E.Sack

Input : - matching parameters
        - RBS
        - lookahead
        - RB under test index
        - RB list
        - NB number
        - class number
        - pointer to class
        - pointer to candidate list
        - RB references
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - updated candidate list
        - control structure updated
Usage: - the program flow contains the unconditioned recall of this 
         function. However an endless loop is avoided by using the parent 
         CNB of the current NB. The recall therefore stops if no further
         parents are available. Furthermore this behaviour ensures any size
         of hierarchy of CNBs to be treated in the same manner.
*/
static void
check_hit_cnb(match_def mat, sector_def rbs_in, int lhead, int rut, 
	      rb_lst rblst, int n, int c, cls_def *cls, cand_lst *klist, 
	      rb_ref rref, run_def *crun, int single_nb, MATCHING_METHOD mtd)
{
  int        i, paren_n, paren_c;
  sector_def rbs;

  for (i = 0; i < cls[c].nbl.nb[n].cnb.cnt; i++) {             /* CNB parent */
    xplab_progress_set_value(crun, -1.0, 0.0);       /* check user interrupt */

    if (*crun->pactive == FALSE) {
      break;                                               /* GTK concession */
    }

    paren_c = cls[c].nbl.nb[n].cnb.cd[i].cdx;                 /* CNB indexes */
    paren_n = cls[c].nbl.nb[n].cnb.cd[i].ndx;

    rbs = check_cnb(rbs_in,                  /* extend current RBS if needed */
		    cls[c].nbl.nb[n].pat.ar,
                    cls[c].nbl.nb[n].cnb.cd[i].x, 
		    cls[c].nbl.nb[n].cnb.cd[i].y,
		    cls[paren_c].nbl.nb[paren_n].pat.ar,
		    rblst.rb[rut].sc, rref.rpic->ar);

    candidate_list_vertical_sd(mat, rbs, lhead, rut, rblst, paren_n, paren_c,
			       cls, klist, rref, crun, single_nb, mtd);
  }
}



/* Calculates the horizontal list of the best candidates. Correlation in
   the spatial domain.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - candidate list
        - RB sector
        - RB under test index
        - RB list
        - class list
        - pointer to RB references
        - pointer to recycling list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - index of recycling list (<0: no entry)
        - recycling list updated
        - control structure updated
*/
static int
candidate_list_horizontal_sd(match_def mat, cand_lst hlist, sector_def rbs,
			     int rut, rb_lst rblst, cls_lst clst, 
			     rb_ref *rref, rc_lst *rlist, run_def *crun, 
			     int single_nb, MATCHING_METHOD mtd)
{
  int        h_idx, i, *idx = 0, lhead, *yet_created = 0, vc, vn;
  double     cv_threshold;
  cand_lst   vlist1, vlist2;                   /* vertical candidate lists */
  sector_def rbs_old;

  assert(rbs.column.start <= rbs.column.end);

  cv_threshold = (mtd == MATCHING_METHOD_2) ? mat.corrbroken : mat.corrlevel;

  vlist1 = vertical_list_part_1_sd(mat, &rbs, rut, rblst, *rref, clst, crun, 
				   single_nb, mtd);

  if (vlist1.cnt) {                              /* filtering changed list */
    sort_out(mat.nborder, cv_threshold, hlist, clst.cls, &vlist1, 
	     *rref->rpic, rblst.rb[rut].gridy.cnt);
  }

  lhead = lookahead(vlist1, clst.cls, MBORDER_DEFAULT+mat.nborder+1,
		    mat.interleave);
  vlist2 = vertical_list_part_2_sd(mat, lhead, rbs, rut, rblst, *rref, 
				   clst, crun, single_nb, mtd);
  cat_cand_lst(&vlist2, &vlist1);

  if (vlist2.cnt) {                              /* filtering changed list */
    sort_out(mat.nborder, cv_threshold, hlist, clst.cls, &vlist1, 
	     *rref->rpic, rblst.rb[rut].gridy.cnt);
  }

  free_cand_lst(&vlist2);
  delete_non_competitors(&vlist1, clst.cls, mat.nb_lap);

  if (!vlist1.cnt) {
    return NO_STRING;                          /* no candidate found in RB */
  }

  CALLOC(int, vlist1.cnt, idx);                        /* apply index list */
  CALLOC(int, vlist1.cnt, yet_created);

  /* calculate best substring per candidate                                */
  for (i = 0, rbs_old = rbs; i < vlist1.cnt; i++) {
    idx[i] = NO_STRING;                                            /* init */
    vc = vlist1.c[i].c;
    vn = vlist1.c[i].n;

    idx[i] = lookup_rc_lst(*rlist, vlist1.c[i].hdx);
    if (idx[i] >= 0) {                                     /* string index */
      yet_created[i] = 1;               /* string exists in recycling list */
    }
    else 
    {
      int step;

      yet_created[i] = 0;     /* string doesn't exist in recycling list,   */
                              /* so look for a substring before enroll it. */

      crop(*rref->rpic, rref->shdw1, vlist1.c[i].knl,/*eliminate V-Candid. */
	   clst.cls[vc].nbl.nb[vn].pat, 0, 0, 
	   vlist1.c[i].knl.column.start - vlist1.c[i].hdx.column.start,
	   vlist1.c[i].knl.line.start - vlist1.c[i].hdx.line.start,
	   0);

      crop(rref->wpic, rref->shdw2, vlist1.c[i].knl,
	   clst.cls[vc].nbl.nb[vn].pat, 0, 0, 
	   vlist1.c[i].knl.column.start - vlist1.c[i].hdx.column.start,
	   vlist1.c[i].knl.line.start - vlist1.c[i].hdx.line.start,
	   0);

      compose_key(vc, vn, &rref->pic_state);

      step = stepping(*rref->rpic, vlist1.c[i].roi,mat.mborder+mat.nborder+1,
		      mat.interleave);

      do {                                     /* enforce substring search */
	rbs.column.start = vlist1.c[i].roi.column.start + step;

        if (rbs.column.start <= rbs_old.column.end) {/* no substring if end*/
	  idx[i] = candidate_list_horizontal_sd(mat, hlist, rbs, rut, rblst,
						clst, rref, rlist, crun, 
						single_nb, mtd);
	}
	else {
	  break;                                  /* end of region reached */
	}

	step += mat.mborder + mat.nborder + 1;
      } while (idx[i] < 0);                          /* no substring found */

      crop(rref->shdw1, *rref->rpic, vlist1.c[i].knl,/*restore V-Candidate */
	   clst.cls[vc].nbl.nb[vn].pat, 0, 0, 
	   vlist1.c[i].knl.column.start - vlist1.c[i].hdx.column.start,
	   vlist1.c[i].knl.line.start - vlist1.c[i].hdx.line.start,
	   0);
      crop(rref->shdw2, rref->wpic, vlist1.c[i].knl,
	   clst.cls[vc].nbl.nb[vn].pat, 0, 0, 
	   vlist1.c[i].knl.column.start - vlist1.c[i].hdx.column.start,
	   vlist1.c[i].knl.line.start - vlist1.c[i].hdx.line.start,
	   0);

      reduce_key(&rref->pic_state);
    }
  }                                     /* vertical candidate list handled */

  sifting_gap(vlist1, idx, yet_created, *rlist, rbs_old, mat.nborder);
  h_idx = add_substring(vlist1, idx, yet_created, rlist, rbs_old, clst);

  free_cand_lst(&vlist1);
  FREEMEM(idx);
  FREEMEM(yet_created);

  return h_idx;
}



/* Calculates the horizontal list of the best Candidates. Correlation in
   the frequency domain.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - candidate list
        - horizontal lag
        - RB sector
        - RB index
        - RB list
        - class list
        - RB references
        - pointer to recycling list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - index of recycling list (-1: no entry)
        - updated recycling list
        - control structure updated
Usage:  - not designed for broken NB.

Bug: not maintained because of no runtime advantages
*/
static int
candidate_list_horizontal_fd(match_def mat, cand_lst hlist, int x, 
			     sector_def rbs, int r, rb_lst rblst, 
			     cls_lst clst, rb_ref rref, rc_lst *rlist, 
			     run_def *crun, int single_nb, 
			     MATCHING_METHOD mtd)
{
  int        h_idx, i, *idx = 0, lhead=0, *yet_created = 0;
  cand_lst   vlist1, vlist2;                   /* vertical candidate lists */
  sector_def rbs_old, org = rblst.rb[r].sc;

  assert(rbs.column.start <= rbs.column.end);

  vlist1 = vertical_list_part_1_fd(mat, &x, &rbs.column.start, r, rblst, 
				   rref, clst, crun, single_nb, mtd);

  if (!vlist1.cnt) {
    return NO_STRING;                          /* no candidate found in RB */
  }

  sort_out(mat.nborder, mat.corrlevel, hlist, clst.cls, &vlist1, *rref.rpic, 
	   rblst.rb[r].gridy.cnt);

  lhead = lookahead(vlist1, clst.cls, mat.mborder+mat.nborder+1,
		    mat.interleave);                          /* fair play */
  lhead = XPMAX(lhead, mat.mborder+mat.nborder); /* gradient approximation */
  vlist2 = vertical_list_part_2_fd(mat, lhead, x, rbs.column, r, rblst, 
				   rref, clst, crun, single_nb, mtd);

  trim_vertical_list(&vlist2, rbs.column.start+mat.mborder+mat.nborder);

  cat_cand_lst(&vlist2, &vlist1);
  free_cand_lst(&vlist2);

  sort_out(mat.nborder, mat.corrlevel, hlist, clst.cls, &vlist1, *rref.rpic, 
	   rblst.rb[r].gridy.cnt);

  CALLOC(int, vlist1.cnt, idx);           /* indices of strings/substrings */
  CALLOC(int, vlist1.cnt, yet_created);/* indicate list of reusable strings*/

  /* calculate best substring per candidate                                */
  for (i = 0, rbs_old = rbs; i < vlist1.cnt; i++) {
    idx[i] = NO_STRING;                                            /* init */

    idx[i] = lookup_rc_lst(*rlist, vlist1.c[i].hdx);
    if (idx[i] >= 0) {                                     /* string index */
      yet_created[i] = 1;               /* string exists in recycling list */
    }
    else                 
    {
      int step, xnew = x;
      yet_created[i] = 0;     /* string doesn't exist in recycling list,   */
                              /* so look for a substring before enroll it. */

      crop(rref.wpic, rref.shdw1, vlist1.c[i].knl,/* eliminate V-Candidate */
	   clst.cls[vlist1.c[i].c].nbl.nb[vlist1.c[i].n].pat, 0, 0, 
	   vlist1.c[i].knl.column.start - vlist1.c[i].hdx.column.start,
	   vlist1.c[i].knl.line.start - vlist1.c[i].hdx.line.start,
	   0);

      step = stepping(rref.wpic, vlist1.c[i].roi, mat.mborder+mat.nborder+1,
		      mat.interleave);

      do {                                     /* enforce substring search */
	xnew = x+step-mat.nborder;
	rbs.column.start = org.column.start+XPMAX(0, xnew);

        if (rbs.column.start <= org.column.end) {   /* no substring if end */
	  idx[i] = candidate_list_horizontal_fd(mat, hlist, 
						xnew,/* look for substring */
						rbs, r, rblst, clst, rref, 
						rlist, crun, single_nb, 
						mtd);
	}
	else break;

	idx[i] = sifting_lap_h(vlist1.c[i], idx[i], yet_created[i], *rlist, 
			       clst, mat.overlap);

	step += mat.mborder+mat.nborder;
      } while (idx[i] < 0);                          /* no substring found */

      crop(rref.shdw1, rref.wpic, vlist1.c[i].knl,  /* restore V-Candidate */
	   clst.cls[vlist1.c[i].c].nbl.nb[vlist1.c[i].n].pat, 0, 0, 
	   vlist1.c[i].knl.column.start - vlist1.c[i].hdx.column.start,
	   vlist1.c[i].knl.line.start - vlist1.c[i].hdx.line.start,
	   0);
    }
  }                                     /* vertical candidate list handled */

  sifting_gap(vlist1, idx, yet_created, *rlist, rbs_old, mat.nborder);
  h_idx = add_substring(vlist1, idx, yet_created, rlist, rbs_old, clst);

  free_cand_lst(&vlist1);
  FREEMEM(idx);
  FREEMEM(yet_created);

  return h_idx;
}



/* The vertical list is trimed by eliminating candidates too far away.

Input : - pointer to vertical list
        - limit defining absolut x-coordinate
Output: - updated vertical list
*/
static void
trim_vertical_list(cand_lst *vlist, int start)
{
  int i;

  for (i = 0; i < vlist->cnt; i++) {
    if (vlist->c[i].knl.column.start > start) {
      delete_cand_def(vlist, i--);                           /* delete NBi */
    }
  }
}



/* Generation of the first part of the vertical candidate list for the
   spatial domain. The list is just a register of candidates in a certain
   column, no substrings are defined. The RB column is searched to the right
   end of the RB. The search is stopped with the first list found.

Revision: 05.08.2005 E.Sack

Input : - matching parameters
        - pointer to RB sector
        - RB under test index
        - RB list
        - RB references
        - class list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - vertical list
        - updated RBS
        - control structure updated
Postcondition: - the vertical list may be empty (no candidates found).
*/
static cand_lst
vertical_list_part_1_sd(match_def mat, sector_def *rbs, int rut, 
			rb_lst rblst, rb_ref rref, cls_lst clst, 
			run_def *crun, int single_nb, MATCHING_METHOD mtd)
{
  int            c, n, positive_idx;
  cand_lst       vlist;
  REGISTER_INDEX reg_index;

  init_cand_lst(&vlist);

  do {
    reg_index = check_register(rref.xreg, rbs->column.start, rbs->line,
			       rref.pic_state, &positive_idx);

    switch (reg_index) {
    case NOT_REGISTERED:     /* creating the vertical candidate list from DB */
      for (c = 0; c < clst.cnt && *crun->pactive == TRUE; c++) {
	for (n = 0; n < clst.cls[c].nbl.cnt && *crun->pactive == TRUE; n++) {
	  xplab_progress_set_value(crun, -1.0, 0.0); /* check user interrupt */
    	  candidate_list_vertical_sd(mat, *rbs, 0, rut, rblst, n, c, 
				     clst.cls, &vlist, rref, crun, single_nb,
				     mtd);
	}
      }

      if (vlist.cnt) {                                             /* a'_{1} */
	register_positive_result(rref.xreg, rbs->column.start, rbs->line,
				 rref.pic_state, vlist);
	return vlist;
      }
      else {
	register_negative_result(rref.xreg, rbs->column.start, rbs->line,
				 rref.pic_state);
      }

      break;

    case REGISTERED_POSITIVE_RESTRICT:   /* reg. VCL found with restrictions */
      assert(positive_idx>=0 && 
	     positive_idx<rref.xreg.yrl[rbs->column.start].cnt);

      restricted_nb_loop(mat, *rbs, 0, rut, rblst, clst, &vlist, rref, 
			 rref.xreg.yrl[rbs->column.start].yrd[positive_idx].
			 vlst, crun, single_nb, mtd);

      if (vlist.cnt) {                                           /* a'_{1} */
	return vlist;
      }

      break;

    case REGISTERED_POSITIVE_IDENTIC:              /* registered VCL found */
      assert(positive_idx>=0 && 
	     positive_idx<rref.xreg.yrl[rbs->column.start].cnt);
      cat_cand_lst(&rref.xreg.yrl[rbs->column.start].yrd[positive_idx].vlst, 
		   &vlist);
      return vlist;
      break;

    case REGISTERED_NEGATIVE:  /* registered: no vertical cand. list found */
      break;

    default:
      assert(0);
      break;
    }

    rbs->column.start++;
  } while (INTERVAL_SIZE(rbs->column) > 0);

  return vlist;
}



/* Creates a VCL from register including all anchored CNB.

Revision: 05.08.2005 E.Sack

Input : - matching parameters
        - RBS
        - lookahead
        - RB under test index
        - RB list
        - class list
        - pointer to candidate list sum
        - RB references
        - vertical candidate list root
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - updated candidate list sum
        - control structure updated
*/
static void
restricted_nb_loop(match_def mat, sector_def rbs, int lhead, int rut, 
		   rb_lst rblst, cls_lst clst, cand_lst *vcl_sum, 
		   rb_ref rref, cand_lst vcl_root, run_def *crun, 
		   int single_nb, MATCHING_METHOD mtd)
{
  int cand, cnr, nnr;

  for (cand = 0; cand < vcl_root.cnt && *crun->pactive == TRUE; cand++) {
    cnr = vcl_root.c[cand].c;                     /* restrict classes to VCL */

    for (nnr = 0; nnr < clst.cls[cnr].nbl.cnt && *crun->pactive == TRUE;nnr++) {
      xplab_progress_set_value(crun, -1.0, 0.0);     /* check user interrupt */
      candidate_list_vertical_sd(mat, rbs, lhead, rut, rblst, nnr, cnr, 
				 clst.cls, vcl_sum, rref, crun, single_nb,
				 mtd);
    }
  }

  /* restrict classes to NNB to get CNB                                      */
  for (cnr = 0; cnr < clst.cnt && *crun->pactive == TRUE; cnr++) {
    for (nnr = 0; nnr < clst.cls[cnr].nbl.cnt && *crun->pactive == TRUE;nnr++) {
      xplab_progress_set_value(crun, -1.0, 0.0);     /* check user interrupt */

      if (!clst.cls[cnr].nbl.nb[nnr].cnb.cnt) {
	continue;                                               /* not a NNB */
      }

      candidate_list_vertical_sd(mat, rbs, lhead, rut, rblst, nnr, cnr, 
				 clst.cls, vcl_sum, rref, crun, single_nb,
				 mtd);
    }
  }
}



/* Registers a negative recognition result. Registering all recognition 
   results shall shorten runtime through their multiple use instead of
   repeated calculation.

   The rule for registering a VCL is:
   - same column && identic vertical range registered
     ==> no registration of VCL,
   - same column && vertical range within registered one
     ==> no registration of VCL,
   - any other case as above (not same column || overlapping vertical ranges
                              || non overlapping vertical ranges)
     ==> registration of VCL.

Revision: 07.12.2002 E.Sack

Input : - vertical candidate list register
        - column to register
        - column start & end
Output: - register updated
*/
static void
register_negative_result(reg_def vregister, int column, ival_def y_range, 
			 char *key)
{
  int            d, yrd_idx, len;
  REGISTER_INDEX key_res;

  if (!vregister.yrl) {
    assert(!vregister.cnt);
    return;                                         /* no register defined */
  }
  else {
    assert(column >= 0);
    assert(column < vregister.cnt);
  }

  if (y_range.end < y_range.start) {
    return;                                                 /* range empty */
  }

  for (yrd_idx = 0; yrd_idx < vregister.yrl[column].cnt; yrd_idx++) {
    if (vregister.yrl[column].yrd[yrd_idx].use != REGISTERED_NEGATIVE) {
      continue;                                 /* not negative registered */
    }

    key_res = compare_image_state(key, 
				    vregister.yrl[column].yrd[yrd_idx].
				    pic_state);

    if (key_res == NOT_REGISTERED) {
      continue;                     /* contents of both keys are different */
    }

    d = join_range_keys(y_range, &vregister.yrl[column].yrd[yrd_idx].y_range);

    if (d) {                                          /* range keys joined */
      return;
    }
  }

  len = vregister.yrl[column].cnt++;
  if (len) {
    REALLOC(y_reg_def, len+1, vregister.yrl[column].yrd);
  }
  else {
    CALLOC(y_reg_def, len+1, vregister.yrl[column].yrd);
  }

  init_y_reg_def(&vregister.yrl[column].yrd[len]);

  vregister.yrl[column].yrd[len].use = REGISTERED_NEGATIVE;
  vregister.yrl[column].yrd[len].y_range = y_range;              /* no VCL */

  add_string(key, &vregister.yrl[column].yrd[len].pic_state);/*may be empty*/
}



/* Registers a positive recognition result, that means an available VCL.

Revision: 22.11.2002 E.Sack

Input : - vertical candidate list register
        - column to register
        - column start & end
        - vertical candidate list
Output: - register updated
*/
static void
register_positive_result(reg_def vregister, int column, ival_def y_range, 
			 char *key, cand_lst vlist)
{
  int            d, len, yrd_idx;
  REGISTER_INDEX key_res;

  if (!vregister.yrl) {
    assert(!vregister.cnt);
    return;                                         /* no register defined */
  }
  else {
    assert(column >= 0);
    assert(column < vregister.cnt);
  }

  if (y_range.end < y_range.start) {
    return;                                                 /* range empty */
  }

  for (yrd_idx = 0; yrd_idx < vregister.yrl[column].cnt; yrd_idx++) {
    if (vregister.yrl[column].yrd[yrd_idx].use != REGISTERED_POSITIVE) {
      continue;                                 /* not positive registered */
    }

    key_res = compare_image_state(key, 
				    vregister.yrl[column].yrd[yrd_idx].
				    pic_state);

    if (key_res == NOT_REGISTERED) {
      continue;                     /* contents of both keys are different */
    }

    d = join_range_keys(y_range, &vregister.yrl[column].yrd[yrd_idx].y_range);

    if (d) {                                          /* range keys joined */
      cat_cand_lst(&vlist, &vregister.yrl[column].yrd[yrd_idx].vlst);
      return;                                    /* candidate lists joined */
    }
  }

  len = vregister.yrl[column].cnt++;
  if (len) {
    REALLOC(y_reg_def, len+1, vregister.yrl[column].yrd);
  }
  else {
    CALLOC(y_reg_def, len+1, vregister.yrl[column].yrd);
  }

  init_y_reg_def(&vregister.yrl[column].yrd[len]);

  vregister.yrl[column].yrd[len].use = REGISTERED_POSITIVE;
  vregister.yrl[column].yrd[len].y_range = y_range;

  cat_cand_lst(&vlist, &vregister.yrl[column].yrd[len].vlst);       /* VCL */
  add_string(key, &vregister.yrl[column].yrd[len].pic_state);/*may be empty*/
}



/* Joins range keys. Succeeds if a interval comprises the other.

Revision: 14.12.2002 E.Sack

Input : - range A
        - pointer to range B
Output: - range B updated
        - join result (0: not joined, 1: joined)
*/
static int
join_range_keys(ival_def range_a, ival_def *range_b)
{
  REGISTER_INDEX range_res;

  range_res = compare_range(range_a, *range_b);

  if (range_res == NOT_REGISTERED) {
    range_res = compare_range(*range_b, range_a);

    if (range_res == NOT_REGISTERED) {
      return 0;                             /* range keys not able to join */
    }
    else {
      *range_b = range_a;                            /* interval exchanged */
    }
  }

  return 1;                                           /* range keys joined */
}



/* Checks register for a recognition result against several keys.

   The rule for reusability of registered VCLs is:

      keys       | comparison result: Input ? Register      
  ---------------+-------+-------+-----------+--------+ all other results
   column        |   !=  |   *   |     *     |   ==   | comparing keys lead
                 |       |       |           |        | to a restricted 
  ---------------+-------+-------+-----------+--------+ search over the
    image        |   *   |   <   |     *     |   ==   | registered VCL's
    state        |       |       |           |        | classes. There cannot
  ---------------+-------+-------+-----------+--------+ be other classes than
   vertical      |   *   |   *   | non/part- |   ==   | in the registered 
    range        |       |       | overlap   |        | VCLs but less.
  ===============+=======+=======+===========+========+
   Usability of  |  no   |  no   |   no      |  copy  |
 registered VCLs |       |       |           |        |
  ---------------+-------+-------+-----------+--------+

      <   smaller than
     ==   identic
     !=   not identic
      *   any result
     no   no usability
   copy   direct use of registered VCL


   The rule for priority of multiple VCLs is:
   - 1. positive index with identic vertical ranges
   - 2. positive index with vertical range within registered one
   - 3. all other indexes


Revision: 08.12.2002 E.Sack

Input : - vertical candidate list register
        - column key
        - vertical range key
        - pointer to image state key
        - pointer to VCL index
Output: - register flag
        - VCL index updated
Precondition:  - keys have to be exclusive
Postcondition: - if flag is non positive, then VCL index is negative.
*/
static REGISTER_INDEX
check_register(reg_def vregister, int column, ival_def y_range, char *key, 
               int *positive_idx)
{
  int            yrd_idx;
  REGISTER_INDEX key_res, range_res, total_res;
  REGISTER_INDEX registered_index = NOT_REGISTERED;
  REGISTER_INDEX max_registered_index = NOT_REGISTERED;
  REGISTER_INDEX max_vcl_quality = NOT_REGISTERED;

  assert(column>=0);
  assert(positive_idx);

  *positive_idx = NOT_REGISTERED;/*input check: destroy positive VCL index */

  if (!vregister.yrl) {
    assert(!vregister.cnt);
    return NOT_REGISTERED;             /* input check: no register defined */
  }
  else {
    assert(column >= 0);
    assert(column < vregister.cnt);
  }

  if (y_range.end < y_range.start) {
    return NOT_REGISTERED;                     /* input check: range empty */
  }

  for (yrd_idx = 0; yrd_idx < vregister.yrl[column].cnt; yrd_idx++) {
    key_res = compare_image_state(key, 
				    vregister.yrl[column].yrd[yrd_idx].
				    pic_state);
    if (key_res == NOT_REGISTERED) {
      continue;
    }

    range_res = compare_range(y_range, 
			      vregister.yrl[column].yrd[yrd_idx].y_range);
    if (range_res == NOT_REGISTERED) {
      continue;
    }

    registered_index = vregister.yrl[column].yrd[yrd_idx].use;

    if (registered_index == NOT_REGISTERED) {
      continue;
    }

    if (key_res   == REGISTERED_RESTRICT ||  /* sum key comparison results */
	range_res == REGISTERED_RESTRICT) {
      total_res = REGISTERED_RESTRICT;             /* weakest key advances */
    }
    else {
      total_res = key_res;
    }

    if (max_registered_index == REGISTERED_POSITIVE) {      /* 2. priority */
      if (registered_index == REGISTERED_POSITIVE && 
	  total_res == REGISTERED_IDENTIC) {
	max_vcl_quality = total_res;
      }
    }
    else {
      max_registered_index = registered_index;
      max_vcl_quality = total_res;
    }

    if (max_registered_index == REGISTERED_POSITIVE) {
      *positive_idx = yrd_idx;

      if (max_vcl_quality == REGISTERED_IDENTIC) {
	break;                    /* 1. priority (key has to be exclusive) */
      }
    }
  }

  switch (max_registered_index) {
  case REGISTERED_POSITIVE:                  /* summarize positive results */
    if (max_vcl_quality == REGISTERED_IDENTIC) {
      max_registered_index = REGISTERED_POSITIVE_IDENTIC;
      break;
    }

    if (max_vcl_quality == REGISTERED_RESTRICT) {
      max_registered_index = REGISTERED_POSITIVE_RESTRICT;
      break;
    }

    assert(0);
    break;

  default:                         /* no discrimination in any other cases */
    break;
  }

  return max_registered_index;
}



/* Compares interval ranges.

Revision: 07.12.2002 E.Sack

Input : - range A and B
Output: - comparison result (range A against range B)
*/
static REGISTER_INDEX
compare_range(ival_def range_a, ival_def range_b)
{
  if (range_a.start == range_b.start && range_a.end == range_b.end) {
    return REGISTERED_IDENTIC;
  }

  if (range_a.start >= range_b.start && range_a.end <= range_b.end) {
    return REGISTERED_RESTRICT;           /* interval A lies in interval B */
  }

  return NOT_REGISTERED;
}



/* Compares image state A against image state B.

   A image state P1 is greater than P2 if
   [(strlen(P1) > strlen(P2)) && (P1 starts with P2)] is true,

   e.g. Smal="ABC", Big1="ABCXYZ", Big2="BCDXYZ", Big3="XABCYZ" ==>
        Big1 > Smal,  Big2 < Smal,  Big3 < Smal.

   'Big1 > Smal' just means that more patterns in 'Big1' are cut off
   than from 'Smal'. The image for 'Big1' cannot have different classes as
   image for 'Smal1', but less.


Revision: 07.12.2002 E.Sack

Input : - key A and B
Output: - comparison result (key A against key B)
*/
static REGISTER_INDEX
compare_image_state(char *key_a, char *key_b)
{
  int          len_key_a, len_key_b, diff;
  unsigned int len_b;

  len_key_a = !!key_a;
  len_key_b = !!key_b;

  if (!len_key_a && !len_key_b) {
    return REGISTERED_IDENTIC;                   /* key A identic to key B */
  }

  if (!len_key_a && len_key_b) {
    return NOT_REGISTERED;                     /* key A shorter than key B */
  }

  if (len_key_a && !len_key_b) {
    return REGISTERED_RESTRICT;                /* key A greater than key B */
  }

  diff = strcmp(key_a, key_b);                 /* both keys have a content */

  if (!diff) {                
    return REGISTERED_IDENTIC;        /* contents of both keys are identic */
  }

  len_b = strlen(key_b);

  if (strlen(key_a) > len_b) {   /* definition of key comparison 'greater' */
    diff = strncmp(key_a, key_b, len_b);

    if (!diff) {
      return REGISTERED_RESTRICT;            /* key A 'greater' than key B */
    }
  }

  return NOT_REGISTERED;
}



/* Generation of the first part of the vertical candidate list for the
   frequence domain. The list is just a register of candidates, no substrings
   are defined.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - pointer to x-lag
        - pointer to RB sector horizontal start
        - RB number
        - RB list
        - RB references
        - class list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - vertical list
        - updated RBS start
        - control structure updated
Usage:  - not designed for broken NB.
*/
static cand_lst
vertical_list_part_1_fd(match_def mat, int *lx, int *start_x, int r, 
			rb_lst rblst, rb_ref rref, cls_lst clst, 
			run_def *crun, int single_nb, MATCHING_METHOD mtd)
{
  int        y, z, zmax;
  cand_lst   vlist;                             /* vertical candidate list */
  cv_def_i    kmark;
  sector_def org = rblst.rb[r].sc;

  init_cand_lst(&vlist);

  *start_x = org.column.start+XPMAX(0, *lx-mat.nborder);     /* consistent */
  zmax = XPMIN(rref.kl.kwh[r].ar.height, org.line.end+1);

  do {
    kmark.x = *lx;
    for (y = 0; y < zmax; y++) {
      assert(&rref.kl.kwh[r].ll[y][*lx]);              /* if fail: inquiry */
      kmark.y = y;

      for (z = 0; z < rref.kl.kwh[r].ll[y][*lx].cnt; z++) {
	kmark.cv = rref.kl.kwh[r].ll[y][*lx].ld[z].kwm;
        candidate_list_vertical_fd(mat, kmark, 0, r, rblst,
				   rref.kl.kwh[r].ll[y][*lx].ld[z].n, 
				   rref.kl.kwh[r].ll[y][*lx].ld[z].c,
				   clst.cls, &vlist, rref, crun, single_nb,
				   mtd);
      }
    }

    if (vlist.cnt) break;

    (*lx)++;                                        /* sampling RB f.l.t.r. */
    *start_x = org.column.start+XPMAX(0, *lx-mat.nborder);
  } while (org.column.end-*start_x+1 > 0);

  return vlist;
}



/* Generation of the second part of the vertical candidate list for the
   spatial domain. The list is just a register of candidates, no substrings
   are defined.

Revision: 05.08.2005 E.Sack

Input : - matching parameters
        - lookahead
        - RB sector
        - RB under test idex
        - RB list
        - RB references
        - class list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - vertical list
        - control structure updated
*/
static cand_lst
vertical_list_part_2_sd(match_def mat, int lhead, sector_def rbs, int rut,
			rb_lst rblst, rb_ref rref, cls_lst clst, 
			run_def *crun, int single_nb, MATCHING_METHOD mtd)
{
  int            c, n, j, positive_idx;
  cand_lst       vlist_part, vlist_total;
  REGISTER_INDEX reg_index;

  init_cand_lst(&vlist_part);
  init_cand_lst(&vlist_total);

  for (j = 0; j < XPMIN(INTERVAL_SIZE(rbs.column), lhead); j++) {
    rbs.column.start++;

    reg_index = check_register(rref.xreg, rbs.column.start, rbs.line,
			       rref.pic_state, &positive_idx);

    switch (reg_index) {
    case NOT_REGISTERED:             /* creating the vertical candidate list */
      for (c = 0; c < clst.cnt && *crun->pactive == TRUE; c++) {
	for (n = 0; n < clst.cls[c].nbl.cnt && *crun->pactive == TRUE; n++) {
	  xplab_progress_set_value(crun, -1.0, 0.0); /* check user interrupt */
	  candidate_list_vertical_sd(mat, rbs, lhead-j, rut, rblst, n, c, 
				     clst.cls, &vlist_part, rref, 
				     crun, single_nb, mtd);
	}
      }

      if (vlist_part.cnt) {
	register_positive_result(rref.xreg, rbs.column.start, rbs.line,
				 rref.pic_state, vlist_part);
	cat_cand_lst(&vlist_part, &vlist_total);
	free_cand_lst(&vlist_part);
      }
      else {
	register_negative_result(rref.xreg, rbs.column.start, rbs.line,
				 rref.pic_state);
      }

      break;

    case REGISTERED_POSITIVE_RESTRICT: /* reg. VCL found with restrictions */
      assert(positive_idx>=0 && 
	     positive_idx<rref.xreg.yrl[rbs.column.start].cnt);

      restricted_nb_loop(mat, rbs, lhead-j, rut, rblst, clst, 
			 &vlist_part, rref,
			 rref.xreg.yrl[rbs.column.start].yrd[positive_idx].
			 vlst, crun, single_nb, mtd);

      if (vlist_part.cnt) {
	cat_cand_lst(&vlist_part, &vlist_total);
	free_cand_lst(&vlist_part);
      }

      break;

    case REGISTERED_POSITIVE_IDENTIC:              /* registered VCL found */
      assert(positive_idx>=0 && 
	     positive_idx<rref.xreg.yrl[rbs.column.start].cnt);
      cat_cand_lst(&rref.xreg.yrl[rbs.column.start].yrd[positive_idx].vlst, 
		   &vlist_total);
      break;

    case REGISTERED_NEGATIVE:  /* registered: no vertical cand. list found */
      break;

    default:
      assert(0);
      break;
    }
  }

  return vlist_total;
}



/* Generation of the second part of the vertical candidate list for the
   frequence domain. The list is just a register of candidates, no substrings
   are defined.

Revision: 17.10.2004 E.Sack

Input : - matching parameters
        - lookahead
        - x-lag
        - RB sector horizontal intervall
        - RB number
        - RB list
        - RB references
        - class list
        - pointer to control structure
        - flag (0: multi NB, 1: single NB)
Output: - vertical list
        - control structure updated
Usage:  - not designed for broken NB.
*/
static cand_lst
vertical_list_part_2_fd(match_def mat, int lhead, int lx, ival_def column,
			int r, rb_lst rblst, rb_ref rref, cls_lst clst, 
			run_def *crun, int single_nb, MATCHING_METHOD mtd)
{
  int        j, y, z, zmax;
  cand_lst   vlist;                           /* vertikale Kandidatenliste */
  cv_def_i    kmark;
  sector_def org = rblst.rb[r].sc;

  init_cand_lst(&vlist);

  zmax = XPMIN(rref.kl.kwh[r].ar.height, org.line.end+1);

  for (j = 0; j < XPMIN(column.end-org.column.start+1, lhead); j++) {
    lx++;
    column.start = org.column.start+XPMAX(0, lx-mat.nborder);
    kmark.x = lx;

    for (y = 0; y < zmax; y++) {
      assert(&rref.kl.kwh[r].ll[y][lx]);             /* falls fail: Abfrage */

      kmark.y = y;

      for (z = 0; z < rref.kl.kwh[r].ll[y][lx].cnt; z++) {
	kmark.cv = rref.kl.kwh[r].ll[y][lx].ld[z].kwm;

        candidate_list_vertical_fd(mat, kmark, lhead-j, r, rblst,
				   rref.kl.kwh[r].ll[y][lx].ld[z].n,
				   rref.kl.kwh[r].ll[y][lx].ld[z].c,
				   clst.cls, &vlist, rref, crun, 
				   single_nb, mtd);
      }
    }
  }

  return vlist;
}



/* Builds a substring. The best candidate is choosen from the vertical
   list based on rules.

Revision: 25.10.2002 E.Sack

Input : - vertical candidate list
        - pointer to string index list
        - pointer to availability list
        - pointer to recycling list
        - RB sector
        - class list
Output: - index list updated 
        - index of recycling list 

Precondition:
    idx    yet   string source
 ----------------------------------------------------------------------------
     <0     0    new string = candidate in V-List ==> new entry in R-List
     <0     1    <combination not possible>
    >=0     0    new string = candidate in V-List + old string 
                                                  ==> new entry in R-List
    >=0     1    new string = old string ==> no new entry in R-List
                                             (string recycling)
*/
static int
add_substring(cand_lst vlist, int *idx, int *yet_created, rc_lst *rlist,
	      sector_def rbs, cls_lst cl)
{
  int    v_idx, h_idx;
  double corr_max = -1.0;

  if (!vlist.cnt) {
    return NO_STRING;                                      /* no candidate */
  }

  v_idx = best_candidate(vlist, idx, *rlist, cl, UDEFLINE, &corr_max);
  reduce_alignment_class(vlist, idx, *rlist, cl, v_idx);

  if (v_idx < 0) {
    return NO_STRING;                                   /* all strings bad */
  }

  h_idx = idx[v_idx];

  if (!yet_created[v_idx]) {                     /* string not yet created */
    int      gap;
    cand_lst zlist;

    init_cand_lst(&zlist);

    if (h_idx >= 0) {                 /* add substring from recycling list */
      cat_cand_lst(&rlist->s[h_idx].r, &zlist);

      gap =            /* gap from string to substring needed for gap-rule */
	rlist->s[h_idx].o.column.start - vlist.c[v_idx].knl.column.end-1;
    }
    else {             /* right gap (better offset) from string to the end */
      gap = rbs.column.end - vlist.c[v_idx].knl.column.end;
    }

    cat_cand_def(vlist.c[v_idx], &zlist);/* add candid. from vertical list */

    h_idx = create_rlist(vlist.c[v_idx], corr_max, &zlist, rlist, gap);

    free_cand_lst(&zlist);
  }

  return h_idx;             /* return new string index from recycling list */
}



/* Reduces vertical candidate list to one candidate per alignment class.
   The reduction to one alignment class with one candidate will be performed
   in the line-sorting calculation, where the needed infos are available.

Improvement: Pattern matching and line calculation should be iteratively
             performed because of (theoretically) mutual dependence. Every
             method used classifies only a subset of all features, however
             all features should be used at once. The straight forward
             order of methods application therefore assumes that features
             are not linked too close, e.g. patterns have not too big
             neighborhoods or relations among others.

Revision: 22.02.2003 E.Sack

Input : - vertical candidate list
        - pointer to string index list
        - pointer to recycling list
        - class list
	- index of best candidate
Output: - best candidate updated
*/
static void
reduce_alignment_class(cand_lst vlist, int *idx, rc_lst rlist, cls_lst cl,
		       int best)
{
  reduce_alignment(vlist, idx, rlist, cl, best, UNDERLINE);
  reduce_alignment(vlist, idx, rlist, cl, best, BASELINE);
  reduce_alignment(vlist, idx, rlist, cl, best, MIDDLELINE);
  reduce_alignment(vlist, idx, rlist, cl, best, UPPERLINE);
}



/* Searches best candidate of alignment class.

Revision: 23.02.2003 E.Sack

Usage: Used for vertical candidate lists.

Input : - vertical candidate list
        - pointer to string index list
        - recycling list
        - class list
	- index of best candidate
	- alignment class
Output: - best candidate updated
*/
static void
reduce_alignment(cand_lst vlist, int *idx, rc_lst rlist, cls_lst clst,
		 int best, ALIGMNT align_class)
{
  int        vdx;
  double     corr_dum;
  sector_def best_roi;

  if (clst.cls[vlist.c[best].c].att.alignment == align_class) {
    return;
  }

  vdx = best_candidate(vlist, idx, rlist, clst, align_class, &corr_dum);

  if (vdx < 0) {
    return;
  }

  best_roi = vlist.c[best].roi;

  if (best_roi.column.start - vlist.c[vdx].roi.column.start +
	best_roi.column.end - vlist.c[vdx].roi.column.end +
	best_roi.line.start - vlist.c[vdx].roi.line.start +
	best_roi.line.end - vlist.c[vdx].roi.line.end) {
    return;  /* unfit <- constrained to same area ensures real competitors */
  }

  if (vlist.c[vdx].m.cv < vlist.c[best].m.cv*vlist.c[best].m.cv) {
    return;                                              /* unfit, too bad */
  }

  if (!vlist.c[best].alt) {
    CALLOC(cand_lst, 1, vlist.c[best].alt);
    init_cand_lst(vlist.c[best].alt);
  }

  cat_cand_def(vlist.c[vdx], vlist.c[best].alt);    /* add candid. to list */
}




/* Searches best candidate in list.

Revision: 22.02.2003 E.Sack

Usage: Used for vertical candidate lists.

Input : - candidate list
        - pointer to string index list
        - pointer to availability list
        - pointer to recycling list
        - RB sector
        - class list
	- alignment constraint
	- pointer to string correlation value
Output: - index of best candidate (-1: fail)
        - string correlation value according to index output
*/
static int
best_candidate(cand_lst vlist, int *idx, rc_lst rlist, cls_lst clst,
	       ALIGMNT align_class, double *corr_max)
{
  int    i, vdx = -1, cand_cnt, cnr;
  double corr_val;

  *corr_max = -1.0;

  for (i = 0; i < vlist.cnt; i++) {
    if (idx[i] == BAD_STRING) {
      continue;                                 /* string/substring failed */
    }

    cnr = vlist.c[i].c;                          /* candidate class number */

    if (align_class != UDEFLINE) {                /* alignment constrained */
      if (align_class != clst.cls[cnr].att.alignment) {/*alignment not fit */
	continue;
      }
    }

    if (idx[i] >= 0) {        /* string/substring exists in recycling list */
      corr_val = rlist.s[idx[i]].k;
      cand_cnt = rlist.s[idx[i]].r.cnt;
    }
    else {                    /* substring doesn't exist in recycling list */
      corr_val = 0.0;
      cand_cnt = 0;
    }

    corr_val = (corr_val*(double)cand_cnt + vlist.c[i].m.cv)/    /* renorm */
                                        (double)(cand_cnt+1);

    if (*corr_max < corr_val) {                 /* true for the first time */
      *corr_max = corr_val;
      vdx  = i;                              /* candidate in vertical list */
    }
    else if (*corr_max == corr_val) {             /* who is the biggest NB */
      if (AREA_SIZE(clst.cls[cnr].nbl.nb[vlist.c[i].n].pat.ar) >
	  AREA_SIZE(clst.cls[vlist.c[vdx].c].nbl.nb[vlist.c[vdx].n].pat.ar)) {
	vdx  = i;
      }
    }
  }

  return vdx;
}



/* Chooses candidates against gap criterion.

Revision: 25.10.2002 E.Sack

Input : - vertical candidate list
        - pointer to candidate index list
        - pointer to candidate availability list
        - recycling list
        - RB sector
        - maximal allowed lag
Output: - index list updated
 */
static void
sifting_gap(cand_lst vlist, int *idx, int *yet_created, rc_lst rlist,
	    sector_def rbs, int noise_border)
{
  int i, j, gap_a, gap_b, ref_end_a, ref_end_b;

  for (i = 0; i < vlist.cnt-1; i++) {
    if (idx[i] == BAD_STRING) {
      continue;
    }

    gap_a = gap_sum(idx[i], yet_created[i], rlist, vlist.c[i].knl.column,
		    rbs.column, &ref_end_a);

    for (j = i+1; j < vlist.cnt; j++) {
      if (idx[j] == BAD_STRING) {
	continue;
      }

      gap_b = gap_sum(idx[j], yet_created[j], rlist, vlist.c[j].knl.column,
		      rbs.column, &ref_end_b);

      if (abs(ref_end_a - ref_end_b) <= noise_border && /* same ref. points */
          abs(gap_a - gap_b) > noise_border) {          /* too restrictive? */
	if (gap_a > gap_b) {
	  idx[i] = BAD_STRING;                                    /* failed */
	}
	else {
	  idx[j] = BAD_STRING;                                    /* failed */
	}
      }
    }
  }
}



/* Sifting candidates regarding overlap criterium:

   eliminate candidate, if he 
   - is laying right to the beginning of a substring,
   - has common pixels with the substring.

Revision: 24.07.2002 E.Sack

Input : - vertical candidate
        - substring index of recycling list
        - availability flag
        - recycling list
        - class list
        - mass list
Output: - substring index
 */
static int
sifting_lap_h(cand_def cand, int idx, int yet_created, rc_lst rlist, 
	      cls_lst clst, double overlap)
{
  int        ovl, left_cls, right_cls, left_nb, right_nb, lx, ly;
  double     min_mass;
  sector_def right, left;

  if (idx >= 0 && !yet_created) {                              /* substring */
    left = cand.knl;                     /* H-List build from right to left */
    right = rlist.s[idx].ls;

    if (left.column.end < right.column.start) {
      return idx;                           /* best case, no overlap at all */
    }

    if (left.column.start >= right.column.start) {
      return BAD_STRING;          /* should appear in V-List, not in H-List */
    }

    if (left.column.end >= right.column.start &&
        (left.line.end < right.line.start   ||
	 left.line.start > right.line.end)) {
      return idx;                              /* even possible, no overlap */
    }

    /* overlapping candidate with substring                                 */
    lx = right.column.start - left.column.start;     /* candidate offset to */
    ly = right.line.start - left.line.start;         /* substring           */

    left_cls = cand.c;
    left_nb = cand.n;
    right_cls = rlist.s[idx].lc;
    right_nb = rlist.s[idx].ln;

    ovl = overlapping_degree(clst.cls[left_cls].nbl.nb[left_nb].pat,
			     clst.cls[right_cls].nbl.nb[right_nb].pat, lx, ly);

    min_mass = (double)XPMIN(clst.cls[left_cls].nbl.nb[left_nb].ref.nbm, 
			     clst.cls[right_cls].nbl.nb[right_nb].ref.nbm);

    if ((double)ovl > min_mass*overlap) {
      return BAD_STRING;                                  /* overlap exists */
    }
  }

  return idx;
}



/* The gap sum of the candidate is computed with the reference point.

Input : - index flag
        - available flag
        - recycling list
        - candidate interval
        - RBS interval
        - pointer to reference point
Output: - updated reference point
        - gap sum
Postcondition: - only positive gaps counted.
*/
static int 
gap_sum(int idx, int available, rc_lst rlist, ival_def cand, ival_def rbs,
	int *ref_end)
{
  int gap;

  if (idx >= 0) {                       /* existing in recycling list is... */
    gap = available ? 
	       rlist.s[idx].g :              /* ...string, gap yet computed */
               rlist.s[idx].o.column.start -     /* ...substring, gap not   */
                                  cand.end - 1;  /* computed.               */
  }
  else {
    gap = rbs.end - cand.end;
  }

  *ref_end = cand.end + gap;                           /* most left start-1 */

  gap = XPMAX(0, gap);                                      /* right gap... */

  gap += XPMAX(0, cand.start-rbs.start);                /* ...plus left gap */

  return gap;
}



/* Investigates the image of unclassified gaps left.

Revision: 19.01.2003 E.Sack

Input : - pointer to hitlist
        - RB list
        - gap criterion
Output: - hitlist updated
Postcondition: - hitlist may be enlarged with gap substitutes 
*/
static void 
check_gap(cand_lst *hlst, rb_lst rbl, int gap_criterion) 
{
  int        i, j, k, most_left, start, end, r, cnt_old = hlst->cnt;
  int        *done = 0, start_tmp;
  int_lst    nl;               /* list of all candidates of one certain RB */
  sector_def gap;

  if (!cnt_old) {
    return;
  }
  else {
    assert(rbl.cnt>0);
  }

  init_int_lst(&nl);

  CALLOC(int, cnt_old, done);

  for (i = 0; i < cnt_old; i++) {
    if (done[i]) {
      continue;                                       /* candidate handled */
    }

    r = hlst->c[i].rbn;                                  /* candidate's RB */

    end   = rbl.rb[r].sc.column.start - 1;/*to point outside image allowed */
    start = hlst->c[i].knl.column.start;              /* candidate's start */
    most_left = 0;                             /* first candidate enrolled */

    free_int_lst(&nl);
    for (j = 0; j < cnt_old; j++) {           /* enroll all candates of RB */
      if (hlst->c[j].rbn != r) {
	continue;                    /* not a candidate of investigated RB */
      }

      start_tmp = hlst->c[j].knl.column.start;
      if (start > start_tmp) {                     /* most left on the fly */
	start = start_tmp;
	most_left = nl.cnt;
      }

      if (nl.cnt++) {
	REALLOC(int, nl.cnt, nl.nr);
      }
      else {
	CALLOC(int, nl.cnt, nl.nr);
      }

      nl.nr[nl.cnt-1] = j;    /* enroll all RB candidates in list (indexes) */
    }

    gap.line = rbl.rb[r].sc.line;
    for (k = 0; k < nl.cnt; k++) {       /* check distance of RB candidates */
      gap.column.start = end   + 1;      /* |--(RB1)--|s-(gap)-e|--(RB2)--| */
      gap.column.end   = start - 1;

      if (INTERVAL_SIZE(gap.column) > gap_criterion) {/* gap not classified */
        cat_cand_def(substitution(r, gap, SUBSTNB_FLAG), hlst); /* substit. */
      }

      end   = hlst->c[nl.nr[most_left]].knl.column.end;
      start = rbl.rb[r].sc.column.end + 1; /*to point outside image allowed */

      done[nl.nr[most_left]] = 1;              /* mark candidate as handled */
      nl.nr[most_left] = -1;                   /* mark candidate as handled */

      for (j = 0; j < nl.cnt; j++) {      /* investigate next RB candidates */
        if (nl.nr[j] == -1) {
	  continue;                                    /* candidate handled */
	}

	start_tmp = hlst->c[nl.nr[j]].knl.column.start;
        if (start > start_tmp) {                      /* sorting on the fly */
	  start = start_tmp;                                   /* next left */
          most_left = j; /* index of next unhandled most left cand. in list */
        }
      }                
    }

    gap.column.start = end   + 1;              /* finish with gap to RB end */
    gap.column.end   = start - 1;

    if (INTERVAL_SIZE(gap.column) > gap_criterion) {  /* gap not classified */
      cat_cand_def(substitution(r, gap, SUBSTNB_FLAG), hlst); /*substitution*/
    }
  }

  free_int_lst(&nl);
  FREEMEM(done);
}



/* Constructs a candidate substitution.

Revision: 31.12.2002 E.Sack

Input : - RB number
        - sector
        - NB number substitution
Output: - candidate substitution
*/
static cand_def 
substitution(int rb_nr, sector_def sec, int sub) 
{
   cand_def cd;

   assert(rb_nr >= 0);
   assert(sub < 0);

   init_cand_def(&cd);

   cd.rbn = rb_nr;                                     /* RB to substitute */
   cd.n = sub;                                   /* flag for NB substitute */
   cd.c = SUBCLS_FLAG;                           /* flag for invalid class */
   cd.roi = sec;
   cd.ext = sec;
   cd.knl = sec;                 /* frame substitution without mask border */
   cd.hdx = sec;

   assert(cd.c < 0);

   return cd;
}



/* Generates the RB references.

Revision: 05.08.2005 E.Sack

Input : - matching parameters
        - pointer to RB image
        - class list
        - RB list
        - domain flag (0: spatial domain, 1: frequency domain)
        - flag (0: multi NB, 1: single NB)
Output: - RB references
*/
static rb_ref
gen_rb_references(match_def mat, pic_def *rpic, cls_lst clst, rb_lst rblst,
		  int fdomain, int single_nb, MATCHING_METHOD mtd)
{
  int    i;
  rb_ref rref;

  init_rb_ref(&rref);

  assert(rpic);
  rref.rpic = rpic;

  rref.wpic = xpic_def_weighting(*rpic, mat.maskweight);

  rref.shdw1.ar = rpic->ar;
  CALLOC(PICT, AREA_SIZE(rref.shdw1.ar), rref.shdw1.mem);

  rref.shdw2.ar = rpic->ar;
  CALLOC(PICT, AREA_SIZE(rref.shdw2.ar), rref.shdw2.mem);

  rref.xreg.cnt = rpic->ar.width; 
  CALLOC(y_reg_lst, rref.xreg.cnt, rref.xreg.yrl);

  for (i = 0; i < rref.xreg.cnt; i++) {
    init_y_reg_lst(&rref.xreg.yrl[i]);
  }

  if (fdomain) {
    rref.kl = gen_kwh(mat, clst, rblst, *rpic, single_nb, mtd);
  }

  return rref;
}



/* Weights a binary image.

Revision: 10.06.2004 E.Sack

Input : - image
Output: - weighted image
*/
static pic_def
xpic_def_weighting(pic_def in, int maskweight)
{
  int     i, len = AREA_SIZE(in.ar);
  pic_def eroded, out;

  assert(len>0);

  init_pic_def(&eroded);
  init_pic_def(&out);

  out.ar = in.ar;
  CALLOC(PICT, len, out.mem);

  eroded.ar = in.ar;
  CALLOC(PICT, len, eroded.mem);

  copy_bmps(in, eroded, in.ar, 0, 0, 0, 0);                    /* 1:1 copy */
  copy_bmps(in, out, in.ar, 0, 0, 0, 0);                       /* 1:1 copy */

  for (i = 0; i < maskweight; i++) {
    erode_in_box(eroded);
    add_grey(eroded, out, eroded.ar, 0, 0, 0, 0);             /* weighting */
  }

  free_pic_def(&eroded);

  out.org = in.org;                                 /* copy absolut origin */

  return out;
}



/* Fills holes of a binary image.

Revision: 15.09.2002 E.Sack

Input : - image
Output: - filled image
*/
static pic_def
xpic_def_filling(pic_def in)
{
  sector_def fr;
  pic_def    out;

  init_pic_def(&out);

  out.ar = in.ar;
  CALLOC(PICT, AREA_SIZE(out.ar), out.mem);

  copy_bmps(in, out, in.ar, 0, 0, 0, 0);                       /* 1:1 copy */

  AREA_TO_SECTOR(out.ar, fr);

  contour_fill_all(out, fr, FOREGROUND, FOREGROUND);

  out.org = in.org;                                 /* copy absolut origin */

  return out;
}



/* Dilates a binary image.

Revision: 15.09.2002 E.Sack

Input : - image
Output: - dilated image
*/
static pic_def
xpic_def_dilating(pic_def in, int border)
{
  int     i;
  pic_def out;

  assert(border>=0);

  init_pic_def(&out);

  out.ar = in.ar;
  CALLOC(PICT, AREA_SIZE(out.ar), out.mem);

  copy_bmps(in, out, in.ar, 0, 0, 0, 0);                       /* 1:1 copy */

  for (i = 0; i < border; i++) {
    dilate_in_box(out);
  }

  out.org = in.org;                                 /* copy absolut origin */

  return out;
}



/* Creates the NB references.

Revision: 08.06.2004 E.Sack

Input : - image
        - mask border
Output: - NB references
*/
static nb_ref
gen_nb_ref(pic_def pat, int mask_border, int maskweight)
{
  double border_mass;
  nb_ref nref;

  assert(mask_border >= 0);

  init_nb_ref(&nref);

  nref.pic = xextend_image(pat, mask_border, mask_border);    /* framed NB */

  nref.wtd = xpic_def_weighting(nref.pic, maskweight);
  nref.fld = xpic_def_filling(nref.pic);
  nref.msk = xpic_def_dilating(nref.fld, mask_border);          /* NB mask */
  nref.dlt = xpic_def_dilating(nref.pic, mask_border);

  nref.dltm = mass_of_image(nref.dlt);
  nref.mskm = mass_of_image(nref.msk);
  nref.fldm = mass_of_image(nref.fld);
  nref.nbm = mass_of_image(pat);                     /* mass of NB kernels */
  nref.wtdm = mass_of_masked_image(nref.msk, nref.wtd, 1); /* BGND neutral */

  border_mass = 1.0 +        /* pixel added to avoid float problems easier */
                nref.mskm - nref.nbm;

  nref.cmin =                  /* avoid small NB effects caused by masking */
              border_mass/(double)nref.mskm;

  return nref;
}



/* Creates the NB references of a class list.

Revision: 03.10.2002 E.Sack

Input : - class list
        - mask border
Output: - class list updated
*/
void
gen_class_references(cls_lst clst, int mask_border, int maskweight)
{
  int cnr;

  assert(mask_border >= 0);

  for (cnr = 0; cnr < clst.cnt; cnr++) {
    gen_nb_references(clst.cls[cnr].nbl, mask_border, maskweight);
  }
}



/* Creates the NB references of a NB list.

Revision: 03.10.2002 E.Sack

Input : - NB list
        - mask border
Output: - NB list updated
*/
static void
gen_nb_references(nb_lst nbl, int mask_border, int maskweight)
{
  int nnr;

  assert(mask_border >= 0);

  for (nnr = 0; nnr < nbl.cnt; nnr++) {
    free_nb_ref(&nbl.nb[nnr].ref);                  /* allows reallocation */
    nbl.nb[nnr].ref = gen_nb_ref(nbl.nb[nnr].pat, mask_border, maskweight);
  }
}



/* Computation of the image mass by adding the masked pixel values.
   Masked background added with selectable value.

Revision: 03.10.2002 E.Sack

Input : - mask
        - image
        - background weight
Output: - mass
Precondition: masked background adding value not greater than pixel values.
*/
static double
mass_of_masked_image(pic_def mask, pic_def pic, int mbv)
{
  int    x, y, h = pic.ar.height, w = pic.ar.width;
  double mass;

  h *= w;

  for (y = 0, mass = 0.0; y < h; y+=w) {
    for (x = 0; x < w; x++) {
      if (mask.mem[x + y]) {
	mass += (double)XPMAX(mbv, pic.mem[x + y]);
      }
    }
  }

  return mass;
}



/* Computation of the image mass by adding the pixel values.

Input : - image
Output: - mass
*/
static double
mass_of_image(pic_def pic)
{
  int    x, y, h = pic.ar.height, w = pic.ar.width;
  double mass;

  h *= w;

  for (y = 0, mass = 0.0; y < h; y+=w) {
    for (x = 0; x < w; x++) {
      mass += (double)pic.mem[x + y];
    }
  }

  return mass;
}



/* Searches the recycling list for a hash index. Image state is not 
   concerned so assumed be consistent. Consistency is generally true if 
   patterns are propounded only once and registered otherwise. Registration 
   uses image state.

Revision: 31.01.2003 E.Sack

Input : - recycling list
        - hash mark searched for
Output: - index (<0 no entry found) 
*/
static int
lookup_rc_lst(rc_lst rlist, sector_def hdx)
{
  int i;

  for (i = 0; i < rlist.cnt; i++) {
    if (rlist.s[i].sid.line.start == hdx.line.start && 
        rlist.s[i].sid.line.end == hdx.line.end && 
        rlist.s[i].sid.column.start == hdx.column.start && 
        rlist.s[i].sid.column.end   == hdx.column.end) {
	return i;
    }
  }

  return NO_STRING;
}



/* Extends recycling list with one entry.

Revision: 22.01.2003 E.Sack

Input : - candidate
        - correlation value
        - pointer to substring candidates
        - pointer to recycling list
        - gap from most left candidate to the next right candidate
Output: - index of recycling list
        - recycling list updated
*/
static int
create_rlist(cand_def kr, double kwert, cand_lst *zlist, rc_lst *rlist, 
	     int gap)
{
  assert(kwert > 0.0);

  if (rlist->cnt) {
    REALLOC(rc_def, rlist->cnt+1, rlist->s);
  }
  else {
    CALLOC(rc_def, rlist->cnt+1, rlist->s);
  }

  init_rc_def(&rlist->s[rlist->cnt]);

  rlist->s[rlist->cnt].sid = kr.hdx;                   /* update hash index */
  rlist->s[rlist->cnt].k = kwert;         /* correlation value of substring */

  copy_cand_lst(zlist, &rlist->s[rlist->cnt].r);

  rlist->s[rlist->cnt].o.column.start = 
    XPMIN(rlist->s[rlist->cnt].o.column.start, kr.knl.column.start);

  rlist->s[rlist->cnt].o.column.end = 
    XPMAX(rlist->s[rlist->cnt].o.column.end, kr.knl.column.end);

  rlist->s[rlist->cnt].g = gap;

  rlist->s[rlist->cnt].ls = kr.knl;                               /* kernel */
  rlist->s[rlist->cnt].lc = kr.c;       /* position bounded by image, but */
  rlist->s[rlist->cnt].ln = kr.n;        /* overlaps cannot exceed image. */

  return rlist->cnt++;
}



/* Checks NB fitness according rules.

Revision: 08.01.2005 E.Sack

Input : - pointer to RB-sector (seek frame)
        - pointer to extended RB-sector
        - lookahead
        - NB area size
        - image
        - mask border
        - maximal allowed lag
        - flag (0: multi NB, 1: single NB)
Output: - RB-sector updated as flagged
        - flag (0: NB doesn't fit or no pattern found, 
                1: NB fits and pattern available)
Postcondition: - start value of RBS unchanged always
*/
static int
check_nb(sector_def *rbs, sector_def *rbse, int lhead, area_def nar, 
	 pic_def rpic, int mborder, int nborder, int single_nb, 
	 MATCHING_METHOD mtd)
{
  int width, ext_x, ext_y, nborder_x, nborder_y, bsx;

  assert(rbs->column.start >= 0);
  assert(rbs->line.start >= 0);

  nborder_x = rule_nb_size(rbs->column, nar.width, nborder, single_nb);
  if (nborder_x < 0) {
    return 0;                        /* NB doesn't fit because of its size */
  }

  nborder_y = rule_nb_size(rbs->line, nar.height, nborder, single_nb);
  if (nborder_y < 0) {
    return 0;                        /* NB doesn't fit because of its size */
  }

  width = XPMAX(lhead, nar.width);
  if (lhead) {                              /* special treatment for width */
    bsx = width;                                       /* full shift range */
  }
  else {
    bsx = XPMAX(1, width-2*nborder_x);             /* minimal shift a'_{2} */
  }

  if (mtd == MATCHING_METHOD_2) {             /* enlarge RBS for broken NB */
    rbs->column.end = XPMAX(rbs->column.end, rbs->column.start+bsx-1);
  }
  else {                                      /* minimize RBS for solid NB */
    rbs->column.end = XPMIN(rbs->column.end, rbs->column.start+bsx-1);
  }

  ext_x = nborder_x + mborder;
  rbse->column.start = XPMAX(0, rbs->column.start-ext_x);          /* e_{1} */
  rbse->column.end = XPMIN(rpic.ar.width-1, rbs->column.end+ext_x);/* e_{2} */

  ext_y = nborder_y + mborder;
  rbse->line.start = XPMAX(0, rbs->line.start-ext_y);             /* e_{1} */
  rbse->line.end = XPMIN(rpic.ar.height-1, rbs->line.end+ext_y);  /* e_{2} */

  return 1;
}



/* Lag transformation from frequency to spatial domain.

Input : - RB intervall
        - pointer to RBS
        - pointer to extended RBS
        - NB size
        - image size
        - mask border size
        - maximal allowed lag
        - pointer to lag 
Output: - updated RBS
        - updated extended RBS
        - updated lag
Precondition:  - input lag with frequency value.
Postcondition: - input lag with spatial value.
*/
static void
trans_lag(ival_def org, ival_def *rbs, ival_def *rbse, int nb_len, 
	  int pic_len, int mask_border, int noise_border, int *lag)
{
  int s1;

  s1 = org.start + XPMAX(0, *lag-noise_border);

  rbs->start = s1;
  rbs->end = XPMIN(pic_len-1, s1+nb_len-1);
  rbse->start = XPMAX(0, s1-mask_border);
  rbse->end = XPMIN(pic_len-1, s1+nb_len-1+mask_border);
  *lag = XPMIN(0, *lag-noise_border) - mask_border;
}



/* Es wird untersucht, ob Regeln verletzt werden.

Input : - NB line range
        - RB line range
        - alignment flag
Output: - Flag (0: no rule failed, 1: rule failed)
*/
static int
line_rule(ival_def nb_row, ival_def rb_row, ALIGMNT alignment) 
{
  double cel_a, bottom, top;

  if (rb_row.start == nb_row.start &&
      rb_row.end   == nb_row.end) {
    return 0;           /* eine Unterscheidung kann nicht getroffen werden */
  }

  cel_a = (double)(rb_row.end - rb_row.start + 1)/4.0;
  bottom = (double)(rb_row.end - nb_row.end);
  top = (double)(nb_row.start - rb_row.start);

  /* Lage testen:                |__a__|__a__|__a__|__a__|   height = 4*a  */
  switch (alignment) {
  case UPPERLINE : if (top >= cel_a) return 1;
    break;
  case MIDDLELINE: if (top >= cel_a*2.0) return 1;
    if (bottom >= cel_a*3.0) return 1;
    break;
  case BASELINE:   if (top >= cel_a*3.0) return 1;
    if (bottom >= cel_a*2.0) return 1;
    break;
  case UNDERLINE : if (bottom >= cel_a) return 1;
    break;
  case UDEFLINE  :
  default: assert(0); break;
  }

  return 0;                                              /* NB position ok */
}



/* A NB aspirant will be investigated and enrolled into the candidate list 
   if he succeeds (matches).

   There are two versions of the routine; one with the correlation made in 
   spatial domain and one with the correlation made in frequency domain.

Revision: 05.01.2005 E.Sack

Input : - matching parameters
        - RB under test index
        - NB number
        - class number
        - pointer to class list
        - RBS
        - extended RBS
        - RB references
        - pointer to candidate list
        - original RBS
        - RB list
        - flag (0: non-broken NB, 1: broken NB)
        - method
Output: - candidate list updated
*/
static void 
check_hit_sd(match_def mat, int rut, int n, int c, cls_def *cls, 
	     sector_def rbs, sector_def rbse, rb_ref rref, cand_lst *klist, 
	     sector_def rbs_org, rb_lst rblst, MATCHING_METHOD mtd)
{
  int        d;
  double     cv_threshold;
  sector_def hdx, knl;
  cv_def_i   kmark;
  int_lst    crb;

  if (!cls[c].att.printable) {
    return;                                           /* printable NB only */
  }

  cv_threshold = (mtd == MATCHING_METHOD_2) ? mat.corrbroken : mat.corrlevel;

  kmark = masked_correlation_ext(cls[c].nbl.nb[n].ref.pic, 
				 cls[c].nbl.nb[n].ref.msk, rbs, rbse,
				 *rref.rpic, cls[c].nbl.nb[n].ref.mskm, 
				 mat.mborder, mat.nborder, cv_threshold);
  kmark.cv /= cls[c].nbl.nb[n].ref.mskm;       /* normalize with mask area */

  if (kmark.cv < cv_threshold) {
    return;
  }

  if (cls[c].nbl.nb[n].nnb.cnt) {          /* correlation extended for CNB */
    crb = cnb_validation(c, n, cls, rbs_org,  /* transform to original RBS */
			 kmark.x+rbs.column.start-rbs_org.column.start, 
			 kmark.y+rbs.line.start-rbs_org.line.start, 
			 rref, rut, rblst, &d, mat.nnbmatch, mat.mborder);
    if (d) {
      return;                                              /* no valid CNB */
    }
  }
  else {
    init_int_lst(&crb);
  }

  kmark.cv = (double)wtd_correlation_ext(cls[c].nbl.nb[n].ref.wtd, 
		 	 		 cls[c].nbl.nb[n].ref.msk, rbs,
					 rbse, rref.wpic, kmark.x, kmark.y);

  kmark.cv /= cls[c].nbl.nb[n].ref.wtdm;       /* normalize with mask area */

  if (kmark.cv < XPMAX(cv_threshold, cls[c].nbl.nb[n].ref.cmin))  {
    free_int_lst(&crb);
    return;
  }

  /* NB location ignoring document borders and without mask border         */
  hdx.column.start = rbs.column.start + kmark.x + mat.mborder;
  hdx.column.end   = hdx.column.start + cls[c].nbl.nb[n].pat.ar.width-1;
  hdx.line.start   = rbs.line.start + kmark.y + mat.mborder;
  hdx.line.end     = hdx.line.start + cls[c].nbl.nb[n].pat.ar.height-1;

  /* NB location considering document borders and without mask border      */
  knl.column.start = XPMAX(0, hdx.column.start);
  knl.line.start   = XPMAX(0, hdx.line.start);
  knl.column.end   = XPMIN(rref.rpic->ar.width-1, hdx.column.end);
  knl.line.end     = XPMIN(rref.rpic->ar.height-1, hdx.line.end);

  if (mtd == MATCHING_METHOD_2) {    /* check for broken pattern fragments */
    pattern_fragments_i(&crb, knl, rut, rblst);
  }

  if (klist->c) {
    REALLOC(cand_def, klist->cnt+1, klist->c);
  }
  else {
    CALLOC(cand_def, klist->cnt+1, klist->c);
  }

  init_cand_def(&klist->c[klist->cnt]);

  klist->c[klist->cnt].mtd = mtd;                 /* method succeeded      */
  klist->c[klist->cnt].m   = kmark;               /* registering candidate */
  klist->c[klist->cnt].rbn = rut;                 /* RB under test index   */
  klist->c[klist->cnt].n   = n;                   /* NB number             */
  klist->c[klist->cnt].c   = c;                   /* class number          */
  klist->c[klist->cnt].line= rbs_org.line;        /* RB line               */
  klist->c[klist->cnt].roi = rbs;                 /* location              */
  klist->c[klist->cnt].ext = rbse;                /* extended location     */
  klist->c[klist->cnt].knl = knl;                 /* NB kernel             */
  klist->c[klist->cnt].hdx = hdx;                 /* hash index            */
  klist->c[klist->cnt].crb = crb;                 /* CRB list              */
  klist->cnt++;
}



/* A NB aspirant will be investigated and enrolled into the candidate list 
   if he succeeds.

   There are two versions of the routine; one with the correlation made in 
   spatial domain and one with the correlation made in frequency domain.

Revision: 03.10.2002 E.Sack

Input : - matching parameters
        - RB number
        - NB number
        - class number
        - pointer to class
        - RBS
        - extended RBS
        - RB references
        - pointer to candidate list
        - original RBS
        - RB list
        - correlation features
Output: - candidate list updated
Usage:  - not designed for broken NB.
*/
static void
check_hit_fd(match_def mat, int r, int n, int c, cls_def *cls, sector_def rbs,
	     sector_def rbse, rb_ref rref, cand_lst *klist, 
	     sector_def rbs_org, rb_lst rblst, cv_def_i kmark)
{
  int        d;
  sector_def hdx, knl;
  int_lst    crb;

  if (!cls[c].att.printable) {
    return;                                           /* only printable NB */
  }

  if (kmark.cv < mat.corrlevel) {
    return;
  }

  if (cls[c].nbl.nb[n].nnb.cnt) {
    crb = cnb_validation(c, n, cls, rbs_org,  /* transform to original RBS */
			 kmark.x+rbs.column.start-rbs_org.column.start, 
			 kmark.y+rbs.line.start-rbs_org.line.start, 
			 rref, r, rblst, &d, mat.nnbmatch, mat.mborder);
    if (d) {
      return;                                              /* no valid CNB */
    }
  }
  else {
    init_int_lst(&crb);
  }

  kmark.cv = (double)wtd_correlation_ext(cls[c].nbl.nb[n].ref.wtd, 
					 cls[c].nbl.nb[n].ref.msk, rbs,
					 rbse, rref.wpic, kmark.x, kmark.y);

  kmark.cv /= cls[c].nbl.nb[n].ref.wtdm;       /* normalize with mask area */

  if (kmark.cv < mat.corrlevel)  {
    free_int_lst(&crb);
    return;
  }

  /* NB location ignoring document borders and without mask border         */
  hdx.column.start = rbs.column.start + kmark.x + mat.mborder;
  hdx.column.end = hdx.column.start + cls[c].nbl.nb[n].pat.ar.width - 1;
  hdx.line.start = rbs.line.start + kmark.y + mat.mborder;
  hdx.line.end = hdx.line.start + cls[c].nbl.nb[n].pat.ar.height - 1;

  /* NB location considering document borders and without mask border      */
  knl.column.start = XPMAX(0, hdx.column.start);
  knl.line.start = XPMAX(0, hdx.line.start);
  knl.column.end = XPMIN(rref.rpic->ar.width-1, hdx.column.end);
  knl.line.end = XPMIN(rref.rpic->ar.height-1, hdx.line.end);

  if (line_rule(knl.line, rbs_org.line, cls[c].att.alignment)) {
    free_int_lst(&crb);
    return;
  }

  if (klist->c) REALLOC(cand_def, klist->cnt+1, klist->c);
  else CALLOC(cand_def, klist->cnt+1, klist->c);

  init_cand_def(&klist->c[klist->cnt]);

  klist->c[klist->cnt].m   = kmark;               /* registering candidate */
  klist->c[klist->cnt].rbn = r;                      /* RB number          */
  klist->c[klist->cnt].n   = n;                      /* NB number          */
  klist->c[klist->cnt].c   = c;                      /* class number       */
  klist->c[klist->cnt].line= rbs_org.line;           /* RB line            */
  klist->c[klist->cnt].roi = rbs;                    /* location           */
  klist->c[klist->cnt].ext = rbse;                   /* extended location  */
  klist->c[klist->cnt].knl = knl;                    /* NB kernel          */
  klist->c[klist->cnt].hdx = hdx;                    /* hash index         */
  klist->c[klist->cnt].crb = crb;                    /* CRB list           */
  klist->cnt++;
}



/* Validates CNB and assigns associated RBs.

Revision: 02.02.2003 E.Sack

Input : - parent CNB class number
        - parent CNB nb number
        - pointer to classes
        - parent CNB sector
        - parent CNB lag in parent CNB sector
        - RB references
        - RB under test index
        - RB list
        - pointer to flag (0: valid CNB, 1: no valid CNB)
        - pixel matching degree
        - mask border
Output: - CRB list as flagged
        - flag updated
Precondition: - parent CNB lies in parent CNB sector,
              - lag for mask bordered CNB
*/
static int_lst
cnb_validation(int c, int n, cls_def *cls, sector_def rbs, int lx, int ly, 
	       rb_ref rref, int rut, rb_lst rblst, int *fail, int nnbmatch,
	       int mborder)
{
  int_lst crb;                                           /* CRB index list */

  crb = cnb_validation_by_pixel(c, n, cls, rbs, lx, ly, rref, rut, rblst,  
				fail, nnbmatch, mborder);
  if (!*fail) {
    *fail = cnb_validation_by_contour(c, n, cls, rbs, lx, ly, *rref.rpic);
  }

  if (*fail) {
    free_int_lst(&crb);
  }

  return crb;                       /* may be empty if CNB contained in RB */
}



/* Assigns unbooked associated RBs to broken NB.

Revision: 04.01.2005 E.Sack

Input : - pointer to RB index list
        - sector of broken NB
        - RB under test index
        - RB list
Output: - index list updated
        - flag updated (0: valid CNB, 1: no valid CNB)
Precondition: - parent CNB lies in parent CNB sector,
              - lag for mask bordered CNB
Usage: - this function performs its task only together with its counterpart
         pattern_fragments_ii().
*/
static void
pattern_fragments_i(int_lst *brb, sector_def rbs_broken, int rut, 
		    rb_lst rblst)
{
  int ridx;

  assert(brb);

  if (brb->nr) {                                      /* check consistency */
    assert(brb->cnt>0);
  }
  else {
    assert(!brb->cnt);
  }

  for (ridx = 0; ridx < rblst.cnt; ridx++) {
    if (ridx == rut) {
      continue;                                        /* listed elsewhere */ 
    }

    if (rblst.rb[ridx].rld == RB_STATUS_INIT) {             /* unbooked RB */
      if (A_WITHIN_B_SECTOR(rblst.rb[ridx].sc, rbs_broken)) {
	if (brb->cnt) {
	  REALLOC(int, brb->cnt+1, brb->nr);
	}
	else {
	  CALLOC(int, brb->cnt+1, brb->nr);
	}

	brb->nr[brb->cnt++] = ridx;                   /* enrolled RB index */
      }
    }
  }

  /* There are booked and unrecognized RB in the hitlist which might be 
     broken fragments as well. This check is postponed and performed 
     later on. It has no influence to the current recognition.             */
}



/* Assigns unrecognized associated RBs to broken NB.

Revision: 06.01.2005 E.Sack

Input : - pointer to RB index list
        - sector of broken NB
        - RB under test index
        - RB list
Output: - index list updated
        - flag updated (0: valid CNB, 1: no valid CNB)
Precondition: - parent CNB lies in parent CNB sector,
              - lag for mask bordered CNB
Usage: - this function performs its task only together with its counterpart
         pattern_fragments_i(). It enrols residual broken fragments into the
         CRB list.
*/
static void
pattern_fragments_ii(cand_lst *hit_lst, rb_lst rblst)
{
  int        br, cd, ridx, idx;
  sector_def rbs_broken;

  for (br = 0; br < hit_lst->cnt; br++) {         /* search for broken NBs */
    if (hit_lst->c[br].mtd != MATCHING_METHOD_2) {
      continue;                                            /* other method */
    }

    if (hit_lst->c[br].c < 0) {
      continue;                                         /* unrecognized RB */
    }

    rbs_broken = hit_lst->c[br].knl;               /* recognized broken NB */

    for (cd = 0; cd < hit_lst->cnt; cd++) { /* search for unrecognized RBs */
      if (cd == br) {
	continue;                                   /* filter source index */
      }

      if (hit_lst->c[cd].n != SUBSTRB_FLAG) {
	continue;                 /* partially recognized NB are not ment. */
      }

      ridx = hit_lst->c[cd].rbn;      /* index of enrolled unrecognized RB */

      if (A_WITHIN_B_SECTOR(rblst.rb[ridx].sc, rbs_broken)) {
	idx = hit_lst->c[br].crb.cnt++;

	if (idx) {
	  REALLOC(int, hit_lst->c[br].crb.cnt, hit_lst->c[br].crb.nr);
	}
	else {
	  CALLOC(int, hit_lst->c[br].crb.cnt, hit_lst->c[br].crb.nr);
	}

	hit_lst->c[br].crb.nr[idx] = ridx;              /* enroll RB index */

	delete_cand_def(hit_lst, cd);     /* erase candidate from hit list */

	if (cd < br) {
	  br = --cd;          /* reset index because of deletion mechanism */
	  break;                  /* leave loop because of disturbed order */
	}

	--cd;                 /* reset index because of deletion mechanism */
      }
    }
  }
}



/* Parent CNB validation: all fragments (seperate parts) must be covered.
   Additionally gathered are the RB which contains the fragments.
   Restrictive used without masking.

Revision: 02.02.2003 E.Sack

Input : - parent class number
        - parent NB number
        - pointer to class list
        - parent sector
        - parent lags in parent sector
        - RB references
        - RB under test index
        - RB list
        - pointer to flag (0: valid CNB, 1: no valid CNB)
        - pixel matching degree
        - mask border
Output: - CRB list
        - flag updated
Precondition: - parent CNB lies in parent CNB sector,
              - lag for mask bordered CNB
*/
static int_lst
cnb_validation_by_pixel(int cdx, int ndx, cls_def *cls, sector_def rbs, 
			int lx, int ly, rb_ref rref, int rut, rb_lst rblst,
			int *fail, int nnbmatch, int mborder)
{
  int     h, ch, nh, xh, yh, grad, i;
  int_lst crb;                                           /* CRB index list */

  assert(cls);

  init_int_lst(&crb);

  for (h = 0; h < cls[cdx].nbl.nb[ndx].nnb.cnt; h++) {/*test all CNB fractions */
    ch = cls[cdx].nbl.nb[ndx].nnb.cd[h].cdx;   /* fraction CNB class number */
    nh = cls[cdx].nbl.nb[ndx].nnb.cd[h].ndx;   /* fraction CNB nb number    */
    xh = cls[cdx].nbl.nb[ndx].nnb.cd[h].x + lx;/*fract. offset plus parent lag */
    yh = cls[cdx].nbl.nb[ndx].nnb.cd[h].y + ly;/* results in fraction position */

    grad = wtd_correlation(cls[ch].nbl.nb[nh].ref.wtd, 
			   cls[ch].nbl.nb[nh].ref.msk, rbs, rref.wpic, 
			   xh, yh).kernel;         /* no mask verification */

    if (grad < nnbmatch) {
      *fail = 1;
      free_int_lst(&crb);
      return crb;                            /* one fail causes total fail */
    }

    /* a (possible) recursion isn't necessary, because it is handled already
       by the vertical listing.                                            */

    xh += rbs.column.start;                        /* absolute coordinates */
    yh += rbs.line.start;

    for (i = 0; i < rblst.cnt; i++) {   /* fraction CNB distributed in RBx */
      if ((i != rut) &&                     /* no mask taking into account */
	  !(rblst.rb[i].sc.line.start  > 
	    yh+cls[ch].nbl.nb[nh].ref.pic.ar.height-1-mborder ||
            rblst.rb[i].sc.column.start > 
	    xh+cls[ch].nbl.nb[nh].ref.pic.ar.width-1-mborder  ||
            rblst.rb[i].sc.line.end   < yh+mborder            ||
            rblst.rb[i].sc.column.end < xh+mborder)) {
	if (crb.cnt) {
	  REALLOC(int, crb.cnt+1, crb.nr);
	}
	else {
	  CALLOC(int, crb.cnt+1, crb.nr);
	}

	crb.nr[crb.cnt++] = i;            /* multiple enrolled RB possible */
      }
    }
  }

  *fail = 0;
  return crb;                       /* may be empty if CNB contained in RB */
}



/* Parent CNB validation: all fragments (separate parts) must be present.

Revision: 30.05.2004 E.Sack

Input : - parent CNB class number
        - parent CNB nb number
        - pointer to classes
        - parent CNB sector
        - parent CNB lag in parent CNB sector
        - image
Output: - flag updated (0: valid CNB, 1: no valid CNB)
Precondition: - parent CNB lies in parent CNB sector,
              - lag for mask bordered CNB
*/
static int
cnb_validation_by_contour(int c, int n, cls_def *cls, sector_def rbs, 
			  int lx, int ly, pic_def pic)
{
  int         d, h, ch, nh, xh, yh, fail = 0;
  c_def       start;
  sector_def  csec;                                /* CNB sector with mask */
  sector_def  nsec;                                /* NNB sector with mask */
  h_contour_lst outer;

  init_h_contour_lst(&outer);

  csec.column.start = rbs.column.start + lx;
  csec.column.end = csec.column.start + cls[c].nbl.nb[n].ref.pic.ar.width - 1;
  csec.line.start = rbs.line.start + ly;
  csec.line.end = csec.line.start + cls[c].nbl.nb[n].ref.pic.ar.height - 1;

  TRIM_SECTOR(csec, pic.ar);

  for (h = 0; h < cls[c].nbl.nb[n].nnb.cnt; h++) {/*test all CNB fractions */
    ch = cls[c].nbl.nb[n].nnb.cd[h].cdx;      /* fragment CNB class number */
    nh = cls[c].nbl.nb[n].nnb.cd[h].ndx;      /* fragment CNB nb number    */
    xh = cls[c].nbl.nb[n].nnb.cd[h].x + lx + rbs.column.start; /* absolute */
    yh = cls[c].nbl.nb[n].nnb.cd[h].y + ly + rbs.line.start;/* coordinates */

    nsec.column.start = xh;
    nsec.column.end = xh + cls[ch].nbl.nb[nh].ref.pic.ar.width - 1;
    nsec.line.start = yh;
    nsec.line.end = yh + cls[ch].nbl.nb[nh].ref.pic.ar.height - 1;

    TRIM_SECTOR(nsec, pic.ar);

    start = scan_bmp_masked(cls[ch].nbl.nb[nh].ref.msk, nsec, 
			    pic, 0, 0);                /* pixel within NNB */
    if (start.x < 0) {
      fail = 1;                  /* no foreground pixel found in NB sector */
      break;
    }

    outer = xgen_h_contour_from_start(pic, csec, start, 0);
    assert(outer.cnt);

    if (outer.type == CONTOUR_OUTER_INC) {
      d = flood_fill_all(outer, XPLAB_RESERVED_PIXEL, pic);
      assert(d);
    }
    else {
      fail = 1;
    }

    free_h_contour_lst(&outer);
  }

  set_pixel(csec, XPLAB_RESERVED_PIXEL, FOREGROUND, pic, 0);

  return fail;
}



/* Fills a contour with a mark.

Revision: 02.02.2003 E.Sack

Input : - single contour
        - mark
        - image
Output: - number of filled pixels
*/
static int
flood_fill_all(h_contour_lst horizontal, PICT to, pic_def pic)
{
  int           filled = 0, fill, width = pic.ar.width;
  PICT          from;
  v_contour_lst vertical;

  if (!horizontal.hcl) {
    assert(!horizontal.cnt);
    return filled;
  }
  else {
    assert(horizontal.cnt>0);
  }

  from = pic.mem[horizontal.hcl[0].xil.start+horizontal.hcl[0].yco*width];
  assert(from!=to);                               /* distinction important */

  vertical = convert_h_contour_lst(horizontal);
  assert(vertical.cnt);

  do {
    fill = 0;
    fill += flood_fill_horizontal(horizontal, from, to, pic);
    fill += flood_fill_vertical(vertical, from, to, pic);
    filled += fill;
  } while (fill);

  free_v_contour_lst(&vertical);

  return filled;
}



/* Fills a horizontal outer contour with a mark starting on a contour border
   or on a mark within the contour up to the corresponding contour border
   or an unmarked pixel.

Revision: 30.05.2004 E.Sack

Usage: Use together with vertical version of the contour to avoid blank areas.

Input : - contour list
        - marking
        - image
Output: - number of filled pixels
*/
static int
flood_fill_horizontal(h_contour_lst horizontal, PICT from, PICT to, 
		      pic_def pic)
{
  int  i, px, py, xstart, x, xend, filled = 0;
  PICT pixel;

  assert(from!=to);

  if (horizontal.type != CONTOUR_OUTER_INC) {
    return filled;                   /* only an outer contour to be filled */
  }

  for (i = 0; i < horizontal.cnt; i++) {
    py = horizontal.hcl[i].yco*pic.ar.width;
    xstart = horizontal.hcl[i].xil.start;
    xend   = horizontal.hcl[i].xil.end;

    for (x = xstart; x <= xend; x++) {                    /* left to right */
      pixel = pic.mem[px = x + py];

      if (pixel == to) {
	continue;                                       /* continuing line */
      }

      if (pixel != from) {
	break;                                          /* filling stopped */
      }

      pic.mem[px] = to;                                         /* filling */
      filled++;
    }

    for (x = xend; x >= xstart; x--) {                    /* right to left */
      pixel = pic.mem[px = x + py];

      if (pixel == to) {
	continue;                                       /* continuing line */
      }

      if (pixel != from) {
	break;                                          /* filling stopped */
      }

      pic.mem[px] = to;                                         /* filling */
      filled++;
    }
  }

  return filled;
}



/* Fills a vertical contour with a mark starting on a contour border or on a
   mark within the contour up to the corresponding contour border or an 
   unmarked pixel.

Usage: Use together with horizontal version of the contour to avoid blank
       areas.

Revision: 30.05.2004 E.Sack

Input : - contour list
        - marking
        - image
Output: - number of filled pixels
*/
static int
flood_fill_vertical(v_contour_lst vertical, PICT from, PICT to, 
		    pic_def pic)
{
  int  i, px, pxx, ystart, y, yend, filled = 0;
  PICT pixel;

  assert(from!=to);

  if (vertical.type != CONTOUR_OUTER_INC) {
    return filled;                   /* only an outer contour to be filled */
  }

  for (i = 0; i < vertical.cnt; i++) {
    px = vertical.vcl[i].xco;
    ystart = vertical.vcl[i].yil.start;
    yend   = vertical.vcl[i].yil.end;

    for (y = ystart; y <= yend; y++) {                    /* top to bottom */
      pixel = pic.mem[pxx = px + y*pic.ar.width];

      if (pixel == to) {
	continue;                                       /* continuing line */
      }

      if (pixel != from) {
	break;                                          /* filling stopped */
      }

      pic.mem[pxx] = to;                                        /* filling */
      filled++;
    }

    for (y = yend; y >= ystart; y--) {                    /* bottom to top */
      pixel = pic.mem[pxx = px + y*pic.ar.width];

      if (pixel == to) {
	continue;                                       /* continuing line */
      }

      if (pixel != from) {
	break;                                          /* filling stopped */
      }

      pic.mem[pxx] = to;                                        /* filling */
      filled++;
    }
  }

  return filled;
}



/* Converts a horizontal contour (x1-x2,y) to a vertical contour (x,y1-y2).

Revision: 30.05.2004 E.Sack

Input : - horizontal single contour
Output: - vertical single contour 
*/
static v_contour_lst
convert_h_contour_lst(h_contour_lst horizontal)
{
  int           i, ctr_x, ctr_y, ctr_idx, column, step_col, sum_start;
  int           new_start;
  v_contour_lst vertical;

  init_v_contour_lst(&vertical);

  if (horizontal.hcl) {                               /* check consistency */
    assert(horizontal.cnt>0);
  }
  else {
    assert(!horizontal.cnt);
    return vertical;                                 /* nothing to convert */
  }

  vertical.type = horizontal.type;

  ctr_y = horizontal.hcl[0].yco;                /* contour may be unsorted */
  ctr_x = horizontal.hcl[0].xil.start;
  for (i = 1; i < horizontal.cnt; i++) {
    if (ctr_y > horizontal.hcl[i].yco) {         /* topmost starting point */
      ctr_y = horizontal.hcl[i].yco;
      ctr_x = horizontal.hcl[i].xil.start;
      continue;
    }
  }

  ctr_idx = vertical.cnt++;
  CALLOC(v_contour, vertical.cnt, vertical.vcl);
  init_v_contour(&vertical.vcl[ctr_idx]);

  vertical.vcl[ctr_idx].xco = ctr_x;                         /* conversion */
  vertical.vcl[ctr_idx].yil.start = ctr_y;    /*start of vertical interval */

  column = ctr_x;                                /* start of vertical scan */
  step_col = 1;                                           /* scan to right */

  do {                                                  /* horizontal loop */
    sum_start = 0;

    do {                                                  /* vertical loop */
      for (i = 0; i < horizontal.cnt; i++) {/*search end of vert. interval */
	if (horizontal.hcl[i].yco == ctr_y+1) {         /* next lower line */
	  if (horizontal.hcl[i].xil.start <= ctr_x && 
	      horizontal.hcl[i].xil.end >= ctr_x) {
	    ctr_y++;                                  /* continuous column */
	  }
	  else {
	    vertical.vcl[ctr_idx].yil.end = ctr_y++;      /* end of column */
	    break;
	  }
	}
      }

      new_start = 0;

      for (i = 0; i < horizontal.cnt; i++) { /* search start of vert. int. */
	if (horizontal.hcl[i].yco == ctr_y+1) {         /* next lower line */
	  if (horizontal.hcl[i].xil.start <= ctr_x && 
	      horizontal.hcl[i].xil.end >= ctr_x) {
	    ctr_idx = vertical.cnt++;
	    REALLOC(v_contour, vertical.cnt, vertical.vcl);
	    init_v_contour(&vertical.vcl[ctr_idx]);

	    vertical.vcl[ctr_idx].xco = ctr_x;               /* conversion */
	    vertical.vcl[ctr_idx].yil.start = ++ctr_y;     /* column start */
	    new_start = 1;
	    sum_start = 1;
	    break;
	  }
	}
      }
    } while (new_start);                           /* as long as new start */

    if (sum_start) {             /* scan next column as long as new starts */
      ctr_x += step_col;
    }
    else if (step_col > 0) {                      /* scan to right stopped */
      ctr_x = column - 1;
      step_col = -1;                                       /* scan to left */
    }
    else if (step_col < 0) {                       /* scan to left stopped */
      break;
    }
  } while (1);

  return vertical;
}



/* Calculates a CRB sector. Minimal sector is input sector.

Revision: 04.01.2005 E.Sack

Input : - RBS containing NNB
        - NNB size
        - horizontal and vertical offset
        - CNB size
        - RB sector
        - image size
Output: - CRB sector
Postcondition: CRB unchanged if CNB lies inner RBS.
*/
static sector_def
check_cnb(sector_def rbs_in, area_def nnb_ar, int x, int y, area_def cnb_ar,
	  sector_def rb_sec, area_def par) 
{
  int        f1x, f1y, f2x, f2y;
  sector_def rbs_out;

  f1x = rbs_in.column.start - x;
  f1y = rbs_in.line.start - y;
  f2x = rbs_in.column.end + cnb_ar.width - x - nnb_ar.width;
  f2y = rbs_in.line.end + cnb_ar.height - y - nnb_ar.height;

  rbs_out.column.start = XPMAX(0, f1x);                  /* limit to image */
  rbs_out.line.start = XPMAX(0, f1y);
  rbs_out.column.end = XPMIN(par.width-1, f2x);
  rbs_out.line.end = XPMIN(par.height-1, f2y);

  if (A_WITHIN_B_SECTOR(rbs_out, rb_sec)) {
    /* CNB lies inner RB: frame doesn't need to be enlarged, because CNB   */
    /* is apprehended in the normal way.                                   */
    return rbs_in;
  }

  return rbs_out;                                              /* CRB frame */
}



/* Enlarges a RB sector to a NB.

Revision: 04.01.2005 E.Sack

Input : - RBS
        - NB size
        - image size
        - mask border
Output: - RB sector for broken NB.
Postcondition: - RB output is minimal the RB input.
*/
static sector_def
check_broken(sector_def rbs_in, area_def nnb_ar, area_def par, int mborder) 
{
  int        coo;
  sector_def rbs_broken;

  coo = rbs_in.column.start + nnb_ar.width - 1 - mborder;/*horizontal right */
  rbs_broken.column.end = XPMAX(coo, rbs_in.column.end);

  coo = rbs_in.column.end - nnb_ar.width + 1 + mborder; /* horizontal left */
  rbs_broken.column.start = XPMIN(coo, rbs_in.column.start);

  coo = rbs_in.line.start + nnb_ar.height - 1 - mborder;/* vertical bottom */
  rbs_broken.line.end = XPMAX(coo, rbs_in.line.end);

  coo = rbs_in.line.end - nnb_ar.height + 1 + mborder;     /* vertical top */
  rbs_broken.line.start = XPMIN(coo, rbs_in.line.start);

  rbs_broken.column.start = XPMAX(0, rbs_broken.column.start);/*limit to image */
  rbs_broken.line.start = XPMAX(0, rbs_broken.line.start);
  rbs_broken.column.end = XPMIN(par.width-1, rbs_broken.column.end);
  rbs_broken.line.end = XPMIN(par.height-1, rbs_broken.line.end);

  return rbs_broken;                                           /* NB frame */
}



/* The candidate list will be filtered according some heuristics.

Revision: 16.03.2003 E.Sack

Input : - candidate list
        - pointer to class list
        - pointer to vertical candidate list
        - jitter
        - image
        - flag (0: multiline RB, 1: singleline RB)
Output: - candidates list updated
*/
static void 
sort_out(int jitter, double corrlevel, cand_lst hlist, cls_def *cls, 
	 cand_lst *vcl, pic_def pic, int single)
{
  sort_out_inhibited_classes(cls, vcl);           /* sort out CNB prior NB */

  sort_out_scions(cls, vcl, jitter);             /* sort out CNB prior NNB */

  sort_out_doublets(cls, vcl, jitter);                /* sort out doublets */

  if (single) {                             /* qualified RB as single line */
    inhibit_classes_by_nb_alignment(cls, vcl, jitter);  /* NB misalignment */
    inhibit_classes_by_rb_alignment(cls, vcl, jitter);  /* RB misalignment */
  }

  inhibit_classes_by_cnb(corrlevel, hlist, cls, vcl, jitter, pic);    /* CNB prior NB */
}



/* Deletes doublets.

Revision: 25.11.2002 E.Sack

Input : - pointer to class list
        - pointer to vertical candidate list
Output: - candidates list updated
*/
static void
sort_out_doublets(cls_def *cls, cand_lst *vcl, int jitter)
{
  int    d, k, i, class, size;
  double max;

  for (i = 0; i < vcl->cnt-1; i++) {                      /* erase doubles */
    max = vcl->c[i].m.cv;
    class = vcl->c[i].c;
    size = AREA_SIZE(cls[class].nbl.nb[vcl->c[i].n].pat.ar);

    for (k = i+1; k < vcl->cnt; k++) {
      if (vcl->c[k].c == class) {
	d = interval_overlap(vcl->c[k].knl.line, vcl->c[i].knl.line, 
			     jitter);
	if (d == OVL_NOT) {
	  continue;                                   /* NB not overlapping */
	}

	d = interval_overlap(vcl->c[k].knl.column, vcl->c[i].knl.column, 
			     jitter);
	if (d == OVL_NOT) {
	  continue;                                   /* NB not overlapping */
	}

	if (max < vcl->c[k].m.cv) {           /* 1. selection rule: K value */
	  delete_cand_def(vcl, i);
	  i--;
	  break;
	}
	else if (max > vcl->c[k].m.cv) {
	  delete_cand_def(vcl, k);
	  k--;
	  continue;
	}
	else if (max == vcl->c[k].m.cv) {        /* 2. selection rule: size */
	  if (size < AREA_SIZE(cls[vcl->c[k].c].nbl.nb[vcl->c[k].n].pat.ar)) {
	    delete_cand_def(vcl, i);
	    i--;
	    break;
	  }
	  else {
	    delete_cand_def(vcl, k);
	    k--;
	    continue;
	  }
	}
      }
    }
  }                              /* list contains only different candidates */
}



/* Deletes inhibited classes.

Revision: 25.11.2002 E.Sack

Input : - pointer to class list
        - pointer to vertical candidate list
Output: - candidates list updated
*/
static void
sort_out_inhibited_classes(cls_def *cls, cand_lst *vcl)
{
  int d, k, ck;

  for (k = 0; k < vcl->cnt; k++) {               /* delete inhibited class */
    ck = vcl->c[k].c;

    if (cls[ck].inl.cnt) {
      d = check_inh_lst(cls[ck].inl, INHIBITED_TRUE, vcl->c[k].knl);

      if (d) {                                          /* class inhibited */
	delete_cand_def(vcl, k);                        /* class inhibited */
	k--;
      }
    }
  }
}



/* Ensures priority of CNB over its NNB (a NNB can be a CNB, too).

Revision: 25.11.2002 E.Sack

Input : - pointer to class list
        - pointer to vertical candidate list
Output: - candidates list updated
*/
static void
sort_out_scions(cls_def *cls, cand_lst *vcl, int jitter)
{
  int d, k, i, j, anc_c, anc_n, cnb_found;

  for (k = 0; k < vcl->cnt; k++) {              /* looking for CNB anchors */
    anc_c = vcl->c[k].c;
    anc_n = vcl->c[k].n;

    if (!cls[anc_c].nbl.nb[anc_n].cnb.cnt) {
      continue;                                          /* NB is not a NNB */
    }

    for (j = cnb_found = 0;                         /* NB is a NNB (anchor) */
	 j < cls[anc_c].nbl.nb[anc_n].cnb.cnt && !cnb_found;
	 j++) {
      for (i = 0; i < vcl->cnt; i++) {       /* looking for all NNB's CNBs */
	d = interval_overlap(vcl->c[i].knl.line, vcl->c[k].knl.line,jitter);

	if (d == OVL_NOT) {
	  continue;                                  /* NB out of CNB frame */
	}

	d = interval_overlap(vcl->c[i].knl.column, vcl->c[k].knl.column, 
			     jitter);

	if (d == OVL_NOT) {
	  continue;                                  /* NB out of CNB frame */
	}

	if (vcl->c[i].c == cls[anc_c].nbl.nb[anc_n].cnb.cd[j].cdx) {
	  /* class of the CNB is sufficient, because all their NBs are CNBs,*/
	  /* it doesn't matter, which NB a CNB of the anchor is.            */
	  delete_cand_def(vcl, k--);              /* NNB's CNB in the list */
	  cnb_found = 1;        /* anchor's CNB found, CNB have precedence */ 
	  break;
	}
      }
    }
  }
}



/* Inhibits classes according CNB over NB priority. The inhibition shall
   allow the recognition of CNB even if their anchor-NNB is inhibited. The 
   region of the rule is strictly limited, that means, there are no further
   conditions to be taken into account. The inhibition is for filtering VCLs
   only, not for recognition. CNBs are treated as an extension of a NB.

Revision: 16.03.2003 E.Sack

Input : - candidate list
        - pointer to class list
        - pointer to vertical candidate list
        - jitter
Output: - class list updated 
Postcondition: - a printable NB can be a NNB (anchor), in which case the 
                 inhibition must be ignored.
*/
static void
inhibit_classes_by_cnb(double corrlevel, cand_lst hlist, cls_def *cls, cand_lst *vcl, 
		       int jitter, pic_def pic)
{
  int d1, d2, i, k, ck, ci, nk, ni, cnbk, cnbi, del_idx, inh_cnr;
  int cnb_idx, fract, ovl, inh_nnr, cnb_nnr, cnb_cnr, lx, ly;

  assert(cls);
  assert(vcl);
  assert(jitter>=0);

  for (k = 0; k < vcl->cnt-1; k++) {        /* ensure priority CNB over NB */
    ck = vcl->c[k].c;                          /* candidate's class number */
    nk = vcl->c[k].n;                             /* candidate's NB number */
    cnbk = cls[ck].nbl.nb[nk].nnb.cnt;                   /* CNB indication */

    for (i = k+1; i < vcl->cnt; i++) {
      ci = vcl->c[i].c;                        /* candidate's class number */

      if (ck == ci) {
	continue;                         /* a class cannot inhibit itself */
      }

      ni = vcl->c[i].n;                           /* candidate's NB number */
      cnbi = cls[ci].nbl.nb[ni].nnb.cnt;                 /* CNB indication */

      if ((cnbk && cnbi) ||                                 /* CNB <=> CNB */
	  (!cnbk && !cnbi)) {                                /* NB <=> NB  */
	continue;                             /* no CNB <=> NB competition */
      }

      d1 = interval_overlap(vcl->c[i].knl.line, vcl->c[k].knl.line, jitter);

      if (d1 == OVL_NOT) {                                /* multi-line RB */
	continue;                                   /* NB out of CNB frame */
      }

      d2 = interval_overlap(vcl->c[i].knl.column, vcl->c[k].knl.column,
			    jitter);

      if (d2 == OVL_NOT) {                               /* single-line RB */
	continue;                                   /* NB out of CNB frame */
      }

      if (cnbk) {
	if ((d1 == OVL_B_IN_A) && (d2 == OVL_B_IN_A)) {
	  continue;                                    /* NB comprises CNB */
	}

	cnb_idx = k;
	del_idx = i;
      }
      else {
	if ((d1 == OVL_A_IN_B) && (d2 == OVL_A_IN_B)) {
	  continue;                                    /* NB comprises CNB */
	}

	cnb_idx = i;
	del_idx = k;
      }

      /* 8-Neighborhood using only                                         */
      fract = xcount_contours(pic, vcl->c[cnb_idx].knl, FOREGROUND, 0);

      if (fract < 2) {         /* fraction analysis should confirm roughly */
	continue;
      }

      inh_cnr = vcl->c[del_idx].c;           /* NB class number to inhibit */
      inh_nnr = vcl->c[del_idx].n;                 /* NB number to inhibit */

      cnb_cnr = vcl->c[cnb_idx].c;            /* CNB class number to check */
      cnb_nnr = vcl->c[cnb_idx].n;                  /* CNB number to check */

      lx = vcl->c[cnb_idx].hdx.column.start-vcl->c[del_idx].hdx.column.start;
      ly = vcl->c[cnb_idx].hdx.line.start - vcl->c[del_idx].hdx.line.start;

      ovl = overlapping_degree(cls[inh_cnr].nbl.nb[inh_nnr].pat,
			       cls[cnb_cnr].nbl.nb[cnb_nnr].pat, lx, ly);
      if ((double)ovl <                 /* no mask effects as no mask used */
	  (double)cls[inh_cnr].nbl.nb[inh_nnr].ref.nbm*corrlevel) {
	continue;       /* CNB too different to NB, cannot be an extension */
      }

      d1 = overlap_list(hlist, cls, vcl->c[cnb_idx], jitter);

      if (d1 > 0) {                           /* verify false overlappings */
	del_idx = cnb_idx;
	inh_cnr = vcl->c[del_idx].c;        /* CNB class number to inhibit */
      }

      update_inh_lst(&cls[inh_cnr].inl, INHIBITED_BY_CNB, 
		     vcl->c[del_idx].knl);          /* location to inhibit */

      delete_cand_def(vcl, del_idx);                        /* sort out NB */

      if (del_idx == k) {
	k--;
	break;
      }
      else {
	assert(del_idx == i);
	i--;
      }
    }
  }
}



/* Seeks maximal overlapping degree in a list.

Revision: 16.03.2003 E.Sack

Input : - candidate list
        - pointer to class list
        - candidate
Output: - number of fractions
*/
static int 
overlap_list(cand_lst hlist, cls_def *cls, cand_def cand, int jitter)
{
  int d, i, max_olap = 0, olap, hit_cnr, hit_nnr, cnd_cnr, cnd_nnr, lx, ly;

  for (i = 0; i < hlist.cnt; i++) {
    hit_cnr = hlist.c[i].c;                  /* hit-candidate class number */

    if (hit_cnr < 0) {
      continue;                                    /* unclassified pattern */
    }

    d = interval_overlap(hlist.c[i].knl.line, cand.knl.line, jitter);

    if (d == OVL_NOT) {
      continue;                   /* hit-candidate out of candidates frame */
    }

    d = interval_overlap(hlist.c[i].knl.column, cand.knl.column, jitter);

    if (d == OVL_NOT) {
      continue;                   /* hit-candidate out of candidates frame */
    }

    hit_nnr = hlist.c[i].n;                     /* hit-candidate NB number */

    cnd_cnr = cand.c;                            /* candidate class number */
    cnd_nnr = cand.n;                                  /* candidate number */

    lx = cand.hdx.column.start-hlist.c[i].hdx.column.start;
    ly = cand.hdx.line.start - hlist.c[i].hdx.line.start;

    olap = overlapping_degree(cls[hit_cnr].nbl.nb[hit_nnr].pat,
			      cls[cnd_cnr].nbl.nb[cnd_nnr].pat, lx, ly);

    max_olap = XPMAX(max_olap, olap);
  }

  return max_olap;
}



/* Counts pattern fractions in a frame.

Revision: 13.09.2004 E.Sack

Input : - image
        - sector
        - pixel label
        - neighborhood
Output: - number of fractions
Precondition: - only one bit set in key
Bug: - no assertion used for precondition test because of runtime.
*/
int
xcount_contours(pic_def rpic, sector_def sec, PIXEL_KEY key, int n4)
{
  int       d, fract = 0;
  PIXEL_KEY test_key, bit_or;

  bit_or = xget_pixel_or(sec, rpic) | key;                 /* all set bits */

  if (!(bit_or & XPLAB_SELECTBL_PIXEL)) {    /* test key must be different */
    test_key = XPLAB_SELECTBL_PIXEL;
  }
  else if (!(bit_or & XPLAB_TRAINED_PIXEL)) {
    test_key = XPLAB_TRAINED_PIXEL;
  }
  else if (!(bit_or & XPLAB_IGNORED_PIXEL)) {
    test_key = XPLAB_IGNORED_PIXEL;
  }
  else if (!(bit_or & XPLAB_SELECTED_PIXEL)) {
    test_key = XPLAB_SELECTED_PIXEL;
  }
  else if (!(bit_or & XPLAB_RESERVED_PIXEL)) {
    test_key = XPLAB_RESERVED_PIXEL;
  }
  else if (!(bit_or & FOREGROUND)) {
    test_key = FOREGROUND;
  }
  else {
    assert(0);                      /* too many bits set, algo cannot work */
  }

  do {
    d = xcontour_mark(rpic, sec, (PICT)key, (PICT)test_key, 0, n4);
    fract++;
  } while (d);

  /* reset keys just used for counting */
  set_pixel(sec, (PICT)test_key, (PICT)key, rpic, 0);

  return fract-1;
}



/* Inhibits classes according alignment. The erasing follows a
   plausibility test, e.g. a dot near the top of the line cannot 
   be a punctuation mark, a horizontal stroke on the bottom of a line
   cannot be a hyphen. Raw rule used by dividing the line in two equal 
   halfes. Inhibition is made only for small classes, which have problems
   for alignment. 

Revision: 15.02.2003 E.Sack

Usage: Should be used only if RB is classified as a single line.

Input : - pointer to class list
        - pointer to vertical candidate list
Output: - class list updated
*/
static void
inhibit_classes_by_rb_alignment(cls_def *cls, cand_lst *vcl, int jitter)
{
  int        j, k, bot_line, small, bot_tall, bot_small, half_line;
  int        height_j, height_k, tall_height, tall, cnr_j, cnr_k, top_line;
  int        half_height, tall_max, height, placeable = 0, half_tall;
  int        misalign, cnr_small;
  sector_def location;
  ALIGMNT    aligmt_j, aligmt_k, aligmt_small, aligmt_tall;

  if (!vcl->cnt) {
    return;                                                     /* no data */
  }

  top_line = vcl->c[0].line.start;                                 /* init */
  bot_line = vcl->c[0].line.end;
  tall_max = INTERVAL_SIZE(vcl->c[0].knl.line);
  for (k = 1; k < vcl->cnt; k++) {
    top_line = XPMIN(top_line, vcl->c[k].line.start);/*overall line borders */
    bot_line = XPMAX(bot_line, vcl->c[k].line.end);
    tall_max = XPMAX(tall_max, INTERVAL_SIZE(vcl->c[k].knl.line));
  }

  height = bot_line - top_line + 1;
  half_height = height >> 1;

  if (tall_max < half_height) {
    return;                                                   /* no big NB */
  }

  for (k = 0; k < vcl->cnt; k++) {
    if (height >= UPPERLINE*INTERVAL_SIZE(vcl->c[k].knl.line)) {
      placeable = 1;                      /* small NB in a big environment */
      break;
    }
  }

  if (!placeable) {
    return;            /* small NB not placable, tall NBs never misaligned */
  }

  half_line = (bot_line + top_line) >> 1;

  for (k = 0; k < vcl->cnt; k++) {
    cnr_k = vcl->c[k].c;
    aligmt_k = cls[cnr_k].att.alignment;
    height_k = INTERVAL_SIZE(vcl->c[k].knl.line);

    misalign = 0;
    if (height >= UPPERLINE*height_k) {
      switch (aligmt_k) {
      case UPPERLINE: case MIDDLELINE:
	if (vcl->c[k].knl.line.end > half_line + jitter) {
	  misalign = 1;
	}
	break;
      case BASELINE: case UNDERLINE:
	if (vcl->c[k].knl.line.end < half_line - jitter) {
	  misalign = 1;
	}
	break;
      default: break;
      }
    }

    if (misalign) {
      location = vcl->c[k].knl;                     /* location to inhibit */
      update_inh_lst(&cls[cnr_k].inl, INHIBITED_BY_MISALIGNMENT, location);
      delete_cand_def(vcl, k);                    /* sort out misalignment */
      k--;
      continue;
    }

    for (j = k+1; j < vcl->cnt; j++) {     /* comparing against each other */
      cnr_j = vcl->c[j].c;
      aligmt_j = cls[cnr_j].att.alignment;
      height_j = INTERVAL_SIZE(vcl->c[j].knl.line);

      if (height_j < height_k>>1) {
	small = j;
	aligmt_small = aligmt_j;
	tall = k;
	tall_height = height_k;
	aligmt_tall = aligmt_k;
      }
      else if (height_k < height_j>>1) {
	small = k;
	aligmt_small = aligmt_k;
	tall = j;
	tall_height = height_j;
	aligmt_tall = aligmt_j;
      }
      else {
	continue;                                  /* no usable references */
      }

      if (tall_height < half_height) {
	continue;                              /* reference not big enaugh */
      }

      bot_small = vcl->c[small].knl.line.end;
      bot_tall = vcl->c[tall].knl.line.end;
      half_tall = (vcl->c[tall].knl.line.start + bot_tall) >> 1;

      if (aligmt_small == aligmt_tall &&                 /* same alignment */
	  (bot_small > bot_tall+jitter ||          /* wrong place, too low */
	   bot_small < half_tall-jitter)) {       /* wrong place, too high */
      }
      else if (aligmt_small < aligmt_tall &&            /* lower alignment */
	       bot_small < bot_tall-jitter) {     /* wrong place, too high */
      }
      else if (aligmt_small > aligmt_tall &&           /* higher alignment */
	       bot_small > bot_tall+jitter) {      /* wrong place, too low */
      }
      else {
	continue;                                       /* no misalignment */
      }

      cnr_small = vcl->c[small].c;                     /* class to inhibit */
      location = vcl->c[small].knl;                 /* location to inhibit */

      update_inh_lst(&cls[cnr_small].inl, INHIBITED_BY_MISALIGNMENT, 
		     location);
      delete_cand_def(vcl, small);                /* sort out misalignment */

      if (small == j) {
	j--;
      }
      else {
	assert(small==k);
	k--;
	break;
      }
    }
  }
}



/* Inhibits classes according NBs mutual location. The inhibition follows a
   plausibility test, e.g. a dot near the top of a low case letter cannot 
   be a punctuation mark, a horizontal stroke on the bottom of a character
   cannot be a hyphen.

Revision: 29.11.2002 E.Sack

Input : - pointer to class list
        - pointer to vertical candidate list
        - jitter
Output: - class list updated
*/
static void
inhibit_classes_by_nb_alignment(cls_def *cls, cand_lst *vcl, int jitter)
{
  int        y, i, k, ck, ci, ya, yb, yk, yi, cand_a, cand_b, cnr_a, cnr_b;
  ival_def   line_a, line_b;
  sector_def location;

  assert(cls);
  assert(vcl);
  assert(jitter>=0);

  for (k = 0; k < vcl->cnt-1; k++) {                /* ensure plausibility */
    ck = vcl->c[k].c;
    yk = INTERVAL_SIZE(vcl->c[k].knl.line);         /* height of candidate */

    for (i = k+1; i < vcl->cnt; i++) {
      if (i == k) {
	continue;
      }

      ci = vcl->c[i].c;

      if (ck == ci) {
	continue;                         /* a class cannot inhibit itself */
      }

      yi = INTERVAL_SIZE(vcl->c[i].knl.line);       /* height of candidate */

      if (yk >= yi) {                     /* ensure candidate size: A >= B */
	cand_a = k;
	cand_b = i;
	ya = yk;
	yb = yi;
      }
      else {
	cand_a = i;
	cand_b = k;
	ya = yi;
	yb = yk;
      }

      cnr_a = vcl->c[cand_a].c;
      cnr_b = vcl->c[cand_b].c;                 /* class number to inhibit */
      location = vcl->c[cand_b].knl;                /* location to inhibit */
      line_a = vcl->c[cand_a].knl.line;     /* searching for candidates in */
      line_b = vcl->c[cand_b].knl.line;     /* same line and investigating */
      y = XPMIN(XPMAX(ya>>1, jitter), yb);  /* alignment against location  */

      if (cls[cnr_a].att.alignment == cls[cnr_b].att.alignment) {
	if ((line_a.end-y > line_b.end) &&      /* candidates in same line */
            (line_a.start <= line_b.end)) {    /* and mismatched alignment */
	  update_inh_lst(&cls[cnr_b].inl, INHIBITED_BY_NB_ALIGNMENT, 
			 location);
	  delete_cand_def(vcl, cand_b);                     /* sort out NB */
	  k = -1;
	  break;
	}
      }
    }
  }
}



/* Updates a inhibition list.

Revision: 29.11.2002 E.Sack

Input : - pointer to inhibition list
        - inhibition flag
        - inhibition location
Output: - list updated
*/
static void
update_inh_lst(inh_lst *il, INH_FLAG iflag, sector_def loc)
{
  int d, size;

  d = check_inh_lst(*il, iflag, loc);

  if (d) {
    return;                                             /* class inhibited */
  }

  size = il->cnt++;

  if (il->ind) {
    REALLOC(inh_def, il->cnt, il->ind);
  }
  else {
    CALLOC(inh_def, il->cnt, il->ind);
  }

  init_inh_def(&il->ind[size]);

  il->ind[size].inh = iflag;
  il->ind[size].isd = loc;
}



/* Sums all inhibitions of a class list.

Revision: 25.11.2002 E.Sack

Input : - class list
Output: - inhibition sum
*/
static int
inhibition_sum(cls_lst clst)
{
  int i, sum = 0;

  for (i = 0; i < clst.cnt; i++) {
    sum += clst.cls[i].inl.cnt;
  }

  return sum;
}



/* Resets all inhibitions of a class list.

Revision: 25.11.2002 E.Sack

Input : - class list
Output: - class list updated
*/
static void
inhibition_reset(cls_lst clst)
{
  int i;

  for (i = 0; i < clst.cnt; i++) {
    free_inh_lst(&clst.cls[i].inl);
  }
}



/* Checks a inhibition list for a certain inhibition.

Revision: 29.11.2002 E.Sack

Input : - inhibition list
        - inhibition flag
        - inhibition sector
Output: - flag (0: inhibition not found, 1: inhibition found)
*/
static int
check_inh_lst(inh_lst ilist, INH_FLAG iflag, sector_def sector)
{
  int i;

  for (i = 0; i < ilist.cnt; i++) {
    if ((iflag == INHIBITED_TRUE) ||               /* unlimited inhibition */
        (iflag == ilist.ind[i].inh)) {
      if (ilist.ind[i].isd.column.start <= sector.column.start &&
	  ilist.ind[i].isd.column.end   >= sector.column.end &&
	  ilist.ind[i].isd.line.start  <= sector.line.start &&
	  ilist.ind[i].isd.line.end    >= sector.line.end) {
	return 1;
      }
    }
  }

  return 0;
}



/* Calculate overlap of intervals with allowed maximum.
   Proposal: instead of sector overlapping a more exact calculation 
             (e.g. simple correlation) should be made if this function is 
             used for competition decisions.

Input : - intervals
        - overlap maximum
Output: - flag (0: no overlap, 1: overlap)
*/
static int
interval_overlap(ival_def iva, ival_def ivb, int jitter)
{
  assert(jitter >= 0);

  if (iva.end-ivb.start < 0 || ivb.end-iva.start < 0) {
    return OVL_NOT;                              /* intervals not overlaped */
  }

  if (((iva.end-ivb.start < jitter && ivb.end-iva.end > jitter) ||
       (ivb.end-iva.start < jitter && iva.start-ivb.start > jitter)) &&
      ((ivb.end-iva.start < jitter && iva.end-ivb.end > jitter) ||
      (iva.end-ivb.start < jitter && ivb.start-iva.start > jitter))) {
    return OVL_NOT;                        /* allowed overlap of each other */
  }

  if (iva.start >= ivb.start-jitter && iva.end <= ivb.start+jitter) {
    return OVL_A_IN_B;                                         /* preferred */
  }
 
  if (ivb.start >= iva.start-jitter && ivb.end <= iva.start+jitter) {
    return OVL_B_IN_A;
  }
 
  return OVL_PART;                                   /* intervals overlaped */
}



/* Extends an image.

Revision: 03.10.2002 E.Sack

Input : - image
        - vertical and horizontal extension (added to both sides)
Output: - image
*/
pic_def 
xextend_image(pic_def in, int spcg_w, int spcg_h)
{
  int     len;
  pic_def out;

  assert(spcg_w>=0);
  assert(spcg_h>=0);

  init_pic_def(&out);

  out.ar.width  = in.ar.width  + 2*spcg_w;
  out.ar.height = in.ar.height + 2*spcg_h;
  len = AREA_SIZE(out.ar);

  CALLOC(PICT, len, out.mem);

  if (in.bvl) {
    CALLOC(PICT, len, out.bvl);                         /* b-values wanted */
  }

  copy_bmps(in, out, in.ar, 0, 0, spcg_w, spcg_h);       /* centered image */

  out.org.x = in.org.x-spcg_w;                     /* adapt absolut origin */
  out.org.y = in.org.y-spcg_h;

  return out;
}



/* Sets all pixels within a contour regardless holes.

Revision: 01.02.2003 E.Sack

Input : - contour
        - mark to set
        - image
Output: - image updated
        - number of set pixels
*/
static int 
h_contour_fill(h_contour_lst ctr, PICT to, pic_def pic)
{
  int i, z, s, marked = 0, width = pic.ar.width;

  for (i = 0; i < ctr.cnt; i++) {
    z = ctr.hcl[i].yco*width;

    for (s = ctr.hcl[i].xil.start; s <= ctr.hcl[i].xil.end; s++) {
      pic.mem[s+z] = to;                                      /* set pixel */
      marked++;
    }
  }

  return marked;
}



/* Searches and fills all contours in a clipping regardless of holes. 

Revision: 01.02.2003 E.Sack

Input : - image
        - clipping
        - mark to search
        - mark to fill
Output: - image updated
        - number of contours
*/
static int
contour_fill_all(pic_def pic, sector_def fr, PICT search, PICT fill)
{
  int           d, found = 0, aux_needed = search == fill;
  PICT          aux;
  h_contour_lst ctr;

  aux = aux_needed ? ~fill : fill;                       /* auxiliary mark */

  while (1) {
    ctr = xgen_h_contour(pic, fr, search, 0, 0);

    if (!ctr.cnt) {
      break;
    }

    found++;                                              /* contour found */

    d = h_contour_fill(ctr, aux, pic);
    assert(d);
    free_h_contour_lst(&ctr);
  }

  if (found && aux_needed) {
    d = set_pixel(fr, aux, fill, pic, 0);          /* reset auxiliary mark */
    assert(d);
  }

  return found;
}



/* Crops or copies reference pattern without mask. A clipping defines
   the size and location of the reference in the source image.

Revision: 04.01.2004 E.Sack

Input : - source image
        - destination image
        - clipping of source image
        - reference pattern
        - destination offset relative to clipping
        - reference offset relative to clipping
        - flag (0: crop, 1: copy)
Output: - images updated
Precondition: - images may be the same memory (on-place operation allowed)
Usage: - this function is used to crop a NB pattern from an image. However
         the function can be used also if the source is the reference
         simultaneously. Then, all foreground pixels in the clipping are
         erased if the reference offsets are 0 and the flag is 0, too.
*/
static void
crop(pic_def src, pic_def dst, sector_def clip, pic_def ref, int dstx, 
     int dsty, int refx, int refy, int cpy)
{
  register int sx;
  int s, z, zz, zzy, zy, dy, dx, ss;

  assert(clip.column.start>=0);
  assert(clip.line.start>=0);
  assert(clip.column.end<src.ar.width);
  assert(clip.line.end<src.ar.height);

  assert(clip.column.start+dstx>=0);
  assert(clip.line.start+dsty>=0);
  assert(clip.column.end+dstx<dst.ar.width);
  assert(clip.line.end+dsty<dst.ar.height);

  assert(refx>=0);
  assert(refy>=0);
  assert(refx+clip.column.end-clip.column.start<=ref.ar.width);
  assert(refy+clip.line.end-clip.line.start<=ref.ar.height);

  assert(src.mem);
  assert(dst.mem);
  assert(ref.mem);

  for (z = clip.line.start, zz = refy; z <= clip.line.end; z++, zz++) {
    zzy = zz*ref.ar.width;
    zy = z*src.ar.width;
    dy = (z+dsty)*dst.ar.width;

    for (s = clip.column.start, ss = refx; s <= clip.column.end; s++, ss++) {
      sx = s + zy;
      dx = s + dy + dstx;

      if (src.mem[sx] != BACKGROUND && ref.mem[ss + zzy] != BACKGROUND) {
         dst.mem[dx] = src.mem[sx];
         src.mem[sx] = cpy ? src.mem[sx] : BACKGROUND;
      }
    }
  }
}



/*  Stores back all internal pattern and empties the internal stores.

Revision: 01.07.2004 E.Sack

Usage : - This function is called once to restore all the patterns to
          their original sources.
        - There is another function 'crop_cand_lst'  used for storing
          internal patterns.

Input : - image
        - weighted image
        - pointer to class list
        - pointer to support for internal store
Output: - image sources updated
        - stores updated
*/
static void
uncrop_cand_lst(pic_def extern_rpic, pic_def extern_wpic, 
		cls_def *extern_cls, support_def *intern)
{
  int        i, intern_hdx_column_start, intern_hdx_line_start;
  sector_def intern_knl;

  assert(extern_rpic.mem);
  assert(extern_wpic.mem);
  assert(extern_rpic.ar.width == extern_wpic.ar.width);
  assert(extern_rpic.ar.height == extern_wpic.ar.height);
  assert(extern_cls);

  /* restore candidates according the internal list */
  if (!intern->rpic) {                                /* consistency check */
    assert(!intern->wpic);
    assert(!intern->cdl);
    assert(!intern->sec);
  }
  else {
    assert(intern->wpic);
    assert(intern->cdl);
    assert(intern->sec);

    for (i = 0; i < intern->cdl->cnt; i++) {      /* restore internal list */
      intern_knl = intern->cdl->c[i].knl;
      intern_knl.column.start -= intern->sec->column.start;
      intern_knl.column.end   -= intern->sec->column.start;
      intern_knl.line.start   -= intern->sec->line.start;
      intern_knl.line.end     -= intern->sec->line.start;

      intern_hdx_column_start = intern->cdl->c[i].hdx.column.start - 
	intern->sec->column.start;
      intern_hdx_line_start  = intern->cdl->c[i].hdx.line.start - 
	intern->sec->line.start;

      crop(*intern->rpic, extern_rpic, intern_knl,
	   extern_cls[intern->cdl->c[i].c].nbl.nb[intern->cdl->c[i].n].pat, 
	   intern->sec->column.start, intern->sec->line.start, 
	   intern_knl.column.start - intern_hdx_column_start,
	   intern_knl.line.start - intern_hdx_line_start,
	   0);
      crop(*intern->wpic, extern_wpic, intern_knl,
	   extern_cls[intern->cdl->c[i].c].nbl.nb[intern->cdl->c[i].n].pat, 
	   intern->sec->column.start, intern->sec->line.start, 
	   intern_knl.column.start - intern_hdx_column_start,
	   intern_knl.line.start - intern_hdx_line_start,
	   0);
    }
  }

  free_support_def(intern); /* no more stores to use, erase all references */
}



/*  Cuts out candidates from an image and stores the pattern internally.
    The candidates are stored internally as well to allow an incremental 
    candidate list as input, which contains candidates already cut. The
    internal stores are allocated dynamically to avoid unnecessary memory
    consumption.

Revision: 05.08.2005 E.Sack

Usage : - This function is called several times according to a candidates 
          lists update.
        - There is another function 'uncrop_cand_lst'  used to free the
          internal stores.

Input : - candidate list
        - image
        - weighted image
        - pointer to class list
        - unique flag 
        - pointer to support for internal store
Output: - image sources updated
        - stores updated
        - key of cut out candidates
*/
static char *
crop_cand_lst(cand_lst extern_cdl, pic_def extern_rpic, pic_def extern_wpic, 
	      cls_def *extern_cls, int uniq, support_def *intern)
{
  int        i, j, extern_cnr, extern_nnr, enrolled;
  int        offset_x, offset_y, end_x, end_y;
  sector_def extern_hdx, max_sec, extern_sec;
  char       *key = 0, buf[MAXCHAR];

  assert(extern_rpic.mem);
  assert(extern_wpic.mem);
  assert(extern_rpic.ar.width == extern_wpic.ar.width);
  assert(extern_rpic.ar.height == extern_wpic.ar.height);
  assert(extern_cls);

  /* crop and store candidates according the internal list */
  if (!extern_cdl.c) {
    assert(!extern_cdl.cnt);                      /* empty candidates list */
    return key;
  }
  else {
    assert(extern_cdl.cnt>0);
  }

  if (!intern->cdl) {
    CALLOC(cand_lst, 1, intern->cdl);
    init_cand_lst(intern->cdl);
  }

  extern_sec = max_frame(extern_cdl);

  if (intern->sec) {
    max_sec = max_sector(*intern->sec, extern_sec);
  }
  else {
    CALLOC(sector_def, 1, intern->sec);
    *intern->sec = max_sec = extern_sec;
  }

  offset_x = intern->sec->column.start - max_sec.column.start;
  offset_y = intern->sec->line.start - max_sec.line.start;
  end_x = intern->sec->column.end - max_sec.column.end;
  end_y = intern->sec->line.end - max_sec.line.end;

  if (offset_x || offset_y || end_x || end_y || !intern->rpic) {
    *intern->sec = max_sec;

    allocate_image_store(&intern->rpic, max_sec, offset_x, offset_y);
    allocate_image_store(&intern->wpic, max_sec, offset_x, offset_y);
  }

  for (i = 0; i < extern_cdl.cnt; i++) {           /* update internal list */
    extern_hdx = extern_cdl.c[i].hdx;
    extern_cnr = extern_cdl.c[i].c;
    extern_nnr = extern_cdl.c[i].n;

    for (j = enrolled = 0; j < intern->cdl->cnt; j++) {
      if (extern_hdx.line.start  == intern->cdl->c[j].hdx.line.start  && 
	  extern_hdx.line.end    == intern->cdl->c[j].hdx.line.end    && 
	  extern_hdx.column.start == intern->cdl->c[j].hdx.column.start &&
	  extern_hdx.column.end   == intern->cdl->c[j].hdx.column.end   &&
	  extern_cnr == intern->cdl->c[j].c &&
	  extern_nnr == intern->cdl->c[j].n) {
	enrolled = 1;
	break;                                /* candidate already in list */
      }
    }

    if (!enrolled) {                              /* candidate not in list */
      cat_cand_def(extern_cdl.c[i], intern->cdl);            /* enroll her */
      crop(extern_rpic, *intern->rpic, extern_cdl.c[i].knl,    /* crop her */
	   extern_cls[extern_cdl.c[i].c].nbl.nb[extern_cdl.c[i].n].pat,
	   -max_sec.column.start, -max_sec.line.start, 
	   extern_cdl.c[i].knl.column.start - 
	   extern_cdl.c[i].hdx.column.start,
	   extern_cdl.c[i].knl.line.start - 
	   extern_cdl.c[i].hdx.line.start,
	   0);
      crop(extern_wpic, *intern->wpic, extern_cdl.c[i].knl,    /* crop him */
	   extern_cls[extern_cdl.c[i].c].nbl.nb[extern_cdl.c[i].n].pat,
	   -max_sec.column.start, -max_sec.line.start, 
	   extern_cdl.c[i].knl.column.start - 
	   extern_cdl.c[i].hdx.column.start,
	   extern_cdl.c[i].knl.line.start - 
	   extern_cdl.c[i].hdx.line.start,
	   0);
    }
  }

    g_snprintf(buf, sizeof(buf), "X%d", uniq);            /* make key unique */
    add_string(buf, &key);

  for (i = 0; i < intern->cdl->cnt; i++) {                    /* build key */
    compose_key(intern->cdl->c[i].c, intern->cdl->c[i].n, &key);
  }

  return key;
}



/* Allocates or expand image store.

Revision: 24.01.2003 E.Sack

Input : - pointer to image
        - image size
        - horizontal & vertical offset
Output: - sector
*/
static void
allocate_image_store(pic_def **expic, sector_def max_sec, int offset_x, 
		     int offset_y)
{  
  pic_def *pictmp = 0;

  CALLOC(pic_def, 1, pictmp);                      /* allocate image store */
  init_pic_def(pictmp);

  SECTOR_TO_AREA(max_sec, pictmp->ar);
  CALLOC(PICT, AREA_SIZE(pictmp->ar), pictmp->mem);

  if (*expic) {
    copy_bmps(**expic, *pictmp, (*expic)->ar, 0, 0, offset_x, offset_y);
    free_pic_def(*expic);
    FREEMEM(*expic);
  }

  *expic = pictmp;
}



/* Calculates the sector over all of two sectors.

Revision: 30.11.2002 E.Sack

Input : - sector A and B
Output: - sector
*/
static sector_def
max_sector(sector_def sec_a, sector_def sec_b)
{
  sector_def sector_over_all;

  sector_over_all.line.start  = XPMIN(sec_a.line.start,  sec_b.line.start);
  sector_over_all.column.start = XPMIN(sec_a.column.start, sec_b.column.start);
  sector_over_all.line.end    = XPMAX(sec_a.line.end,    sec_b.line.end);
  sector_over_all.column.end   = XPMAX(sec_a.column.end,   sec_b.column.end);

  return sector_over_all;
}



/* Calculates the sector over all of a candidate list.

Revision: 30.11.2002 E.Sack

Input : - candidate list
Output: - sector
*/
static sector_def
max_frame(cand_lst clst)
{
  int        i;
  sector_def sector;

  init_sector_def(&sector);

  if (!clst.c) {
    assert(!clst.cnt);                            /* empty candidates list */
    return sector;
  }
  else {
    assert(clst.cnt>0);
  }

  sector = clst.c[0].knl;                                          /* init */
  for (i = 1; i < clst.cnt; i++) {
    sector = max_sector(sector, clst.c[i].knl);
  }

  return sector;
}



/* Calculates the step increment. Basic algorithm is scanning a region for
   pixels.

Revision: 31.01.2003 E.Sack

Input : - image
        - candidate
        - minimal step increment
Output: - step increment
*/
static int
stepping(pic_def src, sector_def roi, int step_increment, double interleave)
{
  int step, x, y, column = roi.column.end;

  assert(step_increment>0);

  /* check columns for foreground pixels from left to right                */
  for (x = roi.column.start+1; x <= roi.column.end; x++) {   /* horizontal */
    for (y = roi.line.start; y <= roi.line.end; y++) {         /* vertical */
      if (src.mem[x + y*src.ar.width] != BACKGROUND) {        /* is pixel? */
	column = x;                                          /* yes - stop */
	x = roi.column.end+1;       /* break outer loop because of runtime */
	break;                                         /* break inner loop */
      }
    }
  }

  step = (int)((1.0-interleave)*(double)(column-roi.column.start));
   
  /* This line is the most sensitive place of runtime influence in spatial
     domain.                                                               */
  step = XPMAX(step, step_increment);

  return step;
}



/* Sorts out candidates, which are not competing, from a candidate list by
   comparing the candidates pattern in pairs. If the minimal overlapping
   degree, which means competition, is not achieved, the candidate with the 
   rightmost location will be deleted.

Revision: 25.01.2003 E.Sack

Input : - pointer to candidate list
        - pointer to class list
Output: - candidate list updated
*/
static void
delete_non_competitors(cand_lst *klist, cls_def *cls, double nb_lap)
{
  int     ci, ni, cj, nj, i, j, km, kw, lx, ly;
  double  fsm_i, fsm_j;
  pic_def fld_i, fld_j;

  for (i = 0; i < klist->cnt-1; i++) {
    ci = klist->c[i].c;
    ni = klist->c[i].n;
    fsm_i = cls[ci].nbl.nb[ni].ref.fldm;
    fld_i = cls[ci].nbl.nb[ni].ref.fld;

    for (j = i+1; j < klist->cnt; j++) {
      cj = klist->c[j].c;
      nj = klist->c[j].n;
      fsm_j = cls[cj].nbl.nb[nj].ref.fldm;
      fld_j = cls[cj].nbl.nb[nj].ref.fld;
      km = XPMIN(fsm_i, fsm_j);
      km = (int)(nb_lap * (double)km);   /* minimal maximal overlap degree */

      /* relative lag of NB kernels                                        */
      lx = klist->c[j].hdx.column.start - klist->c[i].hdx.column.start;
      ly = klist->c[j].hdx.line.start - klist->c[i].hdx.line.start;

      kw = overlapping_degree(fld_i, fld_j, lx, ly);

      if (kw<=km && lx) {                       /* not a competing NB pair */
	if (lx > 0) {                                   /* delete right NB */
	  delete_cand_def(klist, j);
	  j--;                                                   /* repeat */
	}
	else {
	  delete_cand_def(klist, i);
          i--;                                                   /* repeat */
	  break;
	}
      }
    }
  }
}



/* Calculates the width of the best candidate.

Revision: 31.01.2003 E.Sack

Input : - candidate list
        - NB list
Output: - max. width
*/
static int
lookahead(cand_lst klist, cls_def *cls, int step_increment, double interleave)
{
  int    i, c, n, width = 0;
  double max;

  assert(step_increment>0);

  for (i = 0, max = 0; i < klist.cnt; i++) {
    if (max < klist.c[i].m.cv) {
      max = klist.c[i].m.cv;
      c = klist.c[i].c;
      n = klist.c[i].n;
      width = cls[c].nbl.nb[n].pat.ar.width;
    }
  }

  return XPMAX(step_increment, (int)((1.0-interleave)*(double)width));
}



/* RBs, which are too small regarding the class, are marked as invalid.

Revision: 19.06.2004 E.Sack

Input : - RB list
        - class list
Output: - RB list updated
*/
static void
validate_rb_lst(rb_lst rblst, cls_lst clst)
{
  int c, n, min = XPLAB_INIT_AREA, ra;

  if (!rblst.cnt) {
    return;                                               /* empty RB list */
  }

  for (c = 0; c < clst.cnt; c++) {
    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      ra = AREA_SIZE(clst.cls[c].nbl.nb[n].pat.ar);

      if (min == XPLAB_INIT_AREA) {
	min = ra;
	assert(min != XPLAB_INIT_AREA);
	continue;
      }

      min = XPMIN(min, ra);
    }
  }

  if (min == XPLAB_INIT_AREA) {
    return;
  }

  for (c = 0; c < rblst.cnt; c++) {
    ra = SECTOR_SIZE(rblst.rb[c].sc);                           /* RB size */

    if (min >= RB_PURIFY_MULT*ra) {                        /* cleaning mud */
      rblst.rb[c].rld = INVALID;
    }
  }
}



/* Generates RB chain. Even if a class list is used, no recognition results
   are used. Classes control only some distances for neighborhood 
   relations, which again control the RB chain length, which lastly 
   determine the tile image size. Basically one can use any database,
   whichever fullfills the expectations. Best results gives the database
   with patterns trained from that document, of course.

Revision: 24.08.2002 E.Sack

Input : - class list
        - document RB list
        - tiles size wish
        - pointer to image under test
        - chain control flag
        - recognition control flag
        - tile extension
        - pointer to document filename
Output: - RB chain list
*/
tiles_def
rb_chain_list(cls_lst cls, rb_lst rblst, int tile_size, pic_def *rpic, 
	      RB_CHAIN_BIT rb_flag, int match, int ext, char *filename,
	      int rawbit, run_def *crun)
{
  int       mem;
  tiles_def rcl;

  init_tiles_def(&rcl);

  if (!((rb_flag & RB_CHAIN_DUMP_TILES) ||
	(rb_flag & RB_CHAIN_DUMP_COLOR) ||
	(match && (rb_flag & RB_CHAIN_USE)))) {
    return rcl;
  }

  rcl.rch = chain_neighborhoods(cls, rblst, crun);
  chain_frames_lst(rcl.rch);
  chain_collect_rule(&rcl.rch);
  chain_extend_rule(&rcl.rch, rpic->ar, tile_size);

  rcl.border = ext;

  xreport("[%d tile%s]\n", rcl.rch.cnt, (rcl.rch.cnt==1)?"":"s");

  if (*crun->pactive == TRUE) {
    dump_colored_chain_lst(rcl.rch, *rpic, rb_flag, filename, rawbit);
    mem = dump_chain_lst(&rcl, rpic, rb_flag, match, filename);

    xreport("external memory usage %f%%\n",
	   (double)(mem*100)/(double)(AREA_SIZE(rblst.rar)));
  }

  return rcl;
}



/* Generate RB neighborhood, which are the basic of RB chains. Neighborhood
   is a NB feature and applied to RBs.

Revision: 05.08.2005 E.Sack

Input : - class list
        - RB list
Output: - RB chain list
*/
static rb_chn_lst
chain_neighborhoods(cls_lst cls, rb_lst rblst, run_def *crun)
{
  int        c, r;
  rb_chn_lst rch;

  init_rb_chn_lst(&rch);

  if (!rblst.cnt) {
    return rch;
  }

  rch.cnt = rblst.cnt;
  CALLOC(rb_chn_def, rch.cnt, rch.rcd);
  for (r = 0; r < rblst.cnt; r++) {
    init_rb_chn_def(&rch.rcd[r]);
  }

  xplab_progress_configure(crun, "RB chains");
  xplab_progress_set_value(crun, 1.0, (gdouble)rch.cnt);

  for (r = 0; r < rch.cnt && *crun->pactive == TRUE; r++) {/*get all chain seeds */
    xplab_progress_set_value(crun, (gdouble)(r+1), (gdouble)rch.cnt);

    rblst.rb[r].rld = RBN;                   /* starting RB is first entry */

    get_rb_chain(rblst, &rch.rcd[r].rb_chain);           /* get it to list */

    /* neighborhood of NRB, no recursiv function because of deep stack pos.*/
    for (c = 0; c < rch.rcd[r].rb_chain.cnt; c++) {
      neighborhood_of_rb(cls, rblst, rch.rcd[r].rb_chain.rb[c].sc);
      get_rb_chain(rblst, &rch.rcd[r].rb_chain);
    }

    reset_rb_lst(rblst);      /* make RB available for next possible chain */
  }

  xreport("RB chain seeds (same RBs in several chain seeds possible): %d\n", 
	 rch.cnt);

  merge_rb_chains(&rch, crun);
  purge_rb_chains(rch);

  for (r = 0; r < rch.cnt; r++) {
    rb_sort(&rch.rcd[r].rb_chain);
  }

  return rch;
}



/* All RBs in the RB chain list are made mutually exclusive.

Revision: 26.03.2002 E.Sack

Input : - RB chain list
Output: - RB chain list updated
*/
static void
purge_rb_chains(rb_chn_lst rch)
{
  int u, l, i, j, r, c;

  for (c = 0; c < rch.cnt; c++) {                             /* all chains */
    for (l = 0; l < rch.rcd[c].rb_chain.cnt-1; l++) {         /* scan chain */
      r = rch.rcd[c].rb_chain.rb[l].raux;                     /* RB number  */
      for (i = l+1; i < rch.rcd[c].rb_chain.cnt; i++) {       /* scan rest  */
	u = rch.rcd[c].rb_chain.rb[i].raux;        /* search same RB number */
	if (r == u) {                                         /* same found */
	  for (j = i; j < rch.rcd[c].rb_chain.cnt-1; j++) {    /* delete it */
	    rch.rcd[c].rb_chain.rb[j] = rch.rcd[c].rb_chain.rb[j+1];
	  }
	  rch.rcd[c].rb_chain.cnt--;                        /* shorten list */
	  i--;                                         /* repeat comparison */
	}
      }
    }
  }
}



/* All chains in the RB chain list are made mutually exclusive.

Revision: 05.08.2005 E.Sack

Input : - pointer to RB chain list
Output: - RB chain list updated
*/
static void
merge_rb_chains(rb_chn_lst *rch, run_def *crun)
{
  int     u, l, i, j, r, c, cmax = 0, ch_idx, s_idx, restart = 0, rsum = 0;
  gdouble rch_cnt_old = (gdouble)rch->cnt;

  xplab_progress_configure(crun, "RB chains merge");
  xplab_progress_set_value(crun, 1.0, rch_cnt_old);

  for (c = 0; c < rch->cnt-1 && *crun->pactive == TRUE; c++) {    /* chains */
    ch_idx = rch->rcd[c].rb_chain.cnt;

    for (l = 0; l < ch_idx; l++) {                            /* scan chain */
      r = rch->rcd[c].rb_chain.rb[l].raux;                    /* RB number  */

      for (i = c+1; i < rch->cnt; i++) {        /* search for other entries */
	s_idx = rch->rcd[i].rb_chain.cnt;

	for (j = 0; j < s_idx; j++) {                         /* scan chain */
	  u = rch->rcd[i].rb_chain.rb[j].raux;                /* RB number  */

	  if (r == u) {                                      /* entry found */
	    misc_rb_chn_def(rch, i, c);            /* move and delete chain */
	    restart = 1;                    /* repeat comparison from start */
	    rsum++;
	    break;      
	  }
	}

	if (restart) {
	  break;
	}
      }

      if (restart) {
	break;
      }
    }

    cmax = XPMAX(cmax, c);                /* don't let progress bar decrease */
    xplab_progress_set_value(crun, (gdouble)(cmax+2+rsum), rch_cnt_old);

    if (restart) {
      restart = 0;
      c = -1;
    }
  }

  xplab_progress_set_value(crun, 1.0, 1.0);   /* always show 100% at the end */

  xreport("RB chains (same RB in only one chain): %d\n", rch->cnt);
}



/* All but one RB chains of the chain list are temporarily dumped to disk
   thus saving memory space. RB sectors are calculated according to their
   tile images.

Revision: 25.04.2004 E.Sack

Input : - pointer to RB chain list
        - pointer to image under test
        - chain control flag
        - recognition control flag
        - pointer to document filename
Output: - RB chain list updated
        - disk space used
Postcondition: borders of the document are not extended in tiles.
*/
static int
dump_chain_lst(tiles_def *rcl, pic_def *rpic, RB_CHAIN_BIT dflag, int match,
	       char *filename)
{
  int     rdx, amount = 0, verbose;
  pic_def tpic;

  if (!((dflag & RB_CHAIN_DUMP_TILES) || (match && (dflag & RB_CHAIN_USE)))) {
    return 0;
  }

  if (rcl->rch.cnt <= 1) {                              /* nothing to dump */
    return 0;
  }

  init_pic_def(&tpic);

  for (rdx = 0; rdx < rcl->rch.cnt; rdx++) {
    free_pic_def(&tpic);
    tpic = assemble_tile(*rpic, rcl->rch.rcd[rdx], rcl->border);

    if (rdx == rcl->rch.cnt-1) {                              /* last tile */
      if (dflag & RB_CHAIN_DUMP_TILES) { /* tiles dumped for documentation */
	add_string(filename, &rcl->rch.rcd[rdx].chain_name);
	int_to_string(FILE_CHN_LST, rdx, "", &rcl->rch.rcd[rdx].chain_name);

	dump_pic_def(tpic, rcl->rch.rcd[rdx].chain_name, tpic.ar, 0, 0, 
		     xctrl.cfg.verbose > BE_QUIET, PNM_FORMAT_UNDEF);
	rcl->rch.rcd[rdx].caux = PIC_DUMPED;
	amount += AREA_SIZE(tpic.ar);
      }

      if (match && (dflag & RB_CHAIN_USE)) {
	add_string(filename, &rcl->rch.rcd[rdx].chain_name);
	int_to_string(FILE_CHN_TMP, rdx, "", &rcl->rch.rcd[rdx].chain_name);

	temporary_filename(&rcl->doc.chain_name);

	dump_pic_def(*rpic, rcl->doc.chain_name, rpic->ar, 0, 0, 
		     xctrl.cfg.verbose >= VERBOSE_LOT, PNM_FORMAT_UNDEF);
	rcl->doc.caux = PIC_DUMPED;
	amount += AREA_SIZE(rpic->ar);

	free_pic_def(rpic);                       /* delete document image */
	*rpic = tpic;                             /* map first chain image */
	init_pic_def(&tpic);                                 /* make inert */
	rcl->rch.rcd[rdx].caux = BOOKED;             /* chain image loaded */
      }
    }
    else {
      add_string(filename, &rcl->rch.rcd[rdx].chain_name);
      int_to_string(FILE_CHN_LST, rdx, "", &rcl->rch.rcd[rdx].chain_name);

      verbose = xctrl.cfg.verbose > BE_QUIET;

      if (!(dflag & RB_CHAIN_DUMP_TILES)) {  /* avoid ugly pid on filenames */
	add_string(FILE_EXT, &rcl->rch.rcd[rdx].chain_name);
	temporary_filename(&rcl->rch.rcd[rdx].chain_name);
	verbose = 0;
      }

      dump_pic_def(tpic, rcl->rch.rcd[rdx].chain_name, tpic.ar, 0, 0, 
	           verbose, PNM_FORMAT_UNDEF);

      rcl->rch.rcd[rdx].caux = PIC_DUMPED;
      amount += AREA_SIZE(tpic.ar);
    }
  }

  free_pic_def(&tpic);
  return amount;
}



/* Assembles image tile from source image and changes the chain input
   relative to the tile.

Revision: 25.04.2004 E.Sack

Input : - source image
        - RB chain
        - image border
Output: - chain updated
        - image tile
*/
static pic_def
assemble_tile(pic_def rpic, rb_chn_def rcd, int border)
{
  int      i, dx, dy, sx, sy, x_offset, y_offset;
  pic_def  tpic;
  area_def clp;

  init_pic_def(&tpic);

  SECTOR_TO_AREA(rcd.chain_frame, tpic.ar);                  /* unbordered */

  x_offset = XPMIN(rcd.chain_frame.column.start, border);
  tpic.ar.width += x_offset;
  tpic.ar.width += XPMIN(rpic.ar.width-1-rcd.chain_frame.column.end, border);

  y_offset = XPMIN(rcd.chain_frame.line.start, border);
  tpic.ar.height += y_offset;
  tpic.ar.height += XPMIN(rpic.ar.height-1-rcd.chain_frame.line.end, border);

  CALLOC(PICT, AREA_SIZE(tpic.ar), tpic.mem);

  for (i = 0; i < rcd.rb_chain.cnt; i++) {
    SECTOR_TO_AREA(rcd.rb_chain.rb[i].sc, clp);

    sx = rcd.rb_chain.rb[i].sc.column.start;
    sy = rcd.rb_chain.rb[i].sc.line.start;
    dx = sx - rcd.chain_frame.column.start + x_offset;
    dy = sy - rcd.chain_frame.line.start + y_offset;

    copy_bmps(rpic, tpic, clp, sx, sy, dx, dy);           /* assemble tile */

    /* calculate new RB sector relative to tile                            */
    rcd.rb_chain.rb[i].sc.column.start = dx;
    rcd.rb_chain.rb[i].sc.line.start = dy;
    rcd.rb_chain.rb[i].sc.column.end = dx + clp.width - 1;
    rcd.rb_chain.rb[i].sc.line.end = dy + clp.height - 1;
  }

  return tpic;
}



/* For documentation purposes the document image is dumped with colored
   RB chains within their tiles.

Revision: 05.08.2005 E.Sack

Input : - RB chain list
        - image under test
        - chain control flag
        - pointer to document filename
Output: - RB chain list image dumped to disk
*/
static void
dump_colored_chain_lst(rb_chn_lst rch, pic_def rpic, RB_CHAIN_BIT dflag,
		       char *filename, int rawbit)
{
  int     len;
  pic_def pic;
  char    *buf = 0;

  if (!(dflag & RB_CHAIN_DUMP_COLOR)) {
    return;
  }

  if (!rch.cnt) {
    return;
  }

  init_pic_def(&pic);

  switch (COLORING_METHOD) {
  case MINIMUM_COLORS: pic = minimum_of_colors(rch, rpic, rawbit); break;
  case MAXIMUM_COLORS: pic = maximum_of_colors(rch, rpic, rawbit); break;
  case RANDOM_COLORS : pic = random_colors(rch, rpic, rawbit); break;
  default: assert(0); break;
  }

  len = strlen(filename)+strlen(FILE_CHN_COL)+1;
  CALLOC(char, len, buf);
  g_snprintf(buf, len, "%s%s", filename, FILE_CHN_COL);
  dump_pic_def(pic, buf, pic.ar, 0, 0, xctrl.cfg.verbose > BE_QUIET, 
	       PNM_FORMAT_UNDEF);

  FREEMEM(buf);
  free_pic_def(&pic);
}



/* Object colors are generated randomly.

Revision: 28.03.2002 E.Sack

Input : - RB chain list
        - input image
Output: - output image
*/
static pic_def
random_colors(rb_chn_lst rch, pic_def rpic, int rawbit)
{
  int        l, i, sx, sy, len;
  area_def   clp;
  sector_def sec;
  pic_def    pic;
  col_def    icolor = {0,0,0}, backgnd = {PICT_MAX,PICT_MAX,PICT_MAX};
  col_def    *used = 0, defc = {0,0,0};

  init_pic_def(&pic);

  pic.ar = rpic.ar;
  CALLOC(col_def, AREA_SIZE(pic.ar), pic.col);
  AREA_TO_SECTOR(pic.ar, sec);
  change_color(sec, icolor, backgnd, pic);

  len = rch.cnt+1;
  CALLOC(col_def, len, used);

  for (l = 0; l < rch.cnt; l++) {
    for (i = 0; i < rch.rcd[l].rb_chain.cnt; i++) {
      SECTOR_TO_AREA(rch.rcd[l].rb_chain.rb[i].sc, clp);

      sx = rch.rcd[l].rb_chain.rb[i].sc.column.start;
      sy = rch.rcd[l].rb_chain.rb[i].sc.line.start;

      copy_grey_to_color(rpic, pic, clp, icolor, backgnd, sx, sy, sx, sy,
		 	 rawbit);
    }

    icolor = random_color(PICT_MAX, used, len, defc);
  }

  FREEMEM(used);

  return pic;
}



/* Object colors equally distributed over the basic colors. Objects are taken
   from a grey input image and converted to an image with different colors.

Revision: 28.03.2002 E.Sack

Input : - RB chain list
        - input image
Output: - output image
*/
static pic_def
maximum_of_colors(rb_chn_lst rch, pic_def rpic, int rawbit)
{
  int        l, i, sx, sy;
  double     g_mean;
  area_def   clp;
  sector_def sec;
  pic_def    pic;
  rgb_coef   step, fcolor = {0.0, 0.0, 0.0};
  col_def    icolor = {0,0,0}, backgnd = {PICT_MAX,PICT_MAX,PICT_MAX};

  init_pic_def(&pic);

  g_mean = pow((double)(rch.cnt+1), 1.0/3.0);          /* maximum of colors */

  step.r = step.g = step.b = (double)PICT_MAX/g_mean;  /* equal color steps */

  pic.ar = rpic.ar;
  CALLOC(col_def, pic.ar.width*pic.ar.height, pic.col);
  AREA_TO_SECTOR(pic.ar, sec);
  change_color(sec, icolor, backgnd, pic);

  for (l = 0; l < rch.cnt; l++) {
    for (i = 0; i < rch.rcd[l].rb_chain.cnt; i++) {
      SECTOR_TO_AREA(rch.rcd[l].rb_chain.rb[i].sc, clp);

      sx = rch.rcd[l].rb_chain.rb[i].sc.column.start;
      sy = rch.rcd[l].rb_chain.rb[i].sc.line.start;

      copy_grey_to_color(rpic, pic, clp, icolor, backgnd, sx, sy, sx, sy,
			 rawbit);
    }

    icolor = color_counter((double)PICT_MAX, step, &fcolor);
  }

  return pic;
}



/* Minimum of object colors distributed over the basic colors. Objects are 
   taken from a grey input image and converted to an image with different colors.

Revision: 04.08.2002 E.Sack

Input : - RB chain list
        - input image
Output: - output image
*/
static pic_def
minimum_of_colors(rb_chn_lst rch, pic_def rpic, int rawbit)
{
  int        l, i, sx, sy;
  area_def   clp;
  sector_def sec;
  pic_def    pic;
  rgb_coef   max, step, fcolor = {0.0, 0.0, 0.0};
  col_def    icolor = {0,0,0}, backgnd = {PICT_MAX, PICT_MAX, PICT_MAX};

  init_pic_def(&pic);

  max = rgb_subdivision(rch.cnt+1);     /* color for objects and background */

  step.r = (double)PICT_MAX/max.r;                           /* color steps */
  step.g = (double)PICT_MAX/max.g;
  step.b = (double)PICT_MAX/max.b;

  pic.ar = rpic.ar;
  CALLOC(col_def, pic.ar.width*pic.ar.height, pic.col);
  AREA_TO_SECTOR(pic.ar, sec);
  change_color(sec, icolor, backgnd, pic);

  for (l = 0; l < rch.cnt; l++) {
    for (i = 0; i < rch.rcd[l].rb_chain.cnt; i++) {
      SECTOR_TO_AREA(rch.rcd[l].rb_chain.rb[i].sc, clp);

      sx = rch.rcd[l].rb_chain.rb[i].sc.column.start;
      sy = rch.rcd[l].rb_chain.rb[i].sc.line.start;

      copy_grey_to_color(rpic, pic, clp, icolor, backgnd, sx, sy, sx, sy,
			 rawbit);
    }

    icolor = color_counter((double)PICT_MAX, step, &fcolor);
  }

  return pic;
}



/* Searching for RB chain frames which lies inside others and add the 
   according chains to the other chain. This doesn't enlarge the RB image
   but lessen the amount of images being stored temporarily.

Revision: 19.08.2002 E.Sack

Input : - pointer to RB chain list
Output: - RB chain list updated
Postcondition: - RB chain list resetet
*/
static void
chain_collect_rule(rb_chn_lst *rch)
{
  int d, f, tr = -1;

  if (!rch->cnt) {
    return;
  }

  reset_tiles_def(*rch);

  for (f = 0; f < rch->cnt; f++) {
    if (rch->rcd[f].caux == BOOKED) {
      continue;
    }

    tr = max_chain_sector(*rch);                         /* biggest sector */

    assert(tr>=0);
    rch->rcd[tr].caux = BOOKED;        /* according frame and chain booked */

    d = inner_frame_rule(rch, &tr);
    if (d) {
      f = -1;                         /* repeat loop because lists changed */
    }
  }

  reset_tiles_def(*rch);
}



/* Searching for chain frames lying inside a given frame. Found frames 
   are erased and moved to the chain regarding the given frame.

Revision: 26.03.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
Postcondition: - chain list content can be moved completely. 
*/
static int
inner_frame_rule(rb_chn_lst *rch, int *tr)
{
  int        r, found = 0;
  sector_def dut, fram;

  fram = rch->rcd[*tr].chain_frame;

  for (r = 0; r < rch->cnt; r++) {
    if (rch->rcd[r].caux == BOOKED) continue;

    dut = rch->rcd[r].chain_frame;

    if ((fram.column.start <= dut.column.start) &&                  /* rule */
	(fram.line.start  <= dut.line.start) &&
	(fram.column.end   >= dut.column.end) &&
	(fram.line.end    >= dut.line.end)) {
      *tr = misc_rb_chn_def(rch, r, *tr);   /* move and delete inner chain */

      found = 1;
      r--;
    }
  }

  return found;
}



/* Build supersector of two sectors.

Revision: 03.04.2002 E.Sack

Input : - sector a
        - sector b
Output: - supersector
*/
static sector_def
merge_chain_frame(sector_def sec_1, sector_def sec_2)
{
  sector_def merge;

  merge.column.start = XPMIN(sec_1.column.start, sec_2.column.start);
  merge.line.start = XPMIN(sec_1.line.start, sec_2.line.start);
  merge.column.end = XPMAX(sec_1.column.end, sec_2.column.end);
  merge.line.end = XPMAX(sec_1.line.end, sec_2.line.end);

  return merge;
}



/* The optimal tile area regarding given constraints is calculated.

Revision: 02.04.2002 E.Sack

Input : - document area size
        - tile divisor
Output: - tile area
*/
static double
calc_tile_factor(area_def par, int tile)
{
  int    pfac_w, pfac_h;
  double pfac_old, par_size = (double)(par.width*par.height);
  double pfac_new = par_size/(double)(tile*tile);

  assert(tile >= TILE_DIV_MIN);

  tile++;
  do {
    tile--;                                                /* not too small */
    pfac_old = pfac_new;
    pfac_w = (int)ceil((double)(par.width)/(double)(tile));
    pfac_h = (int)ceil((double)(par.height)/(double)(tile));
    pfac_new = par_size/(double)(tile*tile);
  } while ((tile > TILE_DIV_MIN)   &&              /* rule for optimal tile */
           (pfac_w <= TILE_SIZE_MIN) && 
	   (pfac_h <= TILE_SIZE_MIN));

  return pfac_old;
}



/* Searching for a sector lying between two others. Very simple method.

Revision: 05.04.2002 E.Sack

Input : - pointer to chain list
        - sector a
        - sector b
Output: - flag (0: no frame found, 1: frame found)
*/
static int 
spanning_sector(rb_chn_lst rch, sector_def s_a, sector_def s_b)
{
  int        r;
  sector_def dut;

  for (r = 0; r < rch.cnt; r++) {
    if (rch.rcd[r].caux != BOOKED) {
      continue;
    }

    dut = rch.rcd[r].chain_frame;

    if (s_a.column.end < dut.column.start && 
	dut.column.end < s_b.column.start) {
      return 1;                                 /* DUT between left & right */
    }

    if (s_a.line.end < dut.line.start && 
	dut.line.end < s_b.line.start) {
      return 1;                                 /* DUT between top & bottom */
    }

    if (s_b.column.end < dut.column.start &&       /* exchange candidates */
	dut.column.end < s_a.column.start) {
      return 1;                                 /* DUT between left & right */
    }

    if (s_b.line.end < dut.line.start && 
	dut.line.end < s_a.line.start) {
      return 1;                                 /* DUT between top & bottom */
    }
  }

  return 0;
}



/* Searching for chain frames touching a given frame. Found frames are 
   erased and moved to the chain regarding the given frame.

Revision: 26.03.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
        - pointer to not expandable flag
Output: - found flag (0: no frames found, 1: frames found)
        - not expandable flag (0: all frames expandable, 
                               1: some frames not expandable)
        - chain list updated
        - chain number updated
*/
static int
touch_frame_rule(rb_chn_lst *rch, int *tr, sector_def tsector, double pfac)
{
  int        d, r, found = 0;
  sector_def dut, preview;
  area_def   sar;

  for (r = 0; r < rch->cnt; r++) {
    if ((rch->rcd[r].caux == BOOKED) || 
	(rch->rcd[r].caux == TOUCHED)) {
      continue;
    }

    dut = rch->rcd[r].chain_frame;

    if ((tsector.column.start > dut.column.end)   ||       /* touching rule */
	(tsector.line.start  > dut.line.end)    ||
	(tsector.column.end   < dut.column.start) ||
	(tsector.line.end    < dut.line.start)) {
      continue;                                          /* DUT out of area */
    }

    d = spanning_sector(*rch, dut, rch->rcd[*tr].chain_frame);
    if (d) {
      continue;                                      /* a sector is spanned */
    }

    preview = merge_chain_frame(dut,                    /* build superframe */
				rch->rcd[*tr].chain_frame);

    SECTOR_TO_AREA(preview, sar);
    if ((double)(sar.width*sar.height) <= pfac) {      /* sector expandable */
      *tr = misc_rb_chn_def(rch, r, *tr);          /* move and delete chain */
      rch->rcd[*tr].chain_frame = preview;                   /* copy sector */
      found = 1;
      r--;                                                        /* repeat */
    }
    else {
      rch->rcd[r].caux = TOUCHED;          /* touched sector not expandable */
    }
  }

  if (found) {
    inner_frame_rule(rch, tr);                                    /* hungry */
  }

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the left. 

Revision: 02.04.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_left(rb_chn_lst *rch, int *tr, double pfac)
{
  int        d, found = 0, i, x;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    x = XPMAX(0, tsector.column.start-i);                             /* left */
    tsector.column.start = x;

    d = touch_frame_rule(rch, tr, tsector, pfac);          /* search chains */

    if (d) {                       /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while (x > 0 &&                               /* as long as expandable */ 
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the top. 

Revision: 19.08.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_up(rb_chn_lst *rch, int *tr, double pfac)
{
  int        found = 0, i, y, d;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    y = XPMAX(0, tsector.line.start-i);                                /* up */
    tsector.line.start = y;

    d = touch_frame_rule(rch, tr, tsector, pfac);          /* search chains */

    if (d) {                       /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while (y > 0 &&                               /* as long as expandable */ 
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the right. 

Revision: 02.04.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_right(rb_chn_lst *rch, int *tr, double pfac, area_def par)
{
  int        found = 0, i, x, d;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    x = XPMIN(par.width-1, tsector.column.end+i);                  /* right */
    tsector.column.end = x;

    d = touch_frame_rule(rch, tr, tsector, pfac);          /* search chains */

    if (d) {                       /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while (x < par.width-1 &&                     /* as long as expandable */ 
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the bottom. 

Revision: 02.04.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_down(rb_chn_lst *rch, int *tr, double pfac, area_def par)
{
  int        found = 0, i, y, d;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    y = XPMIN(par.height-1, tsector.line.end+i);                    /* down */
    tsector.line.end = y;

    d = touch_frame_rule(rch, tr, tsector, pfac);          /* search chains */

    if (d) {                       /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while (y < par.height-1 &&                    /* as long as expandable */ 
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the left and to the top. 

Revision: 02.04.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_left_up(rb_chn_lst *rch, int *tr, double pfac)
{
  int        found = 0, x, y, dx, dy, i;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    x = XPMAX(0, tsector.column.start-i);                           /* left */
    tsector.column.start = x;

    dx = touch_frame_rule(rch, tr, tsector, pfac);          /* search chains*/

    y = XPMAX(0, tsector.line.start-i);                               /* up */
    tsector.line.start = y;

    dy = touch_frame_rule(rch, tr, tsector, pfac);         /* search chains */

    if (dx || dy) {                /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while ((x > 0 ||                              /* as long as expandable */ 
	    y > 0) &&                                   /* in one direction */
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the top and to the right. 

Revision: 02.04.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
        - document area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_up_right(rb_chn_lst *rch, int *tr, double pfac, area_def par)
{
  int        found = 0, x, y, dx, dy, i;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    y = XPMAX(0, tsector.line.start-i);                               /* up */
    tsector.line.start = y;

    dy = touch_frame_rule(rch, tr, tsector, pfac);         /* search chains */

    x = XPMIN(par.width-1, tsector.column.end+i);                  /* right */
    tsector.column.end = x;

    dx = touch_frame_rule(rch, tr, tsector, pfac);         /* search chains */

    if (dx || dy) {                /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while ((x < par.width-1 ||                    /* as long as expandable */ 
           y > 0) &&                                    /* in one direction */
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the right and to the bottom. 

Revision: 02.04.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
        - document area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_right_down(rb_chn_lst *rch, int *tr, double pfac, area_def par)
{
  int        found = 0, x, y, dx, dy, i;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    x = XPMIN(par.width-1, tsector.column.end+i);                  /* right */
    tsector.column.end = x;

    dx = touch_frame_rule(rch, tr, tsector, pfac);         /* search chains */

    y = XPMIN(par.height-1, tsector.line.end+i);                    /* down */
    tsector.line.end = y;

    dy = touch_frame_rule(rch, tr, tsector, pfac);         /* search chains */

    if (dx || dy) {                /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while ((x < par.width-1 ||                     /* as long as expandable */
           y < par.height-1) &&                         /* in one direction */
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Searching for chain frames touching a frame, which is successively 
   extended to the bottom and to the left. 

Revision: 02.04.2002 E.Sack

Input : - pointer to chain list
        - pointer to chain number
        - tile area
        - document area
Output: - flag (0: no frames found, 1: frames found)
        - chain list updated
        - chain number updated
*/
static int
extend_frame_down_left(rb_chn_lst *rch, int *tr, double pfac, area_def par)
{
  int        found = 0, x, y, dx, dy, i;
  sector_def tsector = rch->rcd[*tr].chain_frame;
  area_def   sar;

  i = 0;
  do {
    i += EXTEND_RULE_STEP;
    y = XPMIN(par.height-1, tsector.line.end+i);                    /* down */
    tsector.line.end = y;

    dy = touch_frame_rule(rch, tr, tsector, pfac);         /* search chains */

    x = XPMAX(0, tsector.column.start-i);                           /* left */
    tsector.column.start = x;

    dx = touch_frame_rule(rch, tr, tsector, pfac);         /* search chains */

    if (dx || dy) {                /* real extension according inner chains */
      found = 1;
      tsector = rch->rcd[*tr].chain_frame;             /* reset test sector */
      i = 0;
    }

    SECTOR_TO_AREA(tsector, sar);
  } while ((x > 0 ||                              /* as long as expandable */ 
           y < par.height-1) &&                         /* in one direction */
	   ((double)(sar.width*sar.height) <= pfac));    /* and not too big */

  set_tiles_def(*rch, TOUCHED, RB_STATUS_INIT);

  return found;
}



/* Calculate the four distances between sector and document frame.

Revision: 03.04.2002 E.Sack

Input : - document size
        - sector
        - pointers to four distances
Output: - four distances updated
*/
static void
calc_extension_distance(area_def par, sector_def dut, int *lw, int *uw, 
			int *rw, int *dw)
{
  *lw = dut.column.start;                         /* distance to the left   */
  *uw = dut.line.start;                          /* distance to the top    */
  *rw = par.width - 1 - dut.column.end;           /* distance to the right  */
  *dw = par.height - 1 - dut.line.end;           /* distance to the bottom */
}



/* First expansion of the sector under test starts with the smallest
   distance to the document frame.

Revision: 03.04.2002 E.Sack

Input : - pointer to RB chain list
        - pointer to chain number
        - tile factor
        - document size
Output: - RB chain list updated
        - chain number updated
        - flag (0: no extension found, 1: extension found)
*/
static int
extension_start(rb_chn_lst *rch, int *tr, double pfac, area_def par)
{
  int found = 0, lw, uw, rw, dw;

  calc_extension_distance(par, rch->rcd[*tr].chain_frame, &lw, &uw, &rw, &dw);

  if ((lw <= uw) && (lw <= rw) && (lw <= dw)) {
    found = extend_frame_left(rch, tr, pfac);
    if (found) {
      return found;
    }
  }

  if ((uw <= rw) && (uw <= dw)) {
    found = extend_frame_up(rch, tr, pfac);
    if (found) {
      return found;
    }
  }

  if (rw <= dw) {
    found = extend_frame_right(rch, tr, pfac, par);
    if (found) {
      return found;
    }
  }

  found = extend_frame_down(rch, tr, pfac, par);

  return found;
}



/* Circulate expansion of the sector under test with the smallest
   distance to the document frame.

Revision: 03.04.2002 E.Sack

Input : - pointer to RB chain list
        - pointer to chain number
        - tile factor
        - document size
Output: - RB chain list updated
        - chain number updated
        - flag (0: no extension found, 1: extension found)
*/
static int
circulate_extension(rb_chn_lst *rch, int *tr, double pfac, area_def par)
{
  int found = 0, lw, uw, rw, dw;

  calc_extension_distance(par, rch->rcd[*tr].chain_frame, &lw, &uw, &rw, &dw);

  if ((lw <= uw) && (lw <= rw) && (lw <= dw)) {
    found = extend_frame_left_up(rch, tr, pfac);
    if (found) {
      return found;
    }
  }

  if ((uw <= rw) && (uw <= dw)) {
    found = extend_frame_up_right(rch, tr, pfac, par);
    if (found) {
      return found;
    }
  }

  if (rw <= dw) {
    found = extend_frame_right_down(rch, tr, pfac, par);
    if (found) {
      return found;
    }
  }

  found = extend_frame_down_left(rch, tr, pfac, par);

  return found;
}



/* Calculate the maximal tile starting from a given sector. It cannot be 
   garanteed that tiles smaller than the document image can be found 
   because RB neighborhood must not be disturbed. The method uses a
   carefully growth of the tile by attempting growing in one or two 
   directions.

Revision: 26.03.2002 E.Sack

Input : - pointer to RB chain list
        - chain number
        - tile factor
        - document size
Output: - RB chain list updated
        - flag (0: no extension found, 1: extension found)
*/
static int
extend_frame_rule(rb_chn_lst *rch, int tr, double pfac, area_def par)
{
  int      i = 0, found;
  area_def aar;

  found = extension_start(rch, &tr, pfac, par);/*only one direction expanded*/

  SECTOR_TO_AREA(rch->rcd[tr].chain_frame, aar);
  if ((double)(aar.width*aar.height) > pfac) {
    return found;                              /* sector no more extendable */
  }

  do {
    i++;
    found = circulate_extension(rch, &tr, pfac, par);
  } while (found);
  assert(i<=5);

  return found;                                     /* circulation finished */
}



/* Starting with the biggest frame, searching for chain frames which lies 
   outside and add the according chains to the big one up to a given amount. 
   The search direction from the starting frame is to the nearest frame
   border.

Revision: 19.08.2002 E.Sack

Input : - pointer to RB chain list
        - document size
        - tile factor
Output: - RB chain list updated
Postcondition: - RB chain list resetet
*/
static void
chain_extend_rule(rb_chn_lst *rch, area_def par, int tile_size)
{
  int    d, f, tr = -1;
  double pfac;

  if (!rch->cnt) {
    return;
  }

  pfac = calc_tile_factor(par, tile_size);

  reset_tiles_def(*rch);

  for (f = 0; f < rch->cnt; f++) {
    if (rch->rcd[f].caux == BOOKED) continue;

    tr = max_chain_sector(*rch);                          /* biggest sector */

    assert(tr>=0);
    rch->rcd[tr].caux = BOOKED;         /* according frame and chain booked */

    d = extend_frame_rule(rch, tr, pfac, par);

    if (d) {
      f = -1;                          /* repeat loop because lists changed */
    }
  }

  reset_tiles_def(*rch);

  xreport("\nCropping, size and number of RBs of RB Chain Tiles:\n");
  for (f = 0; f < rch->cnt; f++) {                    /* document purposes */
    xreport("tile[%d] %d-%d x %d-%d (%d x %d) (%d)\n", f,
	    rch->rcd[f].chain_frame.column.start,
	    rch->rcd[f].chain_frame.column.end,
	    rch->rcd[f].chain_frame.line.start,
	    rch->rcd[f].chain_frame.line.end,
	    rch->rcd[f].chain_frame.column.end-
	    rch->rcd[f].chain_frame.column.start+1,
	    rch->rcd[f].chain_frame.line.end-
	    rch->rcd[f].chain_frame.line.start+1,
	    rch->rcd[f].rb_chain.cnt);
  }
}



/* Searching for the biggest bounding box of all unbooked chains of the list.
 
Revision: 26.03.2002 E.Sack

Input : - RB chain list
Output: - chain index
*/
static int
max_chain_sector(rb_chn_lst rch)
{
  int        r, tr = -1;
  sector_def tsector;

  for (r = 0; r < rch.cnt; r++) {
    if (rch.rcd[r].caux == BOOKED) continue;

    if (tr == -1) {
      tsector = rch.rcd[r].chain_frame;                     /* init sector */
      tr = r;
      continue;
    }

    if ((tsector.column.end-tsector.column.start)*
	(tsector.line.end-tsector.line.start) <=
	(rch.rcd[r].chain_frame.column.end-
         rch.rcd[r].chain_frame.column.start)*
	(rch.rcd[r].chain_frame.line.end-
         rch.rcd[r].chain_frame.line.start)) {
      tsector = rch.rcd[r].chain_frame;                  /* biggest sector */
      tr = r;
    }
  }

  return tr;
}



/* All bounding boxes of the chains in the list are calculated.
 
Revision: 19.08.2002 E.Sack

Input : - RB chain list
Output: - RB chain list updated
*/
static void
chain_frames_lst(rb_chn_lst rch)
{
  int f;

  for (f = 0; f < rch.cnt; f++) {
    rch.rcd[f].chain_frame = chain_frame_comp(rch.rcd[f].rb_chain);
  }
}



/* The bounding box of a chain is calculated.
 
Revision: 26.03.2002 E.Sack

Input : - RB chain
Output: - bounding box
*/
static sector_def
chain_frame_comp(rb_lst rb_chain)
{
  int        r;
  sector_def frame;

  init_sector_def(&frame);

  if (rb_chain.rb) {                                  /* check consistency */
    assert(rb_chain.cnt>0);
  }
  else {
    assert(!rb_chain.cnt);
    return frame;
  }

  frame = rb_chain.rb[0].sc;                                /* init. frame */

  for (r = 1; r < rb_chain.cnt; r++) {
    frame.column.start = XPMIN(frame.column.start,
                                   rb_chain.rb[r].sc.column.start);
    frame.column.end = XPMAX(frame.column.end, 
				   rb_chain.rb[r].sc.column.end);
    frame.line.start = XPMIN(frame.line.start, 
                                   rb_chain.rb[r].sc.line.start);
    frame.line.end = XPMAX(frame.line.end,
				   rb_chain.rb[r].sc.line.end);
  }

  return frame;
}



/* Marked RBs of the input RB list are copied and booked to an output RB list.
   Input is used to be the document's RB list and output is used to be the
   RB list of a RB chain.

Revision: 26.03.2002 E.Sack

Input : - (document) RB list
        - pointer to (chain) RB list
Output: - (chain) RB list updated
*/
static void
get_rb_chain(rb_lst rblst, rb_lst *rb_chain)
{
  int r, rb_idx = rb_chain->cnt;

  for (r = 0; r < rblst.cnt; r++) {
    if (rblst.rb[r].rld == RBN) {
      rblst.rb[r].rld = BOOKED;

      if (rb_idx) REALLOC(rb_def, rb_idx+1, rb_chain->rb);
      else CALLOC(rb_def, rb_idx+1, rb_chain->rb);

      init_rb_def(&rb_chain->rb[rb_idx]);

      copy_rb_def(&rblst.rb[r], &rb_chain->rb[rb_idx]);
      rb_chain->rb[rb_idx].raux = r;                     /* store RB number */
      rb_chain->rb[rb_idx].rld = RB_STATUS_INIT;         /* reinit mark     */

      rb_idx++;
    }
  }

  rb_chain->cnt = rb_idx;
}



/* Searching NRB belonging to a neighborhood. An RB is a neighbored RB if it
   lies within the neighborhhood sector. 

Revision: 26.03.2002 E.Sack

Input : - RB list
        - neighborhood sector
Output: - RB list updated
*/
static void
neighbors_of_rb(rb_lst rblst, sector_def nbh)
{
#define HUNGRY_NEIGHBORHOOD 0
  int r;

  for (r = 0; r < rblst.cnt; r++) {
    if (rblst.rb[r].rld == BOOKED ||
        rblst.rb[r].rld == RBN) {
      continue;                                        /* RB already booked */
    }

    switch (HUNGRY_NEIGHBORHOOD) {
    case 1:
      if (rblst.rb[r].sc.column.start > nbh.column.end ||
	  rblst.rb[r].sc.column.end < nbh.column.start ||
	  rblst.rb[r].sc.line.start > nbh.line.end   ||
	  rblst.rb[r].sc.line.end < nbh.line.start) {
	continue;                             /* RB lies outer neighborhood */
      }

      rblst.rb[r].rld = RBN;                               /* neighbored RB */
      break;

    case 0:
      if (rblst.rb[r].sc.column.start >= nbh.column.start &&
	  rblst.rb[r].sc.column.end <= nbh.column.end     &&
	  rblst.rb[r].sc.line.start >= nbh.line.start   &&
	  rblst.rb[r].sc.line.end <= nbh.line.end) {
	rblst.rb[r].rld = RBN;     /* neighbored RB lies inner neighborhood */
      }
    }
  }
}



/* Within a class list the biggest possible neighborhood of one RB is 
   searched.

Revision: 27.03.2002 E.Sack

Input : - class list
        - RB list
        - RB sector under test
Output: - RB list updated
*/
static void
neighborhood_of_rb(cls_lst cls, rb_lst rblst, sector_def rsec)
{
  int n, c;

  for (c = 0; c < cls.cnt; c++) {
    for (n = 0; n < cls.cls[c].nbl.cnt; n++) {
      neighborhood_of_nnb(cls.cls, rblst, cls.cls[c].nbl.nb[n], rsec);
    }
  }

  reset_nb_aux(cls);                        /* reset bookings for next RB */
}



/* Sets the named mark of all RBs in the list to its initial value.

Revision: 30.10.2003 E.Sack

Input : - RB list
Output: - RB list updated
*/
void
reset_rb_lst(rb_lst rblst)
{
  int r;

  for (r = 0; r < rblst.cnt; r++) {
    rblst.rb[r].raux = RB_STATUS_INIT;
    rblst.rb[r].rld = RB_STATUS_INIT;
  }
}



/* Sets the named mark of all NBs in the class list to its initial value.

Revision: 27.07.2002 E.Sack

Input : - class list
Output: - class list updated
*/
static void
reset_nb_aux(cls_lst clst)
{
  int n, c;

  for (c = 0; c < clst.cnt; c++) {
    for (n = 0; n < clst.cls[c].nbl.cnt; n++) {
      clst.cls[c].nbl.nb[n].aux = 0;
    }
  }
}



/* Sets the mark of all chains in the list to its initial value.

Revision: 26.03.2002 E.Sack

Input : - chain list
Output: - chain list updated
*/
static void
reset_tiles_def(rb_chn_lst rch)
{
  int c;

  for (c = 0; c < rch.cnt; c++) {
    rch.rcd[c].caux = 0;
  }
}



/* Sets the mark of all chains in the list to a certain value.

Revision: 03.04.2002 E.Sack

Input : - chain list
        - mark
Output: - chain list updated
*/
static void
set_tiles_def(rb_chn_lst rch, RB_STATUS from, RB_STATUS to)
{
  int       c;
  RB_STATUS mark;

  for (c = 0; c < rch.cnt; c++) {
    mark = rch.rcd[c].caux;
    rch.rcd[c].caux = (mark==from) ? to : mark;
  }
}



/* A neighborhood of a NNB is calculated.

Revision: 26.03.2002 E.Sack

Input : - pointer to class list
        - RB list
        - NNB
        - sector
Output: - RB list updated
*/
static void
neighborhood_of_nnb(cls_def *cls, rb_lst rblst, nb_def nnb, sector_def in_sec)
{
  int        ch;
  sector_def nbh;

  for (ch = 0; ch < nnb.cnb.cnt; ch++) {
    nbh = neighborhood_of_cnb(cls, nnb.cnb.cd[ch].cdx, nnb.cnb.cd[ch].ndx, 
			      in_sec);
    neighbors_of_rb(rblst, nbh);
  }
}



/* A neighborhood of a sector for a CNB is calculated. Every NNB is
   placed in the four corners of the sector. The placement determines
   the maximal region the NNB's CNB can superimpose. The superimposed 
   regions of all that CNB's shifts defines the neighborhood.

Revision: 26.03.2002 E.Sack

Input : - pointer to class list
        - CNB class number
        - CNB class NB number
        - sector
Output: - neighborhood sector
Postcondition: neighborhood sectors are not bound to image size.
*/
static sector_def
neighborhood_of_cnb(cls_def *cls, int cnb_c, int cnb_n, sector_def in_sec)
{
  int        chn_for, foc_c, foc_n, foc_width, foc_height, foc_x, foc_y;
  int        cnb_width, cnb_height;
  area_def   in_area;
  sector_def nbh = in_sec;

  assert(cls);

  SECTOR_TO_AREA(in_sec, in_area);

  for (chn_for = 0; chn_for < cls[cnb_c].nbl.nb[cnb_n].nnb.cnt; chn_for++) {
    foc_c = cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_for].cdx; /* NNB in sector */
    foc_n = cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_for].ndx;
    foc_width = cls[foc_c].nbl.nb[foc_n].pat.ar.width;
    foc_height = cls[foc_c].nbl.nb[foc_n].pat.ar.height;

    if (cls[foc_c].nbl.nb[foc_n].aux == BOOKED) {
      continue;                                        /* yet been in focus */
    }

    cls[foc_c].nbl.nb[foc_n].aux = BOOKED;

    if (in_area.width < foc_width || in_area.height < foc_height) {
      continue;        /* input sector too small to contain the testing NNB */
    }

    foc_x = cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_for].x;/*test NNB's offset */
    foc_y = cls[cnb_c].nbl.nb[cnb_n].nnb.cd[chn_for].y; /* in CNB          */

    cnb_width = cls[cnb_c].nbl.nb[cnb_n].pat.ar.width;
    cnb_height = cls[cnb_c].nbl.nb[cnb_n].pat.ar.height;

    nbh.column.start = XPMIN(nbh.column.start, in_sec.column.start-foc_x);
    nbh.line.start = XPMIN(nbh.line.start, in_sec.line.start-foc_y);
    nbh.column.end = XPMAX(nbh.column.end, 
			   in_sec.column.end+cnb_width-foc_x-foc_width);
    nbh.line.end = XPMAX(nbh.line.end, 
			 in_sec.line.end+cnb_height-foc_x-foc_height);
  }

  return nbh;
}



/* Composes a key with one class and NB number pair.

Revision: 05.08.2005 E.Sack

Input : - class number
        - NB number
        - pointer to key
Output: - key updated
*/
static void
compose_key(int cnr, int nnr, char **key)
{
  char buf[MAXCHAR];

  g_snprintf(buf, sizeof(buf), "C%dN%d", cnr, nnr);

  add_string(buf, key);
}



/* Reduces a key by one class number and NB number pair.

Revision: 17.11.2002 E.Sack

Input : - pointer to key
Output: - key updated
*/
static void
reduce_key(char **key)
{
  char *last = 0;

  assert(*key);

  last = strrchr(*key, 'C');

  assert(last!=NULL);

  *last = '\0';                                              /* reduce key */

  if (!strlen(*key)) {
    FREEMEM(*key);
  }
}



/* Searches for a contour of a certain mark and changes it. 

Revision: 01.06.2004 E.Sack

Input : - image
        - clipping
        - mark from change
        - mark to change
        - scan direction
        - neighborhood
Output: - flag (0: no mark found, 1: mark found)
        - pixels changed as flagged
*/
int
xcontour_mark(pic_def shdw, sector_def fr, PICT from, PICT to, int bot_up, 
	      int n4) 
{
  int           found = 0;
  h_contour_lst ctr;

  ctr = xgen_h_contour(shdw, fr, from, bot_up, n4);     /* get one contour */

  if (ctr.cnt) {
    set_h_contour(ctr, from, to, shdw);
    found = 1;
  }

  free_h_contour_lst(&ctr);

  return found;
}



/* Searches for a contour of a marked start coordinate and changes it. 

Revision: 13.09.2004 E.Sack

Input : - image
        - clipping
        - mark to change
        - start coordinate
        - neighborhood
Output: - flag (0: no mark found, 1: mark found)
        - pixels changed as flagged
*/
int
xcontour_mark_from_start(pic_def shdw, sector_def fr, PICT to, c_def start,
                         int n4) 
{
  int           found = 0;
  PICT          from = shdw.mem[start.x + start.y*shdw.ar.width];
  h_contour_lst ctr;

  ctr = xgen_h_contour_from_start(shdw, fr, start, n4);  /* get one contour */

  if (ctr.cnt) {
    set_h_contour(ctr, from, to, shdw);
    found = 1;
  }

  free_h_contour_lst(&ctr);

  return found;
}



/* Changes certain pixels within a contour regardless holes.

Revision: 01.06.2004 E.Sack

Input : - contour
        - mark from change
        - mark to change
        - image
Output: - image updated
        - flag (0: no mark found, 1: mark found)
*/
int 
set_h_contour(h_contour_lst ctr, PICT from, PICT to, pic_def pic)
{
  int reg, i, z, s, mark_found = 0, width = pic.ar.width;

  for (i = 0; i < ctr.cnt; i++) {
    z = ctr.hcl[i].yco*width;

    for (s = ctr.hcl[i].xil.start; s <= ctr.hcl[i].xil.end; s++) {
      if (pic.mem[reg = s+z] == from) {
        pic.mem[reg] = to;                                 /* change pixel */
 	mark_found = 1;
      }
    }
  }

  return mark_found;
}



/* Output generation regarding recognition.

Revision: 02.05.2003 E.Sack
 
Input : - RB list
        - candidate list
        - TPS
        - mask border
        - maximal allowed lag
        - RB image
        - pointer to input image filename
        - forced inline method
        - pointer to control structure
        - pointer to log widget
        - renew flag
Output: - candidate list updated
        - marked RB list
        - control structure updated
*/
void
xrecognition_out(rb_lst rblst, cand_lst hit_lst, tps_def nbl, int mborder,
                 int nborder, pic_def rpic, char *filename, 
		 INLINE_METHOD forced_line_method, run_def *crun, 
		 GtkWidget *ilog, int renew)
{
  int     err, terse = xctrl.cfg.report_write & REPORT_WRITE_TERSE;
  int_lst mst;                                /* mini-statistic for output */
  rb_lst  mblst;                        /* for documentation purposes only */
  int_lst docu_lines;

  err = check_name(filename);
  assert(!err);

  init_int_lst(&mst); 
  init_rb_lst(&mblst); 
  init_int_lst(&docu_lines);

  mst = check_quality(rblst, nbl.cl, hit_lst, terse);

  if (!terse) {
    print_hitlist(rpic.ar, nborder, mborder, hit_lst, nbl.cl.cls);
  }

  xgen_ascii_text(nbl.cl, hit_lst, mst, filename, nborder, nbl.media.in_line, 
		  rpic.ar, forced_line_method, &docu_lines, crun, ilog, renew);

  mblst = merge_crb(hit_lst, rblst);                          /* RB -> MRB */

  xdc_source_all(filename, rblst, mblst, hit_lst, nbl.cl.cls, docu_lines);

  free_int_lst(&mst);
  free_int_lst(&docu_lines);
  free_rb_lst(&mblst); 
}



/* With the assistance of the Hit List the MRB (merged RB) List is build. 
   MRB's are the RB's which are expanded by the CRB's.

Input : - hitlist
        - RB List
Output: - MRB List 
*/
static rb_lst
merge_crb(cand_lst hlist, rb_lst rbl) 
{
  int        i, j, k, rno, cno, *done = 0;
  sector_def zs;
  rb_lst     mrbl;                                       /* merged RB List */

  init_rb_lst(&mrbl);

  if (!hlist.cnt) {
    return mrbl;
  }

  mrbl.rar = rbl.rar;                                       /* general info */

  CALLOC(int, rbl.cnt, done);

  /* MRB size build as the biggest spanning frame of all RB/CRB contained.  */
  for (i = 0; i < hlist.cnt; i++) {
    rno = hlist.c[i].rbn;

    if (done[rno]) {
      continue;
    }
    else {
      done[rno] = 1;
    }

    if (!mrbl.cnt) {
      CALLOC(rb_def , mrbl.cnt+1, mrbl.rb);
    }
    else {
      REALLOC(rb_def , mrbl.cnt+1, mrbl.rb);
    }

    init_rb_def(&mrbl.rb[mrbl.cnt]);

    zs = rbl.rb[rno].sc;                            /* frame initialisation */

    for (k = 0; k < hlist.c[i].crb.cnt; k++) {
      cno = hlist.c[i].crb.nr[k];
      zs = expand_sector(rbl.rb[cno].sc, zs);                  /* RB += CRB */
    }

    for (j = i+1;            /* looking for other candidates of the same RB */
	 j < hlist.cnt;
	 j++) {
      if (hlist.c[j].rbn != rno) {
	continue;
      }

      for (k = 0; k < hlist.c[j].crb.cnt; k++) {
	cno = hlist.c[j].crb.nr[k];
        zs = expand_sector(rbl.rb[cno].sc, zs);                /* RB += CRB */
      }
    }

    mrbl.rb[mrbl.cnt].sc = zs;                          /* enroll MRB frame */
    mrbl.cnt++;                                           /* number of MRBs */
  }

  FREEMEM(done);
  return mrbl;
}



/* The spanning frame is build from two RBs.

Input : - source sector
        - destination sector
Output: - destination sector
Postcondition: destination sector overwritten.
*/
static sector_def
expand_sector(sector_def src, sector_def dst) 
{
  dst.line.start  = XPMIN(dst.line.start, src.line.start);
  dst.line.end    = XPMAX(dst.line.end, src.line.end);
  dst.column.start = XPMIN(dst.column.start, src.column.start);
  dst.column.end   = XPMAX(dst.column.end, src.column.end);

  return dst;
}
